/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-async-queue.h - Async message passing between threads.
 *
 * Copyright (C) 2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 *
 * Based on GAsyncQueue from the GLib library.
 */

#ifndef TNT_ASYNC_QUEUE_H
#define TNT_ASYNC_QUEUE_H

#include <glib/gthread.h>
#include "tnt-types.h"

struct _TntAsyncQueue
{
  GMutex *mutex;
  GCond *cond;
  GCond *dispatch_cond;

  GQueue *action_queue;
  GQueue *return_queue;

  GHashTable *action_hash;

  gpointer last_popped_action;
  gpointer dispatching_for_action;

  guint discard_returns          : 1;
  guint cancelling_popped_action : 1;

  guint waiting_threads;
  guint ref_count;

  TntDispatchFunc *return_dispatch_func;
  GHashFunc action_hash_func;
  GEqualFunc action_equal_func;
  GDestroyNotify action_destroy_func;
  GDestroyNotify return_destroy_func;

  guint return_dispatch_id;
  gpointer callback_data;
};

/* Asyncronous Queues, can be used to communicate between threads
 */

/* Get a new TntAsyncQueue with the ref_count 1 */
TntAsyncQueue *tnt_async_queue_new                       (TntDispatchFunc *return_dispatch_func,
                                                          GHashFunc action_hash_func,
                                                          GEqualFunc action_equal_func,
                                                          GDestroyNotify action_destroy_func,
                                                          GDestroyNotify return_destroy_func,
                                                          gpointer callback_data);

/* Lock and unlock a TntAsyncQueue. All functions lock the queue for
 * themselves, but in certain cirumstances you want to hold the lock longer,
 * thus you lock the queue, call the *_unlocked functions and unlock it again.
 */
void           tnt_async_queue_lock                      (TntAsyncQueue *queue);
void           tnt_async_queue_unlock                    (TntAsyncQueue *queue);

/* Ref and unref the TntAsyncQueue. tnt_async_queue_unref_unlocked makes
 * no sense, as after the unreffing the Queue might be gone and can't
 * be unlocked. So you have a function to call, if you don't hold the
 * lock (tnt_async_queue_unref) and one to call, when you already hold
 * the lock (tnt_async_queue_unref_and_unlock). After that however, you
 * don't hold the lock anymore and the Queue might in fact be
 * destroyed, if you unreffed to zero. */
void           tnt_async_queue_ref                       (TntAsyncQueue *queue);
void           tnt_async_queue_ref_unlocked              (TntAsyncQueue *queue);
void           tnt_async_queue_unref                     (TntAsyncQueue *queue);
void           tnt_async_queue_unref_and_unlock          (TntAsyncQueue *queue);

/* Push return data */
void           tnt_async_queue_push_return               (TntAsyncQueue *queue, gpointer data, gpointer action);
void           tnt_async_queue_push_return_unlocked      (TntAsyncQueue *queue, gpointer data, gpointer action);

/* Push data into the async queue. Must not be NULL. */
void           tnt_async_queue_push_action_head          (TntAsyncQueue *queue, gpointer data);
void           tnt_async_queue_push_action_head_unlocked (TntAsyncQueue *queue, gpointer data);
void           tnt_async_queue_push_action_tail          (TntAsyncQueue *queue, gpointer data);
void           tnt_async_queue_push_action_tail_unlocked (TntAsyncQueue *queue, gpointer data);

/* Pop data from the async queue. When no data is there, the thread is blocked
 * until data arrives. */
gpointer       tnt_async_queue_pop_action                (TntAsyncQueue *queue);
gpointer       tnt_async_queue_pop_action_unlocked       (TntAsyncQueue *queue);

/* Try to pop data. NULL is returned in case of empty queue. */
gpointer       tnt_async_queue_try_pop_action            (TntAsyncQueue *queue);
gpointer       tnt_async_queue_try_pop_action_unlocked   (TntAsyncQueue *queue);

/* Wait for data until at maximum end_time is reached. NULL is returned
 * in case of empty queue. */
gpointer       tnt_async_queue_timed_pop_action          (TntAsyncQueue *queue, GTimeVal *end_time);
gpointer       tnt_async_queue_timed_pop_action_unlocked (TntAsyncQueue *queue, GTimeVal *end_time);

/* Cancel action */
gboolean       tnt_async_queue_cancel_action             (TntAsyncQueue *queue, gpointer data);
gboolean       tnt_async_queue_cancel_action_unlocked    (TntAsyncQueue *queue, gpointer data);
gboolean       tnt_async_queue_cancelling_current_action (TntAsyncQueue *queue);
gboolean       tnt_async_queue_cancelling_current_action_unlocked (TntAsyncQueue *queue);

/* Return the length of the queue. Negative values mean that threads
 * are waiting, positve values mean that there are entries in the
 * queue. Actually this function returns the length of the queue minus
 * the number of waiting threads, tnt_async_queue_length == 0 could also
 * mean 'n' entries in the queue and 'n' thread waiting. Such can
 * happen due to locking of the queue or due to scheduling. */
gint           tnt_async_queue_get_pending_action_length (TntAsyncQueue* queue);
gint           tnt_async_queue_get_pending_action_length_unlocked (TntAsyncQueue* queue);

void           tnt_async_queue_begin_shutdown (TntAsyncQueue *queue);

gpointer       tnt_async_queue_new_action_real (gint size);
void           tnt_async_queue_free_action (gpointer action);
gpointer       tnt_async_queue_new_return_real (gint size);
void           tnt_async_queue_free_return (gpointer result);

#define tnt_async_queue_new_action(action_type) tnt_async_queue_new_action_real (sizeof (action_type))
#define tnt_async_queue_new_return(return_type) tnt_async_queue_new_return_real (sizeof (return_type))

void           tnt_async_queue_ref_action (TntAsyncQueue *queue, gpointer action);
void           tnt_async_queue_ref_action_unlocked (TntAsyncQueue *queue, gpointer action);
void           tnt_async_queue_unref_action (TntAsyncQueue *queue, gpointer action);
void           tnt_async_queue_unref_action_unlocked (TntAsyncQueue *queue, gpointer action);

#endif  /* TNT_ASYNC_QUEUE_H */
