/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-bandwidth-group-list.c - Holds a list of bandwidth groups.
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"

#include <string.h>
#include <glib.h>

#ifdef G_PLATFORM_WIN32
# include "win32/win32-compat.h"
#endif

#include "tnt.h"

#define ITERATION_INTERVAL 1

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), TNT_BANDWIDTH_GROUP_LIST_TYPE)

typedef struct
{
  gpointer func;
  gpointer data;
}
GroupListFunc;

static gint
find_group_by_name (GSList *groups, const gchar *name)
{
  GSList *l;
  gint    n = 0;

  for (l = groups; l; l = g_slist_next (l))
  {
    TntBandwidthGroup *group = l->data;
    const gchar       *group_name;

    group_name = tnt_bandwidth_group_get_name (group);
    if (!strcasecmp (name, group_name))
      break;
    n++;
  }

  if (!l)
    return -1;

  return n;
}

static void
run_callbacks (TntBandwidthGroupList *bwgl, GSList *callbacks, gint group_n)
{
  GSList *l;

  for (l = callbacks; l; l = g_slist_next (l))
  {
    GroupListFunc *group_func;

    group_func = l->data;
    ((TntBandwidthGroupListFunc *) group_func->func)
      (bwgl, group_n, group_func->data);
  }
}

static void
run_action_callbacks (TntBandwidthGroupList *bwgl, GSList *callbacks, TntBandwidthGroup *group,
                      TntTransferDirection direction, TntBandwidthAdjustment adjust)
{
  GSList *l;

  for (l = callbacks; l; l = g_slist_next (l))
  {
    GroupListFunc *group_func;

    group_func = l->data;
    ((TntBandwidthActionFunc *) group_func->func)
      (group, direction, adjust, group_func->data);
  }
}

static void
remove_callback (GSList **callbacks, gpointer func, gpointer data)
{
  GSList *l;
  GSList *next;

  for (l = *callbacks; l; l = next)
  {
    GroupListFunc *group_func;

    next = g_slist_next (l);

    group_func = l->data;
    if (group_func->func == func && group_func->data == data)
    {
      g_free (group_func);
      *callbacks = g_slist_delete_link (*callbacks, l);
    }
  }
}

void
tnt_bandwidth_group_list_add_group_add_func (TntBandwidthGroupList *bwgl,
                                             TntBandwidthGroupListFunc *func, gpointer data)
{
  GroupListFunc *group_func;

  assert_types (bwgl);
  g_return_if_fail (func != NULL);

  group_func = g_new (GroupListFunc, 1);
  group_func->func = func;
  group_func->data = data;

  bwgl->group_add_funcs = g_slist_append (bwgl->group_add_funcs, group_func);
}

void
tnt_bandwidth_group_list_add_group_update_func (TntBandwidthGroupList *bwgl,
                                                TntBandwidthGroupListFunc *func, gpointer data)
{
  GroupListFunc *group_func;

  assert_types (bwgl);
  g_return_if_fail (func != NULL);

  group_func = g_new (GroupListFunc, 1);
  group_func->func = func;
  group_func->data = data;

  bwgl->group_update_funcs = g_slist_append (bwgl->group_update_funcs,
                                                           group_func);
}

void
tnt_bandwidth_group_list_add_group_remove_func (TntBandwidthGroupList *bwgl,
                                                TntBandwidthGroupListFunc *func, gpointer data)
{
  GroupListFunc *group_func;

  assert_types (bwgl);
  g_return_if_fail (func != NULL);

  group_func = g_new (GroupListFunc, 1);
  group_func->func = func;
  group_func->data = data;

  bwgl->group_remove_funcs = g_slist_append (bwgl->group_remove_funcs,
                                                           group_func);
}

void
tnt_bandwidth_group_list_add_group_action_func (TntBandwidthGroupList *bwgl,
                                                TntBandwidthActionFunc *func, gpointer data)
{
  GroupListFunc *group_func;

  assert_types (bwgl);
  g_return_if_fail (func != NULL);

  group_func = g_new (GroupListFunc, 1);
  group_func->func = func;
  group_func->data = data;

  bwgl->group_action_funcs = g_slist_append (bwgl->group_action_funcs,
                                             group_func);
}

void
tnt_bandwidth_group_list_remove_group_func (TntBandwidthGroupList *bwgl,
                                            gpointer func, gpointer data)
{
  assert_types (bwgl);
  g_return_if_fail (func != NULL);

  remove_callback (&bwgl->group_add_funcs,       func, data);
  remove_callback (&bwgl->group_update_funcs,    func, data);
  remove_callback (&bwgl->group_remove_funcs,    func, data);
  remove_callback (&bwgl->group_action_funcs,    func, data);
}

gint
tnt_bandwidth_group_list_get_n_groups (TntBandwidthGroupList *bwgl)
{
  assert_types (bwgl);

  return g_slist_length (bwgl->groups);
}

TntBandwidthGroup *
tnt_bandwidth_group_list_get_group_n (TntBandwidthGroupList *bwgl, gint group_n)
{
  assert_types (bwgl);

  return g_slist_nth_data (bwgl->groups, group_n);
}

static void
group_update_cb (TntBandwidthGroup *group, gpointer data)
{
  TntBandwidthGroupList *bwgl = data;
  GSList                *node;
  gint                   pos;

  assert_types (bwgl);
  flux_object_check (FLUX_OBJECT (group), TNT_BANDWIDTH_GROUP_TYPE);

  node = g_slist_find (bwgl->groups, group);
  if (!node)
  {
    g_warning ("Got update callback from group we don't care about.");
    return;
  }

  pos = g_slist_position (bwgl->groups, node);
  tnt_bandwidth_group_list_updated (bwgl, pos);
}

static void
group_action_cb (TntBandwidthGroup *group, TntTransferDirection direction, TntBandwidthAdjustment adjust,
                 gpointer data)
{
  TntBandwidthGroupList *bwgl = data;
  GSList                *node;
  gint                   pos;

  assert_types (bwgl);
  flux_object_check (FLUX_OBJECT (group), TNT_BANDWIDTH_GROUP_TYPE);

  node = g_slist_find (bwgl->groups, group);
  if (!node)
  {
    g_warning ("Got action callback from group we don't care about.");
    return;
  }

  pos = g_slist_position (bwgl->groups, node);
  run_action_callbacks (bwgl, bwgl->group_action_funcs, group, direction, adjust);
}

gint
tnt_bandwidth_group_list_add_group (TntBandwidthGroupList *bwgl, TntBandwidthGroup *group)
{
  gint pos;

  assert_types (bwgl);
  flux_object_check (FLUX_OBJECT (group), TNT_BANDWIDTH_GROUP_TYPE);

  pos = find_group_by_name (bwgl->groups, tnt_bandwidth_group_get_name (group));
  if (pos != -1)
  {
    g_warning ("Tried to add a group we already have.");
    return pos;
  }

  bwgl->groups = g_slist_append (bwgl->groups, group);
  pos = g_slist_length (bwgl->groups) - 1;
  flux_object_ref (FLUX_OBJECT (group));
  tnt_bandwidth_group_set_update_func (group, group_update_cb, bwgl);
  tnt_bandwidth_group_set_action_func (group, group_action_cb, bwgl);

  run_callbacks (bwgl, bwgl->group_add_funcs, pos);
  return pos;
}

void
tnt_bandwidth_group_list_remove_group (TntBandwidthGroupList *bwgl, TntBandwidthGroup *group)
{
  GSList *node;
  gint    pos;

  assert_types (bwgl);
  flux_object_check (FLUX_OBJECT (group), TNT_BANDWIDTH_GROUP_TYPE);

  node = g_slist_find (bwgl->groups, group);
  if (!node)
  {
    g_warning ("Tried to remove a group we don't have.");
    return;
  }

  pos = g_slist_position (bwgl->groups, node);
  run_callbacks (bwgl, bwgl->group_remove_funcs, pos);

  tnt_bandwidth_group_set_update_func (group, NULL, NULL);
  flux_object_unref (FLUX_OBJECT (group));
  bwgl->groups = g_slist_delete_link (bwgl->groups, node);
  return;
}

TntBandwidthGroup *
tnt_bandwidth_group_list_find_group_by_edge (TntBandwidthGroupList *bwgl, TntEdge *edge)
{
  GSList *l;

  assert_types (bwgl);
  flux_object_check (FLUX_OBJECT (edge), TNT_EDGE_TYPE);

  for (l = bwgl->groups; l; l = g_slist_next (l))
  {
    TntBandwidthGroup *group = l->data;

    if (tnt_bandwidth_group_have_edge (group, edge))
      return group;
  }

  return NULL;
}

gboolean
tnt_bandwidth_group_list_add_edge (TntBandwidthGroupList *bwgl, TntEdge *edge)
{
  TntNodeIP4         best_netmask = 0;
  TntBandwidthGroup *best_group   = NULL;
  GSList            *l;

  assert_types (bwgl);
  flux_object_check (FLUX_OBJECT (edge), TNT_EDGE_TYPE);

  for (l = bwgl->groups; l; l = g_slist_next (l))
  {
    TntBandwidthGroup *group = l->data;

    if (tnt_bandwidth_group_have_edge (group, edge))
      return TRUE;

    if (tnt_bandwidth_group_match_edge (group, edge))
    {
      if (!best_group)
      {
        best_group = group;
      }
      else
      {
        gchar      *netmask_str;
        TntNodeIP4  netmask;

        tnt_bandwidth_group_get_accept (group, NULL, &netmask_str);
        g_assert (netmask_str);
        netmask = tnt_string_to_ip (netmask_str);
        g_free (netmask_str);

        if (netmask > best_netmask)
        {
          best_netmask = netmask;
          best_group   = group;
        }
      }
    }
  }

  if (best_group && tnt_bandwidth_group_add_edge (best_group, edge))
    return TRUE;

  return FALSE;
}

gboolean
tnt_bandwidth_group_list_remove_edge (TntBandwidthGroupList *bwgl, TntEdge *edge)
{
  GSList *l;

  assert_types (bwgl);
  flux_object_check (FLUX_OBJECT (edge), TNT_EDGE_TYPE);

  for (l = bwgl->groups; l; l = g_slist_next (l))
  {
    TntBandwidthGroup *group = l->data;

    flux_object_check (FLUX_OBJECT (group), TNT_BANDWIDTH_GROUP_TYPE);

    if (tnt_bandwidth_group_remove_edge (group, edge))
      return TRUE;
  }

  return FALSE;
}

TntBandwidthGroup *
tnt_bandwidth_group_list_find_group_by_name (TntBandwidthGroupList *bwgl, const gchar *name)
{
  gint n;

  assert_types (bwgl);

  n = find_group_by_name (bwgl->groups, name);
  if (n == -1)
    return NULL;

  return tnt_bandwidth_group_list_get_group_n (bwgl, n);
}

TntBandwidthGroup *
tnt_bandwidth_group_list_find_group_by_ip (TntBandwidthGroupList *bwgl, TntNodeIP4 ip)
{
  GSList *l;

  assert_types (bwgl);

  for (l = bwgl->groups; l; l = g_slist_next (l))
  {
    TntBandwidthGroup *group = l->data;

    if (tnt_bandwidth_group_match_ip (group, ip))
      return group;
  }

  return NULL;
}

void
tnt_bandwidth_group_list_updated (TntBandwidthGroupList *bwgl, gint group_n)
{
  assert_types (bwgl);

  run_callbacks (bwgl, bwgl->group_update_funcs, group_n);
}

static gboolean
iterate_cb (gpointer data)
{
  TntBandwidthGroupList *bwgl = data;
  GSList                *l;

  assert_types (bwgl);

  for (l = bwgl->groups; l; l = g_slist_next (l))
  {
    TntBandwidthGroup *group = l->data;

    tnt_bandwidth_group_iterate (group);
  }
  
  return TRUE;
}

void
tnt_bandwidth_group_list_finalize (FluxObject *object)
{
  TntBandwidthGroupList *bwgl = TNT_BANDWIDTH_GROUP_LIST (object);
  GSList                *l;

  assert_types (bwgl);

  for (l = bwgl->group_add_funcs; l; l = g_slist_next (l))
    g_free (l->data);
  
  for (l = bwgl->group_update_funcs; l; l = g_slist_next (l))
    g_free (l->data);
  
  for (l = bwgl->group_remove_funcs; l; l = g_slist_next (l))
    g_free (l->data);
  
  for (l = bwgl->group_action_funcs; l; l = g_slist_next (l))
    g_free (l->data);

  for (l = bwgl->groups; l; l = g_slist_next (l))
  {
    TntBandwidthGroup *bandwidth_group = l->data;

    flux_object_unref (FLUX_OBJECT (bandwidth_group));
  }

  g_slist_free (bwgl->group_add_funcs);
  g_slist_free (bwgl->group_update_funcs);
  g_slist_free (bwgl->group_remove_funcs);
  g_slist_free (bwgl->group_action_funcs);
  g_slist_free (bwgl->groups);

  g_source_remove (bwgl->iterate_id);
}

void
tnt_bandwidth_group_list_init (TntBandwidthGroupList *bwgl)
{
  flux_object_init              (FLUX_OBJECT (bwgl));
  flux_object_set_type          (FLUX_OBJECT (bwgl), TNT_BANDWIDTH_GROUP_LIST_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (bwgl), tnt_bandwidth_group_list_finalize);

  bwgl->iterate_id = g_timeout_add (1000 * ITERATION_INTERVAL, iterate_cb, bwgl);
}

TntBandwidthGroupList *
tnt_bandwidth_group_list_new (void)
{
  TntBandwidthGroupList *bwgl;

  bwgl = g_new0 (TntBandwidthGroupList, 1);
  tnt_bandwidth_group_list_init (bwgl);

  return bwgl;
}
