/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-block-action.c - Describes operation to be done on a data block.
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"

#include <string.h>
#include <glib.h>

#ifdef G_PLATFORM_WIN32
# include "win32/win32-compat.h"
#endif

#include "tnt.h"

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), TNT_BLOCK_ACTION_TYPE)

typedef struct
{
  TntDataDB      *data_db;
  FluxFileIO     *file_io;

  FluxRmd160     *rmd_160;
  guint           done_len;
}
RunContext;

static void
ensure_block_memory (TntBlockAction *block_action)
{
  gpointer block;

  if (!block_action->block_len)
  {
    g_error ("Can't ensure block memory of 0 bytes.\n");
    return;
  }

  block = tnt_alloc_block (block_action->block_len);
  tnt_block_action_set_block (block_action, block);
}

static void
free_run_context (TntBlockAction *block_action)
{
  RunContext *run_context = block_action->run_context;

  if (!run_context)
    return;

  switch (block_action->type)
  {
    case TNT_BLOCK_ACTION_GET_BLOCK_HASH:
      flux_io_set_read_func     (FLUX_IO (run_context->file_io), NULL, NULL);
      flux_file_io_set_eof_func (run_context->file_io, NULL, NULL);
      flux_object_unref (FLUX_OBJECT (run_context->rmd_160));
      break;

    case TNT_BLOCK_ACTION_GET_BLOCK:
      flux_io_set_read_func     (FLUX_IO (run_context->file_io), NULL, NULL);
      flux_file_io_set_eof_func (run_context->file_io, NULL, NULL);
      break;

    case TNT_BLOCK_ACTION_PUT_BLOCK:
      flux_io_set_write_func    (FLUX_IO (run_context->file_io), NULL, NULL);
      flux_file_io_set_eow_func (run_context->file_io, NULL, NULL);
      break;

    default:
      g_assert_not_reached ();
  }

  flux_object_unref (FLUX_OBJECT (run_context->file_io));

  g_free (run_context);
  block_action->run_context = NULL;
}

static void
hash_done (TntBlockAction *block_action)
{
  RunContext *run_context = block_action->run_context;
  guchar      hash [20];
  guint8      block_hash [TNT_BLOCK_HASH_LEN];

  flux_object_ref (FLUX_OBJECT (block_action));
  flux_rmd160_end (run_context->rmd_160);

  if (run_context->done_len != block_action->block_len)
  {
    tnt_log_debug ("Could not hash requested block length! Wanted %d, got %d.",
                   block_action->block_len,
                   run_context->done_len);

    if (run_context->data_db->action_done_func)
      run_context->data_db->action_done_func (run_context->data_db, FALSE);

    ((TntGotBlockHashFunc *) block_action->post_func)(run_context->data_db,
                                                      &block_action->file_id,
                                                      block_action->block_id,
                                                      run_context->done_len,
                                                      NULL,
                                                      block_action->post_func_data);
  }
  else
  {
    flux_rmd160_get_hash (run_context->rmd_160, hash);
    tnt_fold_hash (hash, block_hash, TNT_BLOCK_HASH_LEN);

    if (run_context->data_db->action_done_func)
      run_context->data_db->action_done_func (run_context->data_db, TRUE);

    ((TntGotBlockHashFunc *) block_action->post_func)(run_context->data_db,
                                                      &block_action->file_id,
                                                      block_action->block_id,
                                                      block_action->block_len,
                                                      block_hash,
                                                      block_action->post_func_data);
  }

  flux_object_unref (FLUX_OBJECT (block_action));
}

static void
hash_read_cb (FluxIO *io, guint min_len, gpointer data)
{
  TntBlockAction *block_action = data;
  RunContext     *run_context  = block_action->run_context;
  gchar          *buf;

  buf = g_malloc (min_len);
  flux_dual_fifo_read (FLUX_DUAL_FIFO (io), buf, min_len);
  flux_rmd160_add (run_context->rmd_160, buf, min_len);
  g_free (buf);

  run_context->done_len += min_len;

  if (run_context->done_len == block_action->block_len)
  {
    hash_done (block_action);
  }
  else if (block_action->block_len - run_context->done_len < 1024)
  {
    /* We must do this since the block can be in the middle of a file, and we don't want to
     * read outside the block. */
    flux_io_set_read_threshold (io, block_action->block_len - run_context->done_len);
  }
}

static void
hash_eof_cb (FluxFileIO *file_io, guint len_left, gpointer data)
{
  TntBlockAction *block_action = data;
  RunContext     *run_context  = block_action->run_context;
  gchar          *buf;

  g_return_if_fail (run_context != NULL);
  assert_types (block_action);
  flux_object_check (FLUX_OBJECT (run_context->data_db), TNT_DATA_DB_TYPE);
  flux_object_check (FLUX_OBJECT (run_context->file_io), FLUX_FILE_IO_TYPE);
  g_assert (file_io == run_context->file_io);

  buf = g_malloc (len_left);
  flux_dual_fifo_read (FLUX_DUAL_FIFO (file_io), buf, len_left);
  flux_rmd160_add (run_context->rmd_160, buf, len_left);
  g_free (buf);

  run_context->done_len += len_left;
  hash_done (block_action);
}

static void
read_done (TntBlockAction *block_action)
{
  RunContext *run_context = block_action->run_context;

  flux_object_ref (FLUX_OBJECT (block_action));

#if 0
#ifdef G_PLATFORM_WIN32
  flux_file_io_close (run_context->file_io);
#endif
#endif

  if (run_context->done_len != block_action->block_len)
  {
    tnt_log_debug ("Could not read requested block length! Wanted %d, got %d.",
                   block_action->block_len,
                   run_context->done_len);

    if (run_context->data_db->action_done_func)
      run_context->data_db->action_done_func (run_context->data_db, FALSE);

    ((TntGotBlockFunc *) block_action->post_func)(run_context->data_db,
                                                  &block_action->file_id,
                                                  block_action->block_id,
                                                  run_context->done_len,
                                                  NULL,
                                                  block_action->post_func_data);
  }
  else
  {
    if (run_context->data_db->action_done_func)
      run_context->data_db->action_done_func (run_context->data_db, TRUE);

    /* Hands over ownership of the block */
    ((TntGotBlockFunc *) block_action->post_func)(run_context->data_db,
                                                  &block_action->file_id,
                                                  block_action->block_id,
                                                  block_action->block_len,
                                                  block_action->block,
                                                  block_action->post_func_data);

    block_action->block = NULL;  /* Callee takes ownership */
  }

  flux_object_unref (FLUX_OBJECT (block_action));
}

static void
data_read_cb (FluxIO *io, guint min_len, gpointer data)
{
  TntBlockAction *block_action = data;
  RunContext     *run_context  = block_action->run_context;

  g_return_if_fail (run_context != NULL);
  assert_types (block_action);
  flux_object_check (FLUX_OBJECT (run_context->data_db), TNT_DATA_DB_TYPE);
  flux_object_check (FLUX_OBJECT (run_context->file_io), FLUX_FILE_IO_TYPE);
  g_assert ((gpointer) io == (gpointer) run_context->file_io);

  flux_dual_fifo_read (FLUX_DUAL_FIFO (io),
                       (guchar *) block_action->block + run_context->done_len,
                       min_len);

  run_context->done_len += min_len;

  if (run_context->done_len == block_action->block_len)
  {
    read_done (block_action);
  }
  else if (block_action->block_len - run_context->done_len < 1024)
  {
    /* We must do this since the block can be in the middle of a file, and we don't want to
     * read outside the block. */
    flux_io_set_read_threshold (io, block_action->block_len - run_context->done_len);
  }
}

static void
data_eof_cb (FluxFileIO *file_io, guint len_left, gpointer data)
{
  TntBlockAction *block_action = data;
  RunContext     *run_context  = block_action->run_context;

  g_return_if_fail (run_context != NULL);
  assert_types (block_action);
  flux_object_check (FLUX_OBJECT (run_context->data_db), TNT_DATA_DB_TYPE);
  flux_object_check (FLUX_OBJECT (run_context->file_io), FLUX_FILE_IO_TYPE);
  g_assert (file_io == run_context->file_io);

  flux_dual_fifo_read (FLUX_DUAL_FIFO (file_io),
                       (guchar *) block_action->block + run_context->done_len,
                       len_left);

  run_context->done_len += len_left;
  read_done (block_action);
}

static void
write_done (TntBlockAction *block_action)
{
  RunContext *run_context = block_action->run_context;

  flux_object_ref (FLUX_OBJECT (block_action));

#if 0
#ifdef G_PLATFORM_WIN32
  flux_file_io_close (run_context->file_io);
#endif
#endif

  if (run_context->done_len != block_action->block_len)
  {
    tnt_log_debug ("Could not write requested block length! Wanted %d, got %d.",
                   block_action->block_len,
                   run_context->done_len);

    if (run_context->data_db->action_done_func)
      run_context->data_db->action_done_func (run_context->data_db, FALSE);

    if (block_action->post_func)
    {
      ((TntWroteBlockFunc *) block_action->post_func)(run_context->data_db,
                                                      &block_action->file_id,
                                                      block_action->block_id,
                                                      0,
                                                      block_action->post_func_data);
    }
  }
  else
  {
    if (run_context->data_db->action_done_func)
      run_context->data_db->action_done_func (run_context->data_db, TRUE);

    if (block_action->post_func)
    {
      ((TntWroteBlockFunc *) block_action->post_func)(run_context->data_db,
                                                      &block_action->file_id,
                                                      block_action->block_id,
                                                      block_action->block_len,
                                                      block_action->post_func_data);
    }
  }

  flux_object_unref (FLUX_OBJECT (block_action));
}

static void
data_eow_cb (FluxFileIO *file_io, gpointer data)
{
  TntBlockAction *block_action = data;
  RunContext     *run_context  = block_action->run_context;

  g_return_if_fail (run_context != NULL);
  assert_types (block_action);
  flux_object_check (FLUX_OBJECT (run_context->data_db), TNT_DATA_DB_TYPE);
  flux_object_check (FLUX_OBJECT (run_context->file_io), FLUX_FILE_IO_TYPE);
  g_assert (file_io == run_context->file_io);

  write_done (block_action);
}

static void
data_write_cb (FluxIO *io, guint min_len, gpointer data)
{
  TntBlockAction *block_action = data;
  RunContext     *run_context  = block_action->run_context;

  g_return_if_fail (run_context != NULL);
  assert_types (block_action);
  flux_object_check (FLUX_OBJECT (run_context->data_db), TNT_DATA_DB_TYPE);
  flux_object_check (FLUX_OBJECT (run_context->file_io), FLUX_FILE_IO_TYPE);
  g_assert ((gpointer) io == (gpointer) run_context->file_io);

  flux_dual_fifo_write (FLUX_DUAL_FIFO (io),
                        (guchar *) block_action->block + run_context->done_len,
                        min_len);

  run_context->done_len += min_len;

  if (run_context->done_len == block_action->block_len)
  {
    flux_io_set_write_func (io, NULL, NULL);
    flux_file_io_set_eow_func (FLUX_FILE_IO (io), data_eow_cb, block_action);
  }
  else if (block_action->block_len - run_context->done_len < 1024)
  {
    /* We must do this since the block can be in the middle of a file, and we don't want to
     * write outside the block. */
    flux_io_set_write_threshold (io, block_action->block_len - run_context->done_len);
  }
}

/* The FluxFileIO must be positioned at the start of the block to operate on.
 * You should unref the FluxFileIO after calling this. */
void
tnt_block_action_run (TntBlockAction *block_action, TntDataDB *data_db, FluxFileIO *file_io)
{
  RunContext *run_context;

  assert_types (block_action);
  flux_object_check (FLUX_OBJECT (data_db), TNT_DATA_DB_TYPE);
  flux_object_check (FLUX_OBJECT (file_io), FLUX_FILE_IO_TYPE);
  g_return_if_fail (block_action->run_context == NULL);

  run_context = g_new0 (RunContext, 1);
  block_action = block_action;
  run_context->data_db      = data_db;
  run_context->file_io      = file_io;

  block_action->run_context = run_context;

  switch (block_action->type)
  {
    case TNT_BLOCK_ACTION_GET_BLOCK_HASH:
      run_context->rmd_160 = flux_rmd160_new ();
      flux_rmd160_begin (run_context->rmd_160);
      flux_io_set_read_threshold (FLUX_IO (file_io), MIN (block_action->block_len, 1024));
      flux_io_set_read_func      (FLUX_IO (file_io), hash_read_cb, block_action);
      flux_file_io_set_eof_func  (file_io, hash_eof_cb, block_action);
      break;

    case TNT_BLOCK_ACTION_GET_BLOCK:
      ensure_block_memory (block_action);
      flux_io_set_read_threshold (FLUX_IO (file_io), MIN (block_action->block_len, 1024));
      flux_io_set_read_func      (FLUX_IO (file_io), data_read_cb, block_action);
      flux_file_io_set_eof_func  (file_io, data_eof_cb, block_action);
      break;

    case TNT_BLOCK_ACTION_PUT_BLOCK:
      flux_io_set_write_threshold (FLUX_IO (file_io), MIN (block_action->block_len, 1024));
      flux_io_set_write_func      (FLUX_IO (file_io), data_write_cb, block_action);
      break;

    default:
      g_assert_not_reached ();
      break;
  }

  flux_object_ref (FLUX_OBJECT (file_io));
}

void
tnt_block_action_cancel (TntBlockAction *block_action, TntDataDB *data_db)
{
  assert_types (block_action);

  flux_object_ref (FLUX_OBJECT (block_action));

  if (data_db && data_db->action_done_func)
    data_db->action_done_func (data_db, FALSE);

  /* Signal failure */
  if (block_action->post_func)
  {
    switch (block_action->type)
    {
      case TNT_BLOCK_ACTION_GET_BLOCK_HASH:
        ((TntGotBlockHashFunc *) block_action->post_func)(data_db,
                                                          &block_action->file_id,
                                                          block_action->block_id,
                                                          0,
                                                          NULL,
                                                          block_action->post_func_data);
        break;

      case TNT_BLOCK_ACTION_GET_BLOCK:
        ((TntGotBlockFunc *) block_action->post_func)(data_db,
                                                      &block_action->file_id,
                                                      block_action->block_id,
                                                      0,
                                                      NULL,
                                                      block_action->post_func_data);
        break;

      case TNT_BLOCK_ACTION_PUT_BLOCK:
        ((TntWroteBlockFunc *) block_action->post_func)(data_db,
                                                        &block_action->file_id,
                                                        block_action->block_id,
                                                        0,
                                                        block_action->post_func_data);
        break;

      default:
        g_assert_not_reached ();
        break;
    }
  }

  flux_object_unref (FLUX_OBJECT (block_action));
}

TntBlockActionType
tnt_block_action_get_type (TntBlockAction *block_action)
{
  assert_types (block_action);

  return block_action->type;
}

void
tnt_block_action_get_file_info (TntBlockAction *block_action, TntFileID *file_id)
{
  assert_types (block_action);

  memcpy (file_id, &block_action->file_id, sizeof (TntFileID));
}

void
tnt_block_action_set_file_info (TntBlockAction *block_action, const TntFileID *file_id)
{
  assert_types (block_action);

  memcpy (&block_action->file_id, file_id, sizeof (TntFileID));
}

gint
tnt_block_action_get_block_id (TntBlockAction *block_action)
{
  assert_types (block_action);

  return block_action->block_id;
}

void
tnt_block_action_set_block_id (TntBlockAction *block_action, gint block_id)
{
  assert_types (block_action);
  g_return_if_fail (block_id >= 0);

  block_action->block_id = block_id;
}

guint
tnt_block_action_get_block_len (TntBlockAction *block_action)
{
  assert_types (block_action);

  return block_action->block_len;
}

void
tnt_block_action_set_block_len (TntBlockAction *block_action, guint block_len)
{
  assert_types (block_action);

  block_action->block_len = block_len;
}

gpointer
tnt_block_action_get_block (TntBlockAction *block_action)
{
  assert_types (block_action);

  return block_action->block;
}

void
tnt_block_action_set_block (TntBlockAction *block_action, gpointer block)
{
  assert_types (block_action);

  if (block_action->block)
    tnt_free_block (block_action->block);

  block_action->block = block;
}

void
tnt_block_action_get_post_func (TntBlockAction *block_action, gpointer *func, gpointer *data)
{
  assert_types (block_action);

  *func = block_action->post_func;
  *data = block_action->post_func_data;
}

void
tnt_block_action_set_post_func (TntBlockAction *block_action, gpointer func, gpointer data)
{
  assert_types (block_action);

  block_action->post_func      = func;
  block_action->post_func_data = data;
}

void
tnt_block_action_finalize (FluxObject *object)
{
  TntBlockAction *block_action = TNT_BLOCK_ACTION (object);

  if (block_action->block)
  {
    tnt_free_block (block_action->block);
    block_action->block = NULL;
  }

  free_run_context (block_action);
}

void
tnt_block_action_init (TntBlockAction *block_action, TntBlockActionType type)
{
  flux_object_init              (FLUX_OBJECT (block_action));
  flux_object_set_type          (FLUX_OBJECT (block_action), TNT_BLOCK_ACTION_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (block_action), tnt_block_action_finalize);

  block_action->type = type;
}

TntBlockAction *
tnt_block_action_new (TntBlockActionType type)
{
  TntBlockAction *block_action;

  block_action = g_new0 (TntBlockAction, 1);
  tnt_block_action_init (block_action, type);

  return block_action;
}
