/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-config.c - Configuration storage.
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include <string.h>
#include <glib.h>

#ifdef G_PLATFORM_WIN32
# include "win32/win32-compat.h"
#endif

#include "tnt.h"

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), TNT_CONFIG_TYPE)

static gboolean
maybe_migrate (FluxTree *tree)
{
  FluxTree *t0 = NULL, *t1 = NULL, *t2 = NULL, *t3 = NULL;

  if (!(t0 = flux_tree_find_first_child (tree, FLUX_STRING_WITH_LEN ("storage"))))
    return FALSE;

  /* Determine common data location */

  t1 = flux_tree_find_first_child (t0, FLUX_STRING_WITH_LEN ("metadata-cache"));

  if (t1)
    t2 = flux_tree_find_first_child (t1, FLUX_STRING_WITH_LEN ("location"));
  if (t2)
    t2 = flux_tree_get_first_child (t2);

  if (t2)
  {
    gchar *path;
    gchar *last_slash;

    path = flux_tree_data_get_str (t2);
    last_slash = strrchr (path, G_DIR_SEPARATOR);

    /* If path ends in slash, get the slash before that */
    if (last_slash && *(last_slash + 1) == '\0' && last_slash > path)
      last_slash = strrchr (last_slash - 1, G_DIR_SEPARATOR);

    /* Remove database-specific path element */
    if (last_slash)
    {
      if (!strncasecmp (last_slash + 1, "meta", 4))
        *last_slash = '\0';
    }

    /* Set common data-dir */
    t1 = flux_tree_find_first_child (tree, FLUX_STRING_WITH_LEN ("common"));
    if (!t1)
      t1 = flux_tree_new_with_parent_and_data (tree, FLUX_STRING_WITH_LEN ("common"));

    t2 = flux_tree_find_first_child (t1, FLUX_STRING_WITH_LEN ("data-dir"));
    if (!t2)
      t2 = flux_tree_new_with_parent_and_data (t1, FLUX_STRING_WITH_LEN ("data-dir"));

    flux_tree_new_with_parent_and_data (t2, path, strlen (path));
    g_free (path);
  }

  /* Transfer meta size */

  t1 = flux_tree_find_first_child (t0, FLUX_STRING_WITH_LEN ("metadata-cache"));

  if (t1)
    t2 = flux_tree_find_first_child (t1, FLUX_STRING_WITH_LEN ("size-limit"));
  else
    t2 = NULL;

  if (t2)
    t2 = flux_tree_get_first_child (t2);

  if (t2)
  {
    t3 = flux_tree_dup (t2);

    t1 = flux_tree_find_first_child (tree, FLUX_STRING_WITH_LEN ("filing"));
    if (!t1)
      t1 = flux_tree_new_with_parent_and_data (tree, FLUX_STRING_WITH_LEN ("filing"));

    t2 = flux_tree_find_first_child (t1, FLUX_STRING_WITH_LEN ("meta-cache-size"));
    if (t2)
      flux_tree_destroy (t2);

    t2 = flux_tree_new_with_parent_and_data (t1, FLUX_STRING_WITH_LEN ("meta-cache-size"));

    flux_tree_add (t2, t3);
  }

  /* Transfer data size */

  t1 = flux_tree_find_first_child (t0, FLUX_STRING_WITH_LEN ("data-cache"));

  if (t1)
    t2 = flux_tree_find_first_child (t1, FLUX_STRING_WITH_LEN ("size-limit"));
  else
    t2 = NULL;

  if (t2)
    t2 = flux_tree_get_first_child (t2);

  if (t2)
  {
    t3 = flux_tree_dup (t2);

    t1 = flux_tree_find_first_child (tree, FLUX_STRING_WITH_LEN ("filing"));
    if (!t1)
      t1 = flux_tree_new_with_parent_and_data (tree, FLUX_STRING_WITH_LEN ("filing"));

    t2 = flux_tree_find_first_child (t1, FLUX_STRING_WITH_LEN ("data-cache-size"));
    if (t2)
      flux_tree_destroy (t2);

    t2 = flux_tree_new_with_parent_and_data (t1, FLUX_STRING_WITH_LEN ("data-cache-size"));

    flux_tree_add (t2, t3);
  }

  /* Transfer incoming dir */

  t1 = flux_tree_find_first_child (t0, FLUX_STRING_WITH_LEN ("incoming"));

  if (t1)
    t2 = flux_tree_find_first_child (t1, FLUX_STRING_WITH_LEN ("location"));
  else
    t2 = NULL;

  if (t2)
    t2 = flux_tree_get_first_child (t2);

  if (t2)
  {
    t3 = flux_tree_dup (t2);

    t1 = flux_tree_find_first_child (tree, FLUX_STRING_WITH_LEN ("filing"));
    if (!t1)
      t1 = flux_tree_new_with_parent_and_data (tree, FLUX_STRING_WITH_LEN ("filing"));

    t2 = flux_tree_find_first_child (t1, FLUX_STRING_WITH_LEN ("incoming-dir"));
    if (t2)
      flux_tree_destroy (t2);

    t2 = flux_tree_new_with_parent_and_data (t1, FLUX_STRING_WITH_LEN ("incoming-dir"));

    flux_tree_add (t2, t3);
  }

  /* Transfer shares */

  t0 = flux_tree_find_first_child (tree, FLUX_STRING_WITH_LEN ("shares"));
  if (t0)
    t3 = flux_tree_dup_all (t0);
  else
    t3 = NULL;

  if (t3)
  {
    t1 = flux_tree_find_first_child (tree, FLUX_STRING_WITH_LEN ("filing"));
    if (!t1)
      t1 = flux_tree_new_with_parent_and_data (tree, FLUX_STRING_WITH_LEN ("filing"));

    t2 = flux_tree_find_first_child (t1, FLUX_STRING_WITH_LEN ("shares"));
    if (t2)
      flux_tree_destroy (t2);

    flux_tree_add (t1, t3);
  }

  /* Destroy cruft */

  t0 = flux_tree_find_first_child (tree, FLUX_STRING_WITH_LEN ("shares"));
  if (t0)
    flux_tree_destroy (t0);

  t0 = flux_tree_find_first_child (tree, FLUX_STRING_WITH_LEN ("storage"));
  if (t0)
    flux_tree_destroy (t0);

  t0 = flux_tree_find_first_child (tree, FLUX_STRING_WITH_LEN ("network"));
  if (t0)
    t0 = flux_tree_find_first_child (t0, FLUX_STRING_WITH_LEN ("id"));
  if (t0)
    flux_tree_destroy (t0);

  return TRUE;
}

static gboolean
save_to_file (TntConfig *config)
{
  FILE     *fhd;
  gchar    *path;
  gboolean  result;

  if (!config->path)
    return FALSE;

  path = g_strconcat (config->path, ".tmp", NULL);
  fhd = fopen (path, "wt");

  if (!fhd)
  {
    g_free (path);
    return FALSE;
  }

  result = flux_tree_print_to_file (config->committed_tree, fhd, FLUX_PRINT_ALLMAN);
  fclose (fhd);

  if (result)
  {
    if (tnt_rename (path, config->path) != 0)
      result = FALSE;
  }

  g_free (path);
  return result;
}

static gboolean
load_from_file (TntConfig *config)
{
  FILE     *fhd;
  FluxTree *tree;

  if (!config->path)
    return FALSE;

  fhd = fopen (config->path, "rt");
  if (!fhd)
    return FALSE;

  tree = flux_tree_scan_from_file (fhd);
  if (!tree)
  {
    fclose (fhd);
    return FALSE;
  }

  maybe_migrate (tree);

  flux_tree_destroy (config->working_tree);
  config->working_tree = tree;

  fclose (fhd);

  tnt_config_commit (config);
  return TRUE;
}

void
tnt_config_finalize (FluxObject *object)
{
  TntConfig *config = TNT_CONFIG (object);

  assert_types (object);

  g_free (config->path);
  flux_tree_destroy (config->committed_tree);
  flux_tree_destroy (config->working_tree);
  flux_object_unref (FLUX_OBJECT (config->listeners));
  flux_object_unref (FLUX_OBJECT (config->validators));
}

void
tnt_config_init (TntConfig *config)
{
  flux_object_init              (FLUX_OBJECT (config));
  flux_object_set_type          (FLUX_OBJECT (config), TNT_CONFIG_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (config), tnt_config_finalize);

  config->committed_tree = flux_tree_new ();
  config->working_tree   = flux_tree_new ();
  config->peek_tree      = config->committed_tree;
  config->listeners      = tnt_callback_list_new ();
  config->validators     = tnt_callback_list_new ();
}

TntConfig *
tnt_config_new (void)
{
  TntConfig *config;

  config = g_new0 (TntConfig, 1);
  tnt_config_init (config);

  return config;
}

void
tnt_config_attach_listener (TntConfig *config, TntConfigListenerFunc *func, gpointer data)
{
  assert_types (config);

  tnt_callback_list_add_callback (config->listeners, func, data);
}

void
tnt_config_detach_listener (TntConfig *config, TntConfigListenerFunc *func, gpointer data)
{
  assert_types (config);

  tnt_callback_list_remove_callback (config->listeners, func, data);
}

void
tnt_config_attach_validator (TntConfig *config, TntConfigListenerFunc *func, gpointer data)
{
  assert_types (config);

  tnt_callback_list_add_callback (config->validators, func, data);
}

void
tnt_config_detach_validator (TntConfig *config, TntConfigListenerFunc *func, gpointer data)
{
  assert_types (config);

  tnt_callback_list_remove_callback (config->validators, func, data);
}

GList *
tnt_config_validate (TntConfig *config)
{
  GList *error_list;

  assert_types (config);

  config->validating = TRUE;
  config->peek_tree = config->working_tree;
  tnt_callback_list_run_while_1 (config->validators, tnt_test_always_true, config);
  config->peek_tree = config->committed_tree;
  config->validating = FALSE;

  error_list = config->validate_errors;
  config->validate_errors = NULL;

  return error_list;
}

void
tnt_config_validation_error (TntConfig *config, const gchar *core_id, FluxTree *node, guint error_type,
                             const gchar *pretty_format, ...)
{
  TntConfigError *error;
  gchar          *message;
  va_list         args;

  assert_types (config);
  g_return_if_fail (config->validating == TRUE);

  va_start (args, pretty_format);
  message = g_strdup_vprintf (pretty_format, args);
  va_end (args);

  error = g_new0 (TntConfigError, 1);

  error->core_id = g_strdup (core_id);
  error->config_node = node;
  error->error_type = error_type;
  error->pretty_message = message;

  config->validate_errors = g_list_prepend (config->validate_errors, error);
}

const gchar *
tnt_config_error_get_core_id (TntConfigError *config_error)
{
  g_assert (config_error);

  return config_error->core_id;
}

FluxTree *
tnt_config_error_get_node (TntConfigError *config_error)
{
  g_assert (config_error);

  return config_error->config_node;
}

guint
tnt_config_error_get_error (TntConfigError *config_error)
{
  g_assert (config_error);

  return config_error->error_type;
}

const gchar *
tnt_config_error_get_message (TntConfigError *config_error)
{
  g_assert (config_error);

  return config_error->pretty_message;
}

void
tnt_config_free_errors (GList *error_list)
{
  GList *l;

  for (l = error_list; l; l = g_list_next (l))
  {
    TntConfigError *error = l->data;

    g_free (error->core_id);
    g_free (error->pretty_message);
    g_free (error);
  }

  g_list_free (error_list);
}

void
tnt_config_rollback (TntConfig *config)
{
  assert_types (config);

  flux_tree_destroy (config->working_tree);
  config->working_tree = flux_tree_dup_all (config->committed_tree);
}

void
tnt_config_commit (TntConfig *config)
{
  assert_types (config);

  config->peek_tree = config->working_tree;

  tnt_callback_list_run_while_1 (config->listeners, tnt_test_always_true, config);

  flux_tree_destroy (config->committed_tree);
  config->committed_tree = flux_tree_dup_all (config->working_tree);
  config->peek_tree = config->committed_tree;

  save_to_file (config);
}

gboolean
tnt_config_store_file (TntConfig *config, const gchar *path, gboolean replace_file)
{
  gboolean  result = TRUE;
  gchar    *old_path;

  assert_types (config);
  g_return_val_if_fail (path != NULL, FALSE);

  if (path == config->path)
    return TRUE;

  old_path = config->path;

  if (path)
  {
    config->path = g_strdup (path);

    if (replace_file)
      result = save_to_file (config);
    else
    {
      result = load_from_file (config);
      if (!result)
        result = save_to_file (config);
    }

    if (!result)
    {
      g_free (config->path);
      config->path = old_path;
    }
  }
  else
  {
    config->path = NULL;
  }

  if (result)
    g_free (old_path);

  return result;
}

void
tnt_config_peek_changes_begin (TntConfig *config)
{
  if (!config->peek_changes_ref)
    config->peek_tree = config->working_tree;

  config->peek_changes_ref++;
}

void
tnt_config_peek_changes_end (TntConfig *config)
{
  g_return_if_fail (config->peek_changes_ref > 0);

  config->peek_changes_ref--;

  if (!config->peek_changes_ref)
    config->peek_tree = config->committed_tree;
}

static FluxTree *
find_node (FluxTree *tree, va_list args)
{
  FluxTree    *node;
  const gchar *child_name;

  for (node = tree; (child_name = va_arg (args, const gchar *)) != NULL; )
  {
    node = flux_tree_find_first_child (node, child_name, strlen (child_name));
    if (!node)
      break;
  }

  return node;
}

static FluxTree *
ensure_node (FluxTree *tree, va_list args)
{
  FluxTree    *node;
  const gchar *child_name;

  for (node = tree; (child_name = va_arg (args, const gchar *)) != NULL; )
  {
    FluxTree *next;

    next = flux_tree_find_first_child (node, child_name, strlen (child_name));
    if (next)
      node = next;
    else
      node = flux_tree_new_with_parent_and_data (node, child_name, strlen (child_name));
  }

  return node;
}

static gboolean
goto_child (FluxTree **node)
{
  if (!*node)
    return FALSE;

  *node = flux_tree_get_first_child (*node);
  if (!*node)
    return FALSE;

  return TRUE;
}

static void
ensure_child (FluxTree **node)
{
  FluxTree *child;

  g_assert (*node);

  child = flux_tree_get_first_child (*node);
  if (!child)
  {
    child = flux_tree_new ();
    flux_tree_add_as_first_child (*node, child);
  }

  *node = child;
}

gboolean
tnt_config_is_path (TntConfig *config, FluxTree *node, ...)
{
  va_list   args;
  FluxTree *possible_match;

  assert_types (config);
  g_return_val_if_fail (node != NULL, FALSE);

  va_start (args, node);
  possible_match = find_node (config->committed_tree, args);
  va_end (args);

  if (!possible_match)
  {
    va_start (args, node);
    possible_match = find_node (config->working_tree, args);
    va_end (args);
  }

  if (node == possible_match)
    return TRUE;

  return FALSE;
}

gboolean
tnt_config_is_changed (TntConfig *config, ...)
{
  va_list   args;
  FluxTree *committed_root, *committed_node;
  FluxTree *working_root,   *working_node;

  assert_types (config);

  va_start (args, config);
  committed_root = find_node (config->committed_tree, args);
  va_end (args);

  va_start (args, config);
  working_root = find_node (config->working_tree, args);
  va_end (args);

  if (!committed_root)
  {
    if (!working_root)
      return FALSE;
    else
      return TRUE;
  }
  else if (!working_root)
    return TRUE;

  committed_node = committed_root;
  working_node   = working_root;

  for (;;)
  {
    if (flux_tree_cmp (committed_node, working_node))
      return TRUE;

    committed_node = flux_tree_get_next_infix (committed_node, committed_root);
    working_node   = flux_tree_get_next_infix (working_node,   working_root);

    if (!committed_node)
    {
      if (!working_node)
        break;

      return TRUE;
    }
    else if (!working_node)
      return TRUE;
  }

  return FALSE;
}

FluxTree *
tnt_config_peek_node (TntConfig *config, ...)
{
  va_list   args;
  FluxTree *node;

  assert_types (config);

  va_start (args, config);
  node = find_node (config->peek_tree, args);
  va_end (args);

  return node;
}

FluxTree *
tnt_config_edit_node (TntConfig *config, ...)
{
  va_list   args;
  FluxTree *node;

  assert_types (config);

  va_start (args, config);
  node = ensure_node (config->working_tree, args);
  va_end (args);

  return node;
}

gboolean
tnt_config_get_boolean (TntConfig *config, ...)
{
  va_list   args;
  FluxTree *node;
  gchar    *value;
  gboolean  result;

  assert_types (config);

  va_start (args, config);
  node = find_node (config->peek_tree, args);
  va_end (args);

  if (!goto_child (&node))
    return 0;

  value = flux_tree_data_get_str (node);
  if (!value)
    return 0;

  if (!g_ascii_strcasecmp (value, "true") ||
      !g_ascii_strcasecmp (value, "yes")  ||
      !g_ascii_strcasecmp (value, "on")   ||
      !g_ascii_strcasecmp (value, "1"))
    result = TRUE;
  else
    result = FALSE;

  g_free (value);
  return result;
}

gchar *
tnt_config_get_str (TntConfig *config, ...)
{
  va_list   args;
  FluxTree *node;

  assert_types (config);

  va_start (args, config);
  node = find_node (config->peek_tree, args);
  va_end (args);

  if (!goto_child (&node))
    return NULL;

  return flux_tree_data_get_str (node);
}

gint
tnt_config_get_int (TntConfig *config, ...)
{
  va_list   args;
  FluxTree *node;
  gchar    *value;
  gint64    result;

  assert_types (config);

  va_start (args, config);
  node = find_node (config->peek_tree, args);
  va_end (args);

  if (!goto_child (&node))
    return 0;

  value = flux_tree_data_get_str (node);
  if (!value)
    return 0;

  sscanf (value, "%" G_GINT64_FORMAT, &result);
  g_free (value);

  return (gint) result;
}

guint
tnt_config_get_uint (TntConfig *config, ...)
{
  va_list   args;
  FluxTree *node;
  gchar    *value;
  guint64   result;

  assert_types (config);

  va_start (args, config);
  node = find_node (config->peek_tree, args);
  va_end (args);

  if (!goto_child (&node))
    return 0;

  value = flux_tree_data_get_str (node);
  if (!value)
    return 0;

  sscanf (value, "%" G_GUINT64_FORMAT, &result);
  g_free (value);

  return (guint) result;
}

gint64
tnt_config_get_int64 (TntConfig *config, ...)
{
  va_list   args;
  FluxTree *node;
  gchar    *value;
  gint64    result;

  assert_types (config);

  va_start (args, config);
  node = find_node (config->peek_tree, args);
  va_end (args);

  if (!goto_child (&node))
    return 0;

  value = flux_tree_data_get_str (node);
  if (!value)
    return 0;

  sscanf (value, "%" G_GINT64_FORMAT, &result);
  g_free (value);

  return result;
}

guint64
tnt_config_get_uint64 (TntConfig *config, ...)
{
  va_list   args;
  FluxTree *node;
  gchar    *value;
  guint64   result;

  assert_types (config);

  va_start (args, config);
  node = find_node (config->peek_tree, args);
  va_end (args);

  if (!goto_child (&node))
    return 0;

  value = flux_tree_data_get_str (node);
  if (!value)
    return 0;

  sscanf (value, "%" G_GUINT64_FORMAT, &result);
  g_free (value);

  return result;
}

void
tnt_config_set_str (TntConfig *config, const gchar *value, ...)
{
  va_list   args;
  FluxTree *node;

  assert_types (config);
  g_return_if_fail (value != NULL);

  va_start (args, value);
  node = ensure_node (config->working_tree, args);
  va_end (args);

  ensure_child (&node);

  flux_tree_data_set_str (node, value);
}

void
tnt_config_set_boolean (TntConfig *config, gboolean value, ...)
{
  va_list   args;
  FluxTree *node;

  assert_types (config);

  va_start (args, value);
  node = ensure_node (config->working_tree, args);
  va_end (args);

  ensure_child (&node);

  flux_tree_data_set_str (node, value ? "yes" : "no");
}

void
tnt_config_set_int (TntConfig *config, gint value, ...)
{
  va_list   args;
  FluxTree *node;
  gchar    *str;

  assert_types (config);

  va_start (args, value);
  node = ensure_node (config->working_tree, args);
  va_end (args);

  ensure_child (&node);

  str = g_strdup_printf ("%d", value);
  flux_tree_data_set_str (node, str);
  g_free (str);
}

void
tnt_config_set_uint (TntConfig *config, guint value, ...)
{
  va_list   args;
  FluxTree *node;
  gchar    *str;

  assert_types (config);

  va_start (args, value);
  node = ensure_node (config->working_tree, args);
  va_end (args);

  ensure_child (&node);

  str = g_strdup_printf ("%u", value);
  flux_tree_data_set_str (node, str);
  g_free (str);
}

void
tnt_config_set_int64 (TntConfig *config, gint64 value, ...)
{
  va_list   args;
  FluxTree *node;
  gchar    *str;

  assert_types (config);

  va_start (args, value);
  node = ensure_node (config->working_tree, args);
  va_end (args);

  ensure_child (&node);

  str = g_strdup_printf ("%" G_GINT64_FORMAT, value);
  flux_tree_data_set_str (node, str);
  g_free (str);
}

void
tnt_config_set_uint64 (TntConfig *config, guint64 value, ...)
{
  va_list   args;
  FluxTree *node;
  gchar    *str;

  assert_types (config);

  va_start (args, value);
  node = ensure_node (config->working_tree, args);
  va_end (args);

  ensure_child (&node);

  str = g_strdup_printf ("%" G_GUINT64_FORMAT, value);
  flux_tree_data_set_str (node, str);
  g_free (str);
}

/* ------------- *
 * va_list calls *
 * ------------- */

gboolean
tnt_config_is_path_v (TntConfig *config, FluxTree *node, va_list path)
{
  va_list   args;
  FluxTree *possible_match;

  assert_types (config);
  g_return_val_if_fail (node != NULL, FALSE);

  G_VA_COPY (args, path);
  possible_match = find_node (config->committed_tree, args);
  va_end (args);

  if (!possible_match)
  {
    G_VA_COPY (args, path);
    possible_match = find_node (config->working_tree, args);
    va_end (args);
  }

  if (node == possible_match)
    return TRUE;

  return FALSE;
}

gboolean
tnt_config_is_changed_v (TntConfig *config, va_list path)
{
  va_list   args;
  FluxTree *committed_root, *committed_node;
  FluxTree *working_root,   *working_node;

  assert_types (config);

  G_VA_COPY (args, path);
  committed_root = find_node (config->committed_tree, args);
  va_end (args);

  G_VA_COPY (args, path);
  working_root = find_node (config->working_tree, args);
  va_end (args);

  if (!committed_root)
  {
    if (!working_root)
      return FALSE;
    else
      return TRUE;
  }
  else if (!working_root)
    return TRUE;

  committed_node = committed_root;
  working_node   = working_root;

  for (;;)
  {
    if (flux_tree_cmp (committed_node, working_node))
      return TRUE;

    committed_node = flux_tree_get_next_infix (committed_node, committed_root);
    working_node   = flux_tree_get_next_infix (working_node,   working_root);

    if (!committed_node)
    {
      if (!working_node)
        break;

      return TRUE;
    }
    else if (!working_node)
      return TRUE;
  }

  return FALSE;
}

FluxTree *
tnt_config_peek_node_v (TntConfig *config, va_list path)
{
  va_list   args;
  FluxTree *node;

  assert_types (config);

  G_VA_COPY (args, path);
  node = find_node (config->peek_tree, args);
  va_end (args);

  return node;
}

FluxTree *
tnt_config_edit_node_v (TntConfig *config, va_list path)
{
  va_list   args;
  FluxTree *node;

  assert_types (config);

  G_VA_COPY (args, path);
  node = ensure_node (config->working_tree, args);
  va_end (args);

  return node;
}

gboolean
tnt_config_get_boolean_v (TntConfig *config, va_list path)
{
  va_list   args;
  FluxTree *node;
  gchar    *value;
  gboolean  result;

  assert_types (config);

  G_VA_COPY (args, path);
  node = find_node (config->peek_tree, args);
  va_end (args);

  if (!goto_child (&node))
    return FALSE;

  value = flux_tree_data_get_str (node);
  if (!value)
    return FALSE;

  if (!g_ascii_strcasecmp (value, "true") ||
      !g_ascii_strcasecmp (value, "yes")  ||
      !g_ascii_strcasecmp (value, "on")   ||
      !g_ascii_strcasecmp (value, "1"))
    result = TRUE;
  else
    result = FALSE;

  g_free (value);
  return result;
}

gchar *
tnt_config_get_str_v (TntConfig *config, va_list path)
{
  va_list   args;
  FluxTree *node;

  assert_types (config);

  G_VA_COPY (args, path);
  node = find_node (config->peek_tree, args);
  va_end (args);

  if (!goto_child (&node))
    return NULL;

  return flux_tree_data_get_str (node);
}

gint
tnt_config_get_int_v (TntConfig *config, va_list path)
{
  va_list   args;
  FluxTree *node;
  gchar    *value;
  gint64    result;

  assert_types (config);

  G_VA_COPY (args, path);
  node = find_node (config->peek_tree, args);
  va_end (args);

  if (!goto_child (&node))
    return 0;

  value = flux_tree_data_get_str (node);
  if (!value)
    return 0;

  sscanf (value, "%" G_GINT64_FORMAT, &result);
  g_free (value);

  return (gint) result;
}

guint
tnt_config_get_uint_v (TntConfig *config, va_list path)
{
  va_list   args;
  FluxTree *node;
  gchar    *value;
  guint64   result;

  assert_types (config);

  G_VA_COPY (args, path);
  node = find_node (config->peek_tree, args);
  va_end (args);

  if (!goto_child (&node))
    return 0;

  value = flux_tree_data_get_str (node);
  if (!value)
    return 0;

  sscanf (value, "%" G_GUINT64_FORMAT, &result);
  g_free (value);

  return (guint) result;
}

gint64
tnt_config_get_int64_v (TntConfig *config, va_list path)
{
  va_list   args;
  FluxTree *node;
  gchar    *value;
  gint64    result;

  assert_types (config);

  G_VA_COPY (args, path);
  node = find_node (config->peek_tree, args);
  va_end (args);

  if (!goto_child (&node))
    return 0;

  value = flux_tree_data_get_str (node);
  if (!value)
    return 0;

  sscanf (value, "%" G_GINT64_FORMAT, &result);
  g_free (value);

  return result;
}

guint64
tnt_config_get_uint64_v (TntConfig *config, va_list path)
{
  va_list   args;
  FluxTree *node;
  gchar    *value;
  guint64   result;

  assert_types (config);

  G_VA_COPY (args, path);
  node = find_node (config->peek_tree, args);
  va_end (args);

  if (!goto_child (&node))
    return 0;

  value = flux_tree_data_get_str (node);
  if (!value)
    return 0;

  sscanf (value, "%" G_GUINT64_FORMAT, &result);
  g_free (value);

  return result;
}

void
tnt_config_set_str_v (TntConfig *config, const gchar *value, va_list path)
{
  va_list   args;
  FluxTree *node;

  assert_types (config);
  g_return_if_fail (value != NULL);

  G_VA_COPY (args, path);
  node = ensure_node (config->working_tree, args);
  va_end (args);

  ensure_child (&node);

  flux_tree_data_set_str (node, value);
}

void
tnt_config_set_boolean_v (TntConfig *config, gboolean value, va_list path)
{
  va_list   args;
  FluxTree *node;

  assert_types (config);

  G_VA_COPY (args, path);
  node = ensure_node (config->working_tree, args);
  va_end (args);

  ensure_child (&node);

  flux_tree_data_set_str (node, value ? "yes" : "no");
}

void
tnt_config_set_int_v (TntConfig *config, gint value, va_list path)
{
  va_list   args;
  FluxTree *node;
  gchar    *str;

  assert_types (config);

  G_VA_COPY (args, path);
  node = ensure_node (config->working_tree, args);
  va_end (args);

  ensure_child (&node);

  str = g_strdup_printf ("%d", value);
  flux_tree_data_set_str (node, str);
  g_free (str);
}

void
tnt_config_set_uint_v (TntConfig *config, guint value, va_list path)
{
  va_list   args;
  FluxTree *node;
  gchar    *str;

  assert_types (config);

  G_VA_COPY (args, path);
  node = ensure_node (config->working_tree, args);
  va_end (args);

  ensure_child (&node);

  str = g_strdup_printf ("%u", value);
  flux_tree_data_set_str (node, str);
  g_free (str);
}

void
tnt_config_set_int64_v (TntConfig *config, gint64 value, va_list path)
{
  va_list   args;
  FluxTree *node;
  gchar    *str;

  assert_types (config);

  G_VA_COPY (args, path);
  node = ensure_node (config->working_tree, args);
  va_end (args);

  ensure_child (&node);

  str = g_strdup_printf ("%" G_GINT64_FORMAT, value);
  flux_tree_data_set_str (node, str);
  g_free (str);
}

void
tnt_config_set_uint64_v (TntConfig *config, guint64 value, va_list path)
{
  va_list   args;
  FluxTree *node;
  gchar    *str;

  assert_types (config);

  G_VA_COPY (args, path);
  node = ensure_node (config->working_tree, args);
  va_end (args);

  ensure_child (&node);

  str = g_strdup_printf ("%" G_GUINT64_FORMAT, value);
  flux_tree_data_set_str (node, str);
  g_free (str);
}
