/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-context-factory.c - Factory for generic contexts.
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "tnt.h"
#include <time.h>

#define DEBUG(x)

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), TNT_CONTEXT_FACTORY_TYPE)

static void
destroy_context_cb (gpointer key, gpointer value, gpointer data)
{
  TntContextFactory *context_factory = data;

  if (context_factory->destroy_func)
    ((void (*)(gpointer)) context_factory->destroy_func) ((TntContextFactory **) key + 1);

  g_free (key);
}

void
tnt_context_factory_finalize (FluxObject *object)
{
  TntContextFactory *context_factory = TNT_CONTEXT_FACTORY (object);

  assert_types (context_factory);

  g_hash_table_foreach (context_factory->context_table, destroy_context_cb, context_factory);
  g_hash_table_destroy (context_factory->context_table);
}

void
tnt_context_factory_init (TntContextFactory *context_factory, guint struct_size, gpointer destroy_func)
{
  flux_object_init              (FLUX_OBJECT (context_factory));
  flux_object_set_type          (FLUX_OBJECT (context_factory), TNT_CONTEXT_FACTORY_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (context_factory), tnt_context_factory_finalize);

  context_factory->struct_size   = struct_size;
  context_factory->destroy_func  = destroy_func;
  context_factory->context_table = g_hash_table_new (g_direct_hash, g_direct_equal);
}

TntContextFactory *
tnt_context_factory_new_real (guint struct_size, gpointer destroy_func)
{
  TntContextFactory *context_factory;

  context_factory = g_new0 (TntContextFactory, 1);
  tnt_context_factory_init (context_factory, struct_size, destroy_func);

  return context_factory;
}

gpointer
tnt_context_factory_new_context (TntContextFactory *context_factory)
{
  TntContextFactory **context;

  assert_types (context_factory);

  context = g_malloc0 (context_factory->struct_size + sizeof (TntContextFactory *));

  g_hash_table_insert (context_factory->context_table, context, (gpointer) TRUE);
  *context = context_factory;
  context++;

  return context;
}

void
tnt_context_destroy (gpointer context)
{
  TntContextFactory *context_factory;

  g_return_if_fail (context != NULL);

  context_factory = *((TntContextFactory **) context - 1);

  if (!g_hash_table_remove (context_factory->context_table, (TntContextFactory **) context - 1))
  {
    g_warning ("Tried to destroy unknown context.");
    return;
  }

  if (context_factory->destroy_func)
    ((void (*)(gpointer)) context_factory->destroy_func) (context);

  g_free ((TntContextFactory **) context - 1);
}
