/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-data-db.c - Base class for disk databases.
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include "tnt.h"

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), TNT_DATA_DB_TYPE)

gboolean
tnt_data_db_init (TntDataDB *data_db, TntStorage *private_storage)
{
  flux_object_init              (FLUX_OBJECT (data_db));
  flux_object_set_type          (FLUX_OBJECT (data_db), TNT_DATA_DB_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (data_db), tnt_data_db_finalize);

  data_db->private_storage = private_storage;
  data_db->rmd_160         = flux_rmd160_new ();

  if (data_db->private_storage)
    flux_object_ref (FLUX_OBJECT (data_db->private_storage));

  return TRUE;
}

void
tnt_data_db_finalize (FluxObject *data_db)
{
  assert_types (data_db);

  flux_object_unref (FLUX_OBJECT (TNT_DATA_DB (data_db)->rmd_160));

  if (TNT_DATA_DB (data_db)->private_storage)
    flux_object_unref (FLUX_OBJECT (TNT_DATA_DB (data_db)->private_storage));
}

TntStorage *
tnt_data_db_get_storage (TntDataDB *data_db)
{
  assert_types (data_db);

  return data_db->private_storage;
}

gboolean
tnt_data_db_have_file_full (TntDataDB *data_db, const TntFileID *file_id)
{
  assert_types (data_db);
  g_return_val_if_fail (file_id != NULL, FALSE);
  g_return_val_if_fail (data_db->have_file_full_func != NULL, FALSE);

  return data_db->have_file_full_func (data_db, file_id);
}

gboolean
tnt_data_db_have_file_part (TntDataDB *data_db, const TntFileID *file_id)
{
  assert_types (data_db);
  g_return_val_if_fail (file_id != NULL, FALSE);
  g_return_val_if_fail (data_db->have_file_part_func != NULL, FALSE);

  return data_db->have_file_part_func (data_db, file_id);
}

gboolean
tnt_data_db_have_block (TntDataDB *data_db, const TntFileID *file_id, gint block_id)
{
  assert_types (data_db);
  g_return_val_if_fail (file_id != NULL, FALSE);
  g_return_val_if_fail (block_id >= 0, FALSE);
  g_return_val_if_fail (data_db->have_block_func != NULL, FALSE);

  return data_db->have_block_func (data_db, file_id, block_id);
}

TntFileBitmap *
tnt_data_db_get_file_bitmap (TntDataDB *data_db, const TntFileID *file_id)
{
  assert_types (data_db);
  g_return_val_if_fail (file_id != NULL, FALSE);
  g_return_val_if_fail (data_db->get_file_bitmap_func != NULL, NULL);

  return data_db->get_file_bitmap_func (data_db, file_id);
}

gboolean
tnt_data_db_get_block_hash (TntDataDB *data_db, const TntFileID *file_id, gint block_id,
                            TntPriority priority,
                            TntGotBlockHashFunc *func, gpointer data)
{
  assert_types (data_db);
  g_return_val_if_fail (file_id != NULL, FALSE);
  g_return_val_if_fail (block_id >= 0, FALSE);
  g_return_val_if_fail (func != NULL, FALSE);
  g_return_val_if_fail (data_db->get_block_hash_func != NULL, FALSE);

  return data_db->get_block_hash_func (data_db, file_id, block_id, priority, func, data);
}

gboolean
tnt_data_db_get_block (TntDataDB *data_db, const TntFileID *file_id, gint block_id,
                       TntPriority priority,
                       TntGotBlockFunc *func, gpointer data)
{
  assert_types (data_db);
  g_return_val_if_fail (file_id != NULL, FALSE);
  g_return_val_if_fail (block_id >= 0, FALSE);
  g_return_val_if_fail (func != NULL, FALSE);
  g_return_val_if_fail (data_db->get_block_func != NULL, FALSE);

  return data_db->get_block_func (data_db, file_id, block_id, priority, func, data);
}

gboolean
tnt_data_db_put_block (TntDataDB *data_db, const TntFileID *file_id, gint block_id,
                       gpointer block, guint len, TntPriority priority,
                       TntWroteBlockFunc *func, gpointer data)
{
  assert_types (data_db);
  g_return_val_if_fail (file_id != NULL, FALSE);
  g_return_val_if_fail (block_id >= 0, FALSE);
  g_return_val_if_fail (block != NULL, FALSE);
  g_return_val_if_fail (len > 0, FALSE);
  g_return_val_if_fail (data_db->put_block_func != NULL, FALSE);

  return data_db->put_block_func (data_db, file_id, block_id, block, len, priority, func, data);
}

void
tnt_data_db_cancel (TntDataDB *data_db, gpointer func, gpointer data)
{
  assert_types (data_db);
  g_return_if_fail (func != NULL);

  data_db->cancel_func (data_db, func, data);
}

gboolean
tnt_data_db_pick_random_file (TntDataDB *data_db, TntFileID *file_id)
{
  assert_types (data_db);
  g_return_val_if_fail (file_id != NULL, FALSE);

  return data_db->pick_random_file_func (data_db, file_id);
}
