/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-edge.c - Describes a connection to another node.
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"

#include <string.h>
#include <math.h>
#include "tnt.h"

#define DEBUG(x)

/* Size of routing table for one edge */
#define REACHABLE_NODES_MAX 2500

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), TNT_EDGE_TYPE)

typedef struct
{
  TntNodeID    id;
  TntTimeStamp last_seen;
  TntTimeStamp weight_timestamp;
  gdouble      weight;
}
ReachableNode;

static FluxMemChunk *reachable_nodes_mem_chunk = NULL;

static guint
hash_reachable_node (gconstpointer key)
{
  const ReachableNode *reachable_node = key;

  return (reachable_node->id & 0xffffffff) ^ (reachable_node->id >> 32);
}

static gboolean
compare_reachable_nodes (gconstpointer a, gconstpointer b)
{
  const ReachableNode *reachable_node_a = a, *reachable_node_b = b;

  if (reachable_node_a->id == reachable_node_b->id)
    return TRUE;

  return FALSE;
}

static void
remove_last_reachable_node (TntEdge *edge)
{
  ReachableNode *reachable_node;
  GList         *last;

  g_assert (edge->reachable_nodes_last);
  last = edge->reachable_nodes_last;

  reachable_node = last->data;
  g_hash_table_remove (edge->reachable_nodes_table, reachable_node);

  last = g_list_previous (last);
  edge->reachable_nodes_list = g_list_delete_link (edge->reachable_nodes_list, edge->reachable_nodes_last);
  edge->reachable_nodes_last = last;

  flux_mem_chunk_free (reachable_nodes_mem_chunk, reachable_node);
  edge->reachable_nodes_count--;
}

/* Assumes we don't already have the node */
static ReachableNode *
add_reachable_node (TntEdge *edge, TntNodeID node_id)
{
  ReachableNode *reachable_node;

  if (!reachable_nodes_mem_chunk)
    reachable_nodes_mem_chunk = flux_mem_chunk_new (128, sizeof (ReachableNode));

  reachable_node                   = flux_mem_chunk_alloc (reachable_nodes_mem_chunk);
  reachable_node->id               = node_id;
  reachable_node->last_seen        = tnt_timestamp ();
  reachable_node->weight           = 0.0;
  reachable_node->weight_timestamp = 0;

  edge->reachable_nodes_list = g_list_prepend (edge->reachable_nodes_list, reachable_node);
  g_hash_table_insert (edge->reachable_nodes_table, reachable_node, edge->reachable_nodes_list);

  if (!edge->reachable_nodes_last)
    edge->reachable_nodes_last = edge->reachable_nodes_list;

  edge->reachable_nodes_count++;
  if (edge->reachable_nodes_count > REACHABLE_NODES_MAX)
    remove_last_reachable_node (edge);

  return reachable_node;
}

static GList *
find_reachable_node_list (TntEdge *edge, TntNodeID node_id)
{
  ReachableNode  reachable_node_find;
  GList         *reachable_node_list;

  reachable_node_find.id = node_id;
  reachable_node_list = g_hash_table_lookup (edge->reachable_nodes_table, &reachable_node_find);
  return reachable_node_list;
}

static ReachableNode *
find_reachable_node (TntEdge *edge, TntNodeID node_id)
{
  GList *reachable_node_list;

  reachable_node_list = find_reachable_node_list (edge, node_id);
  if (!reachable_node_list)
    return NULL;

  return reachable_node_list->data;
}

static void
update_reachable_node_weight (ReachableNode *reachable_node)
{
  TntTimeStamp new_timestamp;

  new_timestamp = tnt_timestamp ();

  /* 1.3 ^ 60 ~= 6.800.000 */
  /* 1.1 ^ 60 ~=       300 */

  if (reachable_node->weight_timestamp != 0 && reachable_node->weight_timestamp < new_timestamp)
    reachable_node->weight /= pow (1.1, (gdouble) new_timestamp - (gdouble) reachable_node->weight_timestamp);

  reachable_node->weight_timestamp = new_timestamp;
}

void
tnt_edge_finalize (FluxObject *object)
{
  TntEdge *edge = TNT_EDGE (object);
  GList   *l;

  assert_types (object);

  if (edge->control_tcp_io)
  {
    _tnt_tcp_io_be_dead (edge->control_tcp_io);
    flux_object_unref (FLUX_OBJECT (edge->control_tcp_io));
  }

  if (edge->data_tcp_io)
  {
    _tnt_tcp_io_be_dead (edge->data_tcp_io);
    flux_object_unref (FLUX_OBJECT (edge->data_tcp_io));
  }

  for (l = edge->reachable_nodes_list; l; l = g_list_next (l))
  {
    flux_mem_chunk_free (reachable_nodes_mem_chunk, l->data);
  }

  g_list_free (edge->reachable_nodes_list);
  g_hash_table_destroy (edge->reachable_nodes_table);

  if (reachable_nodes_mem_chunk)
    flux_mem_chunk_clean (reachable_nodes_mem_chunk);
}

gboolean
tnt_edge_init (TntEdge *edge, TntNetworkCore *network_core)
{
  g_return_val_if_fail (network_core != NULL, FALSE);

  flux_object_init              (FLUX_OBJECT (edge));
  flux_object_set_type          (FLUX_OBJECT (edge), TNT_EDGE_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (edge), tnt_edge_finalize);

  edge->network_core = network_core;
  edge->creation_time = tnt_timestamp ();
  edge->reachable_nodes_table = g_hash_table_new (hash_reachable_node, compare_reachable_nodes);
  edge->score = 0.0;

  return TRUE;
}

TntEdge *
tnt_edge_new (TntNetworkCore *network_core)
{
  TntEdge *edge;

  edge = g_new0 (TntEdge, 1);
  if (!tnt_edge_init (edge, network_core))
  {
    g_free (edge);
    return NULL;
  }

  return edge;
}

TntNetworkCore *
tnt_edge_get_network_core (TntEdge *edge)
{
  assert_types (edge);

  return edge->network_core;
}

TntTimeStamp
tnt_edge_get_creation_time (TntEdge *edge)
{
  assert_types (edge);

  return edge->creation_time;
}

guint64
tnt_edge_get_peer_fingerprint (TntEdge *edge)
{
  assert_types (edge);

  return edge->peer_fingerprint;
}

void
tnt_edge_set_peer_fingerprint (TntEdge *edge, guint64 fingerprint)
{
  assert_types (edge);

  edge->peer_fingerprint = fingerprint;
}

static void
promote_reachable_node (TntEdge *edge, TntNodeID node_id)
{
  GList *reachable_node_list;

  assert_types (edge);
  g_return_if_fail (node_id != TNT_NODE_ID_INVALID);

  reachable_node_list = find_reachable_node_list (edge, node_id);

  if (reachable_node_list)
  {
    /* Move the node to the head of the list, if it isn't already the head */
    if (reachable_node_list != edge->reachable_nodes_list)
    {
      if (edge->reachable_nodes_last == reachable_node_list)
        edge->reachable_nodes_last = g_list_previous (reachable_node_list);

      edge->reachable_nodes_list = g_list_remove_link (edge->reachable_nodes_list, reachable_node_list);
      edge->reachable_nodes_list = g_list_concat (reachable_node_list, edge->reachable_nodes_list);

      if (!edge->reachable_nodes_last)
        edge->reachable_nodes_last = edge->reachable_nodes_list;
    }
  }
  else
  {
    add_reachable_node (edge, node_id);
  }
}

gboolean
tnt_edge_have_node (TntEdge *edge, TntNodeID node_id)
{
  assert_types (edge);
  g_return_val_if_fail (node_id != TNT_NODE_ID_INVALID, FALSE);

  if (find_reachable_node_list (edge, node_id))
    return TRUE;

  return FALSE;
}

gdouble
tnt_edge_get_node_weight (TntEdge *edge, TntNodeID node_id)
{
  ReachableNode *reachable_node;

  assert_types (edge);
  g_return_val_if_fail (node_id != TNT_NODE_ID_INVALID, 0.0);

  reachable_node = find_reachable_node (edge, node_id);
  if (!reachable_node)
    return 0.0;

  update_reachable_node_weight (reachable_node);
  return reachable_node->weight;
}

void
tnt_edge_add_node_weight (TntEdge *edge, TntNodeID node_id, gdouble weight)
{
  ReachableNode *reachable_node;

  assert_types (edge);
  g_return_if_fail (node_id != TNT_NODE_ID_INVALID);

  reachable_node = find_reachable_node (edge, node_id);
  if (!reachable_node)
    reachable_node = add_reachable_node (edge, node_id);

  g_assert (reachable_node);

  DEBUG (g_print ("Adding %.2lf\n", weight));

  update_reachable_node_weight (reachable_node);
  reachable_node->weight += weight;
  reachable_node->last_seen = tnt_timestamp ();
  promote_reachable_node (edge, node_id);
}

guint
tnt_edge_get_node_timestamp (TntEdge *edge, TntNodeID node_id)
{
  ReachableNode *reachable_node;

  assert_types (edge);
  g_return_val_if_fail (node_id != TNT_NODE_ID_INVALID, 0);

  reachable_node = find_reachable_node (edge, node_id);
  if (!reachable_node)
    return 0;

  return reachable_node->last_seen;
}

TntTcpIO *
tnt_edge_get_io (TntEdge *edge, TntStreamType type)
{
  assert_types (edge);

  switch (type)
  {
    case TNT_STREAM_TYPE_CONTROL:
      return edge->control_tcp_io;

    case TNT_STREAM_TYPE_DATA:
      return edge->data_tcp_io;

    default:
      return NULL;
  }
}

void
tnt_edge_set_io (TntEdge *edge, TntStreamType type, TntTcpIO *io)
{
  TntTcpIO **io_slot;

  assert_types (edge);

  switch (type)
  {
    case TNT_STREAM_TYPE_CONTROL:
      io_slot = &edge->control_tcp_io;
      break;

    case TNT_STREAM_TYPE_DATA:
      io_slot = &edge->data_tcp_io;
      if (io)
        tnt_tcp_io_set_queue_out_max_len (io, TNT_MESSAGE_UNKNOWN, 3);
      break;

    default:
      return;
  }

  if (io)
    flux_object_ref (FLUX_OBJECT (io));

  if (*io_slot)
    flux_object_unref (FLUX_OBJECT (*io_slot));

  *io_slot = io;
}

gboolean
tnt_edge_get_qualified (TntEdge *edge)
{
  assert_types (edge);

  if (edge->qualification_refs > 0)
    return TRUE;

  return FALSE;
}

void
tnt_edge_qualification_ref (TntEdge *edge)
{
  assert_types (edge);

  edge->qualification_refs++;
}

void
tnt_edge_qualification_unref (TntEdge *edge)
{
  assert_types (edge);

  edge->qualification_refs--;

  /* This sort of sucks. See TntNetworkCore::handle_incoming_data_block() for an explanation */
  if (edge->qualification_refs < 0)
    edge->qualification_refs = 0;
}

TntNodeIP4
tnt_edge_get_peer_ip4 (TntEdge *edge)
{
  TntTcpIO *tcp_io;

  assert_types (edge);

  tcp_io = tnt_edge_get_io (edge, TNT_STREAM_TYPE_CONTROL);
  if (!tcp_io)
    tcp_io = tnt_edge_get_io (edge, TNT_STREAM_TYPE_DATA);
  if (!tcp_io)
    return TNT_NODE_IP_INVALID;

  return tnt_tcp_io_get_remote_ip4 (tcp_io);
}

void
tnt_edge_get_peer_address4 (TntEdge *edge, TntNodeAddress4 *address)
{
  assert_types (edge);

  address->ip = tnt_edge_get_peer_ip4 (edge);

  if (address->ip == TNT_NODE_IP_INVALID)
  {
    address->port = 0;
    return;
  }

  address->port = edge->remote_listening_port;
}

void
tnt_edge_set_remote_port (TntEdge *edge, gint port)
{
  assert_types (edge);

  edge->remote_listening_port = port;
}

void
tnt_edge_send_message (TntEdge *edge, TntMessage *message)
{
  assert_types (edge);
  flux_object_check (FLUX_OBJECT (message), TNT_MESSAGE_TYPE);

  if (edge->control_tcp_io && !tnt_message_has_data_block (message))
    tnt_tcp_io_message_queue_out (edge->control_tcp_io, message);
  else if (edge->data_tcp_io)
    tnt_tcp_io_message_queue_out (edge->data_tcp_io, message);
}

#if 0
gboolean
tnt_edge_get_incoming_file_id (TntEdge *edge, TntFileID *file_id)
{
  TntMessageType  type;
  gpointer        payload, payload_copy;
  guint16         payload_size;
  TntMessage     *message;
  TntDataBlockM  *data_block_m;

  assert_types (edge);
  g_return_val_if_fail (file_id != NULL, FALSE);

  if (!edge->data_tcp_io)
    return FALSE;

  if (!tnt_tcp_io_get_read_info (edge->data_tcp_io, &type, &payload, &payload_size))
    return FALSE;

  if (type != TNT_MESSAGE_DATA_BLOCK)
    return FALSE;

  /* Set up message, assign payload, demarshal it, free message */

  payload_copy = g_malloc (payload_size);
  memcpy (payload_copy, payload, payload_size);
  message = tnt_message_new (type);
  tnt_message_set_payload (message, payload_copy, payload_size);
  data_block_m = tnt_message_get_struct (message);
  flux_object_unref (FLUX_OBJECT (message));

  /* Copy file ID from message struct, free message struct */

  if (!data_block_m)
    return FALSE;

  memcpy (file_id, &data_block_m->file_id, sizeof (TntFileID));
  g_free (data_block_m);
  return TRUE;
}
#endif

void
tnt_edge_change_score (TntEdge *edge, gdouble change)
{
  assert_types (edge);

  edge->score += change;
}

gdouble
tnt_edge_get_score (TntEdge *edge)
{
  assert_types (edge);

  return edge->score;
}

void
tnt_edge_debug_dump_routing_table (TntEdge *edge)
{
  GList *l;

  assert_types (edge);

  tnt_log_debug ("Routing table for %016llx:", (guint64) tnt_edge_get_peer_fingerprint (edge));

  for (l = edge->reachable_nodes_list; l; l = g_list_next (l))
  {
    ReachableNode *reachable_node = l->data;
    guint          age            = tnt_timestamp () - reachable_node->last_seen;

    update_reachable_node_weight (reachable_node);

    tnt_log_debug ("%016llx q=%8.1f t=%4d",
                   (guint64) reachable_node->id, reachable_node->weight,
                   age < 9999 ? age : 9999);
  }
}
