/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-file-bitmap.c -
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"

#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <glib.h>
#include "tnt.h"

#ifndef G_PLATFORM_WIN32
# include <unistd.h>
#else
# include "win32/win32-compat.h"
# include <io.h>
#endif

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), TNT_FILE_BITMAP_TYPE)

gboolean
tnt_file_bitmap_init (TntFileBitmap *file_bitmap, guint len_bits)
{
  g_return_val_if_fail (len_bits > 0, FALSE);

  flux_object_init              (FLUX_OBJECT (file_bitmap));
  flux_object_set_type          (FLUX_OBJECT (file_bitmap), TNT_FILE_BITMAP_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (file_bitmap), tnt_file_bitmap_finalize);

  file_bitmap->len_bits = len_bits;
  file_bitmap->bitmap   = g_malloc (TNT_BITS_TO_BYTES (len_bits));
  return TRUE;
}

TntFileBitmap *
tnt_file_bitmap_new (guint len_bits)
{
  TntFileBitmap *file_bitmap;

  file_bitmap = g_new0 (TntFileBitmap, 1);
  if (!tnt_file_bitmap_init (file_bitmap, len_bits))
  {
    g_free (file_bitmap);
    return NULL;
  }

  memset (file_bitmap->bitmap, 0, TNT_BITS_TO_BYTES (len_bits));
  return file_bitmap;
}

TntFileBitmap *
tnt_file_bitmap_dup (TntFileBitmap *file_bitmap)
{
  TntFileBitmap *new_bitmap;

  assert_types (file_bitmap);

  new_bitmap = g_new (TntFileBitmap, 1);
  if (!tnt_file_bitmap_init (new_bitmap, file_bitmap->len_bits))
  {
    g_free (new_bitmap);
    return NULL;
  }

  memcpy (new_bitmap->bitmap, file_bitmap->bitmap, TNT_BITS_TO_BYTES (new_bitmap->len_bits));
  return new_bitmap;
}

TntFileBitmap *
tnt_file_bitmap_load_from_file (const gchar *path, guint len_bits)
{
  TntFileBitmap *file_bitmap;
  TntStat        tstat;
  gint           fd;
  guint          len_bytes;
  gboolean       success;

  g_return_val_if_fail (path != NULL, NULL);
  g_return_val_if_fail (len_bits > 0, NULL);

  len_bytes = TNT_BITS_TO_BYTES (len_bits);

  if (!tnt_lstat (path, &tstat))
    return NULL;

  if (tstat.len > 0xffff || tstat.len != len_bytes)
    return NULL;

  fd = tnt_open (path, FALSE, FALSE, FALSE);
  if (fd < 0)
    return NULL;

  file_bitmap = tnt_file_bitmap_new (len_bits);
  if (!file_bitmap)
  {
    tnt_close (fd);
    return NULL;
  }

  /* We do a blocking read here - since these files are always small, it shouldn't pose
   * a problem. If this blocks the interface too much, though, I might reconsider. */
  success = tnt_safe_read (fd, (guint8 *) file_bitmap->bitmap, len_bytes);
  tnt_close (fd);

  if (!success)
  {
    flux_object_unref (FLUX_OBJECT (file_bitmap));
    return NULL;
  }

  return file_bitmap;
}

gboolean
tnt_file_bitmap_save_to_file (TntFileBitmap *file_bitmap, const gchar *path)
{
  gint     fd;
  guint    len_bytes;
  gboolean success;

  assert_types (file_bitmap);
  g_return_val_if_fail (path != NULL, FALSE);

  fd = tnt_open (path, TRUE, TRUE, FALSE);
  if (fd < 0)
    return FALSE;

  len_bytes = TNT_BITS_TO_BYTES (file_bitmap->len_bits);

  /* We do a blocking write here - since these files are always small, it shouldn't pose
   * a problem. If this blocks the interface too much, though, I might reconsider. */
  success = tnt_safe_write (fd, (const guint8 *) file_bitmap->bitmap, len_bytes);
  tnt_close (fd);

  if (!success)
    return FALSE;

  return TRUE;
}

void
tnt_file_bitmap_finalize (FluxObject *file_bitmap)
{
  assert_types (file_bitmap);
  g_free (TNT_FILE_BITMAP (file_bitmap)->bitmap);
}

gboolean
tnt_file_bitmap_get_bit (TntFileBitmap *file_bitmap, guint bit_index)
{
  guchar   byte;
  gboolean bit;

  assert_types (file_bitmap);
  g_return_val_if_fail (bit_index < file_bitmap->len_bits, FALSE);

  byte = file_bitmap->bitmap [bit_index >> 3];
  bit  = (byte >> (7 - (bit_index & 0x07))) & 0x01;
  return bit;
}

void
tnt_file_bitmap_set_bit (TntFileBitmap *file_bitmap, guint bit_index, gboolean value)
{
  guchar mask_byte;

  assert_types (file_bitmap);
  g_return_if_fail (bit_index < file_bitmap->len_bits);

  if (value)
  {
    mask_byte = 0x01 << (7 - (bit_index & 0x07));
    file_bitmap->bitmap [bit_index >> 3] |= mask_byte;
  }
  else
  {
    mask_byte = ~(0x01 << (7 - (bit_index & 0x07)));
    file_bitmap->bitmap [bit_index >> 3] &= mask_byte;
  }
}

void
tnt_file_bitmap_set_bitmap (TntFileBitmap *file_bitmap, guchar *bitmap)
{
  assert_types (file_bitmap);

  if (file_bitmap->bitmap && file_bitmap->bitmap != bitmap)
  {
    g_free (file_bitmap->bitmap);
  }

  file_bitmap->bitmap = bitmap;
}

/* Returns a copy of the bitmap */
guchar *
tnt_file_bitmap_get_bitmap (TntFileBitmap *file_bitmap)
{
  guchar *bitmap;

  assert_types (file_bitmap);

  bitmap = g_malloc (TNT_BITS_TO_BYTES (file_bitmap->len_bits));
  memcpy (bitmap, file_bitmap->bitmap, TNT_BITS_TO_BYTES (file_bitmap->len_bits));

  return bitmap;
}

guint
tnt_file_bitmap_get_len_bits (TntFileBitmap *file_bitmap)
{
  assert_types (file_bitmap);

  return file_bitmap->len_bits;
}

void
tnt_file_bitmap_and (TntFileBitmap *target_bitmap, TntFileBitmap *src_bitmap)
{
  gint i;

  assert_types (src_bitmap);
  assert_types (target_bitmap);
  g_return_if_fail (target_bitmap->len_bits == src_bitmap->len_bits);

  for (i = 0; i < TNT_BITS_TO_BYTES (target_bitmap->len_bits); i++)
  {
    target_bitmap->bitmap [i] &= src_bitmap->bitmap [i];
  }
}

void
tnt_file_bitmap_or (TntFileBitmap *target_bitmap, TntFileBitmap *src_bitmap)
{
  gint i;

  assert_types (src_bitmap);
  assert_types (target_bitmap);
  g_return_if_fail (target_bitmap->len_bits == src_bitmap->len_bits);

  for (i = 0; i < TNT_BITS_TO_BYTES (target_bitmap->len_bits); i++)
  {
    target_bitmap->bitmap [i] |= src_bitmap->bitmap [i];
  }
}

void
tnt_file_bitmap_nand (TntFileBitmap *target_bitmap, TntFileBitmap *src_bitmap)
{
  gint i;

  assert_types (src_bitmap);
  assert_types (target_bitmap);
  g_return_if_fail (target_bitmap->len_bits == src_bitmap->len_bits);

  for (i = 0; i < TNT_BITS_TO_BYTES (target_bitmap->len_bits); i++)
  {
    target_bitmap->bitmap [i] &= ~(src_bitmap->bitmap [i]);
  }
}

void
tnt_file_bitmap_fill_true (TntFileBitmap *file_bitmap)
{
  gint i;

  assert_types (file_bitmap);

  for (i = 0; i < TNT_BITS_TO_BYTES (file_bitmap->len_bits); i++)
  {
    file_bitmap->bitmap [i] = 0xff;
  }
}

void
tnt_file_bitmap_fill_false (TntFileBitmap *file_bitmap)
{
  gint i;

  assert_types (file_bitmap);

  for (i = 0; i < TNT_BITS_TO_BYTES (file_bitmap->len_bits); i++)
  {
    file_bitmap->bitmap [i] = 0x00;
  }
}

gboolean
tnt_file_bitmap_is_complete (TntFileBitmap *file_bitmap)
{
  guint  remaining_bits_num;
  guint8 remaining_byte;
  gint   i;

  assert_types (file_bitmap);

  /* NOTE: There are faster ways to do this. */
  for (i = 0; i < TNT_BITS_TO_BYTES (file_bitmap->len_bits + 1) - 1; i++)
  {
    if (file_bitmap->bitmap [i] != 0xff)
      return FALSE;
  }

  remaining_bits_num = file_bitmap->len_bits & 7;
  remaining_byte = (0xff << (8 - remaining_bits_num)) & 0xff;
  if ((file_bitmap->bitmap [TNT_BITS_TO_BYTES (file_bitmap->len_bits) - 1] & remaining_byte) != remaining_byte)
    return FALSE;

  return TRUE;
}

gboolean
tnt_file_bitmap_is_empty (TntFileBitmap *file_bitmap)
{
  gint i;

  assert_types (file_bitmap);

  /* NOTE: There are faster ways to do this. */
  for (i = 0; i < TNT_BITS_TO_BYTES (file_bitmap->len_bits); i++)
  {
    if (file_bitmap->bitmap [i] != 0x00)
      return FALSE;
  }

  return TRUE;
}

guint
tnt_file_bitmap_count_set_bits (TntFileBitmap *file_bitmap)
{
  guint count = 0;
  gint  i;

  assert_types (file_bitmap);

  for (i = 0; i < TNT_BITS_TO_BYTES (file_bitmap->len_bits); i++)
  {
    gint   j;
    guint8 b = file_bitmap->bitmap [i];

    if (!b)
      continue;

    for (j = 7; j >= 0 && (i << 3) + (7 - j) < file_bitmap->len_bits; j--)
    {
      count += (b >> j) & 1;
    }
  }

  return count;
}

gint
tnt_file_bitmap_get_set_bit_n (TntFileBitmap *file_bitmap, guint n)
{
  guint count = 0;
  gint  index = 0;
  gint  i;

  assert_types (file_bitmap);

  for (i = 0; i < TNT_BITS_TO_BYTES (file_bitmap->len_bits); i++)
  {
    gint   j;
    guint8 b = file_bitmap->bitmap [i];

    if (!b)
    {
      index += 8;
      continue;
    }

    for (j = 7; j >= 0; j--)
    {
      count += (b >> j) & 1;

      if (count > n)
        return index;

      index++;
    }
  }

  return -1;
}

guint64
tnt_file_bitmap_calc_data_size (TntFileBitmap *file_bitmap, guint file_len)
{
  guint64 size;
  guint   bits_set;

  assert_types (file_bitmap);
  g_return_val_if_fail (file_bitmap->len_bits * TNT_DATA_BLOCK_SIZE >= file_len, 0);

  bits_set = tnt_file_bitmap_count_set_bits (file_bitmap);
  if (!bits_set)
    return 0;

  size = (guint64) (bits_set - 1) * (guint64) TNT_DATA_BLOCK_SIZE;

  if (tnt_file_bitmap_get_bit (file_bitmap, file_bitmap->len_bits - 1))
  {
    size += TNT_BLOCK_IN_FILE_LEN (file_bitmap->len_bits - 1, file_len);
#if 0
    size -= TNT_DATA_BLOCK_SIZE;
    size += file_len % (TNT_DATA_BLOCK_SIZE + 1);
#endif
  }
  else
    size += TNT_DATA_BLOCK_SIZE;

  return size;
}

void
tnt_file_bitmap_dump (TntFileBitmap *file_bitmap)
{
  gint i;
  gint set_bits, picked_bit;

  assert_types (file_bitmap);

  for (i = 0; i < tnt_file_bitmap_get_len_bits (file_bitmap); i++)
  {
    g_print (tnt_file_bitmap_get_bit (file_bitmap, i) ? "1" : "0");
  }

  set_bits = tnt_file_bitmap_count_set_bits (file_bitmap);
  picked_bit = g_random_int () % set_bits;

  g_print ("\nSet bits: %d\n", set_bits);
  g_print ("Set bit #%d: %d\n", picked_bit, tnt_file_bitmap_get_set_bit_n (file_bitmap, picked_bit));
}
