/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-file-db.h -
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#ifndef TNT_FILE_DB_H
#define TNT_FILE_DB_H

#include <dirent.h>

#define TNT_FILE_DB(object) ((TntFileDB *) object)

typedef enum
{
  TNT_FILE_DB_IDLE,
  TNT_FILE_DB_SCANNING,
  TNT_FILE_DB_HASHING,
  TNT_FILE_DB_ACTION
}
TntFileDBState;

typedef enum
{
  TNT_FILE_DB_HANDLE_ONLINE,
  TNT_FILE_DB_HANDLE_SCANNING,
  TNT_FILE_DB_HANDLE_TALLIED,
  TNT_FILE_DB_HANDLE_HASHING
}
TntFileDBHandleState;

struct _TntFileDB
{
  TntDataDB          data_db;

  TntStorage        *private_storage;

  TntFileUpdateFunc *file_update_func;
  gpointer           file_update_func_data;
  TntFileIDHintFunc *file_id_hint_func;
  gpointer           file_id_hint_func_data;

  TntHandleFunc     *handle_add_func;
  gpointer           handle_add_func_data;
  TntHandleFunc     *handle_remove_func;
  gpointer           handle_remove_func_data;
  TntHandleFunc     *handle_change_func;
  gpointer           handle_change_func_data;

  GNode             *dir_tree;

  GList             *file_list;
  GList             *file_list_last;      /* Speedup pointer to last element for appending */

  GHashTable        *files_by_location;
  GHashTable        *files_by_id;

  GList             *serialization_list;

  GList             *block_action_queue;
  TntBlockAction    *current_block_action;

  gint               dir_count;
  gint               file_count;
  gint               file_hashed_count;
  gint64             file_size;
  gint64             file_hashed_size;

  TntFileDBState     state;
  gboolean           need_scan;
  gboolean           need_hash;
  gboolean           need_stabilization;

  TntTimeStamp       stabilization_time;

  /* State information */
  guint              callback_source_id;
  guint              report_timeout_id;
  gpointer           id_hint_file_stat;
  GList             *stat_node;
  DIR               *dir_hd;
  FluxFileIO        *file_io;
  gpointer           block;
};

void             tnt_file_db_init                     (TntFileDB *file_db);
TntFileDB       *tnt_file_db_new                      (void);
void             tnt_file_db_finalize                 (FluxObject *file_db);

TntFileDBState   tnt_file_db_get_state                (TntFileDB *file_db);

void             tnt_file_db_add_dir                  (TntFileDB *file_db, const gchar *path, gboolean recursive);
void             tnt_file_db_remove_dir               (TntFileDB *file_db, const gchar *path);

GList           *tnt_file_db_get_user_dir_list        (TntFileDB *file_db);
gboolean         tnt_file_db_get_dir_stats            (TntFileDB *file_db, gchar *path,
                                                       guint *files, guint64 *size);
gboolean         tnt_file_db_dir_is_scanned           (TntFileDB *file_db, gchar *path);

guint            tnt_file_db_get_dir_count            (TntFileDB *file_db);
guint            tnt_file_db_get_file_count           (TntFileDB *file_db, gboolean include_unhashed);
guint64          tnt_file_db_get_file_size            (TntFileDB *file_db, gboolean include_unhashed);

void             tnt_file_db_set_file_update_func     (TntFileDB *file_db, TntFileUpdateFunc *func, gpointer data);
void             tnt_file_db_set_file_id_hint_func    (TntFileDB *file_db, TntFileIDHintFunc *func, gpointer data);

gchar           *tnt_file_db_get_file_name            (TntFileDB *file_db, const TntFileID *file_id);
gchar           *tnt_file_db_get_file_path            (TntFileDB *file_db, const TntFileID *file_id);
gchar           *tnt_file_db_get_file_relative_path   (TntFileDB *file_db, const TntFileID *file_id);
TntTimeStamp     tnt_file_db_get_file_timestamp       (TntFileDB *file_db, const TntFileID *file_id);

gboolean         tnt_file_db_begin_serialization      (TntFileDB *file_db, const TntFileID *file_id,
                                                       const gchar *path, const gchar *name);
void             tnt_file_db_end_serialization        (TntFileDB *file_db, const TntFileID *file_id);
void             tnt_file_db_abort_serialization      (TntFileDB *file_db, const TntFileID *file_id);

gboolean         tnt_file_db_pick_random_file         (TntFileDB *file_db, TntFileID *file_id);

/* --- Model helpers --- */

gint             tnt_file_db_n_user_dirs              (TntFileDB *file_db);
TntFileDBHandle  tnt_file_db_nth_user_dir             (TntFileDB *file_db, gint n);

gint             tnt_file_db_handle_n_children        (TntFileDB *file_db, TntFileDBHandle handle);
TntFileDBHandle  tnt_file_db_handle_nth_child         (TntFileDB *file_db, TntFileDBHandle handle, gint n);
TntFileDBHandle  tnt_file_db_handle_parent            (TntFileDB *file_db, TntFileDBHandle handle);

gboolean         tnt_file_db_handle_is_user_dir       (TntFileDB *file_db, TntFileDBHandle handle);
gint             tnt_file_db_handle_get_position      (TntFileDB *file_db, TntFileDBHandle handle);
gchar           *tnt_file_db_handle_get_name          (TntFileDB *file_db, TntFileDBHandle handle);
gchar           *tnt_file_db_handle_get_path          (TntFileDB *file_db, TntFileDBHandle handle);
guint            tnt_file_db_handle_get_file_count    (TntFileDB *file_db, TntFileDBHandle handle);
guint64          tnt_file_db_handle_get_file_size     (TntFileDB *file_db, TntFileDBHandle handle);
TntFileDBHandleState tnt_file_db_handle_get_state     (TntFileDB *file_db, TntFileDBHandle handle);
GList           *tnt_file_db_handle_get_files         (TntFileDB *file_db, TntFileDBHandle handle);

void             tnt_file_db_set_handle_add_func      (TntFileDB *file_db, TntHandleFunc *func,
                                                       gpointer data);
void             tnt_file_db_set_handle_remove_func   (TntFileDB *file_db, TntHandleFunc *func,
                                                       gpointer data);
void             tnt_file_db_set_handle_change_func   (TntFileDB *file_db, TntHandleFunc *func,
                                                       gpointer data);

#endif  /* TNT_FILE_DB_H */
