/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-file-messages.c - Defines messages used by TntFileCore.
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include "tnt.h"
#include "tnt-file-messages.h"

typedef struct
{
  TntMessageType  type;
  gchar          *name;
}
MessageTypeName;

static FluxMarshalledStructEntry file_info_m_entries [] =
{
  FLUX_MARSHAL_STRUCT_FIXED   ("id",     TntFileInfoM, id,    TRUE, "file-id"),
  FLUX_MARSHAL_STRUCT_FIXED   ("meta",   TntFileInfoM, meta,  TRUE, "file-meta"),
  FLUX_MARSHAL_UINT16         ("blocks", TntFileInfoM, local_block_count, TRUE, 0),
  FLUX_MARSHAL_END            ()
};

static FluxMarshalledStructEntry query_m_entries [] =
{
  FLUX_MARSHAL_UINT16         ("id",     TntQueryM, id,     TRUE, 0),
  FLUX_MARSHAL_UINT16         ("flags",  TntQueryM, flags,  TRUE, 0),
  FLUX_MARSHAL_STRING_DYNAMIC ("string", TntQueryM, string, TRUE, NULL),
  FLUX_MARSHAL_END            ()
};

static FluxMarshalledStructEntry query_cancel_m_entries [] =
{
  FLUX_MARSHAL_UINT16         ("id", TntQueryCancelM, id, TRUE, 0),
  FLUX_MARSHAL_END            ()
};

static FluxMarshalledStructEntry data_block_request_m_entries [] =
{
  FLUX_MARSHAL_STRUCT_FIXED   ("file-id",  TntDataBlockRequestM, file_id,  TRUE, "file-id"),
  FLUX_MARSHAL_UINT16         ("block-id", TntDataBlockRequestM, block_id, TRUE, 0),
  FLUX_MARSHAL_END            ()
};

static FluxMarshalledStructEntry data_block_m_entries [] =
{
  FLUX_MARSHAL_STRUCT_FIXED   ("file-id",  TntDataBlockM, file_id,  TRUE, "file-id"),
  FLUX_MARSHAL_UINT16         ("block-id", TntDataBlockM, block_id, TRUE, 0),
  FLUX_MARSHAL_END            ()
};

static FluxMarshalledStructEntry ask_file_avail_m_entries [] =
{
  FLUX_MARSHAL_STRUCT_FIXED   ("file-id", TntAskFileAvailM, file_id, TRUE, "file-id"),
  FLUX_MARSHAL_END            ()
};

static FluxMarshalledStructEntry have_file_m_entries [] =
{
  FLUX_MARSHAL_STRUCT_FIXED   ("file-id", TntHaveFileM, file_id, TRUE, "file-id"),
  FLUX_MARSHAL_END            ()
};

static FluxMarshalledStructEntry have_file_bitmap_m_entries [] =
{
  FLUX_MARSHAL_STRUCT_FIXED   ("file-id",         TntHaveFileBitmapM, file_id,            TRUE, "file-id"),
  FLUX_MARSHAL_RAW_DYNAMIC    ("bitmap",          TntHaveFileBitmapM, bitmap, bitmap_len, TRUE, NULL),
  FLUX_MARSHAL_UINT16         ("bitmap-len-bits", TntHaveFileBitmapM, bitmap_len_bits,    TRUE, 0),
  FLUX_MARSHAL_END            ()
};

static FluxMarshalledStructEntry on_queue_m_entries [] =
{
  FLUX_MARSHAL_STRUCT_FIXED   ("file-id",   TntOnQueueM, file_id,  TRUE, "file-id"),
  FLUX_MARSHAL_UINT16         ("block-id",  TntOnQueueM, block_id, TRUE, 0),
  FLUX_MARSHAL_UINT32         ("timestamp", TntOnQueueM, timestamp, TRUE, 0),
  FLUX_MARSHAL_END            ()
};

static FluxMarshalledStructEntry off_queue_m_entries [] =
{
  FLUX_MARSHAL_STRUCT_FIXED   ("file-id",   TntOffQueueM, file_id,  TRUE, "file-id"),
  FLUX_MARSHAL_UINT16         ("block-id",  TntOffQueueM, block_id, TRUE, 0),
  FLUX_MARSHAL_UINT32         ("timestamp", TntOffQueueM, timestamp, TRUE, 0),
  FLUX_MARSHAL_END            ()
};

static MessageTypeName type_to_name [] =
{
  { TNT_MESSAGE_FILE_INFO,          "file-info"          },
  { TNT_MESSAGE_QUERY,              "query"              },
  { TNT_MESSAGE_QUERY_CANCEL,       "query-cancel"       },
  { TNT_MESSAGE_DATA_BLOCK_REQUEST, "data-block-request" },
  { TNT_MESSAGE_DATA_BLOCK,         "data-block"         },
  { TNT_MESSAGE_ASK_FILE_AVAIL,     "ask-file-avail"     },
  { TNT_MESSAGE_HAVE_FILE,          "have-file"          },
  { TNT_MESSAGE_HAVE_FILE_BITMAP,   "have-file-bitmap"   },
  { TNT_MESSAGE_ON_QUEUE,           "on-queue"           },
  { TNT_MESSAGE_OFF_QUEUE,          "off-queue"          },
  { TNT_MESSAGE_UNKNOWN,            NULL                 }
};

#define F FLUX_MARSHALLER_CLEAR

static FluxMarshalledStruct marshal_info [] =
{
  /* Prerequisites */
  { "file-id",            sizeof (TntFileID),            F, tnt_file_id_m_entries         },
  { "file-meta",          sizeof (TntFileMeta),          F, tnt_file_meta_m_entries       },

  /* Messages */
  { "file-info",          sizeof (TntFileInfoM),         F, file_info_m_entries           },
  { "query",              sizeof (TntQueryM),            F, query_m_entries               },
  { "query-cancel",       sizeof (TntQueryCancelM),      F, query_cancel_m_entries        },
  { "data-block-request", sizeof (TntDataBlockRequestM), F, data_block_request_m_entries  },
  { "data-block",         sizeof (TntDataBlockM),        F, data_block_m_entries          },
  { "ask-file-avail",     sizeof (TntAskFileAvailM),     F, ask_file_avail_m_entries      },
  { "have-file",          sizeof (TntHaveFileM),         F, have_file_m_entries           },
  { "have-file-bitmap",   sizeof (TntHaveFileBitmapM),   F, have_file_bitmap_m_entries    },
  { "on-queue",           sizeof (TntOnQueueM),          F, on_queue_m_entries            },
  { "off-queue",          sizeof (TntOffQueueM),         F, off_queue_m_entries           },
  { NULL }
};

void
tnt_file_messages_register (void)
{
  static gboolean registered = FALSE;
  gint            i;

  if (!registered)
  {
    tnt_message_register_marshal_info (marshal_info);

    for (i = 0; type_to_name [i].name; i++)
      tnt_message_register_type (type_to_name [i].type, type_to_name [i].name);

    registered = TRUE;
  }
}
