/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-file-meta-extract.c -
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"

#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>
#include <glib.h>
#include "tnt.h"

#ifndef G_PLATFORM_WIN32
# include <unistd.h>
#else
# include "win32/win32-compat.h"
# include <io.h>
#endif

#define HEAD_BUF_LEN 384
#define P(x)         ((gpointer) x)

typedef enum
{
  MAGIC_BYTE,
  MAGIC_SHORT_LE,
  MAGIC_SHORT_BE,
  MAGIC_LONG_LE,
  MAGIC_LONG_BE,
  MAGIC_STRING,
  MAGIC_CSTRING,

  MAGIC_LAST
}
MagicType;

typedef struct
{
  MagicType     magic_type;
  gpointer      data;
  gint          offset;
  TntFormatType file_format;
}
HeadMagic;

static const HeadMagic head_magic [] =
{
  { MAGIC_STRING,   P ("\x89PNG"),               0, TNT_FORMAT_PNG },
  { MAGIC_STRING,   P ("GIF8"),                  0, TNT_FORMAT_GIF },
  { MAGIC_SHORT_BE, P (0xffd8),                  0, TNT_FORMAT_JPEG },
  { MAGIC_STRING,   P ("hsi1"),                  0, TNT_FORMAT_JPEG },      /* HSI proprietary */
  { MAGIC_LONG_BE,  P (0x000001b3),              0, TNT_FORMAT_MPEG },
  { MAGIC_LONG_BE,  P (0x000001ba),              0, TNT_FORMAT_MPEG },
  /* { MAGIC_SHORT_BE, P (0xfff0),                  0, TNT_FORMAT_MP3 }, */  /* Doesn't work */
  { MAGIC_STRING,   P ("moov"),                  4, TNT_FORMAT_QUICKTIME },
  { MAGIC_STRING,   P ("mdat"),                  4, TNT_FORMAT_QUICKTIME },

  { MAGIC_STRING,   P ("\177ELF"),               0, TNT_FORMAT_UNKNOWN },
  { MAGIC_SHORT_LE, P (2),                      16, TNT_FORMAT_EXEC_LINUX },

  { MAGIC_STRING,   P ("\177ELF"),               0, TNT_FORMAT_EXEC_HELPER_LINUX },
  { MAGIC_LONG_LE,  P (0x00640107),              0, TNT_FORMAT_EXEC_LINUX },
  { MAGIC_LONG_LE,  P (0x00640108),              0, TNT_FORMAT_EXEC_LINUX },
  { MAGIC_LONG_LE,  P (0x0064010b),              0, TNT_FORMAT_EXEC_LINUX },
  { MAGIC_LONG_LE,  P (0x006400cc),              0, TNT_FORMAT_EXEC_LINUX },
  { MAGIC_STRING,   P ("\007\001\000"),          0, TNT_FORMAT_EXEC_HELPER_LINUX },
  { MAGIC_STRING,   P ("\012("),                 0, TNT_FORMAT_EXEC_LISP },  /* Byte-compiled Emacs 18 LISP */
  { MAGIC_STRING,   P (";ELC"),                  0, TNT_FORMAT_EXEC_LISP },  /* Byte-compiled Emacs 19+ LISP */
  { MAGIC_STRING,   P ("(SYSTEM::VERSION\040"),  0, TNT_FORMAT_EXEC_LISP },
  { MAGIC_STRING,   P ("\372\372\372\372"),      0, TNT_FORMAT_EXEC_LISP },
  { MAGIC_LONG_LE,  P (0xde120495),              0, TNT_FORMAT_GETTEXT },
  { MAGIC_LONG_LE,  P (0x950412de),              0, TNT_FORMAT_GETTEXT },

  { MAGIC_STRING,   P ("RIFF"),                  0, TNT_FORMAT_UNKNOWN },
  { MAGIC_STRING,   P ("RDIB"),                  8, TNT_FORMAT_UNKNOWN },
  { MAGIC_STRING,   P ("BM"),                   16, TNT_FORMAT_BMP },

  { MAGIC_STRING,   P ("RIFF"),                  0, TNT_FORMAT_UNKNOWN },
  { MAGIC_STRING,   P ("RMID"),                  8, TNT_FORMAT_MIDI },

  { MAGIC_STRING,   P ("RIFF"),                  0, TNT_FORMAT_UNKNOWN },
  { MAGIC_STRING,   P ("WAVE"),                  8, TNT_FORMAT_WAV },

  { MAGIC_STRING,   P ("RIFF"),                  0, TNT_FORMAT_UNKNOWN },
  { MAGIC_STRING,   P ("AVI "),                  8, TNT_FORMAT_AVI },

  { MAGIC_STRING,   P ("RIFX"),                  0, TNT_FORMAT_UNKNOWN },
  { MAGIC_STRING,   P ("RDIB"),                  8, TNT_FORMAT_UNKNOWN },
  { MAGIC_STRING,   P ("BM"),                   16, TNT_FORMAT_BMP },

  { MAGIC_STRING,   P ("RIFX"),                  0, TNT_FORMAT_UNKNOWN },
  { MAGIC_STRING,   P ("RMID"),                  8, TNT_FORMAT_MIDI },

  { MAGIC_STRING,   P ("RIFX"),                  0, TNT_FORMAT_UNKNOWN },
  { MAGIC_STRING,   P ("WAVE"),                  8, TNT_FORMAT_WAV },

  { MAGIC_STRING,   P ("RIFX"),                  0, TNT_FORMAT_UNKNOWN },
  { MAGIC_STRING,   P ("AVI "),                  8, TNT_FORMAT_AVI },

  { MAGIC_SHORT_BE, P (0xedab),                  0, TNT_FORMAT_UNKNOWN },
  { MAGIC_SHORT_BE, P (0xeedb),                  2, TNT_FORMAT_RPM },

  { MAGIC_STRING,   P ("<arch>\ndebian"),        0, TNT_FORMAT_DEB },
  { MAGIC_STRING,   P ("ustar"),               257, TNT_FORMAT_TAR },
  { MAGIC_STRING,   P ("PK\003\004"),            0, TNT_FORMAT_ZIP },
  { MAGIC_LONG_LE,  P (0xfdc4a7dc),             20, TNT_FORMAT_ZOO },
  { MAGIC_SHORT_LE, P (0xea60),                  0, TNT_FORMAT_ARJ },
  { MAGIC_STRING,   P ("-lh4-"),                 2, TNT_FORMAT_LHA },
  { MAGIC_STRING,   P ("-lh5-"),                 2, TNT_FORMAT_LHA },
  { MAGIC_STRING,   P ("Rar!"),                  0, TNT_FORMAT_RAR },

  { MAGIC_STRING,   P ("MSCF"),                  0, TNT_FORMAT_UNKNOWN },
  { MAGIC_LONG_LE,  P (0),                       4, TNT_FORMAT_MS_CAB },

  { MAGIC_STRING,   P ("\037\235"),              0, TNT_FORMAT_UNIX_COMPRESS },
  { MAGIC_STRING,   P ("\037\213"),              0, TNT_FORMAT_GZIP },
  { MAGIC_STRING,   P ("BZh"),                   0, TNT_FORMAT_BZIP2 },

  { MAGIC_STRING,   P ("{\\rtf"),                0, TNT_FORMAT_RTF },
  { MAGIC_CSTRING,  P ("<!doctype html"),        0, TNT_FORMAT_HTML },
  { MAGIC_CSTRING,  P ("<head"),                 0, TNT_FORMAT_HTML },
  { MAGIC_CSTRING,  P ("<title"),                0, TNT_FORMAT_HTML },
  { MAGIC_CSTRING,  P ("<html"),                 0, TNT_FORMAT_HTML },
  { MAGIC_CSTRING,  P ("<!--"),                  0, TNT_FORMAT_HTML },
  { MAGIC_CSTRING,  P ("<?xml"),                 0, TNT_FORMAT_XML },
  { MAGIC_CSTRING,  P ("<!doctype"),             0, TNT_FORMAT_SGML },
  { MAGIC_CSTRING,  P ("<!subdoc"),              0, TNT_FORMAT_SGML },

  { MAGIC_STRING,   P ("# This is a shell archive"), 0, TNT_FORMAT_SHELL_ARCHIVE },

  { MAGIC_STRING,   P ("\377WPC\020"),           0, TNT_FORMAT_WORD_PERFECT },
  { MAGIC_LONG_BE,  P (0xe1be0000),              0, TNT_FORMAT_MS_DOC },

  { MAGIC_STRING,   P ("CD001"),             32769, TNT_FORMAT_ISO9660 },     /* Straight ISO9660 */
  { MAGIC_STRING,   P ("BEA01"),             32769, TNT_FORMAT_MISC_CDROM },  /* Some DVD-specific format? Nero? */
  { MAGIC_STRING,   P ("CDROM"),             32776, TNT_FORMAT_MISC_CDROM },  /* High Sierra */
  { MAGIC_STRING,   P ("CD001"),             37633, TNT_FORMAT_ISO9660 },     /* ISO9660, raw 2352 bytes sectors */
  { MAGIC_STRING,   P ("CD001"),             37649, TNT_FORMAT_MISC_CDROM },  /* El Torito FileSystem Builder? */

  { MAGIC_STRING,   P (";;"),                    0, TNT_FORMAT_SOURCE_LISP },
  { MAGIC_STRING,   P ("/*"),                    0, TNT_FORMAT_SOURCE_C },
  { MAGIC_STRING,   P ("#include"),              0, TNT_FORMAT_SOURCE_C },
  { MAGIC_STRING,   P ("//"),                    0, TNT_FORMAT_SOURCE_CPLUSPLUS },
  { MAGIC_STRING,   P ("#!"),                    0, TNT_FORMAT_SCRIPT },
  { MAGIC_STRING,   P ("BEGIN"),                 0, TNT_FORMAT_SCRIPT },
  { MAGIC_STRING,   P ("BZ"),                    0, TNT_FORMAT_BZIP },
  { MAGIC_BYTE,     P (0x1a),                    0, TNT_FORMAT_ARC },
  { MAGIC_STRING,   P ("**ACE**"),               7, TNT_FORMAT_ACE },

  { MAGIC_LAST, NULL, -1, TNT_FORMAT_UNKNOWN }
};

static TntMediaType
format_to_media (gint type)
{
  TntMediaType media;

  if (type == TNT_FORMAT_UNKNOWN)
    media = TNT_MEDIA_UNKNOWN;
  else if (type < TNT_FORMAT_LAST_TEXT)
    media = TNT_MEDIA_TEXT;
  else if (type < TNT_FORMAT_LAST_PICTURE)
    media = TNT_MEDIA_PICTURE;
  else if (type < TNT_FORMAT_LAST_AUDIO)
    media = TNT_MEDIA_AUDIO;
  else if (type < TNT_FORMAT_LAST_VIDEO)
    media = TNT_MEDIA_VIDEO;
  else if (type < TNT_FORMAT_LAST_SOFTWARE)
    media = TNT_MEDIA_SOFTWARE;
  else
    media = TNT_MEDIA_OTHER;

  return media;
}

static gboolean
is_text (const guchar *data, gint len)
{
  gint  i;
  guint text_freq = 0;

  /* Avoid division by zero later */
  if (!len)
    return FALSE;

  for (i = 0; i < len; i++)
  {
    if (isprint (data [i]) && isascii (data [i]))
      text_freq++;
  }

  /* More than 90% text -> is text. */
  if ((text_freq * 1000) / len > 900)
    return TRUE;

  return FALSE;
}

static gboolean
analyze_head (gint fd, const guchar *head_data, gint head_len, TntFileMeta *file_meta)
{
  gint i;
  gint type = TNT_FORMAT_UNKNOWN;

  for (i = 0; head_magic [i].magic_type != MAGIC_LAST && type == TNT_FORMAT_UNKNOWN; i++)
  {
    guchar       *temp_data = NULL;
    const guchar *data;
    gint          magic_len;
    gboolean      match = FALSE;

    switch (head_magic [i].magic_type)
    {
      case MAGIC_BYTE:
        magic_len = 1;
        break;

      case MAGIC_SHORT_LE:
      case MAGIC_SHORT_BE:
        magic_len = 2;
        break;

      case MAGIC_LONG_LE:
      case MAGIC_LONG_BE:
        magic_len = 4;
        break;

      case MAGIC_STRING:
      case MAGIC_CSTRING:
        magic_len = strlen (head_magic [i].data) + 1;
        break;

      default:
        magic_len = 0;
        g_assert_not_reached ();
        break;
    }

    if (head_magic [i].offset + magic_len < head_len)
    {
      /* Use cached header */
      data = head_data + head_magic [i].offset;
    }
    else
    {
      /* Data is outside cached header, read to a temporary buffer */
      if (tnt_lseek (fd, head_magic [i].offset, SEEK_SET) != head_magic [i].offset)
        continue;

      temp_data = g_malloc (magic_len);
      if (!tnt_safe_read (fd, temp_data, magic_len))
      {
        g_free (temp_data);
        continue;
      }

      data = temp_data;
    }

    switch (head_magic [i].magic_type)
    {
      case MAGIC_BYTE:
        if (data [0] == (guint32) head_magic [i].data)
          match = TRUE;
        break;

      case MAGIC_SHORT_LE:
        if (GUINT16_FROM_LE (*((guint16 *) (data))) == (guint32) head_magic [i].data)
          match = TRUE;
        break;

      case MAGIC_SHORT_BE:
        if (GUINT16_FROM_BE (*((guint16 *) (data))) == (guint32) head_magic [i].data)
          match = TRUE;
        break;

      case MAGIC_LONG_LE:
        if (GUINT32_FROM_LE (*((guint32 *) (data))) == (guint32) head_magic [i].data)
          match = TRUE;
        break;

      case MAGIC_LONG_BE:
        if (GUINT32_FROM_BE (*((guint32 *) (data))) == (guint32) head_magic [i].data)
          match = TRUE;
        break;

      case MAGIC_STRING:
        if (!memcmp ((gchar *) data, head_magic [i].data, strlen (head_magic [i].data)))
          match = TRUE;
        break;

      case MAGIC_CSTRING:
        if (!flux_memcasecmp ((gchar *) data, head_magic [i].data, strlen (head_magic [i].data)))
          match = TRUE;
        break;

      default:
        g_assert_not_reached ();
        break;
    }

    if (!match && head_magic [i].file_format == TNT_FORMAT_UNKNOWN)
    {
      /* Skip the rest of this chain */

      while (head_magic [i].file_format == TNT_FORMAT_UNKNOWN)
        i++;

      /* Skip the concluding format in the chain */
      i++;
    }
    else if (match)
    {
      type = head_magic [i].file_format;
    }

    g_free (temp_data);
  }

  if (type == TNT_FORMAT_UNKNOWN)
    return FALSE;

  tnt_file_meta_set_format (file_meta, type);
  return TRUE;
}

typedef struct
{
  gchar         *ext;
  TntFormatType  format;
}
Extension;

static const Extension extension_list [] =
{
  { "txt",  TNT_FORMAT_ASCII },
  { "xml",  TNT_FORMAT_XML },
  { "html", TNT_FORMAT_HTML },
  { "sgml", TNT_FORMAT_SGML },
  { "doc",  TNT_FORMAT_MS_DOC },
  { "rtf",  TNT_FORMAT_RTF },
  { "wp5",  TNT_FORMAT_WORD_PERFECT },
  { "cpp",  TNT_FORMAT_SOURCE_CPLUSPLUS },
  /* { "", TNT_FORMAT_SOURCE_CPLUSPLUS }, */
  { "java", TNT_FORMAT_SOURCE_JAVA },
  /* { "", TNT_FORMAT_SOURCE_DOTNET }, */
  /* { "", TNT_FORMAT_SOURCE_RUBY }, */
  { "py",   TNT_FORMAT_SOURCE_PYTHON },
  { "pl",   TNT_FORMAT_SOURCE_PERL },
  { "sh",   TNT_FORMAT_SOURCE_UNIX_SHELL },
  { "lisp", TNT_FORMAT_SOURCE_LISP },
  { "el",   TNT_FORMAT_SOURCE_LISP },
  { "png",  TNT_FORMAT_PNG },
  { "jpg",  TNT_FORMAT_JPEG },
  { "jpeg", TNT_FORMAT_JPEG },
  { "gif",  TNT_FORMAT_GIF },
  { "bmp",  TNT_FORMAT_BMP },
  { "cbz",  TNT_FORMAT_ZIP_COMIC },
  { "cbr",  TNT_FORMAT_RAR_COMIC },
  { "cbt",  TNT_FORMAT_TAR_COMIC },
  { "cba",  TNT_FORMAT_ACE_COMIC },
  { "au",   TNT_FORMAT_AU },
  { "wav",  TNT_FORMAT_WAV },
  { "ogg",  TNT_FORMAT_OGG },
  { "mp3",  TNT_FORMAT_MP3 },
  { "mid",  TNT_FORMAT_MIDI },
  { "midi", TNT_FORMAT_MIDI },
  { "mpg",  TNT_FORMAT_MPEG },
  { "mpeg", TNT_FORMAT_MPEG },
  { "divx", TNT_FORMAT_DIVX },  /* Sort of bogus, they're MPEG */
  { "qt",   TNT_FORMAT_QUICKTIME },
  { "mov",  TNT_FORMAT_QUICKTIME },
  { "avi",  TNT_FORMAT_AVI },
  { "asf",  TNT_FORMAT_ASF },
  { "wmf",  TNT_FORMAT_WMF },
  { "wmv",  TNT_FORMAT_WMV },
  { "ogm",  TNT_FORMAT_OGM },
  { "so",   TNT_FORMAT_EXEC_HELPER_LINUX },
  { "exe",  TNT_FORMAT_EXEC_WINDOWS },
  { "nes",  TNT_FORMAT_EXEC_NES_ROM },
  { "smc",  TNT_FORMAT_EXEC_SNES_ROM },
  { "n64",  TNT_FORMAT_EXEC_NES64_ROM },
  { "gb",   TNT_FORMAT_EXEC_NGB_ROM },
  { "gba",  TNT_FORMAT_EXEC_NGBA_ROM },
  { "po",   TNT_FORMAT_GETTEXT },
  { "zip",  TNT_FORMAT_ZIP },
  { "tar",  TNT_FORMAT_TAR },
  { "zoo",  TNT_FORMAT_ZOO },
  { "arj",  TNT_FORMAT_ARJ },
  { "lha",  TNT_FORMAT_LHA },
  { "lzh",  TNT_FORMAT_LHA },
  { "rar",  TNT_FORMAT_RAR },
  { "arc",  TNT_FORMAT_ARC },  /* Archaic. Ha-ha. */
  { "gz",   TNT_FORMAT_GZIP },
  { "bz",   TNT_FORMAT_BZIP },
  { "bz2",  TNT_FORMAT_BZIP2 },
  { "cab",  TNT_FORMAT_MS_CAB },
  { "rpm",  TNT_FORMAT_RPM },
  { "deb",  TNT_FORMAT_DEB },
  { "iso",  TNT_FORMAT_ISO9660 },
  { "ace",  TNT_FORMAT_ACE },

  /* Put short ones at bottom to minimize risk of mistaken match */
  { "c",    TNT_FORMAT_SOURCE_C },
  { "h",    TNT_FORMAT_SOURCE_C },

  { NULL,   TNT_FORMAT_UNKNOWN }
};

static gboolean
analyze_extension (const gchar *extension, TntFileMeta *file_meta)
{
  gboolean  result = FALSE;
  gint      i;

  for (i = 0; extension_list [i].ext != NULL; i++)
  {
    if (!strcasecmp (extension, extension_list [i].ext))
    {
      tnt_file_meta_set_format (file_meta, extension_list [i].format);
      result = TRUE;
      break;
    }
  }

  return result;
}

static gboolean
analyze_sequence_number (const gchar *file_path, TntFileMeta *file_meta)
{
  TntMediaType  media;
  gchar        *file_name;
  gchar        *p;
  gint          seq = 0;
  gboolean      result = FALSE;

  /* Do this for audio and video only */
  media = tnt_file_meta_get_media (file_meta);
  if (media != TNT_MEDIA_AUDIO && media != TNT_MEDIA_VIDEO)
    return FALSE;

  /* Find the last path separator */
  file_name = strrchr (file_path, G_DIR_SEPARATOR);
  if (!file_name)
    return FALSE;

  /* Go to char immediately after it (first char of file name), and make
   * sure we have at least two characters in the name. */
  file_name++;
  if (strlen (file_name) < 2)
    return FALSE;

  /* Make sure it's at least two digits */
  if (!g_ascii_isdigit (file_name [0]) || !g_ascii_isdigit (file_name [1]))
    goto out;

  /* Parse */
  seq = atoi (file_name);
  result = TRUE;

  /* Check that separator is not digit or alpha */
  p = file_name + strspn (file_name, "0123456789");
  if (!p)
    goto out;

  if (*p && (g_ascii_isdigit (*p) || g_ascii_isalpha (*p)))
    result = FALSE;

out:
  if (result)
  {
    gchar *seq_str = g_strdup_printf ("%d", seq);

    tnt_file_meta_set_seq (file_meta, seq_str);
    g_free (seq_str);
  }

  return result;
}

/* Input file_path is in local charset */
TntFileMeta *
tnt_file_meta_extract (const gchar *file_path)
{
  TntFileMeta   *file_meta;
  gint           fd;
  guchar         head_buf [HEAD_BUF_LEN];
  gboolean       have_type       = FALSE;
  gboolean       looks_like_text = FALSE;
  TntFormatType  format          = TNT_FORMAT_UNKNOWN;

  g_return_val_if_fail (file_path != NULL, NULL);

  file_meta = tnt_file_meta_new ();

  fd = tnt_open (file_path, FALSE, FALSE, TRUE);
  if (fd >= 0)
  {
    gint len;

    len = tnt_read (fd, head_buf, HEAD_BUF_LEN);
    have_type = analyze_head (fd, head_buf, len, file_meta);

    if (have_type)
    {
      switch (tnt_file_meta_get_format (file_meta))
      {
        case TNT_FORMAT_OGG:
          tnt_ogg_extract_file_meta (fd, file_meta);
          break;
        case TNT_FORMAT_MP3:
          tnt_id3_extract_file_meta (fd, file_meta);
          break;
        case TNT_FORMAT_AVI:
          tnt_avi_extract_file_meta (fd, file_meta);
          break;
#if 0
        case TNT_FORMAT_QUICKTIME:
          tnt_qt_extract_file_meta (file_path, file_meta);
          break;
#endif
        case TNT_FORMAT_MPEG:
          tnt_mpg_extract_file_meta (fd, file_meta);
          break;
        case TNT_FORMAT_JPEG:
          tnt_jpeg_extract_file_meta (file_path, file_meta);
          break;
        default:
          break;
      }
    }
    else if ((have_type = tnt_ogg_extract_file_meta (fd, file_meta)))
    {
      /* OGG Vorbis, not detected by magic */
    }
    else if ((have_type = tnt_id3_extract_file_meta (fd, file_meta)))
    {
      /* MP3 with ID3 tag, not detected by magic */
    }
    else
    {
      /* Text, possibly? */
      looks_like_text = is_text (head_buf, len);
    }

    tnt_close (fd);
  }
  else
  {
    tnt_log (TNT_LOG_SUBSYSTEM_FILE, TNT_LOG_LEVEL_DEBUG, TNT_LOG_MOOD_DISAPPOINTING, 0,
             "Failed to open '%s' for meta extraction.", file_path);
  }

  format = tnt_file_meta_get_format (file_meta);

  if (!have_type || format == TNT_FORMAT_ZIP || format == TNT_FORMAT_RAR ||
                    format == TNT_FORMAT_TAR || format == TNT_FORMAT_ACE)
  {
    const gchar *file_extension = strrchr (file_path, '.');

    if (file_extension)
    {
      file_extension++;
      if (!strlen (file_extension))
        file_extension = NULL;
    }

    if (file_extension)
      have_type = analyze_extension (file_extension, file_meta);
  }

  format = tnt_file_meta_get_format (file_meta);

  if (have_type && (format == TNT_FORMAT_ZIP_COMIC || format == TNT_FORMAT_RAR_COMIC ||
                    format == TNT_FORMAT_TAR_COMIC || format == TNT_FORMAT_ACE_COMIC))
  {
    /* Set the category to comic if the file has a comic format */
    tnt_file_meta_set_category (file_meta, 0, 6);
    tnt_file_meta_set_category (file_meta, 1, 0);
  }

  if (!have_type && looks_like_text)
  {
    /* This catches plain-text files with weird extensions, like .nfo */
    tnt_file_meta_set_format (file_meta, TNT_FORMAT_ASCII);
  }

  if (have_type)
    tnt_file_meta_set_media (file_meta, format_to_media (tnt_file_meta_get_format (file_meta)));

  analyze_sequence_number (file_path, file_meta);

  return file_meta;
}
