/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-hash-collider.c - Hashcash worker class
 *
 * Copyright (C) 2005 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "tnt.h"
#include <string.h>
#include <time.h>

#define DEBUG(x)

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), TNT_HASH_COLLIDER_TYPE)

#define STATE_FILENAME     "hash-collisions"
#define THREAD_STACK_SIZE  65536
#define ATTEMPTS_PER_CYCLE 16384
#define SAVE_INTERVAL_S    30

typedef struct
{
  guint64           target;
  guint8            next_attempt [TNT_HASH_COLLISIONS_LEN];
  TntHashCollisions target_collisions;
}
ColliderState;

static gchar *
full_state_path (TntHashCollider *hash_collider)
{
  return g_strjoin (G_DIR_SEPARATOR_S, tnt_storage_get_path (hash_collider->storage),
                    STATE_FILENAME, NULL);
}

static void
construct_target_hash (TntHashCollider *hash_collider)
{
  flux_rmd160_hash_single (hash_collider->target_hash,
                           &hash_collider->target,
                           sizeof (hash_collider->target));
}

static gint
compare_bits (const guint8 *a, const guint8 *b, gint len)
{
  guint count = 0;
  gint  i;

  for (i = 0; i < len; i++)
  {
    guint8 a_d = a [i];
    guint8 b_d = b [i];
    gint   j;

    for (j = 7; j >= 0; j--)
    {
      if (((a_d >> j) & 1) == ((b_d >> j) & 1))
        count++;
    }
  }

  return count;
}

static gint
collide_value (const guint8 *target_hash, const guint8 *value)
{
  guint8 value_hash [20];

  flux_rmd160_hash_single (value_hash, value, TNT_HASH_COLLISIONS_LEN);
  return compare_bits (target_hash, value_hash, 20);
}

static void
increment_value (guint8 *value)
{
  gint i;

  for (i = 0; i < TNT_HASH_COLLISIONS_LEN; i++)
  {
    guint8 b = value [i] + 1;

    value [i] = b;
    if (b != 0)
      break;
  }
}

static gboolean
insert_value (TntHashCollider *hash_collider, const guint8 *value, gint n_bits)
{
  gint best_i = 0;
  gint i;

  /* Find the poorest collision we currently have */

  for (i = 0; i < TNT_HASH_COLLISIONS_NUM; i++)
  {
    if (hash_collider->target_collisions_n_bits [i] <
        hash_collider->target_collisions_n_bits [best_i])
      best_i = i;
  }

  /* If the new collision is better, replace the poorest one */

  if (n_bits > hash_collider->target_collisions_n_bits [best_i])
  {
    memcpy (hash_collider->target_collisions.v [best_i], value, TNT_HASH_COLLISIONS_LEN);
    hash_collider->target_collisions_n_bits [best_i] = n_bits;
    return TRUE;
  }

  return FALSE;
}

static gboolean
load_state (TntHashCollider *hash_collider)
{
  gchar         *path;
  gint           state_fd;
  ColliderState  state;
  gboolean       result = FALSE;
  gint           i;

  path = full_state_path (hash_collider);
  state_fd = tnt_open (path, FALSE, FALSE, FALSE);
  g_free (path);
  if (state_fd < 0)
    return FALSE;

  /* Read each field separately so platform-specific padding doesn't bite us */

  if (!tnt_safe_read (state_fd, &state.target, sizeof (state.target)))
    goto io_error;
  if (!tnt_safe_read (state_fd, state.next_attempt, TNT_HASH_COLLISIONS_LEN))
    goto io_error;

  for (i = 0; i < TNT_HASH_COLLISIONS_NUM; i++)
  {
    if (!tnt_safe_read (state_fd, state.target_collisions.v [i], TNT_HASH_COLLISIONS_LEN))
      goto io_error;
  }

  /* Successful read; copy values to hash collider and init derived values */

  hash_collider->target = state.target;
  memcpy (hash_collider->next_attempt, state.next_attempt, TNT_HASH_COLLISIONS_LEN);

  construct_target_hash (hash_collider);

  for (i = 0; i < TNT_HASH_COLLISIONS_NUM; i++)
  {
    memcpy (hash_collider->target_collisions.v [i], state.target_collisions.v [i],
            TNT_HASH_COLLISIONS_LEN);
    hash_collider->target_collisions_n_bits [i] =
      collide_value (hash_collider->target_hash, hash_collider->target_collisions.v [i]);
  }

  result = TRUE;

io_error:
  tnt_close (state_fd);
  return result;
}

static gboolean
save_state (TntHashCollider *hash_collider)
{
  gchar    *path;
  gint      state_fd;
  gboolean  result = FALSE;
  gint      i;

  path = full_state_path (hash_collider);
  state_fd = tnt_open (path, TRUE, FALSE, FALSE);
  g_free (path);
  if (state_fd < 0)
    return FALSE;

  /* Write each field separately so platform-specific padding doesn't bite us */

  if (!tnt_safe_write (state_fd, &hash_collider->target, sizeof (hash_collider->target)))
    goto io_error;
  if (!tnt_safe_write (state_fd, &hash_collider->next_attempt, TNT_HASH_COLLISIONS_LEN))
    goto io_error;

  for (i = 0; i < TNT_HASH_COLLISIONS_NUM; i++)
  {
    if (!tnt_safe_write (state_fd, hash_collider->target_collisions.v [i], TNT_HASH_COLLISIONS_LEN))
      goto io_error;
  }

  result = TRUE;

io_error:
  tnt_close (state_fd);
  return result;
}

static gpointer
worker_main (TntHashCollider *hash_collider)
{
  TntTimeStamp last_save;
  gulong       usec_per_cycle = 0;

  g_mutex_lock (hash_collider->mutex);

  last_save = tnt_timestamp ();

  while (hash_collider->running)
  {
    TntTimeStamp now;
    GTimeVal     before;
    GTimeVal     after;
    gint         i;

    if (hash_collider->work_rate < 0.001)
    {
      /* Work rate is at minimum; pause indefinitely */

      g_cond_wait (hash_collider->cond, hash_collider->mutex);
      continue;
    }
    else if (hash_collider->work_rate > 0.999)
    {
      /* Work rate is at maximum; do nothing */
    }
    else
    {
      gdouble factor;
      gdouble usec_to_sleep;

      /* Work rate is somewhere in between min and max; wait for
       * an adequate length of time. */

      factor = ((gdouble) 1.0 / (gdouble) hash_collider->work_rate) - (gdouble) 1.0;
      usec_to_sleep = (gdouble) usec_per_cycle * factor;

      /* Never sleep for more than a second */
      if (usec_to_sleep > 1000000.0)
        usec_to_sleep = 1000000.0;

      g_mutex_unlock (hash_collider->mutex);
      g_usleep ((gulong) usec_to_sleep);
      g_mutex_lock (hash_collider->mutex);
    }

    if (hash_collider->reset)
    {
      memset (&hash_collider->target_collisions, 0, sizeof (hash_collider->target_collisions));
      memset (hash_collider->target_collisions_n_bits, 0, sizeof (hash_collider->target_collisions_n_bits));
      memset (hash_collider->next_attempt, 0, sizeof (hash_collider->next_attempt));

      hash_collider->reset = FALSE;
    }

    g_get_current_time (&before);
    g_mutex_unlock (hash_collider->mutex);

    for (i = 0; i < ATTEMPTS_PER_CYCLE; i++)
    {
      gint n_bits;

      n_bits = collide_value (hash_collider->target_hash, hash_collider->next_attempt);
      if (n_bits >= TNT_HASH_COLLISIONS_MIN_BITS)
      {
        g_mutex_lock (hash_collider->mutex);
        insert_value (hash_collider, hash_collider->next_attempt, n_bits);
        g_mutex_unlock (hash_collider->mutex);
      }

      increment_value (hash_collider->next_attempt);
    }

    g_mutex_lock (hash_collider->mutex);
    g_get_current_time (&after);

    usec_per_cycle = (after.tv_sec - before.tv_sec) * G_USEC_PER_SEC + (after.tv_usec - before.tv_usec);

    /* Save state regularly */

    now = tnt_timestamp ();
    if (now < last_save || now > last_save + SAVE_INTERVAL_S)
    {
      save_state (hash_collider);
      last_save = now;
    }
  }

  g_mutex_unlock (hash_collider->mutex);
  return NULL;
}

void
tnt_hash_collider_finalize (FluxObject *object)
{
  TntHashCollider *hash_collider = TNT_HASH_COLLIDER (object);

  assert_types (hash_collider);

  g_mutex_lock (hash_collider->mutex);
  hash_collider->running = FALSE;
  g_cond_signal (hash_collider->cond);
  g_mutex_unlock (hash_collider->mutex);

  g_thread_join (hash_collider->worker_thread);
  g_mutex_free (hash_collider->mutex);
  g_cond_free (hash_collider->cond);

  flux_object_unref (FLUX_OBJECT (hash_collider->storage));
}

void
tnt_hash_collider_init (TntHashCollider *hash_collider, TntStorage *storage)
{
  /* FIXME: Should really check the storage != NULL here, and return a boolean */

  flux_object_init              (FLUX_OBJECT (hash_collider));
  flux_object_set_type          (FLUX_OBJECT (hash_collider), TNT_HASH_COLLIDER_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (hash_collider), tnt_hash_collider_finalize);

  if (!g_thread_supported ())
    g_thread_init (NULL);

  hash_collider->storage = storage;
  flux_object_ref (FLUX_OBJECT (hash_collider->storage));

  load_state (hash_collider);

  hash_collider->running       = TRUE;
  hash_collider->work_rate     = 0.0;
  hash_collider->mutex         = g_mutex_new ();
  hash_collider->cond          = g_cond_new ();
  hash_collider->worker_thread = g_thread_create_full ((GThreadFunc) worker_main, hash_collider,
                                                       THREAD_STACK_SIZE, TRUE, TRUE,
                                                       G_THREAD_PRIORITY_LOW, NULL);
}

TntHashCollider *
tnt_hash_collider_new (TntStorage *storage)
{
  TntHashCollider *hash_collider;

  g_return_val_if_fail (storage != NULL, NULL);

  hash_collider = g_new0 (TntHashCollider, 1);
  tnt_hash_collider_init (hash_collider, storage);

  return hash_collider;
}

gfloat
tnt_hash_collider_get_work_rate (TntHashCollider *hash_collider)
{
  assert_types (hash_collider);

  return hash_collider->work_rate;
}

void
tnt_hash_collider_set_work_rate (TntHashCollider *hash_collider, gfloat work_rate)
{
  assert_types (hash_collider);

  g_mutex_lock (hash_collider->mutex);

  hash_collider->work_rate = work_rate;

  g_cond_signal (hash_collider->cond);
  g_mutex_unlock (hash_collider->mutex);
}

guint64
tnt_hash_collider_get_target (TntHashCollider *hash_collider)
{
  assert_types (hash_collider);

  return GUINT64_FROM_LE (hash_collider->target);
}

void
tnt_hash_collider_set_target (TntHashCollider *hash_collider, guint64 target)
{
  guint64 target_le;

  assert_types (hash_collider);

  target_le = GUINT64_TO_LE (target);
  if (target_le == hash_collider->target)
    return;

  g_mutex_lock (hash_collider->mutex);

  hash_collider->target = target_le;
  construct_target_hash (hash_collider);
  hash_collider->reset = TRUE;

  g_mutex_unlock (hash_collider->mutex);

  /* TODO: Invalidate current collisions */
}

void
tnt_hash_collider_get_collisions (TntHashCollider *hash_collider,
                                  TntHashCollisions *collisions)
{
  assert_types (hash_collider);

  g_mutex_lock (hash_collider->mutex);
  memcpy (collisions, &hash_collider->target_collisions, sizeof (TntHashCollisions));
  g_mutex_unlock (hash_collider->mutex);
}

void
tnt_hash_collider_verify_collisions (TntHashCollider *hash_collider,
                                     guint64 target, TntHashCollisions *collisions,
                                     gint *n_minimal_collisions, gint *n_extra_bits)
{
  guint64 target_le;
  guint8  target_hash [20];
  gint    n_minimal_collisions_local = 0;
  gint    n_extra_bits_local = 0;
  gint    i;

  assert_types (hash_collider);
  g_return_if_fail (collisions != NULL);

  target_le = GUINT64_TO_LE (target);
  *n_minimal_collisions = 0;
  *n_extra_bits = 0;

  flux_rmd160_hash_single (target_hash, &target_le, sizeof (target_le));

  for (i = 0; i < TNT_HASH_COLLISIONS_NUM; i++)
  {
    gint n_bits;
    gint j;

    /* Skip duplicates by comparing with previous values */

    for (j = 0; j < i; j++)
    {
      if (!memcmp (collisions->v [i], collisions->v [j], TNT_HASH_COLLISIONS_LEN))
        break;
    }

    if (j != i)
      continue;

    n_bits = collide_value (target_hash, collisions->v [i]);

    if (n_bits >= TNT_HASH_COLLISIONS_MIN_BITS)
    {
      n_bits = MIN (n_bits, TNT_HASH_COLLISIONS_MAX_BITS);

      n_minimal_collisions_local++;
      n_extra_bits_local += n_bits - TNT_HASH_COLLISIONS_MIN_BITS;
    }
  }

  if (n_minimal_collisions)
    *n_minimal_collisions = n_minimal_collisions_local;
  if (n_extra_bits)
    *n_extra_bits = n_extra_bits_local;
}
