/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-match.c - Describes a match for a query.
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"

#include <string.h>
#include "tnt.h"

#define DEBUG_MATCH_MEMORY 1

#ifndef DEBUG_MATCH_MEMORY
static FluxMemChunk *match_mem_chunk = NULL;
#endif

#if 0

static gint
count_set_bits (guint32 w)
{
  gint i, n = 0;

  for (i = 0; i < 32; i++)
  {
    n += (w >> i) & 1;
  }

  return n;
}

#endif

TntTimeStamp
tnt_match_determine_grace_timeout (TntTimeStamp seen_time)
{
  TntTimeStamp update_time;

  if (seen_time % TNT_MATCH_GRACE_INTERVAL > TNT_MATCH_GRACE_INTERVAL / 2)
  {
    update_time = seen_time - (seen_time % TNT_MATCH_GRACE_INTERVAL) +
      TNT_MATCH_GRACE_INTERVAL * 2;
  }
  else
  {
    update_time = seen_time - (seen_time % TNT_MATCH_GRACE_INTERVAL) +
      TNT_MATCH_GRACE_INTERVAL;
  }

  return update_time;
}

TntMatch *
tnt_match_new (void)
{
  TntMatch *match;

#ifndef DEBUG_MATCH_MEMORY
  if (!match_mem_chunk)
    match_mem_chunk = flux_mem_chunk_new (256, sizeof (TntMatch));

  match = flux_mem_chunk_alloc0 (match_mem_chunk);
#else
  match = g_new0 (TntMatch, 1);
#endif

  match->first_seen = tnt_timestamp ();

  return match;
}

TntMatch *
tnt_match_dup (TntMatch *match)
{
  TntMatch *duplicate;

  duplicate = tnt_match_new ();
  memcpy (duplicate, match, sizeof (TntMatch));

  duplicate->parent = NULL;
  if (duplicate->file_meta)
    duplicate->file_meta = tnt_file_meta_dup (duplicate->file_meta);
  if (tnt_match_is_group (duplicate))
    duplicate->u.group.children = g_ptr_array_new ();

  return duplicate;
}

static void
free_match (TntMatch *match)
{
  if (match->file_meta)
    tnt_file_meta_free (match->file_meta);

  if (tnt_match_is_group (match))
  {
    GPtrArray *children;
    gint       i;

    children = match->u.group.children;
    g_assert (children);

    for (i = 0; i < children->len; i++)
      free_match (g_ptr_array_index (children, i));

    g_ptr_array_free (children, TRUE);
  }

#ifndef DEBUG_MATCH_MEMORY
  flux_mem_chunk_free (match_mem_chunk, match);
#else
  memset (match, 0, sizeof (*match));
  g_free (match);
#endif
}

void
tnt_match_free (TntMatch *match)
{
  g_return_if_fail (match != NULL);

  /* Remove from parent's list */
  if (match->parent)
  {
    g_assert (match->parent->u.group.children);
    g_ptr_array_remove (match->parent->u.group.children, match);
  }

  free_match (match);
}

void
tnt_match_reclaim_memory (void)
{
#ifndef DEBUG_MATCH_MEMORY
  if (match_mem_chunk)
    flux_mem_chunk_clean (match_mem_chunk);
#endif
}

void
tnt_match_convert_to_group (TntMatch *match)
{
  g_return_if_fail (match != NULL);
  g_return_if_fail (tnt_match_is_file (match));

  match->flags               |= TNT_MATCH_IS_GROUP;
  match->flags               &= ~TNT_MATCH_DATA_IS_LOCAL;

  match->u.group.children         = g_ptr_array_new ();
  match->u.group.total_blocks     = 0;
  match->u.group.hit_blocks       = 0;
  match->u.group.hit_blocks_sum   = 0;
  match->u.group.local_files      = 0;
  match->u.group.quality_children = 0;
  match->u.group.dimension_ban    = FALSE;
}

static gint
get_max_blocks_avail (TntMatch *file_match)
{
  gint max = 0, i;

  g_return_val_if_fail (tnt_match_is_file (file_match), 0);

  for (i = 0; i < TNT_MATCH_HIT_BUCKETS; i++)
    max = MAX (max, file_match->u.file.avail [i]);

  return max;
}

static gint
get_sum_blocks_avail (TntMatch *file_match)
{
  gint i, n = 0;

  g_return_val_if_fail (tnt_match_is_file (file_match), 0);

  for (i = 0; i < TNT_MATCH_HIT_BUCKETS; i++)
    n += file_match->u.file.avail [i];

  return n;
}

#if 0

static gint
get_num_buckets_used (TntMatch *file_match)
{
  gint i, n = 0;

  g_return_val_if_fail (tnt_match_is_file (file_match), 0);

  for (i = 0; i < TNT_MATCH_HIT_BUCKETS; i++)
  {
    if (file_match->u.file.avail [i] > 0)
      n++;
  }

  return n;
}

#endif

static void
update_total_blocks (TntMatch *group, gint num_blocks)
{
  g_return_if_fail (tnt_match_is_group (group));

  group->u.group.total_blocks += num_blocks;
  if (group->parent)
    update_total_blocks (group->parent, num_blocks);
}

static void
update_hit_blocks_sum (TntMatch *group, gint num_blocks)
{
  g_return_if_fail (tnt_match_is_group (group));

  group->u.group.hit_blocks_sum += num_blocks;
  if (group->parent)
    update_hit_blocks_sum (group->parent, num_blocks);
}

static void
update_local_file_count (TntMatch *group, gint num_files)
{
  g_return_if_fail (tnt_match_is_group (group));

  group->u.group.local_files += num_files;
  if (group->parent)
    update_local_file_count (group->parent, num_files);
}

gint
tnt_match_get_n_quality_children (TntMatch *group)
{
  g_return_val_if_fail (group != NULL, 0);
  g_return_val_if_fail (tnt_match_is_group (group), 0);

  return group->u.group.quality_children;
}

void
tnt_match_inc_quality_children (TntMatch *group)
{
  g_return_if_fail (group != NULL);
  g_return_if_fail (tnt_match_is_group (group));

  group->u.group.quality_children++;
}

void
tnt_match_dec_quality_children (TntMatch *group)
{
  g_return_if_fail (group != NULL);
  g_return_if_fail (tnt_match_is_group (group));

  if (!group->u.group.quality_children)
    return;

  group->u.group.quality_children--;
}

gboolean
tnt_match_get_dimension_ban (TntMatch *group)
{
  g_return_val_if_fail (group != NULL, FALSE);
  g_return_val_if_fail (tnt_match_is_group (group), FALSE);

  return group->u.group.dimension_ban ? TRUE : FALSE;
}

void
tnt_match_set_dimension_ban (TntMatch *group, gboolean dimension_ban)
{
  g_return_if_fail (group != NULL);
  g_return_if_fail (tnt_match_is_group (group));

  group->u.group.dimension_ban = dimension_ban;
}

void
tnt_match_add_child (TntMatch *parent, TntMatch *child)
{
  g_return_if_fail (parent != NULL);
  g_return_if_fail (child != NULL);
  g_return_if_fail (tnt_match_is_group (parent));
  g_return_if_fail (child->parent == NULL);

  g_assert (parent->u.group.children);
  g_ptr_array_add (parent->u.group.children, child);

  child->parent = parent;

  if (tnt_match_is_file (child))
  {
    gint total_blocks, hit_blocks, sum_blocks;

    total_blocks = TNT_REAL_LEN_TO_BITMAP_LEN (child->u.file.file_id.len);
    hit_blocks   = get_max_blocks_avail (child);
    sum_blocks   = get_sum_blocks_avail (child);

    update_total_blocks (parent, total_blocks);
    update_hit_blocks_sum (parent, sum_blocks);
    tnt_match_add_hit (parent, TNT_NODE_ID_INVALID, hit_blocks);

    if (tnt_match_is_local (child))
      update_local_file_count (parent, 1);
  }
  else
  {
    update_total_blocks (parent, child->u.group.total_blocks);
    update_hit_blocks_sum (parent, child->u.group.hit_blocks_sum);
    tnt_match_add_hit (parent, TNT_NODE_ID_INVALID, child->u.group.hit_blocks);
  }
}

void
tnt_match_unlink (TntMatch *match)
{
  TntMatch *parent;

  g_return_if_fail (match != NULL);
  g_return_if_fail (match->parent != NULL);

  parent = match->parent;

  g_assert (parent->u.group.children);
  g_ptr_array_remove (parent->u.group.children, match);

  match->parent = NULL;

  if (tnt_match_is_file (match))
  {
    gint total_blocks, hit_blocks, sum_blocks;

    total_blocks = TNT_REAL_LEN_TO_BITMAP_LEN (match->u.file.file_id.len);
    hit_blocks   = get_max_blocks_avail (match);
    sum_blocks   = get_sum_blocks_avail (match);

    update_total_blocks (parent, -total_blocks);
    update_hit_blocks_sum (parent, -sum_blocks);
    tnt_match_add_hit (parent, TNT_NODE_ID_INVALID, -hit_blocks);

    if (tnt_match_is_local (match))
      update_local_file_count (parent, -1);
  }
  else
  {
    update_total_blocks (parent, -match->u.group.total_blocks);
    update_hit_blocks_sum (parent, -match->u.group.hit_blocks_sum);
    tnt_match_add_hit (parent, TNT_NODE_ID_INVALID, -match->u.group.hit_blocks);
  }
}

TntMatch *
tnt_match_get_parent (TntMatch *match)
{
  g_return_val_if_fail (match != NULL, NULL);

  return match->parent;
}

const GPtrArray *
tnt_match_get_children (TntMatch *match)
{
  g_return_val_if_fail (match != NULL, NULL);
  g_return_val_if_fail (tnt_match_is_group (match), NULL);

  return match->u.group.children;
}

gint
tnt_match_get_n_children (TntMatch *match)
{
  g_return_val_if_fail (match != NULL, 0);
  g_return_val_if_fail (tnt_match_is_group (match), 0);

  g_assert (match->u.group.children != NULL);
  return match->u.group.children->len;
}

gint
tnt_match_get_index (TntMatch *match)
{
  const GPtrArray *siblings;
  gint             i;

  g_return_val_if_fail (match != NULL, 0);

  if (!match->parent)
    return 0;

  siblings = tnt_match_get_children (match->parent);
  if (!siblings)
    return 0;

  for (i = 0; i < siblings->len; i++)
  {
    if (g_ptr_array_index (siblings, i) == match)
      break;
  }

  if (i == siblings->len)
  {
    g_warning ("Submatch does not appear in parent list!");
    return 0;
  }

  return i;
}

static TntMatch *
for_all_matches (TntMatch *match, TntMatchTraversalFunc *func, gpointer data)
{
  /* Infix traversal */

  if (func (match, data))
    return match;

  if (tnt_match_is_group (match))
  {
    const GPtrArray *children;
    gint             i;

    children = tnt_match_get_children (match);
    for (i = 0; i < children->len; i++)
    {
      TntMatch *child_match = g_ptr_array_index (children, i);
      TntMatch *terminal_match;

      terminal_match = for_all_matches (child_match, func, data);
      if (terminal_match)
        return terminal_match;
    }
  }

  return NULL;
}

static void
subtract_from_stats (TntMatch *match)
{
  TntMatch *parent;

  if (!tnt_match_is_file (match))
    return;

  parent = match->parent;
  if (!parent)
    return;

  update_total_blocks (parent, -(TNT_REAL_LEN_TO_BITMAP_LEN (match->u.file.file_id.len)));
  update_hit_blocks_sum (parent, -(get_sum_blocks_avail (match)));
  if (match->flags & TNT_MATCH_DATA_IS_LOCAL)
    update_local_file_count (parent, -1);
  tnt_match_add_hit (parent, TNT_NODE_ID_INVALID, -(get_max_blocks_avail (match)));
}

static void
for_all_matches_remove (TntMatch *match, TntMatchTraversalFunc *func, gpointer data)
{
  /* Postfix traversal */

  if (tnt_match_is_group (match))
  {
    GPtrArray *children;
    gint       i;

    children = (GPtrArray *) tnt_match_get_children (match);
    for (i = children->len - 1; i >= 0; i--)
    {
      TntMatch *child_match = g_ptr_array_index (children, i);

      if (tnt_match_is_group (child_match))
      {
        for_all_matches_remove (child_match, func, data);
      }

      /* We remove children when called for their parent, not in their
       * own instance, since we have both the index and match pointer. */
      if (func (child_match, data))
      {
        subtract_from_stats (child_match);
        free_match (child_match);
        g_ptr_array_remove_index (children, i);
      }
    }
  }
}

TntMatch *
tnt_match_for_all (TntMatch *match, TntMatchTraversalFunc *func, gpointer data)
{
  g_return_val_if_fail (tnt_match_is_group (match), NULL);
  g_return_val_if_fail (func != NULL, NULL);

  return for_all_matches (match, func, data);
}

gboolean
tnt_match_for_all_remove (TntMatch *match, TntMatchTraversalFunc *func, gpointer data)
{
  g_return_val_if_fail (func != NULL, FALSE);

  for_all_matches_remove (match, func, data);
  if (func (match, data))
  {
    subtract_from_stats (match);
    tnt_match_free (match);
    return TRUE;
  }

  return FALSE;
}

gboolean
tnt_match_add_hit (TntMatch *match, TntNodeID node_id, gint num_blocks)
{
  gint     parent_hit_change = 0;
  gboolean changed = FALSE;

  g_return_val_if_fail (match != NULL, FALSE);

  if (tnt_match_is_file (match))
  {
    gint    bucket_n;
    gint    old_max;

    g_return_val_if_fail (node_id != TNT_NODE_ID_INVALID, FALSE);

    old_max  = get_max_blocks_avail (match);
    bucket_n = node_id % TNT_MATCH_HIT_BUCKETS;

    if (num_blocks > match->u.file.avail [bucket_n])
    {
      if (match->parent)
        update_hit_blocks_sum (match->parent, num_blocks - match->u.file.avail [bucket_n]);
      match->u.file.avail [bucket_n] = num_blocks;
      changed = TRUE;
    }

    if (num_blocks > old_max)
      parent_hit_change = num_blocks - old_max;
  }
  else /* group */
  {
    match->u.group.hit_blocks += num_blocks;
    parent_hit_change = num_blocks;
    changed = TRUE;
  }

  if (parent_hit_change > 0 && match->parent)
  {
    tnt_match_add_hit (match->parent, node_id, parent_hit_change);
  }

  return changed;
}

TntAvailability
tnt_match_get_availability (TntMatch *match)
{
  TntTimeStamp    now;
  TntAvailability avail;
  gint            hit_blocks;
  gint            hits;

  g_return_val_if_fail (match != NULL, TNT_AVAILABILITY_UNKNOWN);

  now = tnt_timestamp ();

  if (tnt_match_is_file (match))
  {
    hit_blocks = get_sum_blocks_avail (match);
    hits = hit_blocks / TNT_REAL_LEN_TO_BITMAP_LEN (match->u.file.file_id.len);
    if (hit_blocks > 0 && hits == 0)
      hits = 1;
  }
  else
  {
    if (match->u.group.children->len == 0)
      return TNT_AVAILABILITY_UNKNOWN;

    hits = (match->u.group.hit_blocks_sum + match->u.group.total_blocks - 1) /
      match->u.group.total_blocks;
  }

  if (hits == 0)
  {
    if (now < tnt_match_determine_grace_timeout (match->first_seen))
      avail = TNT_AVAILABILITY_UNKNOWN;
    else
      avail = TNT_AVAILABILITY_NONE;
  }
  else if (hits < 2)
    avail = TNT_AVAILABILITY_VERY_LOW;
  else if (hits < 5)
    avail = TNT_AVAILABILITY_LOW;
  else if (hits < 10)
    avail = TNT_AVAILABILITY_MEDIUM;
  else if (hits < 18)
    avail = TNT_AVAILABILITY_HIGH;
  else
    avail = TNT_AVAILABILITY_VERY_HIGH;

  return avail;
}

TntCompleteness
tnt_match_get_completeness (TntMatch *match)
{
  TntCompleteness comp = TNT_COMPLETENESS_NONE;
  TntTimeStamp    now;

  g_return_val_if_fail (match != NULL, TNT_COMPLETENESS_NONE);

  if (tnt_match_is_file (match))
  {
    gint n;

    n = get_max_blocks_avail (match);

    if (n >= TNT_REAL_LEN_TO_BITMAP_LEN (match->u.file.file_id.len))
      comp = TNT_COMPLETENESS_COMPLETE;
    else if (n > 0)
      comp = TNT_COMPLETENESS_PARTIAL;
  }
  else /* group */
  {
    if (match->u.group.hit_blocks >= match->u.group.total_blocks)
      comp = TNT_COMPLETENESS_COMPLETE;
    else if (match->u.group.hit_blocks > 0)
      comp = TNT_COMPLETENESS_PARTIAL;
  }

  now = tnt_timestamp ();

  if (comp == TNT_COMPLETENESS_NONE &&
      now < tnt_match_determine_grace_timeout (match->first_seen))
    return TNT_COMPLETENESS_UNKNOWN;

  return comp;
}

const TntFileID *
tnt_match_get_file_id (TntMatch *match)
{
  g_return_val_if_fail (match != NULL, NULL);
  g_return_val_if_fail (tnt_match_is_file (match), NULL);

  return &match->u.file.file_id;
}

void
tnt_match_set_file_id (TntMatch *match, const TntFileID *file_id)
{
  g_return_if_fail (match != NULL);
  g_return_if_fail (file_id != NULL);
  g_return_if_fail (tnt_match_is_file (match));

  memcpy (&match->u.file.file_id, file_id, sizeof (TntFileID));
}

TntFileMeta *
tnt_match_get_file_meta (TntMatch *match)
{
  g_return_val_if_fail (match != NULL, NULL);

  return match->file_meta;
}

void
tnt_match_set_local (TntMatch *match, gboolean local)
{
  g_return_if_fail (match != NULL);
  g_return_if_fail (tnt_match_is_file (match));

  if (tnt_match_is_local (match) && local == FALSE)
  {
    match->flags &= ~TNT_MATCH_DATA_IS_LOCAL;

    if (match->parent)
      update_local_file_count (match->parent, -1);
  }
  else if (!tnt_match_is_local (match) && local == TRUE)
  {
    match->flags |= TNT_MATCH_DATA_IS_LOCAL;

    if (match->parent)
      update_local_file_count (match->parent, 1);
  }
}

gboolean
tnt_match_is_local (TntMatch *match)
{
  g_return_val_if_fail (match != NULL, FALSE);

  if (tnt_match_is_file (match))
  {
    if (match->flags & TNT_MATCH_DATA_IS_LOCAL)
      return TRUE;
  }
  else  /* group */
  {
    if (match->u.group.local_files + (match->u.group.children->len / 10) >= match->u.group.children->len)
      return TRUE;
  }

  return FALSE;
}
