/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-message-filter.c - Message filter.
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "tnt.h"
#include <time.h>

#define DEBUG(x)

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), TNT_MESSAGE_FILTER_TYPE)

#define DELTA_TIME_SECONDS          30
#define CLEANUP_INTERVAL_ITERATIONS 10000
#define CRC_POLYNOMIAL              0x04c11db7L

static guint32  crc_table [256];
static gboolean crc_table_init = FALSE;

/* Generate the table of CRC remainders for all possible bytes */
static void
generate_crc_table (void)
{
  guint32 crc_accum;
  gint    i, j;

  for (i = 0; i < 256; i++)
  {
    crc_accum = (guint32) i << 24;
    for (j = 0; j < 8; j++)
    {
      if (crc_accum & 0x80000000L)
        crc_accum = (crc_accum << 1) ^ CRC_POLYNOMIAL;
      else
        crc_accum = (crc_accum << 1);
    }
    crc_table [i] = crc_accum;
  }
}

/* Update the CRC on the data block one byte at a time */
static guint32
update_crc (guint32 crc_accum, gpointer data, gint data_len)
{
  gint    i, j;
  guchar *data_blk_ptr = data;

  for (j = 0; j < data_len; j++)
  {
    i = ((gint) (crc_accum >> 24) ^ *(data_blk_ptr++)) & 0xff;
    crc_accum = (crc_accum << 8) ^ crc_table [i];
  }

  return crc_accum;
}

static guint32
checksum_message (TntMessage *message)
{
  guint32        crc = 0;
  TntMessageType type;
  gpointer       payload;
  guint          payload_len;

  type = tnt_message_get_type (message);
  crc = update_crc (crc, &type, sizeof (type));

  tnt_message_get_payload (message, &payload, &payload_len);
  if (payload)
    crc = update_crc (crc, payload, payload_len);

  if (flux_object_get_type (FLUX_OBJECT (message)) == TNT_ROUTED_MESSAGE_TYPE)
  {
    TntNodeID node_id;

    if (tnt_routed_message_has_source (TNT_ROUTED_MESSAGE (message)))
    {
      node_id = tnt_routed_message_get_source (TNT_ROUTED_MESSAGE (message));
      crc = update_crc (crc, &node_id, sizeof (node_id));
    }

    if (tnt_routed_message_has_destination (TNT_ROUTED_MESSAGE (message)))
    {
      node_id = tnt_routed_message_get_destination (TNT_ROUTED_MESSAGE (message));
      crc = update_crc (crc, &node_id, sizeof (node_id));
    }
  }

  return crc;
}

static gboolean
cleanup_remove_cb (gpointer key, gpointer value, gpointer data)
{
  guint now       = GPOINTER_TO_UINT (data);
  guint last_seen = GPOINTER_TO_UINT (value);

  if (last_seen + DELTA_TIME_SECONDS > now)
    return FALSE;  /* Keep */

  return TRUE;  /* Remove */
}

static void
table_cleanup (TntMessageFilter *filter)
{
  guint now;

  now = time (NULL);
  g_hash_table_foreach_remove (filter->table, cleanup_remove_cb, GUINT_TO_POINTER (now));
}

void
tnt_message_filter_finalize (FluxObject *object)
{
  TntMessageFilter *filter = TNT_MESSAGE_FILTER (object);

  assert_types (filter);

  g_hash_table_destroy (filter->table);
}

void
tnt_message_filter_init (TntMessageFilter *filter)
{
  flux_object_init              (FLUX_OBJECT (filter));
  flux_object_set_type          (FLUX_OBJECT (filter), TNT_MESSAGE_FILTER_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (filter), tnt_message_filter_finalize);

  /* Not threadsafe and all that */
  if (!crc_table_init)
  {
    crc_table_init = TRUE;
    generate_crc_table ();
  }

  filter->table = g_hash_table_new (g_direct_hash, g_direct_equal);
  filter->iterations_since_cleanup = 0;
}

TntMessageFilter *
tnt_message_filter_new (void)
{
  TntMessageFilter *filter;

  filter = g_new0 (TntMessageFilter, 1);
  tnt_message_filter_init (filter);

  return filter;
}

gboolean
tnt_message_filter_allow_message (TntMessageFilter *filter, TntMessage *message)
{
  guint    now, last_blocked;
  guint32  checksum;
  gpointer value;
  gboolean result;

  now = time (NULL);
  checksum = checksum_message (message);

  value = g_hash_table_lookup (filter->table, GUINT_TO_POINTER (checksum));
  last_blocked = GPOINTER_TO_UINT (value);

  if (value != NULL && last_blocked + DELTA_TIME_SECONDS > now &&
      last_blocked <= now)
  {
    result = FALSE;
  }
  else
  {
    /* The (last_blocked <= now) in the previous condition ensures that we get here
     * if time moved backwards */

    g_hash_table_insert (filter->table, GUINT_TO_POINTER (checksum), GUINT_TO_POINTER (now));
    result = TRUE;
  }

  filter->iterations_since_cleanup++;
  if (filter->iterations_since_cleanup > CLEANUP_INTERVAL_ITERATIONS)
  {
    table_cleanup (filter);
    filter->iterations_since_cleanup = 0;
  }

  return result;
}
