/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-message.c - Network message.
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "tnt.h"
#include <time.h>

#define DEBUG(x)

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), TNT_MESSAGE_TYPE)

static FluxMarshaller *message_marshaller = NULL;
static GHashTable     *type_table         = NULL;

static const gchar *
get_name_for_type (TntMessageType type)
{
  if (type == 0)
    return NULL;

  return g_hash_table_lookup (type_table, GUINT_TO_POINTER (type));
}

gint
tnt_message_register_marshal_info (FluxMarshalledStruct *marshal_info)
{
  g_return_val_if_fail (marshal_info != NULL, 0);

  if (!message_marshaller)
  {
    message_marshaller = flux_marshaller_new ();
    type_table = g_hash_table_new_full (g_direct_hash, g_direct_equal,
                                        NULL, (GDestroyNotify) g_free);
  }

  return flux_marshaller_register_many (message_marshaller, marshal_info);
}

gboolean
tnt_message_register_type (TntMessageType type, const gchar *marshal_name)
{
  g_return_val_if_fail (type > 0, FALSE);
  g_return_val_if_fail (marshal_name != NULL, FALSE);

  if (!message_marshaller)
  {
    message_marshaller = flux_marshaller_new ();
    type_table = g_hash_table_new_full (g_direct_hash, g_direct_equal,
                                        NULL, (GDestroyNotify) g_free);
  }

  if (g_hash_table_lookup (type_table, GUINT_TO_POINTER (type)))
    return FALSE;

  g_hash_table_insert (type_table, GUINT_TO_POINTER (type), g_strdup (marshal_name));
  return TRUE;
}

void
tnt_message_finalize (FluxObject *message)
{
  assert_types (message);

  DEBUG
  (
    g_print ("[destroy] ");
    tnt_message_dump (TNT_MESSAGE (message));
  )

  if (TNT_MESSAGE (message)->destroy_func)
    TNT_MESSAGE (message)->destroy_func (message, TNT_MESSAGE (message)->destroy_func_data);

  if (TNT_MESSAGE (message)->payload)
    g_free (TNT_MESSAGE (message)->payload);

  if (TNT_MESSAGE (message)->data_block)
    tnt_free_block (TNT_MESSAGE (message)->data_block);
}

void
tnt_message_init (TntMessage *message, TntMessageType type)
{
  flux_object_init              (FLUX_OBJECT (message));
  flux_object_set_type          (FLUX_OBJECT (message), TNT_MESSAGE_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (message), tnt_message_finalize);

  message->type       = type;
  message->ctime      = time (NULL);
  message->ttl        = 0;
  message->payload    = NULL;
  message->data_block = NULL;

  if (!message_marshaller)
  {
    message_marshaller = flux_marshaller_new ();
    type_table = g_hash_table_new_full (g_direct_hash, g_direct_equal,
                                        NULL, (GDestroyNotify) g_free);
  }
}

TntMessage *
tnt_message_new (TntMessageType type)
{
  TntMessage *message;

  message = g_new0 (TntMessage, 1);
  tnt_message_init (message, type);

  return message;
}

TntMessageType
tnt_message_get_type (TntMessage *message)
{
  assert_types (message);

  return message->type;
}

guint8
tnt_message_get_ttl (TntMessage *message)
{
  assert_types (message);

  return message->ttl;
}

guint
tnt_message_get_ctime (TntMessage *message)
{
  assert_types (message);

  return message->ctime;
}

gboolean
tnt_message_get_direct (TntMessage *message)
{
  assert_types (message);

  return message->is_direct;
}

void
tnt_message_get_payload (TntMessage *message, gpointer *payload, guint *len)
{
  assert_types (message);
  g_return_if_fail (payload != NULL);
  g_return_if_fail (len != NULL);

  *payload = message->payload;
  *len     = message->payload_len;
}

void
tnt_message_get_data_block (TntMessage *message, gpointer *data_block, guint *len)
{
  assert_types (message);
  g_return_if_fail (data_block != NULL);
  g_return_if_fail (len != NULL);

  *data_block = message->data_block;
  *len        = message->data_block_len;
}

void
tnt_message_set_ttl (TntMessage *message, guint8 ttl)
{
  assert_types (message);

  message->ttl = ttl;
}

void
tnt_message_set_direct (TntMessage *message, gboolean direct)
{
  assert_types (message);

  message->is_direct = direct;
}

/* Does not copy the data, but assumes ownership of the pointer, which
 * should point to an independently allocated block */
void
tnt_message_set_payload (TntMessage *message, gpointer payload, guint len)
{
  assert_types (message);

  if (message->payload && message->payload != payload)
    g_free (message->payload);

  message->payload     = payload;
  message->payload_len = len;
}

void
tnt_message_set_data_block (TntMessage *message, gpointer data_block, guint len)
{
  assert_types (message);

  if (message->data_block && message->data_block != data_block)
    tnt_free_block (message->data_block);

  message->data_block     = data_block;
  message->data_block_len = len;
} 

gpointer
tnt_message_get_struct (TntMessage *message)
{
  const gchar *name;
  gpointer     struct_ptr;

  assert_types (message);

  if (!message->payload)
    return NULL;

  name = get_name_for_type (message->type);
  if (!name)
    return NULL;

  if (!flux_marshaller_decode_pack (message_marshaller, name, message->payload, message->payload_len,
                                    &struct_ptr))
    return NULL;

  return struct_ptr;
}

/* Copies the struct data (actually, marshals it) */
void
tnt_message_set_struct (TntMessage *message, gpointer struct_ptr)
{
  const gchar *name;
  gpointer     payload;
  guint        payload_len;

  assert_types (message);
  g_return_if_fail (struct_ptr != NULL);

  name = get_name_for_type (message->type);
  if (!name)
  {
    g_warning ("Tried to assign payload to message of unknown type.");
    return;
  }

  if (flux_marshaller_encode_pack (message_marshaller, name, struct_ptr, &payload, &payload_len))
  {
    tnt_message_set_payload (message, payload, payload_len);
  }
  else
  {
    g_warning ("Failed to encode a message payload of type '%s'.", name);
  }
}

void
tnt_message_free_struct (TntMessage *message, gpointer struct_ptr)
{
  const gchar *name;

  assert_types (message);
  g_return_if_fail (struct_ptr != NULL);

  name = get_name_for_type (message->type);
  if (!name)
  {
    g_warning ("Tried to free struct of unknown type.");
    return;
  }

  flux_marshaller_free_struct (message_marshaller, name, struct_ptr);
}

guint
tnt_message_get_size_estimate (TntMessage *message)
{
  return message->data_block_len + message->payload_len + 10;
}

void
tnt_message_set_destroy_func (TntMessage *message, TntDestroyFunc *func, gpointer data)
{
  assert_types (message);

  message->destroy_func      = func;
  message->destroy_func_data = data;
}

gboolean
tnt_message_is_time_warped (TntMessage *message)
{
  assert_types (message);

  return message->is_time_warped ? TRUE : FALSE;
}

void
tnt_message_set_time_warped (TntMessage *message, gboolean is_time_warped)
{
  assert_types (message);

  message->is_time_warped = is_time_warped;
}

void
tnt_message_dump (TntMessage *message)
{
  const gchar *name;

  assert_types (message);

  g_print ("Message: ");
  if (!message)
  {
    g_print ("(NULL)\n");
    return;
  }

  name = get_name_for_type (message->type);
  g_print ("%s ttl=%d %sdata=%d", name ? name : "unknown", message->ttl, message->is_direct ? "[direct] " : "",
           message->data_block ? message->data_block_len : 0);
  g_print ("\n");
}
