/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-meta-db.c - On-disk indexed file metadata.
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"

#include <sys/types.h>
#include <string.h>
#include <fcntl.h>
#include <glib.h>
#include "tnt.h"

#ifdef G_PLATFORM_WIN32
# include "win32/win32-compat.h"
#endif

#define DEBUG(x)
#define CULL_DEBUG(x)
#define MATCH_DEBUG(x)

#define FILE_META_CHUNK_NAME "meta.dat"
#define ID_TO_FILE_META_NAME "meta.idx"
#define STRINGS_FILE_PREFIX  "string"

#define THREAD_STACK_SIZE    65536

/* This is the number of actions that can be queued before we consider it
 * a "critically high" amount. If we exceed this number of pending actions,
 * other work, like running queries, will be stopped until we can bring it
 * back down. */
#define ACTIONS_HIGH_WATER_MARK 1200

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), TNT_META_DB_TYPE)

typedef enum
{
  ACTION_WAKEUP,
  ACTION_GET_ID,
  ACTION_GET_META,
  ACTION_SET_META,
  ACTION_MERGE_META,
  ACTION_EXTRACT_META,
  ACTION_QUERY,
  ACTION_TERMINATE
}
ActionType;

typedef enum
{
  RETURN_GET_ID,
  RETURN_GET_META,
  RETURN_MERGE_META,
  RETURN_EXTRACT_META,
  RETURN_QUERY,
  RETURN_QUERY_DONE
}
ReturnType;

typedef struct
{
  guint type;
}
ActionData;

typedef struct
{
  ActionData     action;
}
WakeupActionData;

typedef struct
{
  ActionData     action;

  gchar         *name;
  guint64        len;
  TntTimeStamp   timestamp;

  TntFileIDFunc *func;
  gpointer       func_data;
}
GetIDActionData;

typedef struct
{
  ActionData       action;

  TntFileID        file_id;

  TntFileMetaFunc *func;
  gpointer         func_data;
}
GetMetaActionData;

typedef struct
{
  ActionData   action;

  TntFileID    file_id;
  TntFileMeta *file_meta;
}
SetMetaActionData;

typedef struct
{
  ActionData       action;

  TntFileID        file_id;
  TntFileMeta     *file_meta;
  gboolean         store_result;

  TntFileMetaFunc *func;
  gpointer         func_data;
}
MergeMetaActionData;

typedef struct
{
  ActionData       action;
  
  TntFileID        file_id;
  gchar           *name;
  gchar           *path;
  gchar           *relative_path;

  TntFileMetaFunc *func;
  gpointer         func_data;
}
ExtractMetaActionData;

typedef struct
{
  ActionData            action;

  TntQuery             *query;
  TntStringIndexCursor *cursor;
  guint                 paused : 1;

  TntQueryDoneFunc     *func;
  gpointer              func_data;
}
QueryActionData;

typedef struct
{
  guint type;
}
ReturnData;

typedef struct
{
  ReturnData     return_data;

  gboolean       found;
  TntFileID      file_id;

  TntFileIDFunc *func;
  gpointer       func_data;
}
GetIDReturnData;

typedef struct
{
  ReturnData       return_data;

  TntFileID        file_id;
  TntFileMeta     *file_meta;

  TntFileMetaFunc *func;
  gpointer         func_data;
}
GetMetaReturnData;

typedef struct
{
  ReturnData       return_data;

  TntFileID        file_id;
  TntFileMeta     *file_meta;

  TntFileMetaFunc *func;
  gpointer         func_data;
}
MergeMetaReturnData;

typedef MergeMetaReturnData ExtractMetaReturnData;

typedef struct
{
  ReturnData   return_data;

  TntQuery    *query;
  TntFileID    file_id;
  TntFileMeta *file_meta;
}
QueryReturnData;

typedef struct
{
  ReturnData        return_data;

  TntQuery         *query;
  TntQueryDoneFunc *func;
  gpointer          func_data;
}
QueryDoneReturnData;

typedef struct
{
  TntTimeStamp    timestamp;

  FluxItem        name;
  FluxItem        title;
  FluxItem        collection;
  FluxItem        performer;
  FluxItem        author;
  FluxItem        keywords;
  FluxItem        seq;
  guint32         duration;
  guint16         width;
  guint16         height;
  guint16         fragment;
  guint16         max_fragment;
  guint16         quality;
  guint16         format;
  gint16          year;
  guint8          month;
  guint8          day;
  guint8          lang [2];
  guint8          cat  [2];
  guint8          media;
  guint8          flags;
}
CodedMeta;

static guint
get_used (TntMetaDB *meta_db)
{
  return flux_storage_get_used_size     (meta_db->string_storage)          +
         flux_storage_get_used_size     (meta_db->id_to_file_meta_storage) +
         tnt_string_index_get_file_size (meta_db->string_to_id_index)      -
         tnt_string_index_get_free_size (meta_db->string_to_id_index);
}

static gchar *
load_string (FluxStorage *storage, FluxItem item)
{
  gchar *str;
  gint   len;

  if (item == FLUX_ITEM_NULL)
    return NULL;

  len = flux_storage_get_item_len (storage, item);
  if (!len)
    return NULL;

  str = g_malloc (len + 1);

  if (!flux_storage_read (storage, item, str))
  {
    g_free (str);
    return NULL;
  }

  str [len] = '\0';
  return str;
}

static FluxItem
save_string (FluxStorage *storage, gchar *str)
{
  FluxItem item;
  gint     len;

  if (!str)
    return FLUX_ITEM_NULL;

  len = strlen (str);
  if (!len)
    return FLUX_ITEM_NULL;

  item = flux_storage_alloc (storage, len);
  if (item == FLUX_ITEM_NULL)
    return FLUX_ITEM_NULL;

  flux_storage_write (storage, item, str);
  return item;
}

static gboolean
free_string (FluxStorage *storage, FluxItem item)
{
  if (item == FLUX_ITEM_NULL)
    return TRUE;

  return flux_storage_free (storage, item);
}

static FluxItem
ensure_string (FluxStorage *storage, gchar *old, gchar *new, FluxItem old_item)
{
  if (old == NULL)
  {
    if (new == NULL)
      return old_item;

    return save_string (storage, new);
  }

  if (new == NULL)
  {
    free_string (storage, old_item);
    return FLUX_ITEM_NULL;
  }

  if (strcmp (old, new))
  {
    free_string (storage, old_item);
    return save_string (storage, new);
  }

  return old_item;
}

/* Warning: Pokes the TntFileMeta struct */
static TntFileMeta *
decode_file_meta (TntMetaDB *meta_db, CodedMeta *coded)
{
  TntFileMeta *file_meta;

  file_meta = tnt_file_meta_new ();

  file_meta->timestamp    = coded->timestamp;

  file_meta->name         = load_string (meta_db->string_storage, coded->name);
  file_meta->title        = load_string (meta_db->string_storage, coded->title);
  file_meta->collection   = load_string (meta_db->string_storage, coded->collection);
  file_meta->performer    = load_string (meta_db->string_storage, coded->performer);
  file_meta->author       = load_string (meta_db->string_storage, coded->author);
  file_meta->keywords     = load_string (meta_db->string_storage, coded->keywords);
  file_meta->seq          = load_string (meta_db->string_storage, coded->seq);

  file_meta->fragment     = coded->fragment;
  file_meta->max_fragment = coded->max_fragment;
  file_meta->quality      = coded->quality;
  file_meta->format       = coded->format;
  file_meta->year         = coded->year;
  file_meta->month        = coded->month;
  file_meta->day          = coded->day;
  file_meta->width        = coded->width;
  file_meta->height       = coded->height;
  file_meta->duration     = coded->duration;
  file_meta->lang [0]     = coded->lang [0];
  file_meta->lang [1]     = coded->lang [1];
  file_meta->cat  [0]     = coded->cat  [0];
  file_meta->cat  [1]     = coded->cat  [1];
  file_meta->media        = coded->media;
  file_meta->flags        = coded->flags;

  return file_meta;
}

/* Warning: Peeks the TntFileMeta struct */
static CodedMeta *
encode_file_meta (TntMetaDB *meta_db, TntFileMeta *file_meta)
{
  CodedMeta *coded;

  coded = g_new (CodedMeta, 1);

  coded->timestamp    = file_meta->timestamp;

  coded->name         = save_string (meta_db->string_storage, file_meta->name);
  coded->title        = save_string (meta_db->string_storage, file_meta->title);
  coded->collection   = save_string (meta_db->string_storage, file_meta->collection);
  coded->performer    = save_string (meta_db->string_storage, file_meta->performer);
  coded->author       = save_string (meta_db->string_storage, file_meta->author);
  coded->keywords     = save_string (meta_db->string_storage, file_meta->keywords);
  coded->seq          = save_string (meta_db->string_storage, file_meta->seq);

  coded->fragment     = file_meta->fragment;
  coded->max_fragment = file_meta->max_fragment;
  coded->quality      = file_meta->quality;
  coded->format       = file_meta->format;
  coded->year         = file_meta->year;
  coded->month        = file_meta->month;
  coded->day          = file_meta->day;
  coded->width        = file_meta->width;
  coded->height       = file_meta->height;
  coded->duration     = file_meta->duration;
  coded->lang [0]     = file_meta->lang [0];
  coded->lang [1]     = file_meta->lang [1];
  coded->cat  [0]     = file_meta->cat  [0];
  coded->cat  [1]     = file_meta->cat  [1];
  coded->media        = file_meta->media;
  coded->flags        = file_meta->flags;

  return coded;
}

static void
release_file_meta (TntMetaDB *meta_db, CodedMeta *coded)
{
  free_string (meta_db->string_storage, coded->name);
  free_string (meta_db->string_storage, coded->title);
  free_string (meta_db->string_storage, coded->collection);
  free_string (meta_db->string_storage, coded->performer);
  free_string (meta_db->string_storage, coded->author);
  free_string (meta_db->string_storage, coded->keywords);
  free_string (meta_db->string_storage, coded->seq);
}

/* Warning: Peeks and pokes the TntFileMeta struct */
static CodedMeta *
reencode_file_meta (TntMetaDB *meta_db, TntFileMeta *old_meta, TntFileMeta *new_meta, CodedMeta *old_coded)
{
  CodedMeta *new_coded;

  new_coded = g_new (CodedMeta, 1);
  memcpy (new_coded, old_coded, sizeof (CodedMeta));

  new_coded->timestamp  = new_meta->timestamp;

  new_coded->name       = ensure_string (meta_db->string_storage, old_meta->name,
                                         new_meta->name, old_coded->name);
  new_coded->title      = ensure_string (meta_db->string_storage, old_meta->title,
                                         new_meta->title, old_coded->title);
  new_coded->collection = ensure_string (meta_db->string_storage, old_meta->collection,
                                         new_meta->collection, old_coded->collection);
  new_coded->performer  = ensure_string (meta_db->string_storage, old_meta->performer,
                                         new_meta->performer, old_coded->performer);
  new_coded->author     = ensure_string (meta_db->string_storage, old_meta->author,
                                         new_meta->author, old_coded->author);
  new_coded->keywords   = ensure_string (meta_db->string_storage, old_meta->keywords,
                                         new_meta->keywords, old_coded->keywords);
  new_coded->seq        = ensure_string (meta_db->string_storage, old_meta->seq,
                                         new_meta->seq, old_coded->seq);

  new_coded->fragment     = new_meta->fragment;
  new_coded->max_fragment = new_meta->max_fragment;
  new_coded->quality      = new_meta->quality;
  new_coded->format       = new_meta->format;
  new_coded->year         = new_meta->year;
  new_coded->month        = new_meta->month;
  new_coded->day          = new_meta->day;
  new_coded->width        = new_meta->width;
  new_coded->height       = new_meta->height;
  new_coded->duration     = new_meta->duration;
  new_coded->lang [0]     = new_meta->lang [0];
  new_coded->lang [1]     = new_meta->lang [1];
  new_coded->cat  [0]     = new_meta->cat  [0];
  new_coded->cat  [1]     = new_meta->cat  [1];
  new_coded->media        = new_meta->media;
  new_coded->flags        = new_meta->flags;

  return new_coded;
}

static void
add_string_to_index (TntMetaDB *meta_db, const TntFileID *file_id, const gchar *string)
{
  if (!string)
    return;

  tnt_string_index_insert (meta_db->string_to_id_index, string, file_id);
  tnt_storage_set_used (meta_db->storage, get_used (meta_db));
}

/* Consumes the strings (frees them) */
static gchar *
concat_str (gchar *text, gchar *str)
{
  gchar *new_text;

  if (!str)
    return text;

  if (text)
  {
    new_text = g_strjoin (" ", text, str, NULL);
    g_free (text);
  }
  else
    new_text = g_strdup (str);

  g_free (str);
  return new_text;
}

static gchar *
concat_meta_text (TntFileMeta *file_meta)
{
  gchar *text = NULL;

  text = concat_str (text, tnt_file_meta_get_name       (file_meta));
  text = concat_str (text, tnt_file_meta_get_title      (file_meta));
  text = concat_str (text, tnt_file_meta_get_collection (file_meta));
  text = concat_str (text, tnt_file_meta_get_performer  (file_meta));
  text = concat_str (text, tnt_file_meta_get_author     (file_meta));
  text = concat_str (text, tnt_file_meta_get_keywords   (file_meta));

  return text;
}

static void
add_text_to_index (TntMetaDB *meta_db, const TntFileID *file_id, TntFileMeta *file_meta)
{
  gchar *text;

  text = concat_meta_text (file_meta);
  add_string_to_index (meta_db, file_id, text);
  g_free (text);
}

static void
calculate_total_data_size (TntMetaDB *meta_db, guint *known_count, guint64 *known_size)
{
  FluxPatTreeCursor *cursor;

  *known_count = 0;
  *known_size = 0;

  cursor = flux_pat_tree_traverse (meta_db->id_to_file_meta_tree, FLUX_ITEM_NULL);
  if (!cursor)
    return;

  do
  {
    TntFileID *file_id;
    guint      file_id_len;

    if (!flux_pat_tree_cursor_read (cursor, (gpointer) &file_id, &file_id_len, NULL, NULL))
      continue;

    if (file_id_len == sizeof (TntFileID))
    {
      *known_size += file_id->len;
      (*known_count)++;
    }

    g_free (file_id);
  }
  while (flux_cursor_next (FLUX_CURSOR (cursor)));

  flux_object_unref (FLUX_OBJECT (cursor));
}

static void
cull_db (TntMetaDB *meta_db)
{
  gint i;

  CULL_DEBUG (g_print ("[TntMetaDB Cull] Free: %lld Used: %lld\n", tnt_storage_get_free (meta_db->storage),
                       tnt_storage_get_used (meta_db->storage)));

  for (i = 0; i < 2; i++)
  {
    TntFileID   *file_id;
    guint        file_id_len;
    CodedMeta   *coded_meta;
    guint        coded_meta_len;
    TntFileMeta *file_meta;
    gchar       *meta_text;
    gboolean     exists;

    exists = flux_pat_tree_get_random (meta_db->id_to_file_meta_tree, FLUX_ITEM_NULL,
                                       (gpointer) &file_id, &file_id_len,
                                       (gpointer) &coded_meta, &coded_meta_len);
    if (!exists)
      continue;

    if (file_id_len != sizeof (TntFileID) || coded_meta_len != sizeof (CodedMeta))
    {
      g_free (file_id);
      g_free (coded_meta);
      continue;
    }

    if (coded_meta->flags & (TNT_FILE_META_LOCAL_USER_INPUT | TNT_FILE_META_LOCAL_FILE))
      continue;

    file_meta = decode_file_meta (meta_db, coded_meta);
    if (!file_meta)
      continue;

    meta_text = concat_meta_text (file_meta);
    if (meta_text)
    {
      tnt_string_index_remove (meta_db->string_to_id_index, meta_text, file_id);
      g_free (meta_text);
    }

    release_file_meta (meta_db, coded_meta);
    flux_pat_tree_remove (meta_db->id_to_file_meta_tree, FLUX_ITEM_NULL, file_id, sizeof (TntFileID));
    tnt_file_meta_free (file_meta);

    if (file_id->len <= meta_db->known_size)
      meta_db->known_size -= file_id->len;
    if (meta_db->known_count > 0)
      meta_db->known_count--;

    g_free (file_id);
    g_free (coded_meta);
  }
}

static TntFileMeta *
get_file_meta (TntMetaDB *meta_db, const TntFileID *file_id)
{
  TntFileMeta *file_meta;
  CodedMeta   *coded_meta;
  guint        coded_meta_len;

  if (!flux_pat_tree_find (meta_db->id_to_file_meta_tree, FLUX_ITEM_NULL,
                           file_id, sizeof (TntFileID), (gpointer) &coded_meta, &coded_meta_len))
    return NULL;

  if (coded_meta_len == sizeof (CodedMeta))
    file_meta = decode_file_meta (meta_db, coded_meta);
  else
    file_meta = NULL;

  g_free (coded_meta);
  return file_meta;
}

static void
set_file_meta (TntMetaDB *meta_db, const TntFileID *file_id, TntFileMeta *file_meta)
{
  CodedMeta   *coded_meta;
  guint        coded_meta_len;
  gboolean     exists;

  if (tnt_storage_get_free (meta_db->storage) < 64 * BYTES_IN_K &&
      (flux_storage_get_arena_size (meta_db->id_to_file_meta_storage) -
       flux_storage_get_used_size (meta_db->id_to_file_meta_storage) < sizeof (CodedMeta) ||
       tnt_string_index_get_free_size (meta_db->string_to_id_index) < sizeof (TntFileID) * 100))
  {
    /* Get random candidates and remove them from database files if they're unimportant. */
    CULL_DEBUG (g_print ("Culling.\n"));
    cull_db (meta_db);
  }

  exists = flux_pat_tree_find (meta_db->id_to_file_meta_tree, FLUX_ITEM_NULL,
                               file_id, sizeof (TntFileID), (gpointer) &coded_meta, &coded_meta_len);

  if (exists && coded_meta_len != sizeof (CodedMeta))
  {
    g_free (coded_meta);
    exists = FALSE;
  }

  if (exists)
  {
    TntFileMeta *old_meta;
    CodedMeta   *new_coded_meta;
    gchar       *old_text, *new_text;

    /* Replace old record, refresh index strings pointing to it */

    old_meta = decode_file_meta (meta_db, coded_meta);

    old_text = concat_meta_text (old_meta);
    new_text = concat_meta_text (file_meta);

    new_coded_meta = reencode_file_meta (meta_db, old_meta, file_meta, coded_meta);

    flux_pat_tree_insert (meta_db->id_to_file_meta_tree, FLUX_ITEM_NULL,
                          file_id, sizeof (TntFileID), new_coded_meta, sizeof (CodedMeta));
    g_free (new_coded_meta);
    tnt_file_meta_free (old_meta);

    tnt_string_index_update (meta_db->string_to_id_index,
                             FLUX_STRING_NOT_NULL (old_text),
                             FLUX_STRING_NOT_NULL (new_text),
                             file_id);

    g_free (old_text);
    g_free (new_text);

    g_free (coded_meta);
  }
  else
  {
    CodedMeta *new_coded_meta;

    /* The record is new; just insert and index it */

    new_coded_meta = encode_file_meta (meta_db, file_meta);
    flux_pat_tree_insert (meta_db->id_to_file_meta_tree, FLUX_ITEM_NULL,
                          file_id, sizeof (TntFileID), new_coded_meta, sizeof (CodedMeta));
    g_free (new_coded_meta);

    meta_db->known_size += file_id->len;
    meta_db->known_count++;

    add_text_to_index (meta_db, file_id, file_meta);
    tnt_storage_set_used (meta_db->storage, get_used (meta_db));
  }
}

/* Called from worker thread only, mutex is locked */
static void
remove_query (TntMetaDB *meta_db, gint running_query_n)
{
  QueryActionData     *running_query;
  QueryDoneReturnData *return_data;

  g_return_if_fail (running_query_n < meta_db->running_queries->len);

  running_query = g_ptr_array_index (meta_db->running_queries, running_query_n);

  return_data = tnt_async_queue_new_return (QueryDoneReturnData);
  return_data->return_data.type = RETURN_QUERY_DONE;
  return_data->query            = running_query->query;
  return_data->func             = running_query->func;
  return_data->func_data        = running_query->func_data;

  flux_object_ref (FLUX_OBJECT (return_data->query));

  if (!running_query->paused)
    meta_db->running_queries_num--;

  g_ptr_array_remove_index_fast (meta_db->running_queries, running_query_n);

  tnt_async_queue_push_return_unlocked (meta_db->queue, return_data, running_query);
  tnt_async_queue_unref_action_unlocked (meta_db->queue, running_query);
}

/* Called from worker thread only, mutex is locked */
static void
add_query (TntMetaDB *meta_db, QueryActionData *action)
{
  const gchar *string;

  meta_db->running_queries_num++;
  action->paused = FALSE;

  tnt_async_queue_ref_action_unlocked (meta_db->queue, action);
  g_ptr_array_add (meta_db->running_queries, action);

  string = tnt_query_get_string (action->query);
  if (!string || string [0] == '\0' ||
      !(action->cursor = tnt_string_index_find (meta_db->string_to_id_index, string)))
  {
    gint n;

    n = meta_db->running_queries->len - 1;
    remove_query (meta_db, n);
  }
}

static void
step_queries (TntMetaDB *meta_db)
{
  gint i;

  for (i = 0; i < meta_db->running_queries->len; i++)
  {
    QueryActionData      *running_query = g_ptr_array_index (meta_db->running_queries, i);
    TntQuery             *query;
    TntStringIndexCursor *cursor;
    TntFileID             file_id;
    TntFileMeta          *file_meta;
    gboolean              read_ok, next_ok;

    query = running_query->query;
    cursor = running_query->cursor;

    flux_object_ref (FLUX_OBJECT (query));
    flux_object_ref (FLUX_OBJECT (cursor));

    tnt_async_queue_unlock (meta_db->queue);

    /* Begin blocking code */

    read_ok = tnt_string_index_cursor_read (cursor, &file_id);
    next_ok = flux_cursor_next (FLUX_CURSOR (cursor));

    if (read_ok)
      file_meta = get_file_meta (meta_db, &file_id);
    else
      file_meta = NULL;

    /* End blocking code */

    tnt_async_queue_lock (meta_db->queue);

    flux_object_unref (FLUX_OBJECT (cursor));
    flux_object_unref (FLUX_OBJECT (query));

    if (i < meta_db->running_queries->len &&
        running_query == g_ptr_array_index (meta_db->running_queries, i))
    {
      /* Query is still running */

      if (!next_ok)
      {
        remove_query (meta_db, i);
        i--;
      }

      if (file_meta)
      {
        QueryReturnData *return_data;

        return_data = tnt_async_queue_new_return (QueryReturnData);
        return_data->return_data.type = RETURN_QUERY;
        return_data->query            = query;
        return_data->file_meta        = file_meta;
        memcpy (&return_data->file_id, &file_id, sizeof (TntFileID));

        flux_object_ref (FLUX_OBJECT (return_data->query));

        tnt_async_queue_push_return_unlocked (meta_db->queue, return_data, running_query);

        /* NOTE: Action, return data may no longer be valid at this point */
      }
    }
    else
    {
      /* Query was removed by another thread */

      if (file_meta)
        tnt_file_meta_free (file_meta);

      i--;
    }
  }
}

static void
run_set_meta (TntMetaDB *meta_db, SetMetaActionData *action)
{
  TntFileMeta *file_meta;
  TntFileID   *file_id;

  file_meta = action->file_meta;
  file_id   = &action->file_id;
  g_assert (file_meta);

  tnt_async_queue_unlock (meta_db->queue);
  set_file_meta (meta_db, file_id, file_meta);
  tnt_async_queue_lock (meta_db->queue);
}

static void
run_get_meta (TntMetaDB *meta_db, GetMetaActionData *action)
{
  TntFileMeta       *file_meta;
  GetMetaReturnData *return_data;

  tnt_async_queue_unlock (meta_db->queue);
  file_meta = get_file_meta (meta_db, &action->file_id);
  tnt_async_queue_lock (meta_db->queue);

  return_data = tnt_async_queue_new_return (GetMetaReturnData);
  return_data->return_data.type = RETURN_GET_META;
  return_data->file_meta = file_meta;
  return_data->func = action->func;
  return_data->func_data = action->func_data;
  memcpy (&return_data->file_id, &action->file_id, sizeof (TntFileID));

  tnt_async_queue_push_return_unlocked (meta_db->queue, return_data, action);
}

static void
run_merge_meta (TntMetaDB *meta_db, MergeMetaActionData *action)
{
  TntFileMeta *file_meta;

  tnt_async_queue_unlock (meta_db->queue);
  file_meta = get_file_meta (meta_db, &action->file_id);
  tnt_async_queue_lock (meta_db->queue);

  if (file_meta)
  {
    TntFileMeta *merged_meta;

    merged_meta = tnt_file_meta_merge (file_meta, action->file_meta);
    tnt_file_meta_free (file_meta);
    file_meta = merged_meta;
  }
  else
  {
    file_meta = tnt_file_meta_dup (action->file_meta);
  }

  if (action->store_result)
  {
    tnt_async_queue_unlock (meta_db->queue);
    set_file_meta (meta_db, &action->file_id, file_meta);
    tnt_async_queue_lock (meta_db->queue);
  }

  if (action->func)
  {
    MergeMetaReturnData *return_data;

    return_data = tnt_async_queue_new_return (MergeMetaReturnData);
    return_data->return_data.type = RETURN_MERGE_META;
    return_data->file_meta = file_meta;
    return_data->func = action->func;
    return_data->func_data = action->func_data;
    memcpy (&return_data->file_id, &action->file_id, sizeof (TntFileID));

    tnt_async_queue_push_return_unlocked (meta_db->queue, return_data, action);
  }
  else
  {
    tnt_file_meta_free (file_meta);
  }
}

static void
run_extract_meta (TntMetaDB *meta_db, ExtractMetaActionData *action)
{
  TntFileID     *file_id = &action->file_id;
  TntFileMeta   *file_meta;
  gchar         *meta_file_name;
  gchar         *meta_keywords;
  gchar         *local_file_name = action->name;
  gchar         *local_file_path = action->path;

  tnt_async_queue_unlock (meta_db->queue);
  file_meta = get_file_meta (meta_db, file_id);
  tnt_async_queue_lock (meta_db->queue);

  if (file_meta)
  {
    meta_file_name = tnt_file_meta_get_name (file_meta);
    meta_keywords  = tnt_file_meta_get_keywords (file_meta);
  }
  else
  {
    meta_file_name = NULL;
    meta_keywords  = NULL;
  }

  if (meta_file_name && !local_file_name)
  {
    DEBUG (g_print (G_STRLOC ": '%s' has no local correspondence.\n", meta_file_name));
  }

  if (local_file_name)
  {
    TntFileMeta  *generated_meta;
    TntFileMeta  *merged_meta;
    TntTimeStamp  timestamp;
    TntStat       tstat;
    gchar        *keywords;
    gchar        *local_file_name_utf8;

    local_file_name_utf8 = tnt_filename_to_utf8_lossy (local_file_name);

    if (local_file_path)
    {
      gchar *full_path;

      full_path = g_strjoin (G_DIR_SEPARATOR_S, local_file_path, local_file_name, NULL);

      tnt_async_queue_unlock (meta_db->queue);

      if (!file_meta || !tnt_file_meta_is_local_file (file_meta))
        generated_meta = tnt_file_meta_extract (full_path);  /* Blocks on file I/O */
      else
        generated_meta = NULL;

      if (tnt_stat (full_path, &tstat))  /* Blocks on file I/O */
        timestamp = tstat.mtime;
      else
        timestamp = 0;

      tnt_async_queue_lock (meta_db->queue);

      g_free (full_path);
    }
    else
    {
      generated_meta = NULL;
      timestamp = 0;
    }

    if (generated_meta)
      keywords = tnt_file_meta_get_keywords (generated_meta);
    else
      keywords = NULL;

    if (!keywords && action->relative_path && (!meta_keywords || !*meta_keywords))
    {
      gchar *keywords_utf8;

      keywords = tnt_path_make_keyword_string (action->relative_path);
      keywords_utf8 = tnt_filename_to_utf8_lossy (keywords);
      tnt_file_meta_set_keywords (generated_meta, keywords_utf8);
      g_free (keywords_utf8);
    }

    if (keywords)
      g_free (keywords);

    if (file_meta)
    {
      if (generated_meta)
      {
        merged_meta = tnt_file_meta_merge (file_meta, generated_meta);
        tnt_file_meta_free (file_meta);
        tnt_file_meta_free (generated_meta);
        file_meta = merged_meta;
      }
    }
    else
    {
      file_meta = generated_meta;
    }

    if (file_meta)
    {
      tnt_file_meta_set_name (file_meta, local_file_name_utf8);

      if (local_file_name && local_file_path)
        tnt_file_meta_set_local_file (file_meta, TRUE);

      tnt_file_meta_set_timestamp (file_meta, timestamp);

      tnt_async_queue_unlock (meta_db->queue);
      set_file_meta (meta_db, file_id, file_meta);  /* Blocks on file I/O */
      tnt_async_queue_lock (meta_db->queue);
    }

    g_free (local_file_name_utf8);
  }

  if (meta_file_name)
    g_free (meta_file_name);
  if (meta_keywords)
    g_free (meta_keywords);

  if (action->func)
  {
    MergeMetaReturnData *return_data;

    return_data = tnt_async_queue_new_return (MergeMetaReturnData);
    return_data->return_data.type = RETURN_EXTRACT_META;
    return_data->file_meta = file_meta;
    return_data->func = action->func;
    return_data->func_data = action->func_data;
    memcpy (&return_data->file_id, file_id, sizeof (TntFileID));

    tnt_async_queue_push_return_unlocked (meta_db->queue, return_data, action);
  }
  else if (file_meta)
  {
    tnt_file_meta_free (file_meta);
  }
}

static void
run_get_id (TntMetaDB *meta_db, GetIDActionData *action)
{
  GetIDReturnData   *return_data;
  TntFileID          file_id;
  TntFileID          file_id_mask;
  FluxPatTreeCursor *cursor;
  gboolean           found = FALSE;

  memset (&file_id, 0, sizeof (TntFileID));
  file_id.len = action->len;

  memset (&file_id_mask, 0, sizeof (TntFileID));
  file_id_mask.len = G_MAXUINT64;

  tnt_async_queue_unlock (meta_db->queue);

  /* Blocks on file I/O */
  cursor = flux_pat_tree_find_with_mask (meta_db->id_to_file_meta_tree, FLUX_ITEM_NULL,
                                         &file_id, &file_id_mask, sizeof (TntFileID));

  if (!cursor)
  {
    DEBUG (g_print ("tnt_meta_db_get_local_file_id(): '%s' returned no matches.\n", action->name));
    goto out;
  }

  do
  {
    TntFileID *file_id_found;
    guint      file_id_found_len;
    CodedMeta *coded_meta;
    guint      coded_meta_len;
    gchar     *found_name = NULL;

    /* Blocks on file I/O */
    if (!flux_pat_tree_cursor_read (cursor, (gpointer) &file_id_found, &file_id_found_len,
                                    (gpointer) &coded_meta, &coded_meta_len))
      break;

    /* Check that key is valid */
    if (!file_id_found || file_id_found_len != sizeof (TntFileID))
    {
      DEBUG (g_print ("tnt_meta_db_get_local_file_id(): Got invalid file ID!\n"));
    }

    /* Check file length - if this doesn't match, the database is on crack */
    else if (file_id_found->len != action->len)
    {
      DEBUG (g_print ("tnt_meta_db_get_local_file_id(): Database mismatched size!\n"));
    }

    /* Check that metadata is valid */
    else if (!coded_meta || coded_meta_len != sizeof (CodedMeta))
    {
      DEBUG (g_print ("tnt_meta_db_get_local_file_id(): Got invalid metadata!\n"));
    }

    /* Check that data is (probably) local */
    else if (!(coded_meta->flags & TNT_FILE_META_LOCAL_FILE))
    {
      DEBUG (g_print ("tnt_meta_db_get_local_file_id(): '%s' match not local.\n", action->name));
    }

    /* Check that the metadata has a file name */
    else if (coded_meta->name == FLUX_ITEM_NULL)
    {
      DEBUG (g_print ("tnt_meta_db_get_local_file_id(): '%s' match missing file name!.\n",
                      action->name));
    }

    /* Read the name and make sure we got it */
    else if (!(found_name = load_string (meta_db->string_storage, coded_meta->name)))
    {
      DEBUG (g_print ("tnt_meta_db_get_local_file_id(): '%s' match has unloadable file name!.\n",
                      action->name));
    }

    /* Check the name */
    else if (strcmp (action->name, found_name))
    {
      DEBUG (g_print ("tnt_meta_db_get_local_file_id(): '%s' does not match '%s'.\n",
                      action->name, found_name));
    }

    /* Check file timestamp */
    else if (coded_meta->timestamp != action->timestamp)
    {
      DEBUG (g_print ("tnt_meta_db_get_local_file_id(): '%s' timestamp mismatch (new=%d vs old=%d).\n",
                      action->name, action->timestamp, coded_meta->timestamp));
    }

    else
    {
      memcpy (&file_id, file_id_found, sizeof (TntFileID));
      found = TRUE;
    }

    g_free (found_name);
    g_free (coded_meta);
    g_free (file_id_found);
  }
  while (!found && flux_cursor_next (FLUX_CURSOR (cursor)));  /* Blocks on file I/O */

  flux_object_unref (FLUX_OBJECT (cursor));

out:
  tnt_async_queue_lock (meta_db->queue);

  return_data = tnt_async_queue_new_return (GetIDReturnData);
  return_data->return_data.type = RETURN_GET_ID;
  return_data->func = action->func;
  return_data->func_data = action->func_data;
  return_data->found = found;
  if (found)
    memcpy (&return_data->file_id, &file_id, sizeof (TntFileID));

  tnt_async_queue_push_return_unlocked (meta_db->queue, return_data, action);  
}

static gpointer
worker_main (TntMetaDB *meta_db)
{
  TntAsyncQueue *queue;
  gboolean       running = TRUE;

  queue = meta_db->queue;
  tnt_async_queue_ref (queue);

  calculate_total_data_size (meta_db, &meta_db->known_count, &meta_db->known_size);

  tnt_async_queue_lock (queue);

  while (running)
  {
    ActionData *action_data;

    if (meta_db->running_queries_num == 0)
    {
      action_data = tnt_async_queue_pop_action_unlocked (queue);
    }
    else
    {
      /* Don't service queries if we have a lot of pending actions */
      if (tnt_async_queue_get_pending_action_length_unlocked (queue) < ACTIONS_HIGH_WATER_MARK)
        step_queries (meta_db);

      action_data = tnt_async_queue_try_pop_action_unlocked (queue);
    }

    if (!action_data)
      continue;

    switch (action_data->type)
    {
      case ACTION_WAKEUP:
        break;

      case ACTION_GET_ID:
        run_get_id (meta_db, (GetIDActionData *) action_data);
        break;

      case ACTION_GET_META:
        run_get_meta (meta_db, (GetMetaActionData *) action_data);
        break;

      case ACTION_SET_META:
        run_set_meta (meta_db, (SetMetaActionData *) action_data);
        break;

      case ACTION_MERGE_META:
        run_merge_meta (meta_db, (MergeMetaActionData *) action_data);
        break;

      case ACTION_EXTRACT_META:
        run_extract_meta (meta_db, (ExtractMetaActionData *) action_data);
        break;

      case ACTION_QUERY:
        add_query (meta_db, (QueryActionData *) action_data);
        break;

      case ACTION_TERMINATE:
        running = FALSE;
        break;

      default:
        g_assert_not_reached ();
    }
  }

  tnt_async_queue_unlock (queue);
  tnt_async_queue_unref (queue);
  return NULL;
}

static void
return_dispatch (ReturnData *return_data, TntMetaDB *meta_db)
{
  switch (return_data->type)
  {
    case RETURN_GET_ID:
      {
        GetIDReturnData *get_id_return_data = (GetIDReturnData *) return_data;

        get_id_return_data->func (get_id_return_data->found ? &get_id_return_data->file_id : NULL,
                                  get_id_return_data->func_data);
      }
      break;

    case RETURN_GET_META:
      {
        GetMetaReturnData *get_meta_return_data = (GetMetaReturnData *) return_data;

        get_meta_return_data->func (&get_meta_return_data->file_id,
                                    get_meta_return_data->file_meta,
                                    get_meta_return_data->func_data);

        get_meta_return_data->file_meta = NULL;
      }
      break;

    case RETURN_MERGE_META:
      {
        MergeMetaReturnData *merge_meta_return_data = (MergeMetaReturnData *) return_data;

        merge_meta_return_data->func (&merge_meta_return_data->file_id,
                                      merge_meta_return_data->file_meta,
                                      merge_meta_return_data->func_data);

        merge_meta_return_data->file_meta = NULL;
      }
      break;

    case RETURN_EXTRACT_META:
      {
        ExtractMetaReturnData *extract_meta_return_data = (ExtractMetaReturnData *) return_data;

        extract_meta_return_data->func (&extract_meta_return_data->file_id,
                                        extract_meta_return_data->file_meta,
                                        extract_meta_return_data->func_data);

        extract_meta_return_data->file_meta = NULL;
      }
      break;

    case RETURN_QUERY:
      {
        QueryReturnData *query_return_data = (QueryReturnData *) return_data;

        if (!tnt_query_try_match (query_return_data->query, &query_return_data->file_id,
                                  query_return_data->file_meta,
                                  TNT_NODE_ID_INVALID, TNT_COMPLETENESS_UNKNOWN))
        {
          DEBUG
          (
            gchar *title;

            title = tnt_file_meta_get_best_title (query_return_data->file_meta);
            MATCH_DEBUG (g_print ("[%s] Mismatch: %s\n", tnt_query_get_string (query_return_data->query), title));
            g_free (title);
          )
        }
      }
      break;

    case RETURN_QUERY_DONE:
      {
        QueryDoneReturnData *query_done_return_data = (QueryDoneReturnData *) return_data;

        MATCH_DEBUG (g_print ("[%s] Query done!\n\n", tnt_query_get_string (query_done_return_data->query)));

        if (query_done_return_data->func)
          query_done_return_data->func (query_done_return_data->query,
                                        query_done_return_data->func_data);
      }
      break;

    default:
      g_assert_not_reached ();
  }
}

static guint
action_hash (ActionData *action_data)
{
  guint value = 0;

  value ^= action_data->type << 24;

  switch (action_data->type)
  {
    case ACTION_GET_META:
      {
        GetMetaActionData *get_meta_action_data = (GetMetaActionData *) action_data;

        value ^= GPOINTER_TO_UINT (get_meta_action_data->func) ^
                 GPOINTER_TO_UINT (get_meta_action_data->func_data);
      }
      break;

    case ACTION_SET_META:
      {
        SetMetaActionData *set_meta_action_data = (SetMetaActionData *) action_data;
        guint              folded_hash;

        tnt_fold_mem (&set_meta_action_data->file_id, sizeof (TntFileID), &folded_hash, sizeof (guint));
        value ^= folded_hash;
      }
      break;

    case ACTION_MERGE_META:
    case ACTION_EXTRACT_META:
      {
        MergeMetaActionData *merge_meta_action_data = (MergeMetaActionData *) action_data;

        value ^= GPOINTER_TO_UINT (merge_meta_action_data->func) ^
                 GPOINTER_TO_UINT (merge_meta_action_data->func_data);
      }
      break;

    case ACTION_QUERY:
      {
        QueryActionData *query_action_data = (QueryActionData *) action_data;

        value ^= GPOINTER_TO_UINT (query_action_data->query);
      }
      break;

    case ACTION_WAKEUP:
    case ACTION_GET_ID:
    case ACTION_TERMINATE:
      break;

    default:
      g_assert_not_reached ();
  }

  return value;
}

static gboolean
action_equal (ActionData *action_data_1, ActionData *action_data_2)
{
  if (action_data_1->type != action_data_2->type)
    return FALSE;

  switch (action_data_1->type)
  {
    case ACTION_GET_META:
      {
        GetMetaActionData *get_meta_action_data_1 = (GetMetaActionData *) action_data_1;
        GetMetaActionData *get_meta_action_data_2 = (GetMetaActionData *) action_data_2;

        if (get_meta_action_data_1->func      == get_meta_action_data_2->func &&
            get_meta_action_data_1->func_data == get_meta_action_data_2->func_data)
          return TRUE;

        return FALSE;
      }

    case ACTION_SET_META:
      {
        SetMetaActionData *set_meta_action_data_1 = (SetMetaActionData *) action_data_1;
        SetMetaActionData *set_meta_action_data_2 = (SetMetaActionData *) action_data_2;

        if (!memcmp (&set_meta_action_data_1->file_id,
                     &set_meta_action_data_2->file_id, sizeof (TntFileID)))
          return TRUE;

        return FALSE;
      }

    case ACTION_MERGE_META:
    case ACTION_EXTRACT_META:
      {
        MergeMetaActionData *merge_meta_action_data_1 = (MergeMetaActionData *) action_data_1;
        MergeMetaActionData *merge_meta_action_data_2 = (MergeMetaActionData *) action_data_2;

        if (merge_meta_action_data_1->func      == merge_meta_action_data_2->func &&
            merge_meta_action_data_1->func_data == merge_meta_action_data_2->func_data)
          return TRUE;

        return FALSE;
      }

    case ACTION_QUERY:
      {
        QueryActionData *query_action_data_1 = (QueryActionData *) action_data_1;
        QueryActionData *query_action_data_2 = (QueryActionData *) action_data_2;

        if (query_action_data_1->query == query_action_data_2->query)
          return TRUE;

        return FALSE;
      }

    case ACTION_WAKEUP:
    case ACTION_GET_ID:
    case ACTION_TERMINATE:
      return TRUE;

    default:
      g_assert_not_reached ();
  }

  g_assert_not_reached ();
  return FALSE;
}

static void
action_destroy (ActionData *action_data)
{
  switch (action_data->type)
  {
    case ACTION_GET_ID:
      {
        GetIDActionData *get_id_action_data = (GetIDActionData *) action_data;

        g_free (get_id_action_data->name);
      }
      break;

    case ACTION_GET_META:
      break;

    case ACTION_SET_META:
      {
        SetMetaActionData *set_meta_action_data = (SetMetaActionData *) action_data;

        tnt_file_meta_free (set_meta_action_data->file_meta);
      }
      break;

    case ACTION_MERGE_META:
      {
        MergeMetaActionData *merge_meta_action_data = (MergeMetaActionData *) action_data;

        tnt_file_meta_free (merge_meta_action_data->file_meta);
      }
      break;

    case ACTION_EXTRACT_META:
      {
        ExtractMetaActionData *extract_meta_action_data = (ExtractMetaActionData *) action_data;

        g_free (extract_meta_action_data->name);
        g_free (extract_meta_action_data->path);
        g_free (extract_meta_action_data->relative_path);
      }
      break;

    case ACTION_QUERY:
      {
        QueryActionData *query_action_data = (QueryActionData *) action_data;

        if (query_action_data->cursor)
          flux_object_unref (FLUX_OBJECT (query_action_data->cursor));
        flux_object_unref (FLUX_OBJECT (query_action_data->query));
      }
      break;

    case ACTION_WAKEUP:
    case ACTION_TERMINATE:
      break;

    default:
      g_assert_not_reached ();
  }

  tnt_async_queue_free_action (action_data);
}

static void
return_destroy (ReturnData *return_data)
{
  switch (return_data->type)
  {
    case RETURN_GET_ID:
      break;

    case RETURN_GET_META:
    {
      GetMetaReturnData *get_meta_return_data = (GetMetaReturnData *) return_data;

      /* Callback assumes ownership of TntFileMeta, and file_meta field is NULLed. However,
       * if we're freed and don't get the chance to call back, we have to free it here. */

      if (get_meta_return_data->file_meta)
        tnt_file_meta_free (get_meta_return_data->file_meta);
    }
    break;

    case RETURN_MERGE_META:
      {
        MergeMetaReturnData *merge_meta_return_data = (MergeMetaReturnData *) return_data;

      /* Callback assumes ownership of TntFileMeta, and file_meta field is NULLed. However,
       * if we're freed and don't get the chance to call back, we have to free it here. */

        if (merge_meta_return_data->file_meta)
          tnt_file_meta_free (merge_meta_return_data->file_meta);
      }
      break;

    case RETURN_EXTRACT_META:
      {
        ExtractMetaReturnData *extract_meta_return_data = (ExtractMetaReturnData *) return_data;

      /* Callback assumes ownership of TntFileMeta, and file_meta field is NULLed. However,
       * if we're freed and don't get the chance to call back, we have to free it here. */

        if (extract_meta_return_data->file_meta)
          tnt_file_meta_free (extract_meta_return_data->file_meta);
      }
      break;

    case RETURN_QUERY:
      {
        QueryReturnData *query_return_data = (QueryReturnData *) return_data;

        flux_object_unref (FLUX_OBJECT (query_return_data->query));
        tnt_file_meta_free (query_return_data->file_meta);
      }
      break;

    case RETURN_QUERY_DONE:
      {
        QueryDoneReturnData *query_done_return_data = (QueryDoneReturnData *) return_data;

        flux_object_unref (FLUX_OBJECT (query_done_return_data->query));
      }
      break;

    default:
      g_assert_not_reached ();
  }

  tnt_async_queue_free_return (return_data);
}

gboolean
tnt_meta_db_init (TntMetaDB *meta_db, TntStorage *storage)
{
  gint   file_meta_chunk_fd, id_to_file_meta_fd;
  guint  id_to_file_meta_node_len;
  gchar *path;

  g_return_val_if_fail (storage != NULL, FALSE);

  flux_object_init              (FLUX_OBJECT (meta_db));
  flux_object_set_type          (FLUX_OBJECT (meta_db), TNT_META_DB_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (meta_db), tnt_meta_db_finalize);

  if (!g_thread_supported ())
    g_thread_init (NULL);

  path = g_strjoin (G_DIR_SEPARATOR_S, tnt_storage_get_path (storage), FILE_META_CHUNK_NAME, NULL);
  file_meta_chunk_fd = tnt_open (path, TRUE, FALSE, FALSE);
  g_free (path);

  if (file_meta_chunk_fd < 0)
    return FALSE;

  path = g_strjoin (G_DIR_SEPARATOR_S, tnt_storage_get_path (storage), ID_TO_FILE_META_NAME, NULL);
  id_to_file_meta_fd = tnt_open (path, TRUE, FALSE, FALSE);
  g_free (path);

  if (id_to_file_meta_fd < 0)
  {
    close (file_meta_chunk_fd);
    return FALSE;
  }

  id_to_file_meta_node_len = flux_pat_tree_calc_node_size (sizeof (TntFileID), sizeof (CodedMeta));

  meta_db->storage = storage;
  flux_object_ref (FLUX_OBJECT (meta_db->storage));

  meta_db->id_to_file_meta_storage =
    FLUX_STORAGE (flux_disk_chunk_new (id_to_file_meta_fd, id_to_file_meta_node_len));
  meta_db->string_storage       = FLUX_STORAGE (flux_disk_heap_new (file_meta_chunk_fd));
  meta_db->id_to_file_meta_tree = flux_pat_tree_new (meta_db->id_to_file_meta_storage, "id-to-file-meta",
                                                     sizeof (TntFileID), sizeof (CodedMeta));
  meta_db->string_to_id_index   = tnt_string_index_new (storage, STRINGS_FILE_PREFIX, sizeof (TntFileID));

  tnt_storage_set_used (storage, get_used (meta_db));

  meta_db->running_queries = g_ptr_array_new ();

  /* Set up worker thread */
  meta_db->queue         = tnt_async_queue_new ((TntDispatchFunc *) return_dispatch,
                                                (GHashFunc)         action_hash,
                                                (GEqualFunc)        action_equal,
                                                (GDestroyNotify)    action_destroy,
                                                (GDestroyNotify)    return_destroy,
                                                meta_db);
  meta_db->worker_thread = g_thread_create_full ((GThreadFunc) worker_main, meta_db,
                                                 THREAD_STACK_SIZE, TRUE, FALSE,
                                                 G_THREAD_PRIORITY_NORMAL, NULL);

  return TRUE;
}

TntMetaDB *
tnt_meta_db_new (TntStorage *storage)
{
  TntMetaDB *meta_db;

  meta_db = g_new0 (TntMetaDB, 1);
  if (!tnt_meta_db_init (meta_db, storage))
  {
    g_free (meta_db);
    return NULL;
  }

  return meta_db;
}

void
tnt_meta_db_finalize (FluxObject *object)
{
  TntMetaDB  *meta_db = TNT_META_DB (object);
  ActionData *action_data;

  assert_types (object);

  action_data = tnt_async_queue_new_action (ActionData);
  action_data->type = ACTION_TERMINATE;
  tnt_async_queue_push_action_head (meta_db->queue, action_data);

  tnt_async_queue_begin_shutdown (meta_db->queue);
  g_thread_join (meta_db->worker_thread);
  tnt_async_queue_unref (meta_db->queue);

  flux_object_unref (FLUX_OBJECT (meta_db->id_to_file_meta_tree));
  flux_object_unref (FLUX_OBJECT (meta_db->string_to_id_index));
  flux_object_unref (FLUX_OBJECT (meta_db->id_to_file_meta_storage));
  flux_object_unref (FLUX_OBJECT (meta_db->string_storage));

  g_ptr_array_free (meta_db->running_queries, TRUE);

  flux_object_unref (FLUX_OBJECT (meta_db->storage));
}

TntStorage *
tnt_meta_db_get_storage (TntMetaDB *meta_db)
{
  assert_types (meta_db);

  return meta_db->storage;
}

void
tnt_meta_db_set_file_meta (TntMetaDB *meta_db, const TntFileID *file_id, TntFileMeta *file_meta)
{
  SetMetaActionData *action_data;

  assert_types (meta_db);
  g_return_if_fail (file_id != NULL);
  g_return_if_fail (file_meta != NULL);

  action_data = tnt_async_queue_new_action (SetMetaActionData);
  action_data->action.type = ACTION_SET_META;
  action_data->file_meta = tnt_file_meta_dup (file_meta);
  memcpy (&action_data->file_id, file_id, sizeof (TntFileID));

  tnt_async_queue_push_action_head (meta_db->queue, action_data);
}

void
tnt_meta_db_get_file_meta (TntMetaDB *meta_db, const TntFileID *file_id,
                           TntFileMetaFunc *func, gpointer data)
{
  GetMetaActionData *action_data;

  assert_types (meta_db);
  g_return_if_fail (file_id != NULL);
  g_return_if_fail (func != NULL);

  action_data = tnt_async_queue_new_action (GetMetaActionData);
  action_data->action.type = ACTION_GET_META;
  action_data->func = func;
  action_data->func_data = data;
  memcpy (&action_data->file_id, file_id, sizeof (TntFileID));

  tnt_async_queue_push_action_head (meta_db->queue, action_data);
}

void
tnt_meta_db_merge_file_meta (TntMetaDB *meta_db, const TntFileID *file_id, TntFileMeta *file_meta,
                             gboolean store, TntFileMetaFunc *func, gpointer data)
{
  MergeMetaActionData *action_data;

  assert_types (meta_db);
  g_return_if_fail (file_id != NULL);
  g_return_if_fail (file_meta != NULL);

  action_data = tnt_async_queue_new_action (MergeMetaActionData);
  action_data->action.type = ACTION_MERGE_META;
  action_data->file_meta = tnt_file_meta_dup (file_meta);
  action_data->store_result = store;
  action_data->func = func;
  action_data->func_data = data;
  memcpy (&action_data->file_id, file_id, sizeof (TntFileID));

  tnt_async_queue_push_action_head (meta_db->queue, action_data);
}

void
tnt_meta_db_extract_file_meta (TntMetaDB *meta_db, const TntFileID *file_id, const gchar *file_name,
                               const gchar *file_path, const gchar *relative_path,
                               TntFileMetaFunc *func, gpointer data)
{
  ExtractMetaActionData *action_data;

  assert_types (meta_db);
  g_return_if_fail (file_id != NULL);

  action_data = tnt_async_queue_new_action (ExtractMetaActionData);
  action_data->action.type = ACTION_EXTRACT_META;
  action_data->func = func;
  action_data->func_data = data;
  memcpy (&action_data->file_id, file_id, sizeof (TntFileID));

  if (file_name)
    action_data->name = g_strdup (file_name);
  else
    action_data->name = NULL;

  if (file_path)
    action_data->path = g_strdup (file_path);
  else
    action_data->path = NULL;

  if (relative_path)
    action_data->relative_path = g_strdup (relative_path);
  else
    action_data->relative_path = NULL;

#if 0
  tnt_async_queue_push_action_head (meta_db->queue, action_data);
#endif
  tnt_async_queue_push_action_tail (meta_db->queue, action_data);
}

void
tnt_meta_db_query (TntMetaDB *meta_db, TntQuery *query, gboolean before,
                   TntQueryDoneFunc *func, gpointer data)
{
  QueryActionData *action;

  assert_types (meta_db);
  flux_object_check (FLUX_OBJECT (query), TNT_QUERY_TYPE);

  flux_object_ref (FLUX_OBJECT (query));

  action = tnt_async_queue_new_action (QueryActionData);
  action->action.type = ACTION_QUERY;
  action->query = query;
  action->func = func;
  action->func_data = data;

#if 0
  if (before)
#endif
    tnt_async_queue_push_action_tail (meta_db->queue, action);
#if 0
  else
    tnt_async_queue_push_action_head (meta_db->queue, action);
#endif
}

static void
remove_running_query_action (TntMetaDB *meta_db, TntQuery *query)
{
  gint i;

  for (i = 0; i < meta_db->running_queries->len; i++)
  {
    QueryActionData *running_query = g_ptr_array_index (meta_db->running_queries, i);

    if (running_query->query == query)
    {
      g_ptr_array_remove_index_fast (meta_db->running_queries, i);
      if (!running_query->paused)
        meta_db->running_queries_num--;
      tnt_async_queue_unref_action_unlocked (meta_db->queue, running_query);
      return;
    }
  }
}

void
tnt_meta_db_cancel_query (TntMetaDB *meta_db, TntQuery *query)
{
  QueryActionData action;

  assert_types (meta_db);
  flux_object_check (FLUX_OBJECT (query), TNT_QUERY_TYPE);

  action.action.type = ACTION_QUERY;
  action.query = query;

  tnt_async_queue_lock (meta_db->queue);
  remove_running_query_action (meta_db, query);
  tnt_async_queue_cancel_action_unlocked (meta_db->queue, &action);
  tnt_async_queue_unlock (meta_db->queue);
}

gboolean
tnt_meta_db_query_get_paused (TntMetaDB *meta_db, TntQuery *query)
{
  gboolean is_paused = FALSE;
  gint     i;

  assert_types (meta_db);

  tnt_async_queue_lock (meta_db->queue);

  for (i = 0; i < meta_db->running_queries->len; i++)
  {
    QueryActionData *running_query = g_ptr_array_index (meta_db->running_queries, i);

    if (running_query->query == query)
    {
      is_paused = running_query->paused;
      break;
    }
  }

  tnt_async_queue_unlock (meta_db->queue);
  return is_paused;
}

void
tnt_meta_db_query_set_paused (TntMetaDB *meta_db, TntQuery *query, gboolean paused)
{
  gint i;

  assert_types (meta_db);

  tnt_async_queue_lock (meta_db->queue);

  for (i = 0; i < meta_db->running_queries->len; i++)
  {
    QueryActionData *running_query = g_ptr_array_index (meta_db->running_queries, i);

    if (running_query->query == query)
    {
      if (running_query->paused && !paused)
      {
        meta_db->running_queries_num++;

        if (meta_db->running_queries_num == 1)
        {
          WakeupActionData *action;

          action = tnt_async_queue_new_action (WakeupActionData);
          action->action.type = ACTION_WAKEUP;
          tnt_async_queue_push_action_head_unlocked (meta_db->queue, action);
        }
      }
      else if (!running_query->paused && paused)
        meta_db->running_queries_num--;

      running_query->paused = paused;
      break;
    }
  }

  tnt_async_queue_unlock (meta_db->queue);
}

void
tnt_meta_db_cancel_callback (TntMetaDB *meta_db, gpointer func, gpointer data)
{
  GetMetaActionData   get_meta_action;
  MergeMetaActionData merge_meta_action;
  GetIDActionData     get_id_action;

  assert_types (meta_db);

  get_meta_action.action.type = ACTION_GET_META;
  get_meta_action.func        = func;
  get_meta_action.func_data   = data;

  tnt_async_queue_cancel_action (meta_db->queue, &get_meta_action);

  merge_meta_action.action.type = ACTION_MERGE_META;
  merge_meta_action.func        = func;
  merge_meta_action.func_data   = data;

  tnt_async_queue_cancel_action (meta_db->queue, &merge_meta_action);

  merge_meta_action.action.type = ACTION_EXTRACT_META;

  tnt_async_queue_cancel_action (meta_db->queue, &merge_meta_action);

  get_id_action.action.type = ACTION_GET_ID;
  get_id_action.func        = func;
  get_id_action.func_data   = data;

  tnt_async_queue_cancel_action (meta_db->queue, &get_id_action);
}

void
tnt_meta_db_get_local_file_id (TntMetaDB *meta_db, const gchar *name, guint64 len,
                               TntTimeStamp timestamp, TntFileIDFunc *func, gpointer data)
{
  GetIDActionData *action_data;

  assert_types (meta_db);
  g_return_if_fail (name != NULL);
  g_return_if_fail (len > 0);
  g_return_if_fail (func != NULL);

  action_data = tnt_async_queue_new_action (GetIDActionData);
  action_data->action.type = ACTION_GET_ID;
  action_data->func = func;
  action_data->func_data = data;
  action_data->name = g_strdup (name);
  action_data->len = len;
  action_data->timestamp = timestamp;

#if 0
  tnt_async_queue_push_action_head (meta_db->queue, action_data);
#endif
  tnt_async_queue_push_action_tail (meta_db->queue, action_data);
}

guint
tnt_meta_db_get_file_count (TntMetaDB *meta_db)
{
  assert_types (meta_db);

  return meta_db->known_count;
}

guint64
tnt_meta_db_get_file_size (TntMetaDB *meta_db)
{
  assert_types (meta_db);

  return meta_db->known_size;
}

gint
tnt_meta_db_get_action_queue_len (TntMetaDB *meta_db)
{
  return tnt_async_queue_get_pending_action_length (meta_db->queue);
}
