/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-network-core.c -
 *
 * Copyright (C) 2002-2004 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <glib.h>

#ifndef G_PLATFORM_WIN32
# include <unistd.h>
#else
# include "win32/win32-compat.h"
#endif

#include "tnt.h"
#include "tnt-network-messages.h"

#define DEBUG(x)

#define BROADCAST_NODE_INFO_TIMEOUT_MIN    (1 * SECONDS_IN_MINUTE * 1000)
#define BROADCAST_NODE_INFO_TIMEOUT_MAX    (4 * SECONDS_IN_MINUTE * 1000)

#define BROADCAST_CONTACT_INFO_TIMEOUT_MIN (1 * SECONDS_IN_MINUTE * 1000)
#define BROADCAST_CONTACT_INFO_TIMEOUT_MAX (5 * SECONDS_IN_MINUTE * 1000)

#define BOOTSTRAP_EDGE_TIMEOUT      20

#define CONNECT_BURST_COUNT         4
#define PENDING_EDGE_MAX            50
#define REMOTE_QUERIES_MAX          100

#define BOOTSTRAP_ADDRESSES_MAX     64

/* The size a message is clamped to in determining its probability of being forwarded
 * in "Forward Own Messages" mode. All messages above this size have the same, minimal
 * probability of being forwarded. */
#define MESSAGE_SIZE_CLAMP 300000

/* Probability range in "Forward Own Messages" mode. */
#define FORWARD_OWN_MESSAGE_PROBABILITY_MIN 0.05
#define FORWARD_OWN_MESSAGE_PROBABILITY_MAX 0.2

/* Delay range in "Introduce Delays" mode, in seconds */
#define FAKE_DELAY_MIN 2.5
#define FAKE_DELAY_MAX 6.0

#define get_pending_edge_count(network_core) (network_core->pending_edge_array->len)

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), TNT_NETWORK_CORE_TYPE)

typedef struct
{
  TntNodeID        node_id;  /* Can be TNT_NODE_ID_INVALID, i.e. unknown */
  TntNodeAddress4  address;
  TntEdge         *edge;
  TntTcpIO        *tcp_io;
  guint            outbound : 1;
}
PendingEdge;

/* --- Absolutely necessary forward declarations --- */

static void      boost_outgoing                 (TntNetworkCore *network_core, TntBandwidthGroup *group);
static void      broadcast_message              (TntNetworkCore *network_core, TntEdge *source_edge,
                                                 TntMessage *message);
static TntEdge  *route_message                  (TntNetworkCore *network_core, TntEdge *edge,
                                                 TntMessage *message);
static void      received_message_from_edge     (TntTcpIO *tcp_io, TntMessage *message, gpointer data);
static void      receiving_data_block_from_edge (TntTcpIO *tcp_io, TntMessage *message, gpointer data);
static void      tcp_link_change                (FluxTcpIO *tio, FluxTcpLinkMessage message, gpointer data);
static gboolean  maybe_connect_to               (TntNetworkCore *network_core, TntNodeID node_id,
                                                 const TntNodeAddress4 *address);
static void      incoming_connect_established   (TntTcpListener *listener, TntTcpIO *tcp_io, gpointer data);
static void      bandwidth_action_cb            (TntBandwidthGroup *group, TntTransferDirection dir,
                                                 TntBandwidthAdjustment adjust, gpointer data);
static void      rate_node_address              (TntNetworkCore *network_core, TntNodeID node_id,
                                                 TntNodeAddress4 *address, gint rating_change);

/* --------------- *
 * Edge Management *
 * --------------- */

static gboolean
have_edge (TntNetworkCore *network_core, TntEdge *edge)
{
  gint n;

  for (n = 0; n < network_core->edge_array->len; n++)
  {
    TntEdge *found_edge;

    found_edge = g_ptr_array_index (network_core->edge_array, n);
    if (found_edge == edge)
      return TRUE;
  }

  return FALSE;
}

static TntEdge *
find_edge_by_peer_fingerprint (TntNetworkCore *network_core, guint64 peer_fingerprint)
{
  gint n;

  for (n = 0; n < network_core->edge_array->len; n++)
  {
    TntEdge *edge;

    edge = g_ptr_array_index (network_core->edge_array, n);

    if (tnt_edge_get_peer_fingerprint (edge) == peer_fingerprint)
      return edge;
  }

  return NULL;
}

static TntEdge *
find_edge_by_address (TntNetworkCore *network_core, const TntNodeAddress4 *address)
{
  gint n;

  if (!network_core->net_db)
    return NULL;

  for (n = 0; n < network_core->edge_array->len; n++)
  {
    TntEdge         *edge;
    TntNodeAddress4  node_address;

    edge = g_ptr_array_index (network_core->edge_array, n);
    tnt_edge_get_peer_address4 (edge, &node_address);

    if (node_address.ip == address->ip && node_address.port == address->port)
      return edge;
  }

  return NULL;
}

static TntEdge *
find_edge_by_tcp_io (TntNetworkCore *network_core, TntTcpIO *tcp_io)
{
  gint n;

  for (n = 0; n < network_core->edge_array->len; n++)
  {
    TntEdge *edge;

    edge = g_ptr_array_index (network_core->edge_array, n);

    if (tnt_edge_get_io (edge, TNT_STREAM_TYPE_CONTROL) == tcp_io ||
        tnt_edge_get_io (edge, TNT_STREAM_TYPE_DATA)    == tcp_io)
    {
      return edge;
    }
  }

  return NULL;
}

static TntEdge *
find_edge_closest_to_node_id (TntNetworkCore *network_core, TntNodeID node_id)
{
  TntEdge *best_edge      = NULL;
  gdouble  weight_sum     = 0.0;
  gdouble  weight_pick    = 0.0;
  gint     weight_n_edges = 0;
  guint    current_time;
  gint     n;

  current_time = tnt_timestamp ();

  for (n = 0; n < network_core->edge_array->len; n++)
  {
    TntEdge *edge = g_ptr_array_index (network_core->edge_array, n);
    guint    timestamp;

    timestamp = tnt_edge_get_node_timestamp (edge, node_id);

    /* If node wasn't seen in a long time, assume not reachable */
    if (timestamp + (BROADCAST_NODE_INFO_TIMEOUT_MAX * 4) < current_time)
      continue;

    weight_n_edges++;
    weight_sum += tnt_edge_get_node_weight (edge, node_id);
  }

  if (!weight_n_edges)
    return NULL;

  weight_pick = g_random_double_range (0.0, weight_sum);

  for (n = 0; n < network_core->edge_array->len; n++)
  {
    TntEdge *edge = g_ptr_array_index (network_core->edge_array, n);
    guint    timestamp;

    timestamp = tnt_edge_get_node_timestamp (edge, node_id);

    /* If node wasn't seen in a long time, assume not reachable */
    if (timestamp + (BROADCAST_NODE_INFO_TIMEOUT_MAX * 4) < current_time)
      continue;

    best_edge = edge;
    weight_pick -= tnt_edge_get_node_weight (edge, node_id);

    if (weight_pick <= 0.0)
      break;
  }

  return best_edge;
}

static TntEdge *
new_edge (TntNetworkCore *network_core)
{
  TntEdge *edge;

  edge = tnt_edge_new (network_core);
  g_ptr_array_add (network_core->edge_array, edge);

  return edge;
}

static void
remove_edge (TntNetworkCore *network_core, TntEdge *edge)
{
  g_assert (edge);

  if (network_core->net_db)
  {
    TntTimeStamp    now;
    TntNodeAddress4 address;

    now = tnt_timestamp ();
    tnt_edge_get_peer_address4 (edge, &address);

    /* If we were connected for a short while only, as is typical with bootstrap nodes,
     * lower the address' rating so we don't hammer it */
    if (tnt_edge_get_creation_time (edge) + (BOOTSTRAP_EDGE_TIMEOUT * 3) > now)
      tnt_net_db_rate_address (network_core->net_db, &address, FALSE);

    tnt_net_db_unblock_address (network_core->net_db, &address);
  }

  if (network_core->bandwidth_group_list)
    tnt_bandwidth_group_list_remove_edge (network_core->bandwidth_group_list, edge);

  if (!g_ptr_array_remove (network_core->edge_array, edge))
  {
    g_warning ("Removed edge not in network_core array.");
  }

  flux_object_unref (FLUX_OBJECT (edge));
}

static void
set_tcp_io_caps (gpointer key, gpointer value, gpointer data)
{
  TntMessageType  type   = GPOINTER_TO_UINT (key);
  guint           cap    = GPOINTER_TO_UINT (value);
  TntTcpIO       *tcp_io = data;

  tnt_tcp_io_set_queue_out_max_len (tcp_io, type, cap);
}

static gboolean
maybe_add_tcp_io_to_edge (TntEdge *edge, TntTcpIO *tcp_io)
{
  TntNetworkCore *network_core;
  gboolean        result = FALSE;

  network_core = tnt_edge_get_network_core (edge);

  if (!tnt_edge_get_io (edge, TNT_STREAM_TYPE_CONTROL))
  {
    g_hash_table_foreach (network_core->outgoing_caps, set_tcp_io_caps, tcp_io);
    tnt_edge_set_io (edge, TNT_STREAM_TYPE_CONTROL, tcp_io);
    tnt_tcp_io_set_message_in_func (tcp_io, received_message_from_edge, edge);
    result = TRUE;
  }
  else if (!tnt_edge_get_io (edge, TNT_STREAM_TYPE_DATA))
  {
    tnt_edge_set_io (edge, TNT_STREAM_TYPE_DATA, tcp_io);
    tnt_tcp_io_set_data_in_begin_func (tcp_io, receiving_data_block_from_edge, edge);
    tnt_tcp_io_set_message_in_func (tcp_io, received_message_from_edge, edge);
    result = TRUE;
  }

  if (network_core && network_core->bandwidth_group_list)
    tnt_bandwidth_group_list_add_edge (network_core->bandwidth_group_list, edge);

  return result;
}

#if 0

/* FIXME: Not aptly named anymore. Will always remove the entire matched edge */
static void
maybe_remove_tcp_io_from_edge (TntNetworkCore *network_core, TntTcpIO *tcp_io)
{
  TntEdge *edge;

  edge = find_edge_by_tcp_io (network_core, tcp_io);
  if (!edge)
    return;

  tnt_edge_set_io (edge, TNT_STREAM_TYPE_CONTROL, NULL);
  tnt_edge_set_io (edge, TNT_STREAM_TYPE_DATA, NULL);

  remove_edge (network_core, edge);

  if (!tnt_network_core_get_n_edges (network_core))
    tnt_log (TNT_LOG_SUBSYSTEM_NETWORK, TNT_LOG_LEVEL_NORMAL, TNT_LOG_MOOD_DISAPPOINTING, 0,
             "Disconnected from network.");
}

#endif

static TntEdge *
get_least_useful_edge (TntNetworkCore *network_core, TntBandwidthGroup *group)
{
  TntEdge *least_useful_edge = NULL;
  gint     i;

  for (i = 0; i < network_core->edge_array->len; i++)
  {
    TntEdge *edge = g_ptr_array_index (network_core->edge_array, i);

    /* Limit to bandwidth group if requested */
    if (group && !tnt_bandwidth_group_have_edge (group, edge))
      continue;

    if (!least_useful_edge)
    {
      least_useful_edge = edge;
    }
    else if (!tnt_edge_get_qualified (edge) &&
             tnt_edge_get_score (edge) < tnt_edge_get_score (least_useful_edge))
    {
      least_useful_edge = edge;
    }
  }

  return least_useful_edge;
}

/* ----------------------- *
 * Pending Edge Management *
 * ----------------------- */

static gint
find_pending_edge_by_address (TntNetworkCore *network_core, const TntNodeAddress4 *address)
{
  gint n;

  for (n = 0; n < network_core->pending_edge_array->len; n++)
  {
    PendingEdge *pending_edge;

    pending_edge = g_ptr_array_index (network_core->pending_edge_array, n);

    if (pending_edge->address.ip   == address->ip &&
        pending_edge->address.port == address->port)
      return n;
  }

  return -1;
}

static gint
find_pending_edge_by_tcp_io (TntNetworkCore *network_core, TntTcpIO *tcp_io)
{
  gint n;

  for (n = 0; n < network_core->pending_edge_array->len; n++)
  {
    PendingEdge *pending_edge;

    pending_edge = g_ptr_array_index (network_core->pending_edge_array, n);

    if (pending_edge->tcp_io == tcp_io)
      return n;
  }

  return -1;
}

static gint
find_pending_edge_by_edge (TntNetworkCore *network_core, TntEdge *edge)
{
  gint n;

  for (n = 0; n < network_core->pending_edge_array->len; n++)
  {
    PendingEdge *pending_edge;

    pending_edge = g_ptr_array_index (network_core->pending_edge_array, n);

    if (pending_edge->edge == edge)
      return n;
  }

  return -1;
}

static PendingEdge *
get_pending_edge_n (TntNetworkCore *network_core, gint n)
{
  if (n < 0)
    return NULL;

  return g_ptr_array_index (network_core->pending_edge_array, n);
}

static gint
find_pending_edge (TntNetworkCore *network_core, PendingEdge *pending_edge)
{
  gint n;

  for (n = 0; n < network_core->pending_edge_array->len; n++)
  {
    if (pending_edge == g_ptr_array_index (network_core->pending_edge_array, n))
      return n;
  }

  return -1;
}

static PendingEdge *
new_pending_edge (TntNetworkCore *network_core, TntTcpIO *tcp_io)
{
  PendingEdge *pending_edge;

  pending_edge = g_new0 (PendingEdge, 1);
  g_ptr_array_add (network_core->pending_edge_array, pending_edge);

  pending_edge->tcp_io = tcp_io;

  return pending_edge;
}

static void
remove_pending_edge (TntNetworkCore *network_core, gint n)
{
  PendingEdge *pending_edge;

  if (n == -1)
    return;

  /* FIXME: Could be remove_index_fast()? */
  pending_edge = g_ptr_array_remove_index (network_core->pending_edge_array, n);

  if (pending_edge->edge)
    flux_object_unref (FLUX_OBJECT (pending_edge->edge));
  if (pending_edge->tcp_io)
    flux_object_unref (FLUX_OBJECT (pending_edge->tcp_io));

  g_free (pending_edge);
}

static void
remove_pending_edges_by_edge (TntNetworkCore *network_core, TntEdge *edge)
{
  gint pending_edge_n;

  while ((pending_edge_n = find_pending_edge_by_edge (network_core, edge)) >= 0)
  {
    remove_pending_edge (network_core, pending_edge_n);
  }
}

/* ---------------- *
 * Message Handling *
 * ---------------- */

/* --- Handlers --- */

static gboolean
handle_incoming_connect_request (TntNetworkCore *network_core, TntEdge *source_edge, TntMessage *message)
{
  TntConnectRequestM *connect_request_m;
  TntNodeAddress4     address;

  if (!tnt_message_has_payload (message))
    return FALSE;

  if (!tnt_message_is_routed (message))
    return FALSE;

  if (tnt_routed_message_get_destination (TNT_ROUTED_MESSAGE (message)) != network_core->my_node_id)
    return TRUE;

  /* Disregard connection requests if we want to remain anonymous */
  if (network_core->hide_my_ip)
    return TRUE;

  connect_request_m = tnt_message_get_struct (message);
  if (!connect_request_m)
    return TRUE;

  /* Make sure we have the address (though we might have a better one already) */
  rate_node_address (network_core, connect_request_m->id, &connect_request_m->address, 0);

  if (network_core->net_db && connect_request_m->id != TNT_NODE_ID_INVALID)
  {
    TntNodeMeta *node_meta;

    node_meta = tnt_net_db_get_node_meta (network_core->net_db, connect_request_m->id);
    if (node_meta)
    {
      tnt_node_meta_get_best_address (node_meta, &address);
      tnt_node_meta_free (node_meta);
    }
    else
      memcpy (&address, &connect_request_m->address, sizeof (TntNodeAddress4));
  }
  else
    memcpy (&address, &connect_request_m->address, sizeof (TntNodeAddress4));

  maybe_connect_to (network_core, connect_request_m->id, &address);

  g_free (connect_request_m);
  return TRUE;
}

static gboolean
handle_incoming_peer_reverse_info (TntNetworkCore *network_core, TntEdge *source_edge, TntMessage *message)
{
  TntPeerReverseInfoM *peer_reverse_info;

  if (!tnt_message_has_payload (message))
    return FALSE;

  peer_reverse_info = tnt_message_get_struct (message);
  if (!peer_reverse_info)
    return TRUE;

  if (network_core->my_node_address.ip == TNT_NODE_IP_INVALID)
  {
    network_core->my_node_address.ip = peer_reverse_info->ip4;
  }
  else if (network_core->my_node_address.ip != peer_reverse_info->ip4)
  {
    FluxIPAddr *new_ip_addr = tnt_ip_to_ip_addr (peer_reverse_info->ip4);

    if (flux_ip_addr_is_internet (new_ip_addr))
      network_core->my_node_address.ip = peer_reverse_info->ip4;

    flux_object_unref (FLUX_OBJECT (new_ip_addr));
  }

  g_free (peer_reverse_info);
  return TRUE;
}

static gboolean
handle_incoming_node_info (TntNetworkCore *network_core, TntEdge *source_edge, TntMessage *message)
{
  TntNodeInfoM *node_info;

  if (!tnt_message_has_payload (message))
    return FALSE;

  node_info = tnt_message_get_struct (message);
  if (!node_info)
    return FALSE;

  /* TODO: Save node name if any */

  tnt_message_free_struct (message, node_info);
  return TRUE;
}

static gboolean
handle_incoming_contact_info (TntNetworkCore *network_core, TntEdge *source_edge, TntMessage *message)
{
  TntContactInfoM *contact_info;

  if (!tnt_message_has_payload (message))
    return FALSE;

  contact_info = tnt_message_get_struct (message);
  if (!contact_info)
    return FALSE;

  if (contact_info->id != TNT_NODE_ID_INVALID)
  {
    /* Non-anonymous peer */
    rate_node_address (network_core, contact_info->id, &contact_info->address, 0);
  }
  else
  {
    /* Anonymous peer */
    if (network_core->net_db && contact_info->address.ip != TNT_NODE_IP_INVALID &&
        contact_info->address.port != TNT_NODE_PORT_INVALID)
      tnt_net_db_add_address (network_core->net_db, &contact_info->address);
  }

  tnt_message_free_struct (message, contact_info);
  return TRUE;
}

/* --- Dispatcher --- */

typedef struct
{
  TntMessageType message_type;
  gpointer       func;
  gboolean       may_generate_replies;
}
MessageHandlerMap;

static MessageHandlerMap message_handler_map [] =
{
  { TNT_MESSAGE_CONNECT_REQUEST,    handle_incoming_connect_request,   TRUE  },
  { TNT_MESSAGE_PEER_REVERSE_INFO,  handle_incoming_peer_reverse_info, FALSE },
  { TNT_MESSAGE_NODE_INFO,          handle_incoming_node_info,         FALSE },
  { TNT_MESSAGE_CONTACT_INFO,       handle_incoming_contact_info,      FALSE },

  { TNT_MESSAGE_UNKNOWN, NULL, FALSE }
};

static void
attach_message_handlers (TntNetworkCore *network_core)
{
  gint i;

  for (i = 0; message_handler_map [i].func; i++)
  {
    tnt_network_core_attach_message_handler (network_core,
                                             message_handler_map [i].message_type,
                                             message_handler_map [i].may_generate_replies,
                                             (TntMessageHandler *) message_handler_map [i].func,
                                             network_core);
  }
}

static void
detach_message_handlers (TntNetworkCore *network_core)
{
  gint i;

  for (i = 0; message_handler_map [i].func; i++)
  {
    tnt_network_core_detach_message_handler (network_core, message_handler_map [i].message_type,
                                             (TntMessageHandler *) message_handler_map [i].func,
                                             network_core);
  }
}

static gboolean
handle_edge_message (TntNetworkCore *network_core, TntEdge *source_edge, TntMessage *message)
{
  TntMessageType   type;
  TntCallbackList *callback_list;
  gboolean         keep_routing;

  type = tnt_message_get_type (message);

  callback_list = g_hash_table_lookup (network_core->message_cb_list_table, GUINT_TO_POINTER (type));
  if (!callback_list)
    return TRUE;

  /* If message may trigger replies, and "Introduce Delays" is on, we stick it
   * in the time warp. It'll come back later, and we'll deal with it then. */
  if (network_core->introduce_delays && !tnt_message_is_time_warped (message) &&
      g_hash_table_lookup (network_core->message_replies_table,
                           GUINT_TO_POINTER (type)))
  {
    DEBUG (g_print ("Time warp in!\n"));
    tnt_time_warp_push_message (network_core->time_warp, message, source_edge);
    return TRUE;  /* Always route */
  }

  tnt_callback_list_run_while_with_result_3 (callback_list, gboolean, keep_routing, TRUE, tnt_test_true,
                                             network_core, source_edge, message);

  return keep_routing;
}

/* ------------------ *
 * Message Processing *
 * ------------------ */

static void
broadcast_message (TntNetworkCore *network_core, TntEdge *source_edge, TntMessage *message)
{
  gint n;

  DEBUG
  (
    g_print ("[broadcast] ");
    tnt_message_dump (message);
  )

  for (n = 0; n < network_core->edge_array->len; n++)
  {
    TntEdge *edge;

    edge = g_ptr_array_index (network_core->edge_array, n);

    /* Don't return along the edge it came from. */
    if (edge == source_edge)
      continue;

#if 0
    /* NOTE: We can no longer know who the edge represents */

    /* Don't return to sender. */
    if (tnt_message_is_routed (message) &&
        tnt_routed_message_get_source (TNT_ROUTED_MESSAGE (message)) == tnt_edge_get_peer_id (edge))
      continue;
#endif

    tnt_edge_send_message (edge, message);
  }
}

static gboolean
filter_incoming_message (TntNetworkCore *network_core, TntEdge *edge, TntMessage *message)
{
  /* Refuse incoming messages if we're just bootstrapping others */
  if (network_core->bootstrap_only)
    return FALSE;

  /* If message says who it's from... */
  if (tnt_message_is_routed (message) && tnt_routed_message_has_source (TNT_ROUTED_MESSAGE (message)))
  {
    TntNodeID source_id;

    source_id = tnt_routed_message_get_source (TNT_ROUTED_MESSAGE (message));

    /* Drop messages that were sent by us and came back, but only if we don't care
     * about privacy. */
    if (source_id == network_core->my_node_id && !network_core->forward_own_messages)
    {
      DEBUG
      (
        g_print ("[drop] ");
        tnt_message_dump (message);
      )
      return FALSE;
    }

    /* Update reverse-routing statistics */
    tnt_edge_add_node_weight (edge, source_id, tnt_message_get_size_estimate (message));
  }

  return TRUE;
}

static gboolean
filter_outgoing_message (TntNetworkCore *network_core, TntMessage *message)
{
  return tnt_message_filter_allow_message (network_core->filter, message);
}

static gboolean
decrement_ttl (TntMessage *message)
{
  gint ttl;
  gint n;

  ttl = tnt_message_get_ttl (message);
  n = g_random_int_range (0, 3);
  ttl = MIN (0, ttl - n);

  tnt_message_set_ttl (message, ttl);

  if (n == 0)
    return FALSE;

  return TRUE;
}

static TntEdge *
route_message (TntNetworkCore *network_core, TntEdge *edge, TntMessage *message)
{
  TntEdge *best_edge = NULL;

  DEBUG
  (
    g_print ("[routing] ");
    tnt_message_dump (message);
  )

  if (tnt_message_get_direct (message) && edge != NULL)
  {
    /* Direct messages are for our eyes only. */
    DEBUG
    (
      g_print ("[direct for me] ");
      tnt_message_dump (message);
    )
    return NULL;
  }

  if (tnt_message_is_routed (message) && tnt_routed_message_has_destination (TNT_ROUTED_MESSAGE (message)))
  {
    /* Drop messages for which we were the final destination. */
    if (tnt_routed_message_get_destination (TNT_ROUTED_MESSAGE (message)) == network_core->my_node_id)
    {
      DEBUG
      (
        g_print ("[routed for me] ");
        tnt_message_dump (message);
      )

      /* However, if we are using "Forward Own Messages" obfuscation of identity, we may
       * decide to forward this message anyway, on the pretense that it wasn't for us. */
      if (network_core->forward_own_messages)
      {
        gdouble message_size;
        gdouble forward_p;

        message_size = tnt_message_get_size_estimate (message);
        message_size = MIN (message_size, (gdouble) MESSAGE_SIZE_CLAMP);

        /* Bigger messages have a lower probability of being forwarded */
        forward_p = FORWARD_OWN_MESSAGE_PROBABILITY_MIN +
          (((gdouble) MESSAGE_SIZE_CLAMP - message_size) *
           (FORWARD_OWN_MESSAGE_PROBABILITY_MAX - FORWARD_OWN_MESSAGE_PROBABILITY_MIN)) /
          (gdouble) MESSAGE_SIZE_CLAMP;

        if (forward_p < g_random_double ())
          return NULL;
      }
      else
      {
        return NULL;
      }
    }

    best_edge = find_edge_closest_to_node_id (network_core,
      tnt_routed_message_get_destination (TNT_ROUTED_MESSAGE (message)));

    if (best_edge)
    {
      /* Send message on through what is hopefully the shortest path. */
      DEBUG
      (
        g_print ("[found route] ");
        tnt_message_dump (message);
      )
      tnt_edge_send_message (best_edge, message);
    }
    else if (decrement_ttl (message))
    {
      /* No viable path; broadcast in hopes that neighbouring nodes will know better. */
      broadcast_message (network_core, edge, message);
    }
  }
  else if (decrement_ttl (message))
  {
    /* No destination specified; broadcast message. */
    broadcast_message (network_core, edge, message);
  }

  return best_edge;
}

static void
received_message_from_edge (TntTcpIO *tcp_io, TntMessage *message, gpointer data)
{
  TntEdge        *edge = data;
  TntNetworkCore *network_core;

  flux_object_check (FLUX_OBJECT (edge), TNT_EDGE_TYPE);
  network_core = tnt_edge_get_network_core (edge);
  assert_types (network_core);
  flux_object_check (FLUX_OBJECT (tcp_io), FLUX_TCP_IO_TYPE);
  flux_object_check (FLUX_OBJECT (message), TNT_MESSAGE_TYPE);

  if (filter_incoming_message (network_core, edge, message))
  {
    if (handle_edge_message (network_core, edge, message) &&
        filter_outgoing_message (network_core, message))
    {
      route_message (network_core, edge, message);
    }
  }

  flux_object_unref (FLUX_OBJECT (message));
}

static gboolean
received_message_from_time_warp (TntNetworkCore *network_core, TntEdge *source_edge,
                                 TntMessage *message, gpointer data)
{
  DEBUG (g_print ("Time warp out!\n"));

  /* The time warp doesn't ref the edge, so we have to check if it still exists. */
  if (!have_edge (network_core, source_edge))
    source_edge = NULL;

  handle_edge_message (network_core, source_edge, message);

  return TRUE;  /* Meaningless */
}

/* We get this callback after a payload has been received on a data socket, and a data block
 * will follow the payload. If the data block appears to be part of a file we're getting,
 * add a qualification ref to the edge, so it will get priority bandwidth. */
static void
receiving_data_block_from_edge (TntTcpIO *tcp_io, TntMessage *message, gpointer data)
{
  TntEdge         *edge = data;
  TntMessageType   type;
  TntNetworkCore  *network_core;
  TntCallbackList *callback_list;

  flux_object_check (FLUX_OBJECT (edge), TNT_EDGE_TYPE);
  network_core = tnt_edge_get_network_core (edge);
  assert_types (network_core);
  flux_object_check (FLUX_OBJECT (tcp_io), FLUX_TCP_IO_TYPE);
  flux_object_check (FLUX_OBJECT (message), TNT_MESSAGE_TYPE);

  type = tnt_message_get_type (message);

  callback_list = g_hash_table_lookup (network_core->pre_data_cb_list_table, GUINT_TO_POINTER (type));
  if (callback_list)
    tnt_callback_list_run_while_3 (callback_list, tnt_test_always_true, network_core, edge, message);

  flux_object_unref (FLUX_OBJECT (message));
}

static void
send_reverse_info_to_edge (TntEdge *edge)
{
  TntNetworkCore      *network_core;
  TntPeerReverseInfoM  peer_reverse_info_m;
  TntMessage          *message;

  network_core = tnt_edge_get_network_core (edge);

  memset (&peer_reverse_info_m, 0, sizeof (TntPeerReverseInfoM));
  peer_reverse_info_m.ip4 = tnt_edge_get_peer_ip4 (edge);

  message = tnt_message_construct_direct (TNT_MESSAGE_PEER_REVERSE_INFO, &peer_reverse_info_m, NULL, 0);
  tnt_edge_send_message (edge, message);
  flux_object_unref (FLUX_OBJECT (message));
}

static void
broadcast_node_info (TntNetworkCore *network_core, TntEdge *edge)
{
  TntConfig        *config;
  TntNodeInfoM      node_info;
  TntRoutedMessage *message;
  gchar            *my_name;

  memset (&node_info, 0, sizeof (TntNodeInfoM));
  node_info.id = network_core->my_node_id;

  config = tnt_core_peek_config (TNT_CORE (network_core));
  if (config)
  {
    my_name = tnt_config_get_str (config, "network", "name", NULL);
    if (my_name)
    {
      tnt_utf8_strlcpy (node_info.name, my_name, TNT_NODE_NAME_LEN_MAX);
      g_free (my_name);
    }
  }

  message = tnt_message_construct_routed (TNT_MESSAGE_NODE_INFO, &node_info, NULL, 0, 10,
                                          network_core->my_node_id, TNT_NODE_ID_INVALID);

  if (edge)
    tnt_edge_send_message (edge, TNT_MESSAGE (message));
  else
    broadcast_message (network_core, NULL, TNT_MESSAGE (message));

  flux_object_unref (FLUX_OBJECT (message));
}

static void
broadcast_contact_info (TntNetworkCore *network_core, TntEdge *edge)
{
  TntContactInfoM   contact_info;
  TntRoutedMessage *message;
  gint              n;

  memset (&contact_info, 0, sizeof (TntContactInfoM));

  if (!network_core->my_node_address.port)
    return;

  if (network_core->hide_my_ip)
    contact_info.id = TNT_NODE_ID_INVALID;
  else
    contact_info.id = network_core->my_node_id;

  n = network_core->my_node_address.ip == TNT_NODE_IP_INVALID ? 0 : 1;
  n += g_list_length (network_core->my_alt_address_list);
  if (!n)
    return;

  n = g_random_int_range (0, n);
  if (network_core->my_node_address.ip != TNT_NODE_IP_INVALID)
    n--;

  if (n < 0)
  {
    memcpy (&contact_info.address, &network_core->my_node_address, sizeof (TntNodeAddress4));
  }
  else
  {
    GList *l;

    for (l = network_core->my_alt_address_list; l; l = g_list_next (l))
    {
      TntNodeAddress4 *address = l->data;

      if (!n)
      {
        memcpy (&contact_info.address, address, sizeof (TntNodeAddress4));
        break;
      }

      n--;
    }
  }

  message = tnt_message_construct_routed (TNT_MESSAGE_CONTACT_INFO, &contact_info, NULL, 0, 10,
                                          contact_info.id, TNT_NODE_ID_INVALID);

  if (edge)
    tnt_edge_send_message (edge, TNT_MESSAGE (message));
  else
    broadcast_message (network_core, NULL, TNT_MESSAGE (message));

  flux_object_unref (FLUX_OBJECT (message));
}

static gboolean
broadcast_node_info_timeout_cb (gpointer data)
{
  TntNetworkCore *network_core = data;
  guint           new_timeout;

  broadcast_node_info (network_core, NULL);

  new_timeout = g_random_int_range (BROADCAST_NODE_INFO_TIMEOUT_MIN,
                                    BROADCAST_NODE_INFO_TIMEOUT_MAX);
  network_core->broadcast_node_info_timeout_id =
    g_timeout_add (new_timeout, broadcast_node_info_timeout_cb, network_core);

  return FALSE;
}

static gboolean
broadcast_contact_info_timeout_cb (gpointer data)
{
  TntNetworkCore *network_core = data;
  guint           new_timeout;

  broadcast_contact_info (network_core, NULL);

  new_timeout = g_random_int_range (BROADCAST_CONTACT_INFO_TIMEOUT_MIN,
                                    BROADCAST_CONTACT_INFO_TIMEOUT_MAX);
  network_core->broadcast_contact_info_timeout_id =
    g_timeout_add (new_timeout, broadcast_contact_info_timeout_cb, network_core);

  return FALSE;
}

static gboolean
bootstrap_edges_drop (gpointer data)
{
  TntNetworkCore *network_core = data;
  GList          *edges        = NULL;
  TntTimeStamp    now;
  GList          *l;
  gint            i;

  now = tnt_timestamp ();

  for (i = 0; i < network_core->edge_array->len; i++)
  {
    TntEdge      *edge = g_ptr_array_index (network_core->edge_array, i);
    TntTimeStamp  edge_created;

    edge_created = tnt_edge_get_creation_time (edge);

    /* We do both + and - so that clock skew won't make edges hang around forever */

    if (edge_created + BOOTSTRAP_EDGE_TIMEOUT < now ||
        edge_created - BOOTSTRAP_EDGE_TIMEOUT > now)
    {
      edges = g_list_prepend (edges, edge);
    }
  }

  for (l = edges; l; l = g_list_next (l))
  {
    TntEdge *edge = l->data;
    remove_edge (network_core, edge);
  }

  g_list_free (edges);
  return TRUE;  /* Keep running */
}

static void
send_bootstrap_address (TntEdge *edge, TntNodeAddress4 *address)
{
  TntContactInfoM  contact_info;
  TntMessage      *message;

  memset (&contact_info, 0, sizeof (TntContactInfoM));
  contact_info.id = TNT_NODE_ID_INVALID;
  memcpy (&contact_info.address, address, sizeof (TntNodeAddress4));

  message = tnt_message_construct_direct (TNT_MESSAGE_CONTACT_INFO, &contact_info, NULL, 0);
  tnt_edge_send_message (edge, TNT_MESSAGE (message));
  flux_object_unref (FLUX_OBJECT (message));
}

static void
do_initial_edge_tasks (TntEdge *edge)
{
  TntNetworkCore *network_core;

  network_core = tnt_edge_get_network_core (edge);

  /* Inform peer about its own apparent address */
  send_reverse_info_to_edge (edge);

  if (network_core->bootstrap_only)
  {
    /* Send a list of randomly picked addresses */

    if (network_core->net_db)
    {
      GList *addresses;
      GList *l;

      addresses = tnt_net_db_pick_addresses (network_core->net_db, 0x00000000, 0x00000000,
                                             BOOTSTRAP_ADDRESSES_MAX);

      for (l = addresses; l; l = g_list_next (l))
      {
        TntNodeAddress4 *address = l->data;
        send_bootstrap_address (edge, address);
      }
      
      tnt_list_free_with_data (addresses);
    }
  }

  if (!network_core->hide_my_ip && !network_core->introduce_delays && !network_core->bootstrap_only)
  {
    /* Broadcast my node info through edge */
    broadcast_node_info (network_core, edge);
    broadcast_contact_info (network_core, edge);
  }

  /* Let other components do their stuff */
  tnt_callback_list_run_while_2 (network_core->new_edge_cb_list, tnt_test_always_true, network_core, edge);
}

static gint
address_bandwidth_rating (TntNetworkCore *network_core, TntNodeAddress4 *address)
{
  TntBandwidthGroup *bandwidth_group;

  if (address->ip == TNT_NODE_IP_INVALID)
    return -1;

  bandwidth_group = tnt_bandwidth_group_list_find_group_by_ip (network_core->bandwidth_group_list,
                                                               address->ip);
  if (!bandwidth_group)
    return 0;

  return tnt_bandwidth_group_get_goal (bandwidth_group, TNT_TRANSFER_DIRECTION_IN) +
         tnt_bandwidth_group_get_goal (bandwidth_group, TNT_TRANSFER_DIRECTION_OUT);
}

static void
rate_node_address (TntNetworkCore *network_core, TntNodeID node_id, TntNodeAddress4 *address,
                   gint rating_change)
{
  TntNodeMeta     *node_meta;
  TntNodeAddress4  pri_address, sec_address;

  if (!network_core->net_db)
    return;

  if (node_id == TNT_NODE_ID_INVALID)
    return;

  node_meta = tnt_net_db_get_node_meta (network_core->net_db, node_id);
  if (!node_meta)
    node_meta = tnt_node_meta_new ();

  tnt_node_meta_get_primary_address   (node_meta, &pri_address);
  tnt_node_meta_get_secondary_address (node_meta, &sec_address);

  if (pri_address.ip == address->ip && pri_address.port == address->port)
  {
    tnt_node_meta_rate_primary_address (node_meta, rating_change);
  }
  else if (sec_address.ip == address->ip && sec_address.port == address->port)
  {
    tnt_node_meta_rate_secondary_address (node_meta, rating_change);
  }
  else
  {
    gint new_rating, pri_rating, sec_rating;

    /* New address */

    new_rating = address_bandwidth_rating (network_core, address);
    pri_rating = address_bandwidth_rating (network_core, &pri_address);
    sec_rating = address_bandwidth_rating (network_core, &sec_address);

    if (new_rating >= pri_rating)
    {
      if (new_rating > pri_rating && pri_rating > 0)
        tnt_node_meta_set_secondary_address (node_meta, &pri_address);

      tnt_node_meta_set_primary_address (node_meta, address);
      tnt_node_meta_rate_primary_address (node_meta, rating_change);
    }
    else
    {
      tnt_node_meta_set_secondary_address (node_meta, address);
      tnt_node_meta_rate_secondary_address (node_meta, rating_change);
    }
  }

  /* If primary address is failing, nullify it so secondary address can be used */
  if (tnt_node_meta_get_primary_rating (node_meta) < 0)
  {
    pri_address.ip   = TNT_NODE_IP_INVALID;
    pri_address.port = TNT_NODE_PORT_INVALID;
    tnt_node_meta_set_primary_address (node_meta, &pri_address);
  }

  tnt_net_db_set_node_meta (network_core->net_db, node_id, node_meta);
  tnt_node_meta_free (node_meta);
}

static void
convert_pending_to_edge (TntNetworkCore *network_core, gint pending_edge_n,
                         guint64 fingerprint, TntNodeID node_id, gint listening_port)
{
  TntEdge         *edge;
  PendingEdge     *pending_edge;
  TntNodeAddress4  address;
  TntNodeID        pending_node_id;
  gboolean         outbound;

  pending_edge = get_pending_edge_n (network_core, pending_edge_n);
  g_assert (pending_edge);

  outbound = pending_edge->outbound;
  pending_node_id = pending_edge->node_id;
  memcpy (&address, &pending_edge->address, sizeof (TntNodeAddress4));

  /* Make sure we have an edge for this peer */

  edge = find_edge_by_peer_fingerprint (network_core, fingerprint);
  if (!edge)
  {
    if (!tnt_network_core_get_n_edges (network_core))
      tnt_log (TNT_LOG_SUBSYSTEM_NETWORK, TNT_LOG_LEVEL_NORMAL, TNT_LOG_MOOD_PLEASING, 0,
               "Connected to network.");
    edge = new_edge (network_core);
    tnt_edge_set_peer_fingerprint (edge, fingerprint);
    tnt_edge_set_remote_port (edge, outbound ? address.port : listening_port);
  }

  /* Update stored node metadata */

  if (node_id != TNT_NODE_ID_INVALID)
    rate_node_address (network_core, node_id, &address, 1);

  /* Maybe add connection to edge, removing the pending edge afterwards.
   * The edge will ref it, the pending edge will unref. */

  if (maybe_add_tcp_io_to_edge (edge, pending_edge->tcp_io) &&
      tnt_edge_get_io (edge, TNT_STREAM_TYPE_CONTROL) &&
      tnt_edge_get_io (edge, TNT_STREAM_TYPE_DATA))
  {
    /* If we just finished connecting to the edge, perform initial tasks,
     * like sending running queries, asking for blocks, broadcasting node
     * info, etc. */
    do_initial_edge_tasks (edge);
  }

  if (network_core->net_db && (outbound || network_core->bootstrap_only))
    tnt_net_db_rate_address (network_core->net_db, &address, TRUE);

  remove_pending_edge (network_core, pending_edge_n);

  /* If the edge doesn't have both control and data connections,
   * initiate another connection. The connecting node is responsible for
   * establishing both connections. */

  if (outbound && !(tnt_edge_get_io (edge, TNT_STREAM_TYPE_CONTROL) &&
                    tnt_edge_get_io (edge, TNT_STREAM_TYPE_DATA)))
    maybe_connect_to (network_core, pending_node_id, &address);
}

static void
received_message_from_pending_edge (TntTcpIO *tcp_io, TntMessage *message, gpointer data)
{
  TntNetworkCore *network_core = data;
  TntPeerInfoM   *peer_info;

  assert_types (network_core);
  flux_object_check (FLUX_OBJECT (tcp_io), TNT_TCP_IO_TYPE);
  flux_object_check (FLUX_OBJECT (message), TNT_MESSAGE_TYPE);

  switch (tnt_message_get_type (message))
  {
    case TNT_MESSAGE_PEER_INFO:
      DEBUG (g_print ("Got peer info from pending edge.\n"));

      peer_info = tnt_message_get_struct (message);
      if (peer_info)
      {
        gint         pending_edge_n;
        PendingEdge *pending_edge;

        /* Find the pending edge and fish out some parameters we'll need
         * after it goes away. */

        pending_edge_n = find_pending_edge_by_tcp_io (network_core, tcp_io);
        g_assert (pending_edge_n >= 0);

        pending_edge = get_pending_edge_n (network_core, pending_edge_n);

        if (peer_info->fingerprint != network_core->my_fingerprint)
        {
          TntNodeAddress4 address;

          memcpy (&address, &pending_edge->address, sizeof (TntNodeAddress4));
          if (!pending_edge->outbound)
            address.port = peer_info->listening_port;
          if (network_core->net_db && address.port != 0)
            tnt_net_db_add_address (network_core->net_db, &address);

          convert_pending_to_edge (network_core, pending_edge_n, peer_info->fingerprint,
                                   peer_info->id, peer_info->listening_port);
        }
        else
        {
          /* Don't accept connections from ourself */
          if (network_core->net_db && pending_edge->outbound)
          {
            tnt_net_db_rate_address (network_core->net_db, &pending_edge->address, FALSE);
            tnt_net_db_unblock_address (network_core->net_db, &pending_edge->address);
          }

          _tnt_tcp_io_be_dead (tcp_io);
          remove_pending_edge (network_core, pending_edge_n);
        }

        g_free (peer_info);
      }
      break;

    default:
      DEBUG (g_print ("Got unknown message from pending edge!\n"));
      break;
  }

  flux_object_unref (FLUX_OBJECT (message));
}

/* --------------------- *
 * Connection Management *
 * --------------------- */

static void
send_peer_info (TntNetworkCore *network_core, TntTcpIO *tcp_io)
{
  TntPeerInfoM  peer_info;
  TntMessage   *message;

  if (network_core->hide_my_ip)
    peer_info.id = TNT_NODE_ID_INVALID;
  else
    peer_info.id = network_core->my_node_id;

  peer_info.fingerprint    = network_core->my_fingerprint;
  peer_info.listening_port = network_core->my_node_address.port;

  message = tnt_message_construct_direct (TNT_MESSAGE_PEER_INFO, &peer_info, NULL, 0);
  tnt_tcp_io_message_queue_out (tcp_io, message);
  flux_object_unref (FLUX_OBJECT (message));
}

static void
incoming_connect_established (TntTcpListener *listener, TntTcpIO *tcp_io, gpointer data)
{
  TntNetworkCore *network_core = data;
  PendingEdge    *pending_edge;
  TntNodeIP4      remote_ip4;

  assert_types (network_core);
  flux_object_check (FLUX_OBJECT (listener), TNT_TCP_LISTENER_TYPE);
  flux_object_check (FLUX_OBJECT (tcp_io), TNT_TCP_IO_TYPE);

  /* Check that remote IP matches a bandwidth group */
  remote_ip4 = tnt_tcp_io_get_remote_ip4 (tcp_io);
  if (!network_core->bandwidth_group_list ||
      !tnt_bandwidth_group_list_find_group_by_ip (network_core->bandwidth_group_list, remote_ip4))
  {
    flux_object_unref (FLUX_OBJECT (tcp_io));
    return;
  }

  pending_edge = new_pending_edge (network_core, tcp_io);
  pending_edge->node_id    = TNT_NODE_ID_INVALID;
  pending_edge->address.ip = remote_ip4;

  flux_tcp_io_set_link_func (FLUX_TCP_IO (pending_edge->tcp_io), tcp_link_change, network_core);
  tnt_tcp_io_set_message_in_func (tcp_io, received_message_from_pending_edge, network_core);
  send_peer_info (network_core, tcp_io);
}

static void
outgoing_connect_established (TntNetworkCore *network_core, TntTcpIO *tcp_io)
{
  TntNodeIP4 remote_ip4;

  assert_types (network_core);
  flux_object_check (FLUX_OBJECT (tcp_io), TNT_TCP_IO_TYPE);

  /* Check that remote IP matches a bandwidth group */
  remote_ip4 = tnt_tcp_io_get_remote_ip4 (tcp_io);
  if (!network_core->bandwidth_group_list ||
      !tnt_bandwidth_group_list_find_group_by_ip (network_core->bandwidth_group_list, remote_ip4))
  {
    _tnt_tcp_io_be_dead (tcp_io);
    flux_object_unref (FLUX_OBJECT (tcp_io));
    return;
  }

  tnt_tcp_io_set_message_in_func (tcp_io, received_message_from_pending_edge, network_core);
  send_peer_info (network_core, tcp_io);
}

static void
tcp_link_change (FluxTcpIO *tio, FluxTcpLinkMessage message, gpointer data)
{
  TntNetworkCore   *network_core   = data;
  TntTcpIO        *tcp_io       = TNT_TCP_IO (tio);
  gint             pending_edge_n;
  PendingEdge     *pending_edge;
  TntEdge         *edge;
  TntNodeAddress4  address;

  assert_types (network_core);
  flux_object_check (FLUX_OBJECT (tcp_io), TNT_TCP_IO_TYPE);

  /* Get edge related to I/O */

  edge           = find_edge_by_tcp_io (network_core, tcp_io);
  pending_edge_n = find_pending_edge_by_tcp_io (network_core, tcp_io);

  if (pending_edge_n >= 0)
  {
    pending_edge = get_pending_edge_n (network_core, pending_edge_n);
    memcpy (&address, &pending_edge->address, sizeof (TntNodeAddress4));
  }
  else
  {
    pending_edge = NULL;
  }

  switch (message)
  {
    case FLUX_TCP_CONNECTION_ESTABLISHED:
#if 0
      tnt_log (TNT_LOG_SUBSYSTEM_NETWORK, TNT_LOG_LEVEL_DEBUG, TNT_LOG_MOOD_PLEASING, 0,
               "Established outbound connection.");
#endif

      outgoing_connect_established (network_core, tcp_io);
      break;

    case FLUX_TCP_CONNECTION_FAILED:
#if 0
      tnt_log (TNT_LOG_SUBSYSTEM_NETWORK, TNT_LOG_LEVEL_DEBUG, TNT_LOG_MOOD_DISAPPOINTING, 0,
               "An outbound connection attempt failed.");

      /* If an outbound connection attempt fails, forget about that address */
      if (pending_edge && network_core->net_db)
      {
        tnt_net_db_remove_ip_address (network_core->net_db, &pending_edge->address);
      }
#endif

      if (pending_edge)
      {
        if (pending_edge->node_id != TNT_NODE_ID_INVALID)
          rate_node_address (network_core, pending_edge->node_id, &pending_edge->address, -2);

        if (network_core->net_db && pending_edge->outbound)
        {
          tnt_net_db_rate_address (network_core->net_db, &pending_edge->address, FALSE);
          tnt_net_db_unblock_address (network_core->net_db, &pending_edge->address);
        }
      }

      _tnt_tcp_io_be_dead (tcp_io);
      remove_pending_edge (network_core, pending_edge_n);
      break;

    case FLUX_TCP_CONNECTION_CLOSED:
#if 0
      tnt_log (TNT_LOG_SUBSYSTEM_NETWORK, TNT_LOG_LEVEL_DEBUG, TNT_LOG_MOOD_DISAPPOINTING, 0,
               "A connection closed.");
#endif

      if (pending_edge)
      {
        if (network_core->net_db && pending_edge->outbound)
        {
          tnt_net_db_rate_address (network_core->net_db, &pending_edge->address, FALSE);
          tnt_net_db_unblock_address (network_core->net_db, &pending_edge->address);
        }

        _tnt_tcp_io_be_dead (tcp_io);
        remove_pending_edge (network_core, pending_edge_n);
      }

      if (edge)
      {
        _tnt_tcp_io_be_dead (tcp_io);
        remove_pending_edges_by_edge (network_core, edge);
        remove_edge (network_core, edge);

        if (!tnt_network_core_get_n_edges (network_core))
          tnt_log (TNT_LOG_SUBSYSTEM_NETWORK, TNT_LOG_LEVEL_NORMAL, TNT_LOG_MOOD_DISAPPOINTING, 0,
                   "Disconnected from network.");
      }

      break;

    default:
      g_warning ("Unknown link status from Flux layer.");
  }
}

/* Returns TRUE if the call resulted in a new connection attempt */
static gboolean
maybe_connect_to (TntNetworkCore *network_core, TntNodeID node_id, const TntNodeAddress4 *address)
{
  TntEdge     *edge;
  gchar       *ip_str;
  PendingEdge *pending_edge;
  FluxCipher  *cipher;

  /* If we're just bootstrapping others, skip */

  if (network_core->bootstrap_only)
    return FALSE;

  /* If we maxed out our connections, skip */

  if (get_pending_edge_count (network_core) >= PENDING_EDGE_MAX)
    return FALSE;

  /* If we have no matching bandwidth group, skip */

  if (!network_core->bandwidth_group_list ||
      !tnt_bandwidth_group_list_find_group_by_ip (network_core->bandwidth_group_list, address->ip))
    return FALSE;

  /* Block the address so it doesn't come up again for the duration of this attempt
   * and the subsequently established connection, if any. */

  if (network_core->net_db)
    tnt_net_db_block_address (network_core->net_db, address);

  /* If we are already trying to make a connection, skip */

  if (find_pending_edge_by_address (network_core, address) >= 0)
  {
    return FALSE;
  }

  /* If we already have a fully opened connection, skip */

  edge = find_edge_by_address (network_core, address);
  if (edge)
  {
    if (tnt_edge_get_io (edge, TNT_STREAM_TYPE_CONTROL) &&
        tnt_edge_get_io (edge, TNT_STREAM_TYPE_DATA))
      return FALSE;
  }

  /* Start connection attempt */

  cipher = flux_cipher_dup (network_core->cipher);
  flux_cipher_randomize (cipher);

  pending_edge = new_pending_edge (network_core, tnt_tcp_io_new_with_cipher (cipher));
  pending_edge->outbound = TRUE;
  memcpy (&pending_edge->address, address, sizeof (TntNodeAddress4));
  pending_edge->node_id = node_id;
  pending_edge->edge = edge;
  if (edge)
    flux_object_ref (FLUX_OBJECT (edge));

  flux_object_unref (FLUX_OBJECT (cipher));

  flux_tcp_io_set_link_func (FLUX_TCP_IO (pending_edge->tcp_io), tcp_link_change, network_core);

  ip_str = tnt_ip_to_string (pending_edge->address.ip);
  flux_tcp_io_connect_begin (FLUX_TCP_IO (pending_edge->tcp_io), ip_str, pending_edge->address.port);
  g_free (ip_str);

  return TRUE;
}

/* Returns TRUE if the call resulted in a new connection attempt */
static gboolean
ensure_edge_to_node (TntNetworkCore *network_core, TntNodeID node_id)
{
  TntNodeMeta        *node_meta;
  TntConnectRequestM  connect_request_m;
  TntRoutedMessage   *rmessage;
  TntNodeAddress4     address;

  if (!network_core->net_db)
    return FALSE;

  node_meta = tnt_net_db_get_node_meta (network_core->net_db, node_id);
  if (!node_meta)
    return FALSE;  /* Address not known */

  tnt_node_meta_get_best_address (node_meta, &address);

  if (!maybe_connect_to (network_core, node_id, &address))
    return FALSE;  /* Already connected or too many pending edges */

  connect_request_m.id = network_core->my_node_id;
  memcpy (&connect_request_m.address, &network_core->my_node_address, sizeof (TntNodeAddress4));
  rmessage = tnt_message_construct_routed (TNT_MESSAGE_CONNECT_REQUEST, &connect_request_m, NULL, 0, 5,
                                           network_core->my_node_id, node_id);
  route_message (network_core, NULL, TNT_MESSAGE (rmessage));
  flux_object_unref (FLUX_OBJECT (rmessage));

  tnt_node_meta_free (node_meta);
  return TRUE;
}

/* Initiate a new connection to a node picked randomly from the net db. */
static void
maybe_start_connect (TntNetworkCore *network_core, TntBandwidthGroup *group)
{
  TntNodeAddress4 node_address;
  TntNodeIP4      match_ip, match_mask;
  gint            i;

  if (!network_core->net_db)
    return;

  if (group)
  {
    gchar *ip, *netmask;

    tnt_bandwidth_group_get_accept (group, &ip, &netmask);
    match_ip   = tnt_string_to_ip (ip);
    match_mask = tnt_string_to_ip (netmask);

    g_free (ip);
    g_free (netmask);
  }
  else
  {
    match_ip   = 0x00000000;
    match_mask = 0x00000000;
  }

  for (i = 0; i < CONNECT_BURST_COUNT; i++)
  {
    if (!tnt_net_db_pick_address (network_core->net_db, match_ip, match_mask,
                                  &node_address))
      return;

    maybe_connect_to (network_core, TNT_NODE_ID_INVALID, &node_address);
  }
}

static void
maybe_drop_edge (TntNetworkCore *network_core, TntBandwidthGroup *group)
{
  TntEdge *preferred_edge;

  preferred_edge = get_least_useful_edge (network_core, group);
  if (!preferred_edge)
    return;

  remove_edge (network_core, preferred_edge);
}

static void
boost_outgoing (TntNetworkCore *network_core, TntBandwidthGroup *group)
{
  gboolean handled;

  if (network_core->bootstrap_only)
    return;

  tnt_callback_list_run_while_with_result_3 (network_core->starvation_cb_list,
                                             gboolean, handled, FALSE, tnt_test_false,
                                             network_core, group, TNT_TRANSFER_DIRECTION_OUT);

  if (!handled)
    maybe_start_connect (network_core, group);
}

static void
boost_incoming (TntNetworkCore *network_core, TntBandwidthGroup *group)
{
  gboolean handled;

  if (network_core->bootstrap_only)
    return;

  tnt_callback_list_run_while_with_result_3 (network_core->starvation_cb_list,
                                             gboolean, handled, FALSE, tnt_test_false,
                                             network_core, group, TNT_TRANSFER_DIRECTION_IN);

  if (!handled)
    maybe_start_connect (network_core, group);
}

/* We get this callback if utilization too low/high. It means we should do something about it. */
static void
bandwidth_action_cb (TntBandwidthGroup *group, TntTransferDirection dir, TntBandwidthAdjustment adjust,
                     gpointer data)
{
  TntNetworkCore *network_core = data;

  assert_types (network_core);
  flux_object_check (FLUX_OBJECT (group), TNT_BANDWIDTH_GROUP_TYPE);

  DEBUG (g_print ("*** %s: Asked to %s %s data.\n", tnt_bandwidth_group_get_name (group),
                  adjust == TNT_BANDWIDTH_ADJUSTMENT_INCREASE ? "increase" : "decrease",
                  dir == TNT_TRANSFER_DIRECTION_IN ? "incoming" : "outgoing"));

  if (dir == TNT_TRANSFER_DIRECTION_OUT)
  {
    if (adjust == TNT_BANDWIDTH_ADJUSTMENT_INCREASE)
    {
      boost_outgoing (network_core, group);
    }
    else if (adjust == TNT_BANDWIDTH_ADJUSTMENT_DECREASE)
    {
      maybe_drop_edge (network_core, group);
    }
  }
  else if (dir == TNT_TRANSFER_DIRECTION_IN)
  {
    if (adjust == TNT_BANDWIDTH_ADJUSTMENT_INCREASE)
    {
      boost_incoming (network_core, group);
    }
    else if (adjust == TNT_BANDWIDTH_ADJUSTMENT_DECREASE)
    {
      maybe_drop_edge (network_core, group);
    }
  }
}

/* ----------------- *
 * Config management *
 * ----------------- */

#define PUBLIC_KEY_FILE_NAME   "public-key"
#define SECRET_KEY_FILE_NAME   "secret-key"

#define PUBLIC_KEY_FILE_HEADER "TntPublicKey-00"
#define SECRET_KEY_FILE_HEADER "TntSecretKey-00"

static gboolean
save_secret_key (TntConfig *config, FluxRSASecretKey *secret_key)
{
  gchar   *base_path, *keyfile_path;
  gint     keyfile_fd;
  gpointer raw_key;
  gsize    raw_key_length;

  g_return_val_if_fail (secret_key != NULL, FALSE);

  base_path = tnt_config_get_str (config, "common", "data-dir", NULL);
  if (!base_path)
    return FALSE;

  keyfile_path = g_build_path (G_DIR_SEPARATOR_S, base_path, SECRET_KEY_FILE_NAME, NULL);
  g_free (base_path);

  keyfile_fd = tnt_open (keyfile_path, TRUE, TRUE, FALSE);
  g_free (keyfile_path);

  if (keyfile_fd < 0)
    return FALSE;

  flux_rsa_export_secret_key (secret_key, &raw_key, &raw_key_length);

  tnt_safe_write (keyfile_fd, FLUX_STRING_WITH_LEN (SECRET_KEY_FILE_HEADER));
  tnt_safe_write (keyfile_fd, raw_key, raw_key_length);

  tnt_close (keyfile_fd);
  g_free (raw_key);
  return TRUE;
}

static gboolean
save_public_key (TntConfig *config, FluxRSAPublicKey *public_key)
{
  gchar   *base_path, *keyfile_path;
  gint     keyfile_fd;
  gpointer raw_key;
  gsize    raw_key_length;

  g_return_val_if_fail (public_key != NULL, FALSE);

  base_path = tnt_config_get_str (config, "common", "data-dir", NULL);
  if (!base_path)
    return FALSE;

  keyfile_path = g_build_path (G_DIR_SEPARATOR_S, base_path, PUBLIC_KEY_FILE_NAME, NULL);
  g_free (base_path);

  keyfile_fd = tnt_open (keyfile_path, TRUE, TRUE, FALSE);
  g_free (keyfile_path);

  if (keyfile_fd < 0)
    return FALSE;

  flux_rsa_export_public_key (public_key, &raw_key, &raw_key_length);

  tnt_safe_write (keyfile_fd, FLUX_STRING_WITH_LEN (PUBLIC_KEY_FILE_HEADER));
  tnt_safe_write (keyfile_fd, raw_key, raw_key_length);

  tnt_close (keyfile_fd);
  g_free (raw_key);
  return TRUE;
}

static FluxRSASecretKey *
load_secret_key (TntConfig *config)
{
  FluxRSASecretKey *secret_key;
  gchar            *base_path, *keyfile_path;
  gchar            *raw_key;
  gsize             raw_key_length;
  gboolean          result;

  base_path = tnt_config_get_str (config, "common", "data-dir", NULL);
  if (!base_path)
    return NULL;

  keyfile_path = g_build_path (G_DIR_SEPARATOR_S, base_path, SECRET_KEY_FILE_NAME, NULL);
  g_free (base_path);

  result = g_file_get_contents (keyfile_path, &raw_key, &raw_key_length, NULL);
  g_free (keyfile_path);

  if (!result || raw_key == NULL || raw_key_length == 0)
    return NULL;

  if (raw_key_length <= FLUX_STATIC_STRING_LEN (SECRET_KEY_FILE_HEADER) ||
      memcmp (raw_key, FLUX_STRING_WITH_LEN (SECRET_KEY_FILE_HEADER)))
  {
    g_free (raw_key);
    return NULL;
  }

  secret_key = flux_rsa_import_secret_key (raw_key + FLUX_STATIC_STRING_LEN (SECRET_KEY_FILE_HEADER),
                                           raw_key_length - FLUX_STATIC_STRING_LEN (SECRET_KEY_FILE_HEADER));
  g_free (raw_key);
  return secret_key;
}

static FluxRSAPublicKey *
load_public_key (TntConfig *config)
{
  FluxRSAPublicKey *public_key;
  gchar            *base_path, *keyfile_path;
  gchar            *raw_key;
  gsize             raw_key_length;
  gboolean          result;

  base_path = tnt_config_get_str (config, "common", "data-dir", NULL);
  if (!base_path)
    return NULL;

  keyfile_path = g_build_path (G_DIR_SEPARATOR_S, base_path, PUBLIC_KEY_FILE_NAME, NULL);
  g_free (base_path);

  result = g_file_get_contents (keyfile_path, &raw_key, &raw_key_length, NULL);
  g_free (keyfile_path);

  if (!result || raw_key == NULL || raw_key_length == 0)
  {
    return NULL;
  }

  if (raw_key_length <= FLUX_STATIC_STRING_LEN (PUBLIC_KEY_FILE_HEADER) ||
      memcmp (raw_key, PUBLIC_KEY_FILE_HEADER, FLUX_STATIC_STRING_LEN (PUBLIC_KEY_FILE_HEADER)))
  {
    g_free (raw_key);
    return NULL;
  }

  public_key = flux_rsa_import_public_key (raw_key + FLUX_STATIC_STRING_LEN (PUBLIC_KEY_FILE_HEADER),
                                           raw_key_length - FLUX_STATIC_STRING_LEN (PUBLIC_KEY_FILE_HEADER));
  g_free (raw_key);
  return public_key;
}

static void
setup_local_keypair (TntNetworkCore *network_core, TntConfig *config)
{
  FluxRSAPublicKey *public_key;
  FluxRSASecretKey *secret_key;

  public_key = load_public_key (config);
  secret_key = load_secret_key (config);

  if (!public_key || !secret_key)
  {
    if (public_key)
      flux_object_unref (FLUX_OBJECT (public_key));
    if (secret_key)
      flux_object_unref (FLUX_OBJECT (secret_key));

    if (network_core->my_public_key && network_core->my_secret_key)
    {
      public_key = network_core->my_public_key;
      secret_key = network_core->my_secret_key;
    }
    else
      flux_rsa_generate_keypair (1024, &public_key, &secret_key);
    
    save_public_key (config, public_key);
    save_secret_key (config, secret_key);
  }

  if (public_key != network_core->my_public_key &&
      secret_key != network_core->my_secret_key)
  {
    tnt_network_core_set_keypair (network_core, public_key, secret_key);
    flux_object_unref (FLUX_OBJECT (public_key));
    flux_object_unref (FLUX_OBJECT (secret_key));
  }
}

static void
add_all_edges_to_bandwidth_groups (TntNetworkCore *network_core)
{
  GSList *dead_edges = NULL, *l;
  gint    n;

  if (!network_core->bandwidth_group_list)
    return;

  for (n = 0; n < network_core->edge_array->len; n++)
  {
    TntEdge  *edge;
    gboolean  result;

    edge = g_ptr_array_index (network_core->edge_array, n);
    result = tnt_bandwidth_group_list_add_edge (network_core->bandwidth_group_list, edge);

    if (!result)
      dead_edges = g_slist_prepend (dead_edges, edge);
  }

  for (l = dead_edges; l; l = g_slist_next (l))
  {
    TntEdge *edge = l->data;

    remove_edge (network_core, edge);

    if (!tnt_network_core_get_n_edges (network_core))
      tnt_log (TNT_LOG_SUBSYSTEM_NETWORK, TNT_LOG_LEVEL_NORMAL, TNT_LOG_MOOD_DISAPPOINTING, 0,
               "Disconnected from network.");
  }

  g_slist_free (dead_edges);
}

static void
limit_pending_edges_to_bandwidth_groups (TntNetworkCore *network_core)
{
  GSList *dead_edges = NULL, *l;
  gint    n;

  /* If we don't have a bandwidth group list, all pending edges will die */

  for (n = 0; n < network_core->pending_edge_array->len; n++)
  {
    PendingEdge *pending_edge;

    pending_edge = g_ptr_array_index (network_core->pending_edge_array, n);

    if (!network_core->bandwidth_group_list ||
        !tnt_bandwidth_group_list_find_group_by_ip (network_core->bandwidth_group_list,
                                                    pending_edge->address.ip))
      dead_edges = g_slist_prepend (dead_edges, pending_edge);
  }

  for (l = dead_edges; l; l = g_slist_next (l))
  {
    PendingEdge *pending_edge = l->data;
    gint         pending_edge_n;

    pending_edge_n = find_pending_edge (network_core, pending_edge);
    g_assert (pending_edge_n >= 0);

    if (network_core->net_db && pending_edge->outbound)
      tnt_net_db_unblock_address (network_core->net_db, &pending_edge->address);

    _tnt_tcp_io_be_dead (pending_edge->tcp_io);
    remove_pending_edge (network_core, pending_edge_n);
  }

  g_slist_free (dead_edges);
}

static void
configure_db (TntNetworkCore *network_core, TntConfig *config)
{
  gchar      *base_path;
  gchar      *sub_path;
  TntStorage *storage;

  base_path = tnt_config_get_str (config, "common", "data-dir", NULL);

  if (network_core->net_db)
  {
    flux_object_unref (FLUX_OBJECT (network_core->net_db));
    network_core->net_db = NULL;
  }

  if (network_core->hash_collider)
  {
    flux_object_unref (FLUX_OBJECT (network_core->hash_collider));
    network_core->hash_collider = NULL;
  }

  if (!base_path)
    return;

  sub_path = g_build_path (G_DIR_SEPARATOR_S, base_path, "node", NULL);
  storage = tnt_storage_new (sub_path, TRUE);
  g_free (sub_path);

  if (storage)
  {
    network_core->net_db = tnt_net_db_new (storage);
    network_core->hash_collider = tnt_hash_collider_new (storage);
    flux_object_unref (FLUX_OBJECT (storage));
  }

  g_free (base_path);  

  if (network_core->net_db)
    maybe_start_connect (network_core, NULL);

  setup_local_keypair (network_core, config);

  if (network_core->hash_collider)
  {
    guint work_rate;

    tnt_hash_collider_set_target (network_core->hash_collider, network_core->my_node_id);
    work_rate = tnt_config_get_uint (config, "network", "hashcash-work-rate", NULL);
    tnt_hash_collider_set_work_rate (network_core->hash_collider, (gdouble) work_rate / (gdouble) 100.0);
  }
}

static void
validate_db (TntNetworkCore *network_core, TntConfig *config)
{
  gchar *base_path;
  gchar *nice_path;

  base_path = tnt_config_get_str (config, "common", "data-dir", NULL);
  nice_path = tnt_nice_path (base_path);

  if (!base_path || !strlen (base_path))
  {
    tnt_config_validation_error (config, "network",
                                 tnt_config_peek_node (config, "common", "data-dir", NULL),
                                 TNT_NETWORK_ERROR_DATA_DIR_UNSET,
                                 "The database directory is not set.");
  }
  else if (!nice_path)
  {
    tnt_config_validation_error (config, "network",
                                 tnt_config_peek_node (config, "common", "data-dir", NULL),
                                 TNT_NETWORK_ERROR_DATA_DIR_RELATIVE,
                                 "The database directory is not absolute.");
  }
  else if (!tnt_path_exists (nice_path))
  {
    tnt_config_validation_error (config, "network",
                                 tnt_config_peek_node (config, "common", "data-dir", NULL),
                                 TNT_NETWORK_ERROR_DATA_DIR_NONEXISTENT,
                                 "The database directory doesn't exist.");
  }
  else if (!tnt_path_is_dir (nice_path))
  {
    tnt_config_validation_error (config, "network",
                                 tnt_config_peek_node (config, "common", "data-dir", NULL),
                                 TNT_NETWORK_ERROR_DATA_DIR_NONEXISTENT,
                                 "The database path is not a directory.");
  }
  else if (!tnt_dir_have_full_access (nice_path))
  {
    tnt_config_validation_error (config, "network",
                                 tnt_config_peek_node (config, "common", "data-dir", NULL),
                                 TNT_NETWORK_ERROR_DATA_DIR_PERMS,
                                 "The database path does not provide read/write access.");
  }

  g_free (base_path);
  g_free (nice_path);
}

static void
configure_listening_port (TntNetworkCore *network_core, TntConfig *config)
{
  GList    *l;
  gboolean  enabled;
  guint     port;

  enabled = tnt_config_get_boolean (config, "network", "accept-connections", NULL);
  port = tnt_config_get_uint (config, "network", "listening-port", NULL);

  if (network_core->tcp_listener)
  {
    flux_object_unref (FLUX_OBJECT (network_core->tcp_listener));
    network_core->tcp_listener = NULL;
  }

  if (!enabled)
    port = 0;

  network_core->my_node_address.port = port;

  for (l = network_core->my_alt_address_list; l; l = g_list_next (l))
  {
    TntNodeAddress4 *address = l->data;
    address->port = port;
  }

  if (port != 0)
    network_core->tcp_listener = tnt_tcp_listener_new_with_cipher (port, network_core->cipher);

  if (network_core->tcp_listener)
    tnt_tcp_listener_set_connection_func (network_core->tcp_listener, incoming_connect_established, network_core);
}

static void
validate_listening_port (TntNetworkCore *network_core, TntConfig *config)
{
  gboolean enabled;
  guint    current_port, port;

  enabled = tnt_config_get_boolean (config, "network", "accept-connections", NULL);
  port = tnt_config_get_uint (config, "network", "listening-port", NULL);

  if (network_core->tcp_listener)
    current_port = flux_tcp_listener_get_port (FLUX_TCP_LISTENER (network_core->tcp_listener));
  else
    current_port = 0;

  if (enabled && port == 0)
  {
    tnt_config_validation_error (config, "network",
                                 tnt_config_peek_node (config, "network", "listening-port", NULL),
                                 TNT_NETWORK_ERROR_LISTENING_PORT_UNSET,
                                 "The listening port is unset or invalid.");
  }
  else if (!enabled || port == current_port)
  {
    /* This is ok */
  }
  else
  {
    FluxTcpListener *tcp_listener;

    tcp_listener = flux_tcp_listener_new (port);
    if (!tcp_listener)
    {
      tnt_config_validation_error (config, "network",
                                   tnt_config_peek_node (config, "network", "listening-port", NULL),
                                   TNT_NETWORK_ERROR_LISTENING_PORT_UNAVAIL,
                                   "Could not listen on port %d.", port);
    }
    else
      flux_object_unref (FLUX_OBJECT (tcp_listener));
  }
}

static gchar *
get_str_value (FluxTree *root, const gchar *name)
{
  FluxTree *node;

  node = flux_tree_find_first_child (root, name, strlen (name));
  if (!node)
    return NULL;

  node = flux_tree_get_first_child (node);
  if (!node)
    return NULL;

  return flux_tree_data_get_str (node);
}

static guint64
get_uint64_value (FluxTree *root, const gchar *name)
{
  FluxTree *node;
  gchar    *str;
  guint64   value;

  node = flux_tree_find_first_child (root, name, strlen (name));
  if (!node)
    return 0;

  node = flux_tree_get_first_child (node);
  if (!node)
    return 0;

  str = flux_tree_data_get_str (node);
  if (!str)
    return 0;

  sscanf (str, "%" G_GUINT64_FORMAT, &value);
  g_free (str);

  return value;
}

static TntBandwidthGroup *
bandwidth_group_from_tree (FluxTree *root)
{
  TntBandwidthGroup *group = NULL;
  gchar             *name;
  gchar             *ip;
  gchar             *netmask;

  name    = get_str_value (root, "name");
  ip      = get_str_value (root, "ip");
  netmask = get_str_value (root, "netmask");

  if (!name || !ip || !netmask)
    goto out;

  group = tnt_bandwidth_group_new ();
  tnt_bandwidth_group_set_name (group, name);
  tnt_bandwidth_group_set_accept (group, ip, netmask);

  tnt_bandwidth_group_set_goal (group, TNT_TRANSFER_DIRECTION_IN,  get_uint64_value (root, "cps-in"));
  tnt_bandwidth_group_set_goal (group, TNT_TRANSFER_DIRECTION_OUT, get_uint64_value (root, "cps-out"));

out:
  g_free (name);
  g_free (ip);
  g_free (netmask);
  return group;
}

static void
configure_bandwidth_groups (TntNetworkCore *network_core, TntConfig *config)
{
  FluxTree *group_root, *group_node;
  gint      n_groups;
  gboolean  changed = FALSE;

  group_root = tnt_config_peek_node (config, "network", "bandwidth-groups", NULL);

  /* No configuration node? Create empty list */

  if (!group_root)
  {
    if (network_core->bandwidth_group_list)
    {
      tnt_bandwidth_group_list_remove_group_func (network_core->bandwidth_group_list,
                                                  bandwidth_action_cb, network_core);
      flux_object_unref (FLUX_OBJECT (network_core->bandwidth_group_list));
      network_core->bandwidth_group_list = NULL;
    }

    network_core->bandwidth_group_list = tnt_bandwidth_group_list_new ();
    tnt_bandwidth_group_list_add_group_action_func (network_core->bandwidth_group_list,
                                                    bandwidth_action_cb, network_core);
    return;
  }

  /* Ensure list */

  if (!network_core->bandwidth_group_list)
  {
    network_core->bandwidth_group_list = tnt_bandwidth_group_list_new ();
    tnt_bandwidth_group_list_add_group_action_func (network_core->bandwidth_group_list,
                                                    bandwidth_action_cb, network_core);
  }

  /* See if group list changed at all */

  /* NOTE: N^2. Could beat this by making bandwidth group list keep lookup hash(es) */

  n_groups = 0;

  for (group_node = flux_tree_get_first_child (group_root); !changed && group_node;
       group_node = flux_tree_get_next (group_node))
  {
    TntBandwidthGroup *group;
    gchar             *node_name, *node_ip, *node_netmask;
    gchar             *group_ip, *group_netmask;

    node_name = get_str_value (group_node, "name");
    if (!node_name)
      continue;

    group = tnt_bandwidth_group_list_find_group_by_name (network_core->bandwidth_group_list, node_name);
    g_free (node_name);
    if (!group)
    {
      changed = TRUE;
      break;
    }

    node_ip = get_str_value (group_node, "ip");
    if (!node_ip)
    {
      changed = TRUE;
      break;
    }

    node_netmask = get_str_value (group_node, "netmask");
    if (!node_netmask)
    {
      g_free (node_ip);
      changed = TRUE;
      break;
    }

    tnt_bandwidth_group_get_accept (group, &group_ip, &group_netmask);
    if (!group_ip || !group_netmask || strcmp (node_ip, group_ip) || strcmp (node_netmask, group_netmask))
      changed = TRUE;

    g_free (group_ip);
    g_free (group_netmask);
    g_free (node_ip);
    g_free (node_netmask);

    /* Change group parameters on the fly (these options don't mandate a new group list) */
    tnt_bandwidth_group_set_goal (group, TNT_TRANSFER_DIRECTION_IN,  get_uint64_value (group_node, "cps-in"));
    tnt_bandwidth_group_set_goal (group, TNT_TRANSFER_DIRECTION_OUT, get_uint64_value (group_node, "cps-out"));

    n_groups++;
  }

  if (n_groups != tnt_bandwidth_group_list_get_n_groups (network_core->bandwidth_group_list))
    changed = TRUE;

  /* If bandwidth groups were added or removed, recreate the bandwidth group list */

  if (changed)
  {
    TntBandwidthGroupList *old_bwgl;

    /* Create new list with groups */

    old_bwgl = network_core->bandwidth_group_list;
    network_core->bandwidth_group_list = tnt_bandwidth_group_list_new ();

    tnt_bandwidth_group_list_add_group_action_func (network_core->bandwidth_group_list,
                                                    bandwidth_action_cb, network_core);

    for (group_node = flux_tree_get_first_child (group_root); group_node;
         group_node = flux_tree_get_next (group_node))
    {
      TntBandwidthGroup *group;

      group = bandwidth_group_from_tree (group_node);
      if (!group)
        continue;

      tnt_bandwidth_group_list_add_group (network_core->bandwidth_group_list, group);
      flux_object_unref (FLUX_OBJECT (group));
    }

    /* Sync edges/pending edges with groups */

    add_all_edges_to_bandwidth_groups (network_core);
    limit_pending_edges_to_bandwidth_groups (network_core);

    flux_object_unref (FLUX_OBJECT (old_bwgl));

    /* Make a slack attempt to stay above 0 connections */

    if (!tnt_network_core_get_n_edges (network_core) &&
        !tnt_network_core_get_n_pending_edges (network_core))
    {
      /* FIXME: We should probably iterate over the bandwidth groups and
       * start a connect for each one */
      maybe_start_connect (network_core, NULL);
    }
  }
}

static void
validate_bandwidth_groups (TntNetworkCore *network_core, TntConfig *config)
{
  FluxTree   *group_root, *group_node;
  GHashTable *seen_names, *seen_masked_ips;
  gint        n_groups = 0;

  group_root = tnt_config_peek_node (config, "network", "bandwidth-groups", NULL);
  if (!group_root)
  {
    tnt_config_validation_error (config, "network",
                                 tnt_config_peek_node (config, "network", "bandwidth-groups", NULL),
                                 TNT_NETWORK_ERROR_BANDWIDTH_GROUPS_UNSET,
                                 "No bandwidth groups configured.");
    return;
  }

  seen_names      = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
  seen_masked_ips = g_hash_table_new (g_direct_hash, g_direct_equal);

  for (group_node = flux_tree_get_first_child (group_root); group_node;
       group_node = flux_tree_get_next (group_node))
  {
    gchar      *node_name, *node_ip, *node_netmask;
    TntNodeIP4  node_masked_ip;

    node_name = get_str_value (group_node, "name");
    node_ip = get_str_value (group_node, "ip");
    node_netmask = get_str_value (group_node, "netmask");

    if (node_ip && node_netmask)
      node_masked_ip = tnt_string_to_ip (node_ip) & tnt_string_to_ip (node_netmask);
    else
      node_masked_ip = TNT_NODE_IP_INVALID;

    if (!node_name || !strlen (node_name))
    {
      /* Ignore group */
    }
    else if (g_hash_table_lookup (seen_names, node_name))
    {
      tnt_config_validation_error (config, "network", group_node,
                                   TNT_NETWORK_ERROR_BANDWIDTH_GROUP_DUP_NAME,
                                   "Duplicate bandwidth group '%s'.", node_name);
    }
    else if (!node_ip || !strlen (node_ip))
    {
      tnt_config_validation_error (config, "network", group_node,
                                   TNT_NETWORK_ERROR_BANDWIDTH_GROUP_NO_IP,
                                   "Bandwidth group '%s' does not specify a base IP.", node_name);
    }
    else if (!node_netmask || !strlen (node_netmask))
    {
      tnt_config_validation_error (config, "network", group_node,
                                   TNT_NETWORK_ERROR_BANDWIDTH_GROUP_NO_NETMASK,
                                   "Bandwidth group '%s' does not specify a netmask.", node_name);
    }
    else if (!tnt_string_is_valid_ipv4_address (node_ip))
    {
      tnt_config_validation_error (config, "network", group_node,
                                   TNT_NETWORK_ERROR_BANDWIDTH_GROUP_BAD_IP,
                                   "Bandwidth group '%s' specifies an invalid base IP.", node_name);
    }
    else if (!tnt_string_is_valid_ipv4_address (node_netmask))
    {
      tnt_config_validation_error (config, "network", group_node,
                                   TNT_NETWORK_ERROR_BANDWIDTH_GROUP_BAD_NETMASK,
                                   "Bandwidth group '%s' specifies an invalid netmask.", node_name);
    }
    else if (g_hash_table_lookup (seen_masked_ips, GUINT_TO_POINTER (node_masked_ip)))
    {
      tnt_config_validation_error (config, "network", group_node,
                                   TNT_NETWORK_ERROR_BANDWIDTH_GROUP_DUP_RANGE,
                                   "Bandwidth group '%s' IP range is not unique.", node_name);
    }
    else
    {
      g_hash_table_insert (seen_masked_ips, GUINT_TO_POINTER (node_masked_ip), GUINT_TO_POINTER (1));
      n_groups++;
    }

    if (node_name)
    {
      if (strlen (node_name))
        g_hash_table_insert (seen_names, node_name, GUINT_TO_POINTER (1));
      else
        g_free (node_name);
    }

    g_free (node_ip);
    g_free (node_netmask);
  }

  if (!n_groups)
  {
    tnt_config_validation_error (config, "network",
                                 tnt_config_peek_node (config, "network", "bandwidth-groups", NULL),
                                 TNT_NETWORK_ERROR_BANDWIDTH_GROUPS_UNSET,
                                 "No valid bandwidth groups configured.");
  }

  g_hash_table_destroy (seen_names);
  g_hash_table_destroy (seen_masked_ips);
}

static void
configure_privacy_options (TntNetworkCore *network_core, TntConfig *config)
{
  network_core->hide_my_ip =
    tnt_config_get_boolean (config, "network", "privacy", "hide-ip", NULL);
  network_core->introduce_delays =
    tnt_config_get_boolean (config, "network", "privacy", "fake-delays", NULL);
  network_core->forward_own_messages =
    tnt_config_get_boolean (config, "network", "privacy", "fake-forwarding", NULL);
}

static void
configure_bootstrap (TntNetworkCore *network_core, TntConfig *config)
{
  network_core->bootstrap_only =
    tnt_config_get_boolean (config, "common", "bootstrap-only", NULL);

  if (network_core->bootstrap_only)
  {
    if (!network_core->bootstrap_edges_timeout_id)
      network_core->bootstrap_edges_timeout_id = g_timeout_add (BOOTSTRAP_EDGE_TIMEOUT * 1000,
                                                                bootstrap_edges_drop, network_core);
  }
  else
  {
    if (network_core->bootstrap_edges_timeout_id)
    {
      g_source_remove (network_core->bootstrap_edges_timeout_id);
      network_core->bootstrap_edges_timeout_id = 0;
    }
  }
}

static void
config_set (TntNetworkCore *network_core)
{
  TntConfig *config = tnt_core_peek_config (TNT_CORE (network_core));

  configure_db               (network_core, config);
  configure_listening_port   (network_core, config);
  configure_bandwidth_groups (network_core, config);
  configure_privacy_options  (network_core, config);
  configure_bootstrap        (network_core, config);
}

static void
config_changed (TntNetworkCore *network_core)
{
  TntConfig *config = tnt_core_peek_config (TNT_CORE (network_core));

  /* DB / keys */
  if (tnt_config_is_changed (config, "common", "data-dir", NULL) ||
      tnt_config_is_changed (config, "network", "hashcash-work-rate", NULL))
    configure_db (network_core, config);

  /* Listening port */
  if (tnt_config_is_changed (config, "network", "accept-connections", NULL) ||
      tnt_config_is_changed (config, "network", "listening-port", NULL))
    configure_listening_port (network_core, config);

  /* Bandwidth groups */
  if (tnt_config_is_changed (config, "network", "bandwidth-groups", NULL))
    configure_bandwidth_groups (network_core, config);

  /* Privacy */
  configure_privacy_options (network_core, config);

  /* Bootstrap */
  if (tnt_config_is_changed (config, "common", "bootstrap-only", NULL))
    configure_bootstrap (network_core, config);
}

static void
config_changed_cb (TntConfig *config, TntNetworkCore *network_core)
{
  config_changed (network_core);
}

static void
config_validate (TntNetworkCore *network_core)
{
  TntConfig *config = tnt_core_peek_config (TNT_CORE (network_core));

  validate_db               (network_core, config);
  validate_listening_port   (network_core, config);
  validate_bandwidth_groups (network_core, config);
}

static void
config_validate_cb (TntConfig *config, TntNetworkCore *network_core)
{
  config_validate (network_core);
}

/* --------------------------- *
 * Initialization and Shutdown *
 * --------------------------- */

static void
init_periodic_callbacks (TntNetworkCore *network_core)
{
  guint node_info_timeout;
  guint contact_info_timeout;

  node_info_timeout    = g_random_int_range (BROADCAST_NODE_INFO_TIMEOUT_MIN,
                                             BROADCAST_NODE_INFO_TIMEOUT_MAX);
  contact_info_timeout = g_random_int_range (BROADCAST_NODE_INFO_TIMEOUT_MIN,
                                             BROADCAST_NODE_INFO_TIMEOUT_MAX);

  network_core->broadcast_node_info_timeout_id =
    g_timeout_add (node_info_timeout, broadcast_node_info_timeout_cb, network_core);
  network_core->broadcast_contact_info_timeout_id =
    g_timeout_add (contact_info_timeout, broadcast_contact_info_timeout_cb, network_core);
}

static void
stop_periodic_callbacks (TntNetworkCore *network_core)
{
  g_source_remove (network_core->broadcast_node_info_timeout_id);
  g_source_remove (network_core->broadcast_contact_info_timeout_id);
}

static void
recalculate_node_id (TntNetworkCore *network_core)
{
  if (network_core->my_public_key)
  {
    network_core->my_node_id = tnt_public_key_to_node_id (network_core->my_public_key);
    DEBUG (g_print ("Generated node ID = %08llx.\n", (guint64) network_core->my_node_id));
  }
  else
  {
    network_core->my_node_id = TNT_NODE_ID_INVALID;
    DEBUG (g_print ("Setting node ID invalid.\n"));
  }
}

static void
generate_my_node_info (TntNetworkCore *network_core)
{
  FluxIPAddr *ip_addr;
  gchar      *ip_str;
  GList      *interface_list, *l;

  network_core->my_secret_key = NULL;
  network_core->my_public_key = NULL;
  recalculate_node_id (network_core);

  /* Get Interet interface if present */

  ip_addr = flux_ip_addr_find_internet_interface ();

  if (ip_addr)
  {
    ip_str = flux_ip_addr_get_string (ip_addr);
    DEBUG (g_print ("Found internet interface with IP: %s\n", ip_str));
    flux_object_unref (FLUX_OBJECT (ip_addr));
    network_core->my_node_address.ip = tnt_string_to_ip (ip_str);
    g_free (ip_str);
  }
  else
    network_core->my_node_address.ip = TNT_NODE_IP_INVALID;

  /* Collect alternate interfaces */

  interface_list = flux_ip_addr_get_interfaces ();

  for (l = interface_list; l; l = g_list_next (l))
  {
    ip_addr = l->data;

    if (flux_ip_addr_get_type (ip_addr) == FLUX_IP_ADDR_IPV4 && !flux_ip_addr_is_loopback (ip_addr) &&
        !flux_ip_addr_is_multicast (ip_addr) && !flux_ip_addr_is_reserved (ip_addr))
    {
      TntNodeIP4       ip;
      TntNodeAddress4 *address;

      ip_str = flux_ip_addr_get_string (ip_addr);
      DEBUG (g_print ("Found alternate interface with IP: %s\n", ip_str));
      ip = tnt_string_to_ip (ip_str);
      g_free (ip_str);

      if (ip != network_core->my_node_address.ip)
      {
        address = g_new0 (TntNodeAddress4, 1);
        address->ip = ip;
        network_core->my_alt_address_list = g_list_prepend (network_core->my_alt_address_list, address);
      }
    }

    flux_object_unref (FLUX_OBJECT (ip_addr));
  }

  g_list_free (interface_list);
}

void
tnt_network_core_get_keypair (TntNetworkCore *network_core,
                              FluxRSAPublicKey **public_key, FluxRSASecretKey **secret_key)
{
  assert_types (network_core);

  if (public_key)
    *public_key = network_core->my_public_key;
  if (secret_key)
    *secret_key = network_core->my_secret_key;
}

void
tnt_network_core_set_keypair (TntNetworkCore *network_core,
                              FluxRSAPublicKey *public_key, FluxRSASecretKey *secret_key)
{
  assert_types (network_core);
  g_return_if_fail ((public_key != NULL && secret_key != NULL) || (public_key == NULL && secret_key == NULL));

  DEBUG (g_print ("Setting keypair.\n"));

  if (public_key)
  {
    flux_object_check (FLUX_OBJECT (public_key), FLUX_RSA_PUBLIC_KEY_TYPE);
    flux_object_check (FLUX_OBJECT (secret_key), FLUX_RSA_SECRET_KEY_TYPE);
    flux_object_ref (FLUX_OBJECT (public_key));
    flux_object_ref (FLUX_OBJECT (secret_key));
  }

  if (network_core->my_public_key)
    flux_object_unref (FLUX_OBJECT (network_core->my_public_key));
  if (network_core->my_secret_key)
    flux_object_unref (FLUX_OBJECT (network_core->my_secret_key));

  network_core->my_public_key = public_key;
  network_core->my_secret_key = secret_key;

  recalculate_node_id (network_core);

  if (network_core->my_secret_key && network_core->my_public_key)
  {
    TntConfig *config = tnt_core_peek_config (TNT_CORE (network_core));

    save_secret_key (config, network_core->my_secret_key);
    save_public_key (config, network_core->my_public_key);
  }
}

TntNodeID
tnt_network_core_get_node_id (TntNetworkCore *network_core)
{
  assert_types (network_core);

  return network_core->my_node_id;
}

gboolean
tnt_network_core_init (TntNetworkCore *network_core, TntConfig *config)
{
  g_return_val_if_fail (config != NULL, FALSE);
  g_return_val_if_fail (tnt_core_registry_get_core ("network") == NULL, FALSE);

  if (!tnt_core_init (TNT_CORE (network_core), config))
    return FALSE;

  flux_object_set_type          (FLUX_OBJECT (network_core), TNT_NETWORK_CORE_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (network_core), tnt_network_core_finalize);

  /* Messages */

  tnt_network_messages_register ();

  /* Config */

  tnt_config_attach_listener  (config, (TntConfigListenerFunc *)  config_changed_cb,  network_core);
  tnt_config_attach_validator (config, (TntConfigValidatorFunc *) config_validate_cb, network_core);

  /* Communications */

  network_core->pending_edge_array  = g_ptr_array_new ();
  network_core->edge_array          = g_ptr_array_new ();
  network_core->filter              = tnt_message_filter_new ();
  network_core->outgoing_caps       = g_hash_table_new (g_direct_hash, g_direct_equal);
  network_core->time_warp           = tnt_time_warp_new (network_core);

  tnt_network_core_set_outgoing_cap (network_core, TNT_MESSAGE_UNKNOWN, 200);
  tnt_network_core_set_outgoing_cap (network_core, TNT_MESSAGE_NODE_INFO, 100);
  tnt_network_core_set_outgoing_cap (network_core, TNT_MESSAGE_CONNECT_REQUEST, 50);

  tnt_time_warp_set_message_handler (network_core->time_warp, received_message_from_time_warp, NULL);

  /* Callbacks */

  network_core->new_edge_cb_list       = tnt_callback_list_new ();
  network_core->starvation_cb_list     = tnt_callback_list_new ();
  network_core->message_cb_list_table  = g_hash_table_new_full (g_direct_hash, g_direct_equal,
                                                                NULL, (GDestroyNotify) flux_object_unref);
  network_core->pre_data_cb_list_table = g_hash_table_new_full (g_direct_hash, g_direct_equal,
                                                                NULL, (GDestroyNotify) flux_object_unref);
  network_core->message_replies_table  = g_hash_table_new (g_direct_hash, g_direct_equal);

  attach_message_handlers (network_core);

  /* Misc */

  network_core->my_fingerprint = (guint64) g_random_int () | ((guint64) g_random_int () << 32);

  network_core->cipher = flux_cipher_new ();
  flux_cipher_generate (network_core->cipher);

  generate_my_node_info (network_core);
  init_periodic_callbacks (network_core);

  config_set (network_core);

  /* Register self */

  tnt_core_registry_register_core ("network", network_core);

  return TRUE;
}

void
tnt_network_core_finalize (FluxObject *object)
{
  TntNetworkCore *network_core = TNT_NETWORK_CORE (object);
  TntConfig      *config;
  GList          *l;
  gint            i;

  assert_types (network_core);

  config = tnt_core_peek_config (TNT_CORE (network_core));

  /* TODO: Disconnect database callbacks */
  stop_periodic_callbacks (network_core);
  if (network_core->bootstrap_edges_timeout_id)
    g_source_remove (network_core->bootstrap_edges_timeout_id);

  /* Just a formality */
  detach_message_handlers (network_core);

  if (network_core->bandwidth_group_list)
    flux_object_unref (FLUX_OBJECT (network_core->bandwidth_group_list));

  if (network_core->net_db)
    flux_object_unref (FLUX_OBJECT (network_core->net_db));

  if (network_core->tcp_listener)
    flux_object_unref (FLUX_OBJECT (network_core->tcp_listener));

  if (network_core->cipher)
    flux_object_unref (FLUX_OBJECT (network_core->cipher));

  if (network_core->my_public_key)
    flux_object_unref (FLUX_OBJECT (network_core->my_public_key));

  if (network_core->my_secret_key)
    flux_object_unref (FLUX_OBJECT (network_core->my_secret_key));

  /* FIXME: Unblock pending edge and edge addresses in TntNetDB */
  while (network_core->pending_edge_array->len)
    remove_pending_edge (network_core, 0);
  g_ptr_array_free (network_core->pending_edge_array, TRUE);

  /* FIXME: Use remove_edge? */
  for (i = 0; i < network_core->edge_array->len; i++)
  {
    TntEdge *edge = g_ptr_array_index (network_core->edge_array, i);
    flux_object_unref (FLUX_OBJECT (edge));
  }
  g_ptr_array_free (network_core->edge_array, TRUE);

  g_hash_table_destroy (network_core->outgoing_caps);

  flux_object_unref (FLUX_OBJECT (network_core->new_edge_cb_list));
  flux_object_unref (FLUX_OBJECT (network_core->starvation_cb_list));
  g_hash_table_destroy (network_core->message_cb_list_table);
  g_hash_table_destroy (network_core->pre_data_cb_list_table);
  g_hash_table_destroy (network_core->message_replies_table);

  flux_object_unref (FLUX_OBJECT (network_core->filter));
  flux_object_unref (FLUX_OBJECT (network_core->time_warp));

  for (l = network_core->my_alt_address_list; l; l = g_list_next (l))
  {
    TntNodeAddress4 *address = l->data;
    g_free (address);
  }

  g_list_free (network_core->my_alt_address_list);

  tnt_config_detach_listener  (config, (TntConfigListenerFunc *) config_changed_cb, network_core);
  tnt_config_detach_validator (config, (TntConfigValidatorFunc *) config_validate_cb, network_core);

  tnt_core_registry_unregister_core ("network", network_core);
}

TntNetworkCore *
tnt_network_core_new (TntConfig *config)
{
  TntNetworkCore *network_core;

  network_core = g_new0 (TntNetworkCore, 1);
  if (!tnt_network_core_init (network_core, config))
  {
    g_free (network_core);
    network_core = NULL;
  }

  return network_core;
}

TntNetDB *
tnt_network_core_peek_net_db (TntNetworkCore *network_core)
{
  assert_types (network_core);

  return network_core->net_db;
}

TntHashCollider *
tnt_network_core_peek_hash_collider (TntNetworkCore *network_core)
{
  assert_types (network_core);

  return network_core->hash_collider;
}

TntBandwidthGroupList *
tnt_network_core_peek_bandwidth_groups (TntNetworkCore *network_core)
{
  assert_types (network_core);

  return network_core->bandwidth_group_list;
}

gint
tnt_network_core_get_n_edges (TntNetworkCore *network_core)
{
  assert_types (network_core);

  return network_core->edge_array->len;
}

gint
tnt_network_core_get_n_pending_edges (TntNetworkCore *network_core)
{
  assert_types (network_core);

  return network_core->pending_edge_array->len;
}

const GPtrArray *
tnt_network_core_peek_edge_array (TntNetworkCore *network_core)
{
  assert_types (network_core);

  return network_core->edge_array;
}

TntEdge *
tnt_network_core_find_route (TntNetworkCore *network_core, TntNodeID node_id)
{
  assert_types (network_core);
  g_return_val_if_fail (node_id != TNT_NODE_ID_INVALID, NULL);

  return find_edge_closest_to_node_id (network_core, node_id);
}

void
tnt_network_core_connect_to_node (TntNetworkCore *network_core, TntNodeID node_id)
{
  assert_types (network_core);
  g_return_if_fail (node_id != TNT_NODE_ID_INVALID);

  ensure_edge_to_node (network_core, node_id);
}

TntEdge *
tnt_network_core_route_message (TntNetworkCore *network_core, TntMessage *message)
{
  assert_types (network_core);
  flux_object_check (FLUX_OBJECT (message), TNT_MESSAGE_TYPE);

  return route_message (network_core, NULL, message);
}

gboolean
tnt_network_core_send_message (TntNetworkCore *network_core, TntMessage *message, TntEdge *edge)
{
  /* Edge may not exist anymore */
  if (!have_edge (network_core, edge))
    return FALSE;

  tnt_edge_send_message (edge, message);
  return TRUE;
}

gboolean
tnt_network_core_have_edge (TntNetworkCore *network_core, TntEdge *edge)
{
  assert_types (network_core);

  /* Edge may already be freed, don't follow the pointer */

  return have_edge (network_core, edge);
}

void
tnt_network_core_set_outgoing_cap (TntNetworkCore *network_core, TntMessageType type, guint cap)
{
  assert_types (network_core);

  g_hash_table_insert (network_core->outgoing_caps, GUINT_TO_POINTER (type), GUINT_TO_POINTER (cap));
}

void
tnt_network_core_attach_new_edge_handler (TntNetworkCore *network_core, TntNewEdgeHandler *func,
                                          gpointer data)
{
  assert_types (network_core);

  tnt_callback_list_add_callback (network_core->new_edge_cb_list, func, data);
}

void
tnt_network_core_detach_new_edge_handler (TntNetworkCore *network_core, TntNewEdgeHandler *func,
                                          gpointer data)
{
  assert_types (network_core);

  tnt_callback_list_remove_callback (network_core->new_edge_cb_list, func, data);
}

void
tnt_network_core_attach_starvation_handler (TntNetworkCore *network_core, TntStarvationHandler *func,
                                            gpointer data)
{
  assert_types (network_core);

  tnt_callback_list_add_callback (network_core->starvation_cb_list, func, data);
}

void
tnt_network_core_detach_starvation_handler (TntNetworkCore *network_core, TntStarvationHandler *func,
                                            gpointer data)
{
  assert_types (network_core);

  tnt_callback_list_remove_callback (network_core->starvation_cb_list, func, data);
}

void
tnt_network_core_attach_message_handler (TntNetworkCore *network_core, TntMessageType type,
                                         gboolean may_generate_replies,
                                         TntMessageHandler *func, gpointer data)
{
  TntCallbackList *callback_list;

  assert_types (network_core);

  callback_list = g_hash_table_lookup (network_core->message_cb_list_table, GUINT_TO_POINTER (type));
  if (!callback_list)
  {
    callback_list = tnt_callback_list_new ();
    g_hash_table_insert (network_core->message_cb_list_table, GUINT_TO_POINTER (type), callback_list);
  }

  /* FIXME: This is pretty lame. We should store this boolean with the callbacks instead.
   * It would allow us to reset it to FALSE if there's a specific reply-generating callback
   * that goes away. Currently, this doesn't happen, though. */
  if (may_generate_replies)
  {
    g_hash_table_insert (network_core->message_replies_table,
                         GUINT_TO_POINTER (type), GUINT_TO_POINTER (TRUE));
  }

  tnt_callback_list_add_callback (callback_list, func, data);
}

void
tnt_network_core_detach_message_handler (TntNetworkCore *network_core, TntMessageType type,
                                         TntMessageHandler *func, gpointer data)
{
  TntCallbackList *callback_list;

  assert_types (network_core);

  callback_list = g_hash_table_lookup (network_core->message_cb_list_table, GUINT_TO_POINTER (type));
  if (!callback_list)
    return;

  tnt_callback_list_remove_callback (callback_list, func, data);
}

void
tnt_network_core_attach_pre_data_handler (TntNetworkCore *network_core, TntMessageType type,
                                          TntPreDataHandler *func, gpointer data)
{
  TntCallbackList *callback_list;

  assert_types (network_core);

  callback_list = g_hash_table_lookup (network_core->pre_data_cb_list_table, GUINT_TO_POINTER (type));
  if (!callback_list)
  {
    callback_list = tnt_callback_list_new ();
    g_hash_table_insert (network_core->pre_data_cb_list_table, GUINT_TO_POINTER (type), callback_list);
  }

  tnt_callback_list_add_callback (callback_list, func, data);
}

void
tnt_network_core_detach_pre_data_handler (TntNetworkCore *network_core, TntMessageType type,
                                          TntPreDataHandler *func, gpointer data)
{
  TntCallbackList *callback_list;

  assert_types (network_core);

  callback_list = g_hash_table_lookup (network_core->pre_data_cb_list_table, GUINT_TO_POINTER (type));
  if (!callback_list)
    return;

  tnt_callback_list_remove_callback (callback_list, func, data);
}
