/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-query.c - Abstract class for queries.
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"

#include <string.h>
#include "tnt.h"

#define DEBUG(x)

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), TNT_QUERY_TYPE)

static void
filter_matches (TntQuery *query)
{
  if (query->filter_func)
    query->filter_func (query);
}

void
tnt_query_finalize (FluxObject *object)
{
  TntQuery *query = TNT_QUERY (object);

  assert_types (query);

  if (query->string)
    g_free (query->string);
}

void
tnt_query_init (TntQuery *query)
{
  flux_object_init              (FLUX_OBJECT (query));
  flux_object_set_type          (FLUX_OBJECT (query), TNT_QUERY_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (query), tnt_query_finalize);

  query->match_local   = TRUE;
  query->match_partial = TRUE;
}

guint16
tnt_query_get_id (TntQuery *query)
{
  assert_types (query);

  return query->id;
}

void
tnt_query_set_id (TntQuery *query, guint16 id)
{
  assert_types (query);

  query->id = id;
}

gint
tnt_query_compare (TntQuery *query_1, TntQuery *query_2)
{
  const gchar *str_1, *str_2;
  gint         result;

  assert_types (query_1);
  assert_types (query_2);

  if (query_1 == query_2)
    return 0;

  str_1 = tnt_query_get_string (query_1);
  str_2 = tnt_query_get_string (query_2);

  if (!str_1 && !str_2)
    result = 0;
  else if (!str_1)
    result = -1;
  else if (!str_2)
    result = 1;
  else
    result = strcmp (str_1, str_2);

  if (result == 0)
  {
    if (query_1->match_field < query_2->match_field)
      return -1;
    if (query_1->match_field > query_2->match_field)
      return 1;
    if (query_1->media < query_2->media)
      return -1;
    if (query_1->media > query_2->media)
      return 1;
    if (query_1->category < query_2->category)
      return -1;
    if (query_1->category > query_2->category)
      return 1;
    if (tnt_query_get_match_local (query_1) < tnt_query_get_match_local (query_2))
      return -1;
    if (tnt_query_get_match_local (query_1) > tnt_query_get_match_local (query_2))
      return 1;
    if (tnt_query_get_match_partial (query_1) < tnt_query_get_match_partial (query_2))
      return -1;
    if (tnt_query_get_match_partial (query_1) > tnt_query_get_match_partial (query_2))
      return 1;
  }

  return result;
}

const gchar *
tnt_query_get_string (TntQuery *query)
{
  assert_types (query);

  return query->string;
}

void
tnt_query_set_string (TntQuery *query, const gchar *string)
{
  assert_types (query);
  g_return_if_fail (string != NULL);

  if (string != query->string)
  {
    if (query->string)
      g_free (query->string);
    query->string = g_utf8_strdown (string, -1);

    filter_matches (query);
  }
}

TntMatchField
tnt_query_get_match_field (TntQuery *query)
{
  assert_types (query);

  return query->match_field;
}

void
tnt_query_set_match_field (TntQuery *query, TntMatchField match_field)
{
  assert_types (query);

  if (query->match_field != match_field)
  {
    query->match_field = match_field;
    filter_matches (query);
  }
}

TntMediaType
tnt_query_get_media (TntQuery *query)
{
  assert_types (query);

  return query->media;
}

void
tnt_query_set_media (TntQuery *query, TntMediaType media)
{
  assert_types (query);

  if (query->media != media)
  {
    query->media = media;
    filter_matches (query);
  }
}

gint
tnt_query_get_category (TntQuery *query)
{
  assert_types (query);

  return query->category;
}

void
tnt_query_set_category (TntQuery *query, gint category)
{
  assert_types (query);

  if (query->category != category)
  {
    query->category = category;
    filter_matches (query);
  }
}

gboolean
tnt_query_get_match_local (TntQuery *query)
{
  assert_types (query);

  if (query->match_local)
    return TRUE;

  return FALSE;
}

void
tnt_query_set_match_local (TntQuery *query, gboolean match_local)
{
  gboolean old_value;

  assert_types (query);

  old_value = query->match_local;
  query->match_local = match_local;

  if (old_value != query->match_local)
    filter_matches (query);
}

gboolean
tnt_query_get_match_partial (TntQuery *query)
{
  assert_types (query);

  if (query->match_partial)
    return TRUE;

  return FALSE;
}

void
tnt_query_set_match_partial (TntQuery *query, gboolean match_partial)
{
  gboolean old_value;

  assert_types (query);

  old_value = query->match_partial;
  query->match_partial = match_partial;

  if (old_value != query->match_partial)
    filter_matches (query);
}

gint
tnt_query_get_n_matches (TntQuery *query)
{
  assert_types (query);

  if (query->get_n_matches_func)
    return query->get_n_matches_func (query);

  return 0;
}

gboolean
tnt_query_try_match (TntQuery *query, const TntFileID *file_id, TntFileMeta *file_meta,
                     TntNodeID node_id, gint remote_blocks)
{
  assert_types (query);
  g_return_val_if_fail (file_id != NULL, FALSE);
  g_return_val_if_fail (file_meta != NULL, FALSE);

  if (query->try_match_func)
    return query->try_match_func (query, file_id, file_meta, node_id, remote_blocks);

  return FALSE;
}
