/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-retrieval-list.c - Holds a list of file retrievals.
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"

#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <glib.h>

#ifdef G_PLATFORM_WIN32
# include "win32/win32-compat.h"
#endif

#include "config.h"
#include "tnt.h"

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), TNT_RETRIEVAL_LIST_TYPE)

#define RETRIEVAL_LIST_FILENAME "retrievals"
#define RETRIEVAL_LIST_VERSION  1

#define POLL_INTERVAL           (SECONDS_IN_MINUTE / 2)

typedef struct
{
  TntRetrievalListFunc *func;
  gpointer              data;
}
RetrievalFunc;

typedef struct
{
  TntFileID    file_id;
  TntFileMeta  file_meta;
  gboolean     is_active;
  gchar       *target_path;
}
RetrievalRecord;

FluxMarshalledStructEntry retrieval_record_m_entries_current [] =
{
  FLUX_MARSHAL_STRUCT_FIXED   ("file-id",     RetrievalRecord, file_id,     TRUE, "file-id"),
  FLUX_MARSHAL_STRUCT_FIXED   ("meta",        RetrievalRecord, file_meta,   TRUE, "file-meta"),
  FLUX_MARSHAL_BOOLEAN        ("is-active",   RetrievalRecord, is_active,   TRUE, FALSE),
  FLUX_MARSHAL_STRING_DYNAMIC ("target-path", RetrievalRecord, target_path, TRUE, NULL),
  FLUX_MARSHAL_END            ()
};

#define F FLUX_MARSHALLER_CLEAR

static FluxMarshalledStruct retrieval_record_marshal_info [] =
{
  /* Name                 Size                       Flags  Entries             */
  { "file-id",            sizeof (TntFileID),            F, tnt_file_id_m_entries },
  { "file-meta",          sizeof (TntFileMeta),          F, tnt_file_meta_m_entries },
  { "retrieval-record",   sizeof (RetrievalRecord),      F, retrieval_record_m_entries_current },
  { NULL }
};

static FluxMarshaller *retrieval_marshaller = NULL;

static void retrievals_changed (TntRetrievalList *retrieval_list);

static gint
find_retrieval_by_file_id (GPtrArray *retrievals, const TntFileID *file_id)
{
  gint i;

  for (i = 0; i < retrievals->len; i++)
  {
    TntRetrieval *retrieval = g_ptr_array_index (retrievals, i);

    g_assert (retrieval);

    if (!memcmp (file_id, &retrieval->file_id, sizeof (TntFileID)))
      return i;
  }

  return -1;
}

static GSList *
get_busy_sources (GPtrArray *retrievals)
{
  GSList *busy_list = NULL;
  gint    i;

  for (i = 0; i < retrievals->len; i++)
  {
    TntRetrieval *retrieval = g_ptr_array_index (retrievals, i);

    busy_list = g_slist_concat (busy_list, tnt_retrieval_get_busy_sources (retrieval));
  }

  return busy_list;
}

static void
free_busy_sources (GSList *busy_list)
{
  tnt_slist_free_with_data (busy_list);
}

static GSList *
add_callback_position (TntRetrievalList *retrieval_list, gint pos)
{
  retrieval_list->callback_pos_list =
    g_slist_append (retrieval_list->callback_pos_list, GINT_TO_POINTER (pos));

  return g_slist_last (retrieval_list->callback_pos_list);
}

static void
remove_callback_position (TntRetrievalList *retrieval_list, GSList *pos)
{
  retrieval_list->callback_pos_list =
    g_slist_delete_link (retrieval_list->callback_pos_list, pos);
}

static void
update_callback_positions (TntRetrievalList *retrieval_list, gint first_pos, gint offset)
{
  GSList *l;

  for (l = retrieval_list->callback_pos_list; l; l = g_slist_next (l))
  {
    gint pos = GPOINTER_TO_INT (l->data);

    if (pos >= 0 && pos >= first_pos)
      l->data = GINT_TO_POINTER (pos + offset);
  }
}

static void
pop_callback_position (TntRetrievalList *retrieval_list, gint removed_pos)
{
  GSList *l;

  for (l = retrieval_list->callback_pos_list; l; l = g_slist_next (l))
  {
    gint pos = GPOINTER_TO_INT (l->data);

    if (pos == removed_pos)
      l->data = GINT_TO_POINTER (-1);
  }

  update_callback_positions (retrieval_list, removed_pos, -1);
}

#if 0

static void
push_callback_position (TntRetrievalList *retrieval_list, gint added_pos)
{
  update_callback_positions (retrieval_list, added_pos, 1);
}

#endif

static void
move_callback_position (TntRetrievalList *retrieval_list, gint old_pos, gint new_pos)
{
  GSList   *l;
  gboolean  move_down;

  move_down = new_pos > old_pos ? TRUE : FALSE;

  for (l = retrieval_list->callback_pos_list; l; l = g_slist_next (l))
  {
    gint pos = GPOINTER_TO_INT (l->data);

    if (pos < 0)
      continue;

    if (pos == old_pos)
      pos = new_pos;
    else
    {
      if (pos > old_pos)
        pos--;
      if (pos > new_pos)
        pos++;
    }

    l->data = GINT_TO_POINTER (pos);
  }
}

static gint
run_callbacks (TntRetrievalList *retrieval_list, GSList *callbacks, gint retrieval_n)
{
  GSList *l;
  GSList *pos_l;
  gint    new_pos;

  pos_l = add_callback_position (retrieval_list, retrieval_n);

  for (l = callbacks; l; l = g_slist_next (l))
  {
    RetrievalFunc *retrieval_func;

    retrieval_func = l->data;
    g_assert (retrieval_func != NULL);
    retrieval_func->func (retrieval_list, GPOINTER_TO_INT (pos_l->data), retrieval_func->data);

    /* If retrieval_n was removed by callee, break. */
    if (GPOINTER_TO_INT (pos_l->data) < 0)
      break;
  }

  new_pos = GPOINTER_TO_INT (pos_l->data);
  remove_callback_position (retrieval_list, pos_l);

  return new_pos;
}

static void
remove_callback (GSList **callbacks, TntRetrievalListFunc *func, gpointer data)
{
  GSList *l;
  GSList *next;

  for (l = *callbacks; l; l = next)
  {
    RetrievalFunc *retrieval_func;

    next = g_slist_next (l);

    retrieval_func = l->data;
    g_assert (retrieval_func != NULL);
    if (retrieval_func->func == func && retrieval_func->data == data)
    {
      g_free (retrieval_func);
      *callbacks = g_slist_delete_link (*callbacks, l);
    }
  }
}

static gchar *
full_retrievals_path (TntRetrievalList *retrieval_list)
{
  return g_strjoin (G_DIR_SEPARATOR_S, tnt_storage_get_path (retrieval_list->storage),
                    RETRIEVAL_LIST_FILENAME, NULL);
}

static void
clear_retrievals (TntRetrievalList *retrieval_list)
{
  while (retrieval_list->retrievals->len > 0)
  {
    TntRetrieval *retrieval = g_ptr_array_index (retrieval_list->retrievals, 0);
    gint          pos = 0;

    pos = run_callbacks (retrieval_list, retrieval_list->retrieval_remove_funcs, pos);

    if (pos >= 0)
    {
      tnt_retrieval_set_update_func (retrieval, NULL, NULL);
      flux_object_unref (FLUX_OBJECT (retrieval));
      g_ptr_array_remove_index (retrieval_list->retrievals, pos);

      pop_callback_position (retrieval_list, pos);
    }
  }
}

static gint
ptr_array_find (GPtrArray *ptr_array, gpointer data)
{
  gint i;

  for (i = 0; i < ptr_array->len; i++)
  {
    if (g_ptr_array_index (ptr_array, i) == data)
      return i;
  }

  return -1;
}

static void
retrieval_update_cb (TntRetrieval *retrieval, gpointer data)
{
  TntRetrievalList *retrieval_list = data;
  gint              pos;

  assert_types (retrieval_list);
  flux_object_check (FLUX_OBJECT (retrieval), TNT_RETRIEVAL_TYPE);

  pos = ptr_array_find (retrieval_list->retrievals, retrieval);
  if (pos < 0)
  {
    g_warning ("Got update callback from retrieval we don't care about.");
    return;
  }

  tnt_retrieval_list_updated (retrieval_list, pos);
}

static void
retrieval_completed_cb (TntRetrieval *retrieval, gpointer data)
{
  TntRetrievalList *retrieval_list = data;
  gint              pos;

  assert_types (retrieval_list);
  flux_object_check (FLUX_OBJECT (retrieval), TNT_RETRIEVAL_TYPE);

  pos = ptr_array_find (retrieval_list->retrievals, retrieval);
  if (pos < 0)
  {
    g_warning ("Got completed callback from retrieval we don't care about.");
    return;
  }

  run_callbacks (retrieval_list, retrieval_list->retrieval_completed_funcs, pos);
  retrievals_changed (retrieval_list);
}

static void
retrieval_status_change_cb (TntRetrieval *retrieval, gpointer data)
{
  TntRetrievalList *retrieval_list = data;
  gint              pos;

  assert_types (retrieval_list);
  flux_object_check (FLUX_OBJECT (retrieval), TNT_RETRIEVAL_TYPE);

  pos = ptr_array_find (retrieval_list->retrievals, retrieval);
  if (pos < 0)
  {
    g_warning ("Got status change from retrieval we don't care about.");
    return;
  }

#if 0
  tnt_retrieval_list_updated (retrieval_list, pos);
#endif
  retrievals_changed (retrieval_list);
}

static FluxMarshaller *
get_marshaller (void)
{
  if (!retrieval_marshaller)
  {
    retrieval_marshaller = flux_marshaller_new ();
    flux_marshaller_register_many (retrieval_marshaller, retrieval_record_marshal_info);
  }

  return retrieval_marshaller;
}

static gboolean
load_retrieval_0 (gint fd, TntFileID *file_id, TntFileMeta **file_meta, gboolean *is_active)
{
  guint32      len;
  TntFileMeta *file_meta_m;
  gpointer     buf;

  if (!tnt_safe_read (fd, is_active, sizeof (gboolean)))
    return FALSE;

  if (!tnt_safe_read (fd, file_id, sizeof (TntFileID)))
    return FALSE;

  if (!tnt_safe_read (fd, &len, sizeof (guint32)))
    return FALSE;

  if (len > 400000)
    return FALSE;

  buf = g_malloc (len);
  if (!tnt_safe_read (fd, buf, len))
    return FALSE;

  if (!flux_marshaller_decode_pack (get_marshaller (), "file-meta", buf, len, (gpointer) &file_meta_m))
    return FALSE;

  g_free (buf);

  *file_meta = tnt_file_meta_new ();
  memcpy (*file_meta, file_meta_m, sizeof (TntFileMeta));

  g_free (file_meta_m);
  return TRUE;
}

static gboolean
load_retrieval_modern (gint fd, RetrievalRecord **record)
{
  guint32  len;
  gpointer buf;
  gint     t;

  if (!tnt_safe_read (fd, &len, sizeof (guint32)))
    return FALSE;

  if (len > 400000)
    return FALSE;

  buf = g_malloc (len);
  if (!tnt_safe_read (fd, buf, len))
  {
    g_free (buf);
    return FALSE;
  }

  t = flux_marshaller_decode_pack (get_marshaller (), "retrieval-record",
                                   buf, len, (gpointer *) record);

  g_free (buf);
  return t;
}

static gboolean
load_retrieval (gint fd, RetrievalRecord **record, gint version)
{
  TntFileMeta *file_meta;
  gboolean     result;

  switch (version)
  {
    case 0:
      *record = g_new0 (RetrievalRecord, 1);
      result = load_retrieval_0 (fd, &(*record)->file_id, &file_meta, &(*record)->is_active);

      if (result)
      {
        memcpy (&(*record)->file_meta, file_meta, sizeof (TntFileMeta));
        g_free (file_meta);
      }
      else
        g_free (*record);

      return result;

    case RETRIEVAL_LIST_VERSION:
      return load_retrieval_modern (fd, record);

    default:
      g_assert_not_reached ();
  }

  return FALSE;
}

static gboolean
save_retrieval (gint fd, TntRetrieval *retrieval)
{
  RetrievalRecord record;
  gpointer        buf;
  guint           len;
  guint32         len_u32;

  memcpy (&record.file_id,       tnt_retrieval_get_file_id   (retrieval), sizeof (TntFileID));
  memcpy (&record.file_meta,     tnt_retrieval_get_file_meta (retrieval), sizeof (TntFileMeta));
  record.is_active   =           tnt_retrieval_is_active     (retrieval);
  record.target_path = (gchar *) tnt_retrieval_get_path      (retrieval);

  if (!flux_marshaller_encode_pack (get_marshaller (), "retrieval-record", &record, &buf, &len))
    return FALSE;

  len_u32 = len;
  tnt_safe_write (fd, &len_u32, sizeof (guint32));
  tnt_safe_write (fd, buf, len);

  g_free (buf);
  return TRUE;
}

static void
load_retrievals (TntRetrievalList *retrieval_list, TntBlockDB *block_db)
{
  gchar           *path;
  gint             retrieval_file;
  FluxDBHeader    *header_0, *current_header;
  FluxDBHeader    *found_header;
  gint             version;
  RetrievalRecord *record;

  assert_types (retrieval_list);
  flux_object_check (FLUX_OBJECT (block_db), TNT_BLOCK_DB_TYPE);

  path = full_retrievals_path (retrieval_list);
  retrieval_file = tnt_open (path, FALSE, FALSE, FALSE);
  g_free (path);
  if (retrieval_file < 0)
    return;

  header_0       = flux_db_header_create ("TntRetrievalList", 0);
  current_header = flux_db_header_create ("TntRetrievalList", RETRIEVAL_LIST_VERSION);
  found_header   = flux_db_header_read (retrieval_file);

  if (!memcmp (current_header, found_header, sizeof (FluxDBHeader)))
    version = RETRIEVAL_LIST_VERSION;
  else if (!memcmp (header_0, found_header, sizeof (FluxDBHeader)))
    version = 0;
  else
    version = -1;

  flux_db_header_free (header_0);
  flux_db_header_free (current_header);
  flux_db_header_free (found_header);

  if (version < 0)
  {
    tnt_close (retrieval_file);
    return;
  }

  clear_retrievals (retrieval_list);

  while (load_retrieval (retrieval_file, &record, version))
  {
    TntRetrieval  *retrieval   = tnt_retrieval_new (&record->file_id);
    TntFileBitmap *file_bitmap = tnt_data_db_get_file_bitmap (TNT_DATA_DB (block_db), &record->file_id);
    gint           pos;

    tnt_retrieval_set_file_meta (retrieval, &record->file_meta);
    tnt_retrieval_set_active    (retrieval, record->is_active);
    if (record->target_path)
      tnt_retrieval_set_path (retrieval, record->target_path);
    else if (retrieval_list->incoming_storage)
      tnt_retrieval_set_path (retrieval, tnt_storage_get_path (retrieval_list->incoming_storage));

    flux_marshaller_free_struct (get_marshaller (), "retrieval-record", record);

    if (file_bitmap)
    {
      tnt_retrieval_set_local_bitmap (retrieval, file_bitmap);
      flux_object_unref (FLUX_OBJECT (file_bitmap));
    }

    g_ptr_array_add (retrieval_list->retrievals, retrieval);
    pos = retrieval_list->retrievals->len - 1;
    tnt_retrieval_set_update_func        (retrieval, retrieval_update_cb, retrieval_list);
    tnt_retrieval_set_completed_func     (retrieval, retrieval_completed_cb, retrieval_list);
    tnt_retrieval_set_status_change_func (retrieval, retrieval_status_change_cb, retrieval_list);
    run_callbacks (retrieval_list, retrieval_list->retrieval_add_funcs, pos);
  }

  tnt_close (retrieval_file);
}

static void
save_retrievals (TntRetrievalList *retrieval_list)
{
  gchar        *path;
  gint          retrieval_file;
  gint          n_retrievals;
  gint          i;
  FluxDBHeader *header;

  assert_types (retrieval_list);

  path = full_retrievals_path (retrieval_list);
  retrieval_file = tnt_open (path, TRUE, TRUE, FALSE);
  g_free (path);
  if (retrieval_file < 0)
    return;

  header = flux_db_header_create ("TntRetrievalList", RETRIEVAL_LIST_VERSION);
  flux_db_header_write (header, retrieval_file);
  flux_db_header_free (header);

  n_retrievals = tnt_retrieval_list_get_n_retrievals (retrieval_list);
  for (i = 0; i < n_retrievals; i++)
  {
    TntRetrieval *retrieval = tnt_retrieval_list_get_retrieval_n (retrieval_list, i);

    save_retrieval (retrieval_file, retrieval);
  }

  tnt_close (retrieval_file);
}

static gboolean
retrievals_changed_save_cb (TntRetrievalList *retrieval_list)
{
  retrieval_list->save_timeout_id = -1;
  save_retrievals (retrieval_list);
  return FALSE;
}

static void
retrievals_changed (TntRetrievalList *retrieval_list)
{
  if (retrieval_list->save_timeout_id != -1)
    return;

  retrieval_list->save_timeout_id = g_timeout_add (1000, (GSourceFunc) retrievals_changed_save_cb,
                                                   retrieval_list);
}

void
tnt_retrieval_list_add_retrieval_add_func (TntRetrievalList *retrieval_list,
                                           TntRetrievalListFunc *func, gpointer data)
{
  RetrievalFunc *retrieval_func;

  assert_types (retrieval_list);
  g_return_if_fail (func != NULL);

  retrieval_func = g_new (RetrievalFunc, 1);
  retrieval_func->func = func;
  retrieval_func->data = data;

  retrieval_list->retrieval_add_funcs = g_slist_append (retrieval_list->retrieval_add_funcs, retrieval_func);
}

void
tnt_retrieval_list_add_retrieval_update_func (TntRetrievalList *retrieval_list,
                                              TntRetrievalListFunc *func, gpointer data)
{
  RetrievalFunc *retrieval_func;

  assert_types (retrieval_list);
  g_return_if_fail (func != NULL);

  retrieval_func = g_new (RetrievalFunc, 1);
  retrieval_func->func = func;
  retrieval_func->data = data;

  retrieval_list->retrieval_update_funcs = g_slist_append (retrieval_list->retrieval_update_funcs,
                                                           retrieval_func);
}

void
tnt_retrieval_list_add_retrieval_remove_func (TntRetrievalList *retrieval_list,
                                              TntRetrievalListFunc *func, gpointer data)
{
  RetrievalFunc *retrieval_func;

  assert_types (retrieval_list);
  g_return_if_fail (func != NULL);

  retrieval_func = g_new (RetrievalFunc, 1);
  retrieval_func->func = func;
  retrieval_func->data = data;

  retrieval_list->retrieval_remove_funcs = g_slist_append (retrieval_list->retrieval_remove_funcs,
                                                           retrieval_func);
}

void
tnt_retrieval_list_add_retrieval_completed_func (TntRetrievalList *retrieval_list,
                                                 TntRetrievalListFunc *func, gpointer data)
{
  RetrievalFunc *retrieval_func;

  assert_types (retrieval_list);
  g_return_if_fail (func != NULL);

  retrieval_func = g_new (RetrievalFunc, 1);
  retrieval_func->func = func;
  retrieval_func->data = data;

  retrieval_list->retrieval_completed_funcs = g_slist_append (retrieval_list->retrieval_completed_funcs,
                                                              retrieval_func);
}

void
tnt_retrieval_list_remove_retrieval_func (TntRetrievalList *retrieval_list,
                                          TntRetrievalListFunc *func, gpointer data)
{
  assert_types (retrieval_list);
  g_return_if_fail (func != NULL);

  remove_callback (&retrieval_list->retrieval_add_funcs,       func, data);
  remove_callback (&retrieval_list->retrieval_update_funcs,    func, data);
  remove_callback (&retrieval_list->retrieval_remove_funcs,    func, data);
  remove_callback (&retrieval_list->retrieval_completed_funcs, func, data);
}

gint
tnt_retrieval_list_get_n_retrievals (TntRetrievalList *retrieval_list)
{
  assert_types (retrieval_list);

  return retrieval_list->retrievals->len;
}

gint
tnt_retrieval_list_get_n_incomplete_retrievals (TntRetrievalList *retrieval_list)
{
  gint n = 0;
  gint i;

  assert_types (retrieval_list);

  for (i = 0; i < retrieval_list->retrievals->len; i++)
  {
    TntRetrieval *retrieval = g_ptr_array_index (retrieval_list->retrievals, i);

    if (tnt_retrieval_is_active (retrieval) && !tnt_retrieval_is_complete (retrieval))
      n++;
  }

  return n;
}

gint
tnt_retrieval_list_find_retrieval (TntRetrievalList *retrieval_list, const TntFileID *file_id)
{
  assert_types (retrieval_list);
  g_return_val_if_fail (file_id != NULL, -1);

  return find_retrieval_by_file_id (retrieval_list->retrievals, file_id);
}

TntRetrieval *
tnt_retrieval_list_get_retrieval_n (TntRetrievalList *retrieval_list, gint retrieval_n)
{
  assert_types (retrieval_list);
  g_return_val_if_fail (retrieval_n >= 0, NULL);

  if (retrieval_n >= retrieval_list->retrievals->len)
    return NULL;

  return g_ptr_array_index (retrieval_list->retrievals, retrieval_n);
}

gint
tnt_retrieval_list_add_retrieval (TntRetrievalList *retrieval_list, TntRetrieval *retrieval)
{
  gint pos;

  assert_types (retrieval_list);
  g_return_val_if_fail (retrieval != NULL, -1);

  pos = find_retrieval_by_file_id (retrieval_list->retrievals, &retrieval->file_id);
  if (pos != -1)
  {
    g_warning ("Tried to add a retrieval we already have.");
    return pos;
  }

  if (!tnt_retrieval_get_path (retrieval) && retrieval_list->incoming_storage)
    tnt_retrieval_set_path (retrieval, tnt_storage_get_path (retrieval_list->incoming_storage));

  g_ptr_array_add (retrieval_list->retrievals, retrieval);
  pos = retrieval_list->retrievals->len - 1;
  flux_object_ref (FLUX_OBJECT (retrieval));
  tnt_retrieval_set_update_func        (retrieval, retrieval_update_cb, retrieval_list);
  tnt_retrieval_set_completed_func     (retrieval, retrieval_completed_cb, retrieval_list);
  tnt_retrieval_set_status_change_func (retrieval, retrieval_status_change_cb, retrieval_list);

  retrievals_changed (retrieval_list);
  pos = run_callbacks (retrieval_list, retrieval_list->retrieval_add_funcs, pos);
  return pos;
}

void
tnt_retrieval_list_remove_retrieval (TntRetrievalList *retrieval_list, TntRetrieval *retrieval)
{
  gint pos;

  assert_types (retrieval_list);
  g_return_if_fail (retrieval != NULL);

  pos = ptr_array_find (retrieval_list->retrievals, retrieval);
  if (pos < 0)
  {
    g_warning ("Tried to remove a retrieval we don't have.");
    return;
  }

  pos = run_callbacks (retrieval_list, retrieval_list->retrieval_remove_funcs, pos);

  if (pos >= 0)
  {
    tnt_retrieval_set_update_func (retrieval, NULL, NULL);
    flux_object_unref (FLUX_OBJECT (retrieval));
    g_ptr_array_remove_index (retrieval_list->retrievals, pos);

    pop_callback_position (retrieval_list, pos);
  }

  retrievals_changed (retrieval_list);
  return;
}

static void
ptr_array_insert (GPtrArray *ptr_array, gint index, gpointer data)
{
  g_ptr_array_add (ptr_array, data);

  if (ptr_array->len - 1 > index)
  {
    /* Move elements down, overwriting added element */
    memmove (ptr_array->pdata + index + 1, ptr_array->pdata + index,
             (ptr_array->len - index - 1) * sizeof (gpointer));
  }

  *(ptr_array->pdata + index) = data;
}

gboolean
tnt_retrieval_list_move_retrieval (TntRetrievalList *retrieval_list, gint old_n, gint new_n)
{
  TntRetrieval *retrieval;

  assert_types (retrieval_list);
  g_return_val_if_fail (old_n >= 0, FALSE);
  g_return_val_if_fail (new_n >= 0, FALSE);

  if (new_n > retrieval_list->retrievals->len)
    new_n = retrieval_list->retrievals->len;

  if (old_n == new_n)
    return FALSE;

  if (old_n >= retrieval_list->retrievals->len)
    return FALSE;

  if (new_n > old_n)
    new_n--;

  retrieval = g_ptr_array_index (retrieval_list->retrievals, old_n);
  g_ptr_array_remove_index (retrieval_list->retrievals, old_n);

  /* The removal shrinks the list */
  if (new_n > retrieval_list->retrievals->len)
    new_n = retrieval_list->retrievals->len;

  ptr_array_insert (retrieval_list->retrievals, new_n, retrieval);

  move_callback_position (retrieval_list, old_n, new_n);

  retrievals_changed (retrieval_list);
  return TRUE;
}

void
tnt_retrieval_list_updated (TntRetrievalList *retrieval_list, gint retrieval_n)
{
  assert_types (retrieval_list);
  g_return_if_fail (retrieval_n >= 0);

  run_callbacks (retrieval_list, retrieval_list->retrieval_update_funcs, retrieval_n);
}

gboolean
tnt_retrieval_list_ask_block (TntRetrievalList *retrieval_list, TntFileID *file_id, TntNodeID *node_id,
                              gint *block_id)
{
  gboolean  result = FALSE;
  GSList   *busy_list;
  gint      i;

  assert_types (retrieval_list);
  g_return_val_if_fail (file_id != NULL, FALSE);
  g_return_val_if_fail (node_id != NULL, FALSE);
  g_return_val_if_fail (block_id != NULL, FALSE);

  busy_list = get_busy_sources (retrieval_list->retrievals);

  for (i = 0; i < retrieval_list->retrievals->len; i++)
  {
    TntRetrieval *retrieval = g_ptr_array_index (retrieval_list->retrievals, i);

    if (!tnt_retrieval_is_active (retrieval))
      continue;

    result = tnt_retrieval_ask_block (retrieval, busy_list, node_id, block_id);
    if (result)
    {
      memcpy (file_id, tnt_retrieval_get_file_id (retrieval), sizeof (TntFileID));
      break;
    }
  }

  free_busy_sources (busy_list);
  return result;
}

void
tnt_retrieval_list_on_queue (TntRetrievalList *retrieval_list, const TntFileID *file_id,
                             TntNodeID node_id, gint block_id,
                             TntTimeStamp remote_timestamp)
{
  gint          retrieval_n;
  TntRetrieval *retrieval;

  assert_types (retrieval_list);
  g_return_if_fail (file_id != NULL);
  g_return_if_fail (node_id != TNT_NODE_ID_INVALID);

  retrieval_n = find_retrieval_by_file_id (retrieval_list->retrievals, file_id);
  if (retrieval_n == -1)
    return;

  retrieval = tnt_retrieval_list_get_retrieval_n (retrieval_list, retrieval_n);
  if (!retrieval)
    return;

  tnt_retrieval_queued (retrieval, node_id, block_id, remote_timestamp);
}

void
tnt_retrieval_list_off_queue (TntRetrievalList *retrieval_list, TntNodeID node_id,
                              gint block_id, TntTimeStamp remote_timestamp)
{
  gint i;

  assert_types (retrieval_list);
  g_return_if_fail (node_id != TNT_NODE_ID_INVALID);

  for (i = 0; i < retrieval_list->retrievals->len; i++)
  {
    TntRetrieval *retrieval = g_ptr_array_index (retrieval_list->retrievals, i);

    tnt_retrieval_unqueued (retrieval, node_id, block_id, remote_timestamp);
  }
}

void
tnt_retrieval_list_got_block (TntRetrievalList *retrieval_list, const TntFileID *file_id, gint block_id)
{
  gint          retrieval_n;
  TntRetrieval *retrieval;

  assert_types (retrieval_list);
  g_return_if_fail (file_id != NULL);

  retrieval_n = find_retrieval_by_file_id (retrieval_list->retrievals, file_id);
  if (retrieval_n == -1)
    return;

  retrieval = tnt_retrieval_list_get_retrieval_n (retrieval_list, retrieval_n);
  if (!retrieval)
    return;

  tnt_retrieval_got_block (retrieval, block_id);
}

static gboolean
poll_retrievals_cb (gpointer data)
{
  TntRetrievalList *retrieval_list = data;
  gint              i;

  assert_types (retrieval_list);

  for (i = 0; i < retrieval_list->retrievals->len; i++)
  {
    TntRetrieval *retrieval = g_ptr_array_index (retrieval_list->retrievals, i);

    tnt_retrieval_poll (retrieval);
  }

  return TRUE;
}

void
tnt_retrieval_list_finalize (FluxObject *object)
{
  TntRetrievalList *retrieval_list = TNT_RETRIEVAL_LIST (object);

  assert_types (retrieval_list);

  if (retrieval_list->save_timeout_id != -1)
    g_source_remove (retrieval_list->save_timeout_id);

  g_source_remove (retrieval_list->poll_timeout_id);

  /* TODO: Free stuff */
}

gboolean
tnt_retrieval_list_init (TntRetrievalList *retrieval_list, TntStorage *storage)
{
  g_return_val_if_fail (storage != NULL, FALSE);

  flux_object_init              (FLUX_OBJECT (retrieval_list));
  flux_object_set_type          (FLUX_OBJECT (retrieval_list), TNT_RETRIEVAL_LIST_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (retrieval_list), tnt_retrieval_list_finalize);

  retrieval_list->storage          = storage;
  flux_object_ref (FLUX_OBJECT (retrieval_list->storage));

  retrieval_list->incoming_storage = NULL;
  retrieval_list->save_timeout_id  = -1;

  retrieval_list->retrievals = g_ptr_array_new ();

  retrieval_list->poll_timeout_id  =
    g_timeout_add (1000 * POLL_INTERVAL, poll_retrievals_cb, retrieval_list);

  return TRUE;
}

TntRetrievalList *
tnt_retrieval_list_new (TntStorage *storage)
{
  TntRetrievalList *retrieval_list;

  retrieval_list = g_new0 (TntRetrievalList, 1);
  if (!tnt_retrieval_list_init (retrieval_list, storage))
  {
    g_free (retrieval_list);
    return NULL;
  }

  return retrieval_list;
}

void
tnt_retrieval_list_load (TntRetrievalList *retrieval_list, TntBlockDB *block_db)
{
  assert_types (retrieval_list);
  flux_object_check (FLUX_OBJECT (block_db), TNT_BLOCK_DB_TYPE);

  load_retrievals (retrieval_list, block_db);
}

void
tnt_retrieval_list_set_incoming_storage (TntRetrievalList *retrieval_list, TntStorage *storage)
{
  assert_types (retrieval_list);

  if (retrieval_list->incoming_storage != storage)
  {
    if (retrieval_list->incoming_storage)
      flux_object_unref (FLUX_OBJECT (retrieval_list->incoming_storage));

    retrieval_list->incoming_storage = storage;

    if (retrieval_list->incoming_storage)
      flux_object_ref (FLUX_OBJECT (retrieval_list->incoming_storage));
  }
}
