/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-retrieval.c - Describes a running file retrieval.
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"

#include <string.h>
#include "tnt.h"

#define DEBUG(x)

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), TNT_RETRIEVAL_TYPE)

#define REMOTE_SOURCE_TIMEOUT (20 * SECONDS_IN_MINUTE)
#define ASKED_TIMEOUT         (1 * SECONDS_IN_MINUTE)
#define QUEUED_TIMEOUT        (10 * SECONDS_IN_MINUTE)
#define UNQUEUED_TIMEOUT      (3 * SECONDS_IN_MINUTE)

#define NO_BLOCK              -1

#define HISTORY_LENGTH               16
#define HISTORY_AGE_RELEVANCE_HALVED 300
#define ETA_RECALC_LIMIT             30

typedef enum
{
  BLOCK_IDLE,
  BLOCK_ASKED,
  BLOCK_QUEUED
}
BlockStatus;

typedef struct
{
  TntTimeStamp block_timestamp;
}
HistoryEntry;

typedef struct
{
  TntNodeID      node_id;
  TntTimeStamp   last_seen;
  TntFileBitmap *file_bitmap;

  gint           unqueued_block_id;
  TntTimeStamp   unqueued_timestamp;

  gint           block_id;
  BlockStatus    block_status;
  TntTimeStamp   status_change_timestamp;

  TntTimeStamp   remote_timestamp;
}
RemoteSource;

static RemoteSource *
find_remote_source_by_node_id (TntRetrieval *retrieval, TntNodeID node_id)
{
  GSList *l;

  g_return_val_if_fail (node_id != TNT_NODE_ID_INVALID, NULL);

  for (l = retrieval->remote_sources; l; l = g_slist_next (l))
  {
    RemoteSource *rs = l->data;

    if (rs->node_id == node_id)
      return rs;
  }

  return NULL;
}

static void
update_remote_bitmap (TntRetrieval *retrieval, TntNodeID node_id, TntFileBitmap *file_bitmap)
{
  RemoteSource *remote_source;

  assert_types (retrieval);
  g_return_if_fail (node_id != TNT_NODE_ID_INVALID);

  if (retrieval->local_bitmap)
    g_return_if_fail (tnt_file_bitmap_get_len_bits (retrieval->local_bitmap) ==
                      tnt_file_bitmap_get_len_bits (file_bitmap));

  remote_source = find_remote_source_by_node_id (retrieval, node_id);
  if (remote_source)
  {
    flux_object_ref (FLUX_OBJECT (file_bitmap));
    flux_object_unref (FLUX_OBJECT (remote_source->file_bitmap));
    remote_source->file_bitmap = file_bitmap;
    remote_source->last_seen = tnt_timestamp ();
  }
  else
  {
    remote_source = g_new0 (RemoteSource, 1);
    remote_source->node_id           = node_id;
    remote_source->last_seen         = tnt_timestamp ();
    remote_source->file_bitmap       = file_bitmap;
    remote_source->unqueued_block_id = NO_BLOCK;
    remote_source->block_status      = BLOCK_IDLE;
    flux_object_ref (FLUX_OBJECT (remote_source->file_bitmap));

    retrieval->remote_sources = g_slist_prepend (retrieval->remote_sources, remote_source);
  }
}

static gboolean
poll_timeouts_cb (gpointer data)
{
  TntRetrieval *retrieval = data;
  GSList       *l, *next;
  TntTimeStamp  time_now;
  gboolean      changed = FALSE;

  assert_types (retrieval);
  time_now = tnt_timestamp ();

  for (l = retrieval->remote_sources; l; l = next)
  {
    RemoteSource *remote_source;

    next = g_slist_next (l);
    remote_source = l->data;
    g_assert (remote_source);

    if (remote_source->block_status == BLOCK_ASKED &&
        remote_source->status_change_timestamp + ASKED_TIMEOUT < time_now)
    {
      remote_source->block_status = BLOCK_IDLE;
      changed = TRUE;
    }

    if (remote_source->block_status == BLOCK_QUEUED &&
        remote_source->status_change_timestamp + QUEUED_TIMEOUT < time_now)
    {
      remote_source->block_status = BLOCK_IDLE;
      changed = TRUE;
    }

    if (remote_source->unqueued_block_id != NO_BLOCK &&
        remote_source->unqueued_timestamp + UNQUEUED_TIMEOUT < time_now)
    {
      remote_source->unqueued_block_id = NO_BLOCK;
      changed = TRUE;
    }

    if (remote_source->last_seen + REMOTE_SOURCE_TIMEOUT < time_now)
    {
      /* If we haven't seen the source in a while, remove the whole thing */
      flux_object_unref (FLUX_OBJECT (remote_source->file_bitmap));
      g_free (remote_source);
      retrieval->remote_sources = g_slist_delete_link (retrieval->remote_sources, l);
      changed = TRUE;
    }
  }

  if (changed && retrieval->update_func)
  {
    retrieval->update_func (retrieval, retrieval->update_func_data);
  }

  return TRUE;
}

static void
add_block_to_history (GArray *history)
{
  HistoryEntry entry;

  entry.block_timestamp = tnt_timestamp ();
  g_array_prepend_val (history, entry);
  g_array_set_size (history, HISTORY_LENGTH);
}

static gint
calc_cps_from_history (GArray *history)
{
  TntTimeStamp now;
  gdouble      total_length = 0.0, avg_length;
  gdouble      total_scale  = 0.0;
  gint         i;

  now = tnt_timestamp ();

  for (i = 0; i < HISTORY_LENGTH - 1; i++)
  {
    TntTimeStamp begin, end;
    gint         interval_length, interval_age;
    gdouble      interval_scale;

    begin = g_array_index (history, HistoryEntry, i + 1).block_timestamp;
    end   = g_array_index (history, HistoryEntry, i).block_timestamp;

    if (begin == 0)
      break;

    interval_length = end - begin;
    interval_age    = ((now - end) + (now - begin)) / 2;
    interval_scale = (gdouble) HISTORY_AGE_RELEVANCE_HALVED /
                     ((gdouble) HISTORY_AGE_RELEVANCE_HALVED + (gdouble) interval_age);

    total_length += (gdouble) interval_length * interval_scale;
    total_scale  += interval_scale;
  }

  /* Need at least one interval */
  if (i < 1)
    return -1;

  avg_length = total_length / total_scale;
  return (gint) ((gdouble) TNT_DATA_BLOCK_SIZE / avg_length);
}

static void
update_eta (TntRetrieval *retrieval)
{
  gint remaining_blocks;

  assert_types (retrieval);

  retrieval->eta_timestamp = tnt_timestamp ();

  remaining_blocks =
    tnt_file_bitmap_get_len_bits   (retrieval->local_bitmap) -
    tnt_file_bitmap_count_set_bits (retrieval->local_bitmap);

  if (g_array_index (retrieval->history, HistoryEntry, 0).block_timestamp == 0)
  {
    retrieval->eta = TNT_RETRIEVAL_ETA_NOT_STARTED;
  }
  else if (!tnt_retrieval_is_active (retrieval))
  {
    retrieval->eta = g_array_index (retrieval->history, HistoryEntry, 0).block_timestamp - tnt_timestamp ();
    if (retrieval->eta > TNT_RETRIEVAL_LOWEST_VALUE)
      retrieval->eta = TNT_RETRIEVAL_LOWEST_VALUE - 1;
  }
  else
  {
    gint cps = calc_cps_from_history (retrieval->history);

    if (cps < 50)
    {
      retrieval->eta = TNT_RETRIEVAL_ETA_UNKNOWN;
    }
    else
    {
      retrieval->eta = ((guint64) remaining_blocks * (guint64) TNT_DATA_BLOCK_SIZE) / (guint64) cps;
    }
  }
}

void
tnt_retrieval_finalize (FluxObject *object)
{
  TntRetrieval *retrieval = TNT_RETRIEVAL (object);
  GSList       *l;

  assert_types (retrieval);

  /* Free sources */

  for (l = retrieval->remote_sources; l; l = g_slist_next (l))
  {
    RemoteSource *remote_source;

    remote_source = l->data;
    g_assert (remote_source);

    if (remote_source->file_bitmap)
      flux_object_unref (FLUX_OBJECT (remote_source->file_bitmap));
    g_free (remote_source);
  }

  g_slist_free (retrieval->remote_sources);

  /* Free other stuff */

  if (retrieval->file_meta)
    tnt_file_meta_free (retrieval->file_meta);
  if (retrieval->local_bitmap)
    flux_object_unref (FLUX_OBJECT (retrieval->local_bitmap));
  if (retrieval->target_path)
    g_free (retrieval->target_path);

  g_array_free (retrieval->history, TRUE);
}

gboolean
tnt_retrieval_init (TntRetrieval *retrieval, const TntFileID *file_id)
{
  g_return_val_if_fail (file_id != NULL, FALSE);

  flux_object_init              (FLUX_OBJECT (retrieval));
  flux_object_set_type          (FLUX_OBJECT (retrieval), TNT_RETRIEVAL_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (retrieval), tnt_retrieval_finalize);

  memcpy (&retrieval->file_id, file_id, sizeof (TntFileID));
  retrieval->local_bitmap = tnt_file_bitmap_new (TNT_REAL_LEN_TO_BITMAP_LEN (file_id->len));
  tnt_file_bitmap_fill_false (retrieval->local_bitmap);

  retrieval->eta = TNT_RETRIEVAL_ETA_NOT_STARTED;
  retrieval->history = g_array_sized_new (FALSE, TRUE, sizeof (HistoryEntry), HISTORY_LENGTH);
  g_array_set_size (retrieval->history, HISTORY_LENGTH);

  retrieval->is_active = TRUE;
  return TRUE;
}

TntRetrieval *
tnt_retrieval_new (const TntFileID *file_id)
{
  TntRetrieval *retrieval;

  retrieval = g_new0 (TntRetrieval, 1);
  if (!tnt_retrieval_init (retrieval, file_id))
  {
    g_free (retrieval);
    return NULL;
  }

  return retrieval;
}

const TntFileID *
tnt_retrieval_get_file_id (TntRetrieval *retrieval)
{
  assert_types (retrieval);

  return &retrieval->file_id;
}

const gchar *
tnt_retrieval_get_path (TntRetrieval *retrieval)
{
  assert_types (retrieval);

  return retrieval->target_path;
}

void
tnt_retrieval_set_path (TntRetrieval *retrieval, const gchar *path)
{
  assert_types (retrieval);

  if (retrieval->target_path)
    g_free (retrieval->target_path);

  retrieval->target_path = tnt_nice_path (path);
}

TntFileBitmap *
tnt_retrieval_get_local_bitmap (TntRetrieval *retrieval)
{
  assert_types (retrieval);

  return retrieval->local_bitmap;
}

void
tnt_retrieval_set_local_bitmap (TntRetrieval *retrieval, TntFileBitmap *file_bitmap)
{
  assert_types (retrieval);
  if (file_bitmap)
    flux_object_check (FLUX_OBJECT (file_bitmap), TNT_FILE_BITMAP_TYPE);

  if (retrieval->local_bitmap && file_bitmap)
    g_return_if_fail (tnt_file_bitmap_get_len_bits (retrieval->local_bitmap) ==
                      tnt_file_bitmap_get_len_bits (file_bitmap));

  if (file_bitmap)
    flux_object_ref (FLUX_OBJECT (file_bitmap));

  if (retrieval->local_bitmap)
    flux_object_unref (FLUX_OBJECT (retrieval->local_bitmap));

  retrieval->local_bitmap = file_bitmap;
  update_eta (retrieval);

  if (retrieval->update_func)
  {
    retrieval->update_func (retrieval, retrieval->update_func_data);
  }

  if (retrieval->is_active && tnt_file_bitmap_is_complete (retrieval->local_bitmap) &&
      retrieval->completed_func)
  {
    retrieval->completed_func (retrieval, retrieval->completed_func_data);
  }
}

void
tnt_retrieval_update_remote_bitmap (TntRetrieval *retrieval, TntNodeID node_id, TntFileBitmap *file_bitmap)
{
  assert_types (retrieval);
  g_return_if_fail (node_id != TNT_NODE_ID_INVALID);
  g_return_if_fail (file_bitmap != NULL);
  flux_object_check (FLUX_OBJECT (file_bitmap), TNT_FILE_BITMAP_TYPE);

  update_remote_bitmap (retrieval, node_id, file_bitmap);

  if (retrieval->update_func)
  {
    retrieval->update_func (retrieval, retrieval->update_func_data);
  }
}

TntFileBitmap *
tnt_retrieval_get_remote_bitmap (TntRetrieval *retrieval, TntNodeID node_id)
{
  RemoteSource *remote_source;

  assert_types (retrieval);
  g_return_val_if_fail (node_id != TNT_NODE_ID_INVALID, NULL);

  remote_source = find_remote_source_by_node_id (retrieval, node_id);
  if (remote_source)
    return remote_source->file_bitmap;

  return NULL;
}

void
tnt_retrieval_queued (TntRetrieval *retrieval, TntNodeID node_id, gint block_id, TntTimeStamp remote_timestamp)
{
  RemoteSource *remote_source;

  assert_types (retrieval);
  g_return_if_fail (node_id != TNT_NODE_ID_INVALID);

  /* FIXME: Maybe we should add a source entry if it doesn't exist */
  remote_source = find_remote_source_by_node_id (retrieval, node_id);
  if (!remote_source)
    return;

  if (remote_timestamp < remote_source->remote_timestamp)
    return;

  if (block_id >= tnt_file_bitmap_get_len_bits (remote_source->file_bitmap))
  {
    g_warning ("Queued block outside file! %d/%d, len=%llu",
               block_id, tnt_file_bitmap_get_len_bits (remote_source->file_bitmap),
               retrieval->file_id.len);
    return;
  }

  update_eta (retrieval);

  remote_source->block_id                = block_id;
  remote_source->block_status            = BLOCK_QUEUED;
  remote_source->remote_timestamp        = remote_timestamp;
  remote_source->last_seen               = tnt_timestamp ();
  remote_source->status_change_timestamp = tnt_timestamp ();

  if (retrieval->update_func)
  {
    retrieval->update_func (retrieval, retrieval->update_func_data);
  }
}

void
tnt_retrieval_unqueued (TntRetrieval *retrieval, TntNodeID node_id, gint block_id, TntTimeStamp remote_timestamp)
{
  RemoteSource *remote_source;

  assert_types (retrieval);
  g_return_if_fail (node_id != TNT_NODE_ID_INVALID);
  g_return_if_fail (block_id >= 0);

  /* FIXME: Maybe we should add a source entry if it doesn't exist */
  remote_source = find_remote_source_by_node_id (retrieval, node_id);
  if (!remote_source)
    return;

  if (remote_timestamp < remote_source->remote_timestamp)
    return;

  if (remote_source->block_status != BLOCK_QUEUED)
    return;

  if (block_id >= tnt_file_bitmap_get_len_bits (remote_source->file_bitmap))
  {
    tnt_log_debug ("Unqueued block outside file! %d/%d, file len=%llu",
                   block_id,
                   tnt_file_bitmap_get_len_bits (remote_source->file_bitmap),
                   retrieval->file_id.len);
    return;
  }

  update_eta (retrieval);

  remote_source->unqueued_block_id  = block_id;
  remote_source->block_status       = BLOCK_IDLE;
  remote_source->remote_timestamp   = remote_timestamp;
  remote_source->last_seen          = tnt_timestamp ();
  remote_source->unqueued_timestamp = tnt_timestamp ();

  if (retrieval->update_func)
  {
    retrieval->update_func (retrieval, retrieval->update_func_data);
  }
}

void
tnt_retrieval_got_block (TntRetrieval *retrieval, gint block_id)
{
  GSList *l;

  assert_types (retrieval);

  if (block_id >= tnt_file_bitmap_get_len_bits (retrieval->local_bitmap))
    return;

  if (!retrieval->is_active)
    return;

  /* Already have it? */
  if (tnt_file_bitmap_get_bit (retrieval->local_bitmap, block_id) == TRUE)
    return;

  add_block_to_history (retrieval->history);
  update_eta (retrieval);

  /* Update local bitmap */
  tnt_file_bitmap_set_bit (retrieval->local_bitmap, block_id, TRUE);

  /* Update sources that are servicing this block */
  for (l = retrieval->remote_sources; l; l = g_slist_next (l))
  {
    RemoteSource *remote_source = l->data;

    g_assert (remote_source != NULL);

    if (remote_source->unqueued_block_id == block_id)
      remote_source->unqueued_block_id = NO_BLOCK;

    if (remote_source->block_id == block_id)
    {
      remote_source->block_status = BLOCK_IDLE;
    }
  }

  if (retrieval->update_func)
  {
    retrieval->update_func (retrieval, retrieval->update_func_data);
  }

  if (tnt_file_bitmap_is_complete (retrieval->local_bitmap) && retrieval->completed_func)
  {
    retrieval->completed_func (retrieval, retrieval->completed_func_data);
  }
}

void
tnt_retrieval_set_active (TntRetrieval *retrieval, gboolean state)
{
  assert_types (retrieval);

  if (state != retrieval->is_active)
  {
    retrieval->is_active = state;

    if (!retrieval->is_active && tnt_file_bitmap_is_complete (retrieval->local_bitmap))
      update_eta (retrieval);

    if (retrieval->update_func)
    {
      retrieval->update_func (retrieval, retrieval->update_func_data);
    }
    if (retrieval->status_change_func)
    {
      retrieval->status_change_func (retrieval, retrieval->status_change_func_data);
    }
  }
}

gboolean
tnt_retrieval_is_active (TntRetrieval *retrieval)
{
  assert_types (retrieval);

  return retrieval->is_active;
}

gboolean
tnt_retrieval_is_complete (TntRetrieval *retrieval)
{
  assert_types (retrieval);

  return tnt_file_bitmap_is_complete (retrieval->local_bitmap);
}

TntFileMeta *
tnt_retrieval_get_file_meta (TntRetrieval *retrieval)
{
  assert_types (retrieval);

  return retrieval->file_meta;
}

void
tnt_retrieval_set_file_meta (TntRetrieval *retrieval, TntFileMeta *file_meta)
{
  assert_types (retrieval);
  g_return_if_fail (file_meta != NULL);

  if (retrieval->file_meta)
    tnt_file_meta_free (retrieval->file_meta);

  if (file_meta == NULL)
    retrieval->file_meta = NULL;
  else
  {
    retrieval->file_meta = tnt_file_meta_dup (file_meta);
  }

  if (retrieval->update_func)
  {
    retrieval->update_func (retrieval, retrieval->update_func_data);
  }
}

guint
tnt_retrieval_get_source_count (TntRetrieval *retrieval)
{
  assert_types (retrieval);

  return g_slist_length (retrieval->remote_sources);
}

static gint
count_block_availability (TntRetrieval *retrieval, gint block_id)
{
  GSList *l;
  gint    count = 0;

  for (l = retrieval->remote_sources; l; l = g_slist_next (l))
  {
    RemoteSource *rs = l->data;

    if (!rs->file_bitmap)
      continue;

    if (tnt_file_bitmap_get_bit (rs->file_bitmap, block_id))
      count++;
  }

  return count;
}

guchar
tnt_retrieval_get_block_saturation (TntRetrieval *retrieval, gint block_id)
{
  gint count;

  assert_types (retrieval);

  if (tnt_file_bitmap_get_bit (retrieval->local_bitmap, block_id))
    return 255;

  count = count_block_availability (retrieval, block_id);
  return MIN (count, 254);
}

gdouble
tnt_retrieval_get_file_saturation (TntRetrieval *retrieval, gboolean remaining)
{
  gdouble  blocks = 0.0;
  GSList  *l;

  for (l = retrieval->remote_sources; l; l = g_slist_next (l))
  {
    RemoteSource *rs = l->data;

    if (!rs->file_bitmap)
      continue;

    if (remaining)
    {
      TntFileBitmap *temp_bitmap = tnt_file_bitmap_dup (rs->file_bitmap);

      tnt_file_bitmap_nand (temp_bitmap, retrieval->local_bitmap);
      blocks += tnt_file_bitmap_count_set_bits (temp_bitmap);
      flux_object_unref (FLUX_OBJECT (temp_bitmap));
    }
    else
    {
      blocks += tnt_file_bitmap_count_set_bits (rs->file_bitmap);
    }
  }

  if (remaining)
  {
    gint blocks_left;

    blocks_left =
      tnt_file_bitmap_get_len_bits (retrieval->local_bitmap) -
      tnt_file_bitmap_count_set_bits (retrieval->local_bitmap);

    if (blocks_left)
    {
      blocks /= (gdouble) blocks_left;
    }
    else
    {
      blocks = 0.0;
    }
  }
  else
  {
    blocks /= (gdouble) tnt_file_bitmap_get_len_bits (retrieval->local_bitmap);
  }

  return blocks;
}

TntNodeID
tnt_retrieval_get_block_asked (TntRetrieval *retrieval, gint block_id)
{
  GSList *l;

  assert_types (retrieval);

  for (l = retrieval->remote_sources; l; l = g_slist_next (l))
  {
    RemoteSource *rs = l->data;

    g_assert (rs != NULL);

    if (rs->block_id == block_id && rs->block_status == BLOCK_ASKED)
      return rs->node_id;
  }

  return TNT_NODE_ID_INVALID;
}

TntNodeID
tnt_retrieval_get_block_queued (TntRetrieval *retrieval, gint block_id)
{
  GSList *l;

  assert_types (retrieval);

  for (l = retrieval->remote_sources; l; l = g_slist_next (l))
  {
    RemoteSource *rs = l->data;

    g_assert (rs != NULL);

    if (rs->block_id == block_id && rs->block_status == BLOCK_QUEUED)
      return rs->node_id;
  }

  return TNT_NODE_ID_INVALID;
}

TntNodeID
tnt_retrieval_get_block_unqueued (TntRetrieval *retrieval, gint block_id)
{
  GSList *l;

  assert_types (retrieval);

  for (l = retrieval->remote_sources; l; l = g_slist_next (l))
  {
    RemoteSource *rs = l->data;

    g_assert (rs != NULL);

    if (rs->unqueued_block_id == block_id)
      return rs->node_id;
  }

  return TNT_NODE_ID_INVALID;
}

GSList *
tnt_retrieval_get_busy_sources (TntRetrieval *retrieval)
{
  GSList *busy_list = NULL;
  GSList *l;

  assert_types (retrieval);

  for (l = retrieval->remote_sources; l; l = g_slist_next (l))
  {
    RemoteSource *rs = l->data;

    g_assert (rs != NULL);

    if (rs->block_status != BLOCK_IDLE)
    {
      TntNodeID *node_id;

      node_id = g_new (TntNodeID, 1);
      *node_id = rs->node_id;

      busy_list = g_slist_prepend (busy_list, node_id);
    }
  }

  return busy_list;
}

static gboolean
source_is_busy (GSList *busy_list, TntNodeID source_id)
{
  GSList *l;

  g_return_val_if_fail (source_id != TNT_NODE_ID_INVALID, FALSE);

  for (l = busy_list; l; l = g_slist_next (l))
  {
    TntNodeID *node_id = l->data;

    if (source_id == *node_id)
      return TRUE;
  }

  return FALSE;
}

static guchar *
generate_saturation_map (TntRetrieval *retrieval, TntFileBitmap *eligible_bitmap)
{
  guchar *sat_map;
  gint    len;
  gint    i;

  len = tnt_file_bitmap_get_len_bits (retrieval->local_bitmap);
  sat_map = g_malloc0 (len);

  for (i = 0; i < len; i++)
  {
    gint count;

    if (!tnt_file_bitmap_get_bit (eligible_bitmap, i))
      continue;

    /* FIXME: Poor locality... */
    count = count_block_availability (retrieval, i);
    if (count > 255)
      count = 255;

    sat_map [i] = count;
  }

  return sat_map;
}

static gint
count_lowest_saturation (guchar *sat_map, gint len)
{
  gint   i;
  guchar lowest_sat = 255;
  gint   count      = 0;

  for (i = 0; i < len; i++)
  {
    if (sat_map [i] == 0)
      continue;

    if (sat_map [i] < lowest_sat)
    {
      lowest_sat = sat_map [i];
      count = 1;
    }
    else if (sat_map [i] == lowest_sat)
    {
      count++;
    }
  }

  return count;
}

static gint
get_lowest_saturation_n (guchar *sat_map, gint len, gint n)
{
  gint   i;
  guchar lowest_sat = 255;
  gint   count      = 0;
  gint   found      = -1;

  for (i = 0; i < len; i++)
  {
    if (sat_map [i] == 0)
      continue;

    if (sat_map [i] < lowest_sat)
    {
      lowest_sat = sat_map [i];
      count = 1;
    }
    else if (sat_map [i] == lowest_sat)
    {
      count++;
    }

    if (count == n + 1)
      found = i;
  }

  return found;
}

static gint
pick_ask_block (TntRetrieval *retrieval, RemoteSource *rs)
{
  TntFileBitmap *eligible_bitmap;
  GSList        *l;
  gint           set_bits_count, set_bit_n;
  gint           block_id = -1;
  guchar        *saturation_map;

  if (!rs->file_bitmap)
    return -1;

  g_assert (rs->file_bitmap);
  eligible_bitmap = tnt_file_bitmap_dup (rs->file_bitmap);

  /* Eliminate blocks we already have */
  tnt_file_bitmap_nand (eligible_bitmap, retrieval->local_bitmap);

  /* Generate saturation map */
  saturation_map = generate_saturation_map (retrieval, eligible_bitmap);

  set_bits_count = tnt_file_bitmap_count_set_bits (eligible_bitmap);

  DEBUG (g_print ("%d ", set_bits_count));

  /* Only run elimination if there are more outstanding blocks than sources.
   * This lets us ask the same block from multiple sources if there are just a
   * few blocks left, but not otherwise. */
  if (set_bits_count > g_slist_length (retrieval->remote_sources))
  {
    DEBUG (g_print ("(elim"));

    /* Eliminate blocks we're getting from other sources */
    for (l = retrieval->remote_sources; l; l = g_slist_next (l))
    {
      RemoteSource *other_source = l->data;

      g_assert (other_source);

      DEBUG (g_print ("."));

      if (other_source->block_status != BLOCK_IDLE)
      {
        tnt_file_bitmap_set_bit (eligible_bitmap, other_source->block_id, FALSE);
        saturation_map [other_source->block_id] = 0;
      }

      if (other_source->unqueued_block_id != NO_BLOCK)
      {
        tnt_file_bitmap_set_bit (eligible_bitmap, other_source->unqueued_block_id, FALSE);
        saturation_map [other_source->block_id] = 0;
      }
    }

    DEBUG (g_print (") "));
  }

  /* Count lowest saturation available blocks */
  set_bits_count = count_lowest_saturation (saturation_map, tnt_file_bitmap_get_len_bits (eligible_bitmap));

  if (set_bits_count > 0)
  {
    DEBUG (g_print ("%d ", set_bits_count));

    /* Pick a random block from the set of remaining eligible blocks */
    set_bit_n = g_random_int_range (0, set_bits_count);
    block_id  = get_lowest_saturation_n (saturation_map, tnt_file_bitmap_get_len_bits (eligible_bitmap), set_bit_n);
    if (block_id == -1)
      g_warning ("(pick_ask_block) get_lowest_saturation_n() returned -1!");
  }
  DEBUG
  (
    else
      g_print ("0 ");
  )

  flux_object_unref (FLUX_OBJECT (eligible_bitmap));
  g_free (saturation_map);
  return block_id;
}

gboolean
tnt_retrieval_ask_block (TntRetrieval *retrieval, GSList *busy_list, TntNodeID *node_id, gint *block_id)
{
  GSList *l;

  assert_types (retrieval);
  g_return_val_if_fail (node_id != NULL, FALSE);
  g_return_val_if_fail (block_id != NULL, FALSE);

  /* Find the first source that's idle */
  for (l = retrieval->remote_sources; l; l = g_slist_next (l))
  {
    RemoteSource *rs = l->data;

    DEBUG (g_print ("Considering %08llx... ", rs->node_id));

    if (rs->block_status == BLOCK_IDLE && !source_is_busy (busy_list, rs->node_id))
    {
      gint block_pick_id;

      /* Pick a block to ask for, possibly eliminating blocks that other sources
       * are servicing */
      block_pick_id = pick_ask_block (retrieval, rs);
      if (block_pick_id == -1)
      {
        DEBUG (g_print ("no block picked\n"));
        continue;
      }

      DEBUG (g_print ("ok!\n"));

      update_eta (retrieval);

      *node_id                    = rs->node_id;
      *block_id                   = block_pick_id;

      rs->block_id                = block_pick_id;
      rs->block_status            = BLOCK_ASKED;
      rs->status_change_timestamp = tnt_timestamp ();
      return TRUE;
    }
    DEBUG
    (
      else if (source_is_busy (busy_list, rs->node_id))
        g_print ("in global busy list\n");
      else if (rs->block_status != BLOCK_IDLE)
        g_print ("busy status\n");
    )
  }

  return FALSE;
}

void
tnt_retrieval_set_cps_in (TntRetrieval *retrieval, guint cps_in)
{
  assert_types (retrieval);

  if (retrieval->cps_in != cps_in)
  {
    retrieval->cps_in = cps_in;
    if (retrieval->update_func)
    {
      retrieval->update_func (retrieval, retrieval->update_func_data);
    }
  }
}

guint
tnt_retrieval_get_cps_in (TntRetrieval *retrieval)
{
  assert_types (retrieval);

  return retrieval->cps_in;
}

gint
tnt_retrieval_get_eta (TntRetrieval *retrieval)
{
  assert_types (retrieval);

  if (retrieval->eta_timestamp + ETA_RECALC_LIMIT < tnt_timestamp ())
    update_eta (retrieval);

  return retrieval->eta;
}

void
tnt_retrieval_poll (TntRetrieval *retrieval)
{
  poll_timeouts_cb (retrieval);
}

void
tnt_retrieval_set_update_func (TntRetrieval *retrieval, TntRetrievalFunc *func, gpointer data)
{
  assert_types (retrieval);

  retrieval->update_func      = func;
  retrieval->update_func_data = data;
}

void
tnt_retrieval_set_completed_func (TntRetrieval *retrieval, TntRetrievalFunc *func, gpointer data)
{
  assert_types (retrieval);

  retrieval->completed_func      = func;
  retrieval->completed_func_data = data;
}

void
tnt_retrieval_set_status_change_func (TntRetrieval *retrieval, TntRetrievalFunc *func, gpointer data)
{
  assert_types (retrieval);

  retrieval->status_change_func      = func;
  retrieval->status_change_func_data = data;
}
