/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-storage.c -
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"

#include "tnt.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef HAVE_SYS_VFS_H
# include <sys/vfs.h>
#else
# ifndef G_PLATFORM_WIN32
#  include <sys/param.h>
#  include <sys/mount.h>
# endif
#endif

#ifndef G_PLATFORM_WIN32
# include <unistd.h>
#else
# include <io.h>
# include <direct.h>
#endif

static gchar *
make_storage_path (const gchar *path)
{
  if (!g_path_is_absolute (path))
  {
    gchar *cwd = g_get_current_dir ();
    gchar *path_abs;

    path_abs = g_build_path (G_DIR_SEPARATOR_S, cwd, path, NULL);
    g_free (cwd);

    return path_abs;
  }

  return tnt_nice_path (path);
}

gboolean
tnt_storage_init (TntStorage *storage, const gchar *base_path, gboolean allow_create)
{
  g_return_val_if_fail (base_path != NULL, FALSE);

  flux_object_init              (FLUX_OBJECT (storage));
  flux_object_set_type          (FLUX_OBJECT (storage), TNT_STORAGE_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (storage), tnt_storage_finalize);

  storage->base_path = make_storage_path (base_path);
  if (!storage->base_path)
    return FALSE;

  storage->limit = G_MAXINT64;
  storage->used  = 0;

  if (tnt_path_exists (storage->base_path))
  {
    /* Path exists */

    if (!tnt_path_is_dir (storage->base_path) ||
        !tnt_dir_have_full_access (storage->base_path))
    {
      g_free (storage->base_path);
      return FALSE;
    }
  }
  else if (allow_create)
  {
    gint result;

    /* Create the path */

#ifndef G_PLATFORM_WIN32
    result = mkdir (storage->base_path, TNT_STORAGE_DIR_MODE);
#else
    result = mkdir (storage->base_path);
#endif
    if (result < 0)
    {
      g_free (storage->base_path);
      return FALSE;
    }
  }
  else
  {
    /* Path doesn't exist and we can't create it */

    g_free (storage->base_path);
    return FALSE;
  }

  return TRUE;
}

TntStorage *
tnt_storage_new (const gchar *base_path, gboolean allow_create)
{
  TntStorage *storage;

  storage = g_new0 (TntStorage, 1);
  if (!tnt_storage_init (storage, base_path, allow_create))
  {
    g_free (storage);
    return NULL;
  }

  return storage;
}

void
tnt_storage_finalize (FluxObject *object)
{
  TntStorage *storage = TNT_STORAGE (object);

  g_free (storage->base_path);
}

const gchar *
tnt_storage_get_path (TntStorage *storage)
{
  g_return_val_if_fail (storage != NULL, NULL);

  return storage->base_path;
}

gboolean
tnt_storage_set_path (TntStorage *storage, const gchar *path)
{
  gchar *new_path;

  g_return_val_if_fail (storage != NULL, FALSE);
  g_return_val_if_fail (path != NULL, FALSE);

  new_path = make_storage_path (path);
  if (!tnt_dir_have_full_access (new_path))
  {
    g_free (new_path);
    return FALSE;
  }

  g_free (storage->base_path);
  storage->base_path = new_path;
  return TRUE;
}

gint64
tnt_storage_get_limit (TntStorage *storage)
{
  g_return_val_if_fail (storage != NULL, -1);

  return storage->limit;
}

void
tnt_storage_set_limit (TntStorage *storage, gint64 limit)
{
  g_return_if_fail (storage != NULL);
  g_return_if_fail (limit >= 0);

  storage->limit = limit;
}

gint64
tnt_storage_get_used (TntStorage *storage)
{
  g_return_val_if_fail (storage != NULL, -1);

  return storage->used;
}

void
tnt_storage_set_used (TntStorage *storage, gint64 used)
{
  g_return_if_fail (storage != NULL);

  if (used < 0)
    used = 0;

  storage->used = used;
}

static gint64
fs_get_free (const gchar *path)
{
#ifndef G_PLATFORM_WIN32
  struct statfs fsbuf;
#endif
  gint64        free_bytes;

#ifndef G_PLATFORM_WIN32
  if (statfs (path, &fsbuf) != 0)
    free_bytes = -1;
  else
    free_bytes = fsbuf.f_bsize * fsbuf.f_bavail;
#else
  free_bytes = -1;
#endif

  return free_bytes;
}

gint64
tnt_storage_get_free (TntStorage *storage)
{
  gint64 fs_free_bytes;
  gint64 free_bytes;

  g_return_val_if_fail (storage != NULL, -1);

  free_bytes = storage->limit - storage->used;
  if (free_bytes < 0)
    return 0;

  fs_free_bytes = fs_get_free (storage->base_path);
  if (fs_free_bytes < 0)
    return free_bytes;

  return MIN (free_bytes, fs_free_bytes);
}

gboolean
tnt_storage_same_location (TntStorage *storage_a, TntStorage *storage_b)
{
  g_return_val_if_fail (storage_a != NULL, FALSE);
  g_return_val_if_fail (storage_b != NULL, FALSE);

  if (!strcmp (storage_a->base_path, storage_b->base_path))
    return TRUE;

  return FALSE;
}

#ifdef TEST

int
main (int argc, char *argv [])
{
  TntStorage *storage;

  storage = tnt_storage_init (argv [1]);
  if (!storage)
    g_error ("Unable to open storage.");

  g_print ("Storage path: %s %s%s", storage->base_path,
           tnt_path_exists (storage->base_path) ? "(exists) " : "",
           tnt_path_is_dir (storage->base_path) ? "(dir) " : "");

  if (tnt_path_is_dir (storage->base_path))
    g_print ("%s", tnt_dir_have_full_access (storage->base_path) ? "(axx)" : "");
  else if (tnt_path_is_file (storage->base_path))
    g_print ("%s", tnt_file_have_full_access (storage->base_path) ? "(axx)" : "");

  g_print ("\n");

  return 0;
}

#endif
