/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-string-index.c -
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"

#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "tnt.h"

#ifndef G_PLATFORM_WIN32
# include <unistd.h>
#else
# include "win32/win32-compat.h"
#endif

#define KEY_FILE_EXT   ".dat"
#define VALUE_FILE_EXT ".idx"

/* These are probably not safe to change. Read code first. */
#define MAXIMUM_SUBSTRING_LEN 4
#define MINIMUM_STRING_LEN    3

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), TNT_STRING_INDEX_TYPE)

static gint
insert_list (TntStringIndex *string_index, GList *string_list, gconstpointer value)
{
  gchar     sub [MAXIMUM_SUBSTRING_LEN + 1];
  gint      n_insertions = 0;
  GList    *l;

  for (l = string_list; l; l = g_list_next (l))
  {
    memset (sub, 0, sizeof (sub));
    strncpy (sub, l->data, MAXIMUM_SUBSTRING_LEN);

    if (flux_ext_pat_tree_insert (string_index->ext_pat_tree,
                                  sub, MAXIMUM_SUBSTRING_LEN,
                                  value, string_index->value_len))
      n_insertions++;
  }

  return n_insertions;
}

static gint
remove_list (TntStringIndex *string_index, GList *string_list, gconstpointer value)
{
  gchar   sub [MAXIMUM_SUBSTRING_LEN + 1];
  gint    n_removals = 0;
  GList  *l;

  for (l = string_list; l; l = g_list_next (l))
  {
    memset (sub, 0, sizeof (sub));
    strncpy (sub, l->data, MAXIMUM_SUBSTRING_LEN);

    if (flux_ext_pat_tree_remove_value (string_index->ext_pat_tree,
                                        sub, MAXIMUM_SUBSTRING_LEN,
                                        value, string_index->value_len))
      n_removals++;
  }

  return n_removals;
}

#if 0

/* Need to keep this at top level, since otherwise the hash functions would
 * not know the element length */
static gint hash_elem_len;

static guint
hash_element (gconstpointer elem)
{
  gint    i;
  guint32 val = 0;

  g_return_val_if_fail (elem != NULL, 0);

  for (i = 0; i < hash_elem_len; i++)
  {
    guint32 t;

    val ^= ((const guchar *) elem) [i];
    t = val & 0x80000000;
    val <<= 1;
    val |= t >> 31;
  }

  return val;
}

static gint
compare_elements (gconstpointer elem_0, gconstpointer elem_1)
{
  g_return_val_if_fail (elem_0 != NULL, 1);
  g_return_val_if_fail (elem_1 != NULL, -1);

  if (!memcmp (elem_0, elem_1, hash_elem_len))
    return TRUE;

  return FALSE;
}

static GArray *
array_merge_and (GArray *array_0, GArray *array_1, guint elem_len)
{
  GHashTable *hash_1;
  gint        i;

  if (array_0 == NULL)
  {
    if (array_1)
      g_array_free (array_1, TRUE);
    return NULL;
  }
  if (array_1 == NULL)
  {
    g_array_free (array_0, TRUE);
    return NULL;
  }

  hash_elem_len = elem_len;
  hash_1        = g_hash_table_new (hash_element, compare_elements);

  for (i = 0; i < array_1->len; i++)
  {
    g_hash_table_insert (hash_1, array_1->data + (i * elem_len), array_1->data + (i * elem_len));
  }

  for (i = 0; i < array_0->len; i++)
  {
    if (!g_hash_table_lookup (hash_1, array_0->data + (i * elem_len)))
    {
      /* Not found. The element is removed from the array that is not
       * referenced by the hash table, thus this is safe. */
      g_array_remove_index_fast (array_0, i--);
    }
  }

  g_hash_table_destroy (hash_1);
  g_array_free (array_1, TRUE);
  return array_0;
}

static GArray *
array_merge_or (GArray *array_0, GArray *array_1, guint elem_len)
{
  GHashTable *hash_0;
  gint        array_0_real_len;
  gint        i;

  if (array_0 == NULL)
    return array_1;
  if (array_1 == NULL)
    return array_0;

  hash_elem_len = elem_len;
  hash_0        = g_hash_table_new (hash_element, compare_elements);

  array_0_real_len = array_0->len;
  g_array_set_size (array_0, array_0->len + array_1->len);

  for (i = 0; i < array_0_real_len; i++)
  {
    g_hash_table_insert (hash_0, array_0->data + (i * elem_len), array_0->data + (i * elem_len));
  }

  for (i = 0; i < array_1->len; i++)
  {
    if (!g_hash_table_lookup (hash_0, array_1->data + (i * elem_len)))
    {
      memcpy (array_0->data + (array_0_real_len * elem_len), array_1->data + (i * elem_len), elem_len);
      array_0_real_len++;
    }
  }

  g_hash_table_destroy (hash_0);
  g_array_free (array_1, TRUE);
  g_array_set_size (array_0, array_0_real_len);
  return array_0;
}

#endif

gint
tnt_string_index_insert (TntStringIndex *string_index, const gchar *string, gconstpointer value)
{
  gchar    *stripped;
  GList    *substrings;
  gint      n_insertions;

  assert_types (string_index);
  g_return_val_if_fail (string != NULL, 0);
  g_return_val_if_fail (value != NULL, 0);

  stripped = tnt_utf8_make_matchable (string);
  if (!stripped)
    return 0;

  substrings = tnt_string_extract_substrings (stripped, MINIMUM_STRING_LEN, MAXIMUM_SUBSTRING_LEN, FALSE);
  n_insertions = insert_list (string_index, substrings, value);
  tnt_list_free_with_data (substrings);

  g_free (stripped);
  return n_insertions;
}

static gboolean
cursor_next (FluxCursor *fcursor)
{
  TntStringIndexCursor *cursor = TNT_STRING_INDEX_CURSOR (fcursor);
  gboolean              result;

  g_return_val_if_fail (cursor != NULL, FALSE);

  result = flux_cursor_next (FLUX_CURSOR (cursor->ext_pat_cursor));

  while (!result && cursor->search_mask [MAXIMUM_SUBSTRING_LEN - 1] == 0x00)
  {
    FluxExtPatTreeCursor *ext_pat_cursor;

    memmove (cursor->search_key + 1,  cursor->search_key,  MAXIMUM_SUBSTRING_LEN - 1);
    memmove (cursor->search_mask + 1, cursor->search_mask, MAXIMUM_SUBSTRING_LEN - 1);
    cursor->search_mask [0] = 0x00;

    ext_pat_cursor = flux_ext_pat_tree_find_with_mask (cursor->string_index->ext_pat_tree,
                                                       cursor->search_key, cursor->search_mask,
                                                       MAXIMUM_SUBSTRING_LEN);
    if (ext_pat_cursor)
    {
      flux_object_unref (FLUX_OBJECT (cursor->ext_pat_cursor));
      cursor->ext_pat_cursor = ext_pat_cursor;

      /* Get results in random order */
      flux_ext_pat_tree_cursor_set_randomized (cursor->ext_pat_cursor, TRUE);

      result = TRUE;
    }
  }

  return result;
}

gboolean
tnt_string_index_cursor_init (TntStringIndexCursor *cursor,
                              TntStringIndex *string_index, const gchar *string)
{
  gchar *search_key = NULL, *search_mask = NULL;
  gchar *stripped;
  GList *substrings;

  flux_cursor_init              (FLUX_CURSOR (cursor));
  flux_object_set_type          (FLUX_OBJECT (cursor), TNT_STRING_INDEX_CURSOR_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (cursor), tnt_string_index_cursor_finalize);

  stripped = tnt_utf8_make_matchable (string);
  if (!stripped)
    return FALSE;

  substrings = tnt_string_extract_substrings (stripped, MINIMUM_STRING_LEN, MAXIMUM_SUBSTRING_LEN, FALSE);
  string = tnt_string_list_get_least_common (substrings);

  if (string && strlen (string))
  {
    gint i;

    search_key  = g_malloc0 (MAXIMUM_SUBSTRING_LEN);
    search_mask = g_malloc0 (MAXIMUM_SUBSTRING_LEN);

    for (i = 0; string [i]; i++)
    {
      search_key  [i] = string [i];
      search_mask [i] = 0xff;
    }
  }

  tnt_list_free_with_data (substrings);
  g_free (stripped);

  if (!search_key)
    return FALSE;

  cursor->string_index  = string_index;
  cursor->search_key    = search_key;
  cursor->search_mask   = search_mask;
  flux_object_ref (FLUX_OBJECT (string_index));

  cursor->ext_pat_cursor = flux_ext_pat_tree_find_with_mask (string_index->ext_pat_tree,
                                                             cursor->search_key,
                                                             cursor->search_mask,
                                                             MAXIMUM_SUBSTRING_LEN);

  while (!cursor->ext_pat_cursor && cursor->search_mask [MAXIMUM_SUBSTRING_LEN - 1] == 0x00)
  {
    memmove (cursor->search_key + 1,  cursor->search_key,  MAXIMUM_SUBSTRING_LEN - 1);
    memmove (cursor->search_mask + 1, cursor->search_mask, MAXIMUM_SUBSTRING_LEN - 1);
    cursor->search_mask [0] = 0x00;

    cursor->ext_pat_cursor = flux_ext_pat_tree_find_with_mask (cursor->string_index->ext_pat_tree,
                                                               cursor->search_key,
                                                               cursor->search_mask,
                                                               MAXIMUM_SUBSTRING_LEN);
  }

  if (!cursor->ext_pat_cursor)
  {
    g_free (cursor->search_key);
    g_free (cursor->search_mask);
    flux_object_unref (FLUX_OBJECT (cursor->string_index));
    return FALSE;
  }

  /* Get results in random order */
  flux_ext_pat_tree_cursor_set_randomized (cursor->ext_pat_cursor, TRUE);

  FLUX_CURSOR (cursor)->next = cursor_next;

  return TRUE;
}

TntStringIndexCursor *
tnt_string_index_find (TntStringIndex *string_index, const gchar *string)
{
  TntStringIndexCursor *cursor;

  assert_types (string_index);
  g_return_val_if_fail (string != NULL, NULL);

  cursor = g_new0 (TntStringIndexCursor, 1);
  if (!tnt_string_index_cursor_init (cursor, string_index, string))
  {
    g_free (cursor);
    return NULL;
  }

  return cursor;
}

gint
tnt_string_index_remove (TntStringIndex *string_index, const gchar *string, gconstpointer value)
{
  GList  *substrings;
  gchar  *stripped;
  gint    n_removals = 0;

  assert_types (string_index);
  g_return_val_if_fail (string != NULL, 0);

  stripped = tnt_utf8_make_matchable (string);
  substrings = tnt_string_extract_substrings (stripped, MINIMUM_STRING_LEN, MAXIMUM_SUBSTRING_LEN, FALSE);
  g_free (stripped);

  n_removals = remove_list (string_index, substrings, value);

  tnt_list_free_with_data (substrings);
  return n_removals;
}

gint
tnt_string_index_update (TntStringIndex *string_index, const gchar *before, const gchar *after,
                         gconstpointer value)
{
  GList *sub_1, *sub_2;
  GList *substrings;
  gchar *before_stripped, *after_stripped;
  gint   n_additions = 0;

  assert_types (string_index);
  g_return_val_if_fail (before != NULL, 0);
  g_return_val_if_fail (after != NULL, 0);
  g_return_val_if_fail (value != NULL, 0);

  before_stripped = tnt_utf8_make_matchable (before);
  after_stripped  = tnt_utf8_make_matchable (after);

  sub_1 = tnt_string_extract_substrings (before_stripped, MINIMUM_STRING_LEN, MAXIMUM_SUBSTRING_LEN, FALSE);
  sub_2 = tnt_string_extract_substrings (after_stripped, MINIMUM_STRING_LEN, MAXIMUM_SUBSTRING_LEN, FALSE);

  g_free (before_stripped);
  g_free (after_stripped);

  substrings = tnt_string_list_get_additions (sub_1, sub_2);
  if (substrings)
    n_additions = insert_list (string_index, substrings, value);
  tnt_list_free_with_data (substrings);

  substrings = tnt_string_list_get_additions (sub_2, sub_1);

  tnt_list_free_with_data (sub_1);
  tnt_list_free_with_data (sub_2);

  if (substrings)
    n_additions -= remove_list (string_index, substrings, value);

  tnt_list_free_with_data (substrings);
  return n_additions;
}

guint
tnt_string_index_get_file_size (TntStringIndex *string_index)
{
  assert_types (string_index);

  return flux_storage_get_arena_size (string_index->key_storage) +
         flux_storage_get_arena_size (string_index->value_storage);
}

guint
tnt_string_index_get_free_size (TntStringIndex *string_index)
{
  assert_types (string_index);

  return flux_storage_get_arena_size (string_index->key_storage) +
         flux_storage_get_arena_size (string_index->value_storage) -
         flux_storage_get_used_size (string_index->key_storage) -
         flux_storage_get_used_size (string_index->value_storage);
}

gboolean
tnt_string_index_cursor_read (TntStringIndexCursor *cursor, gpointer value)
{
  gpointer value_ret;
  guint    value_ret_len;

  g_return_val_if_fail (cursor != NULL, FALSE);

  if (!flux_ext_pat_tree_cursor_read (cursor->ext_pat_cursor, NULL, NULL, &value_ret, &value_ret_len))
    return FALSE;

  if (value_ret_len != cursor->string_index->value_len)
  {
    g_free (value_ret);
    return FALSE;
  }

  memcpy (value, value_ret, cursor->string_index->value_len);
  g_free (value_ret);

  return TRUE;
}

void
tnt_string_index_cursor_finalize (FluxObject *object)
{
  TntStringIndexCursor *cursor = TNT_STRING_INDEX_CURSOR (object);

  g_return_if_fail (cursor != NULL);

  g_free (cursor->search_key);
  g_free (cursor->search_mask);
  if (cursor->ext_pat_cursor)
    flux_object_unref (FLUX_OBJECT (cursor->ext_pat_cursor));

  flux_object_unref (FLUX_OBJECT (cursor->string_index));

  flux_cursor_finalize (object);
}

gboolean
tnt_string_index_init (TntStringIndex *string_index, TntStorage *storage, const gchar *base_name, guint value_len)
{
  gint         tree_fd, chunk_fd;
  guint        key_node_len, value_node_len;
  gchar       *path, *path_temp;

  g_return_val_if_fail (string_index != NULL, FALSE);
  g_return_val_if_fail (storage != NULL, FALSE);
  g_return_val_if_fail (base_name != NULL, FALSE);
  g_return_val_if_fail (value_len > 0, FALSE);

  flux_object_init              (FLUX_OBJECT (string_index));
  flux_object_set_type          (FLUX_OBJECT (string_index), TNT_STRING_INDEX_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (string_index), tnt_string_index_finalize);

  path_temp = g_strjoin (G_DIR_SEPARATOR_S, tnt_storage_get_path (storage), base_name, NULL);
  path = g_strconcat (path_temp, KEY_FILE_EXT, NULL);
  g_free (path_temp);

  tree_fd = tnt_open (path, TRUE, FALSE, FALSE);
  g_free (path);

  if (tree_fd < 0)
    return FALSE;

  path_temp = g_strjoin (G_DIR_SEPARATOR_S, tnt_storage_get_path (storage), base_name, NULL);
  path = g_strconcat (path_temp, VALUE_FILE_EXT, NULL);
  g_free (path_temp);

  chunk_fd = tnt_open (path, TRUE, FALSE, FALSE);
  g_free (path);

  if (chunk_fd < 0)
  {
    tnt_close (tree_fd);
    return FALSE;
  }

  string_index->value_len    = value_len;

  flux_ext_pat_tree_calc_node_sizes (MAXIMUM_SUBSTRING_LEN, value_len,
                                     &key_node_len, &value_node_len);

  string_index->key_storage   = FLUX_STORAGE (flux_disk_chunk_new (tree_fd, key_node_len));
  string_index->value_storage = FLUX_STORAGE (flux_disk_chunk_new (chunk_fd, value_node_len));

  string_index->ext_pat_tree = flux_ext_pat_tree_new (string_index->key_storage,
                                                      string_index->value_storage,
                                                      "string-index",
                                                      MAXIMUM_SUBSTRING_LEN, value_len);

  return TRUE;
}

TntStringIndex *
tnt_string_index_new (TntStorage *storage, const gchar *base_name, guint value_len)
{
  TntStringIndex *string_index;

  string_index = g_new (TntStringIndex, 1);
  if (!tnt_string_index_init (string_index, storage, base_name, value_len))
  {
    g_free (string_index);
    return NULL;
  }

  return string_index;
}

void
tnt_string_index_finalize (FluxObject *object)
{
  TntStringIndex *string_index = TNT_STRING_INDEX (object);

  flux_object_unref (FLUX_OBJECT (string_index->ext_pat_tree));
  flux_object_unref (FLUX_OBJECT (string_index->key_storage));
  flux_object_unref (FLUX_OBJECT (string_index->value_storage));
}
