/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* tnt-tcp-io.h - TCP I/O channel.
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#ifndef TNT_TCP_IO_H
#define TNT_TCP_IO_H

#define DEBUG_CALLBACKS(x)

/* NOTE: Changing any of these sizes is extremely hazardous */
#define TNT_MESSAGE_FLAGS_SIZE           1
#define TNT_MESSAGE_TYPE_SIZE            2
#define TNT_MESSAGE_TTL_SIZE             1
#define TNT_MESSAGE_SIGNATURE_SIZE_SIZE  1
#define TNT_MESSAGE_PAYLOAD_SIZE_SIZE    2
#define TNT_MESSAGE_DATA_BLOCK_SIZE_SIZE 4

#define TNT_DATA_STEP_SIZE               1024

#define TNT_PAYLOAD_SIZE_MAX             4096  /* Payload may not exceed this limit */

#define TNT_TCP_IO(object) ((TntTcpIO *) object)

typedef enum
{
  TNT_TCP_IO_READ_FLAGS,            /* Reading initial flag byte */
  TNT_TCP_IO_READ_MESSAGE_HEADER,   /* TntMessage - src, dest, ttl, type and payload size */
  TNT_TCP_IO_READ_SIGNATURE,        /* Optional signature */
  TNT_TCP_IO_READ_MESSAGE_PAYLOAD,  /* Contents specific to message type; marshalled */
  TNT_TCP_IO_READ_DATA_BLOCK        /* Reading a data block (follows payload of TNT_MESSAGE_DATA_BLOCK) */
}
TntTcpIOReadState;

typedef enum
{
  TNT_TCP_IO_WRITE_IDLE,            /* No message to write */
  TNT_TCP_IO_WRITE_HEADERS,         /* Writing flags, header (all at once) */
  TNT_TCP_IO_WRITE_PAYLOAD,         /* Writing payload (piecemeal) */
  TNT_TCP_IO_WRITE_DATA_BLOCK       /* Writing data block (piecemeal) */
}
TntTcpIOWriteState;

typedef struct
{
  TntTcpIOReadState state;

  guchar            flags;
  TntMessageType    type;
  guint             ttl;

  TntNodeID         source_id;
  TntNodeID         dest_id;

  gpointer          signature;
  guint8            signature_size;

  gpointer          payload;
  guint16           payload_size;

  gpointer          data_block;
  guint32           data_block_size;

  guint             block_pos;
}
TntTcpIOReadInfo;

typedef struct
{
  TntTcpIOWriteState  state;

  TntMessage         *message;

  guchar              flags;
  TntMessageType      type;
  guint8              ttl;

  TntNodeID           source_id;
  TntNodeID           dest_id;

  gpointer            block;
  guint               block_size;
  guint               block_pos;
}
TntTcpIOWriteInfo;

struct _TntTcpIO
{
  FluxTcpIO              tcp_io;

  TntTcpIOReadInfo       read_info;
  TntTcpIOWriteInfo      write_info;

  TntMessageInFunc      *message_in_func;
  gpointer               message_in_data;

  TntMessageInFunc      *data_in_begin_func;
  gpointer               data_in_begin_data;

  GList                 *queue_out, *queue_out_last;
  guint                  queue_out_len;

  GHashTable            *queue_class_table;
};

#if DEBUG_CALLBACKS == 'x'
# define _tnt_tcp_io_be_dead(tcp_io) _tnt_tcp_io_be_dead_real (tcp_io, G_GNUC_PRETTY_FUNCTION)
#else
# define _tnt_tcp_io_be_dead(tcp_io)
#endif

TntTcpIO        *tnt_tcp_io_new                        (void);
TntTcpIO        *tnt_tcp_io_new_with_cipher            (FluxCipher *cipher);
void             tnt_tcp_io_init                       (TntTcpIO *tcp_io, FluxCipher *cipher);
void             tnt_tcp_io_init_self                  (TntTcpIO *tcp_io);
void             tnt_tcp_io_finalize                   (FluxObject *tcp_io);

TntNodeIP4       tnt_tcp_io_get_remote_ip4             (TntTcpIO *tcp_io);
void             tnt_tcp_io_get_remote_address4        (TntTcpIO *tcp_io, TntNodeAddress4 *address);

void             tnt_tcp_io_reset                      (TntTcpIO *tcp_io);

gint             tnt_tcp_io_get_queue_out_max_len      (TntTcpIO *tcp_io, TntMessageType message_type);
void             tnt_tcp_io_set_queue_out_max_len      (TntTcpIO *tcp_io, TntMessageType message_type, gint max_len);
gint             tnt_tcp_io_get_queue_out_len          (TntTcpIO *tcp_io, TntMessageType message_type);
void             tnt_tcp_io_message_queue_out          (TntTcpIO *tcp_io, TntMessage *message);

void             tnt_tcp_io_set_message_in_func        (TntTcpIO *tcp_io, TntMessageInFunc *func, gpointer data);
void             tnt_tcp_io_set_data_in_begin_func     (TntTcpIO *tcp_io, TntMessageInFunc *func, gpointer data);

guint            tnt_tcp_io_get_cps                    (TntTcpIO *tcp_io, TntTransferDirection dir);

guint            tnt_tcp_io_get_cps_goal               (TntTcpIO *tcp_io, TntTransferDirection dir);
void             tnt_tcp_io_set_cps_goal               (TntTcpIO *tcp_io, TntTransferDirection dir, guint cps);

gboolean         tnt_tcp_io_get_read_info              (TntTcpIO *tcp_io, TntMessageType *type,
                                                        gpointer *payload, guint16 *payload_size);
TntMessage      *tnt_tcp_io_get_write_message          (TntTcpIO *tcp_io);

#endif  /* TNT_TCP_IO_H */
