/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* test-meta-db.c - Test TntMetaDB
 *
 * Copyright (C) 2005 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include <glib.h>
#include <tnt.h>
#include <string.h>

#define SCRATCH_DIR      "tnt-meta-db-test"
#define QUEUE_LEN_MIN    32
#define ACTION_BURST_NUM 32
#define QUERIES_MAX      8
#define QUERY_SKIP_BACK  16

static gboolean   opt_verbose  = FALSE;
static TntMetaDB *meta_db      = NULL;
static GMainLoop *loop         = NULL;
static GList     *file_list    = NULL;
static GList     *extract_next = NULL;
static GList     *queries      = NULL;

static void
extract_file_done (const TntFileID *file_id, TntFileMeta *file_meta, gpointer data)
{
  if (file_meta)
  {
    if (opt_verbose)
    {
      gchar *path;

      path = *((gpointer *) file_id->hash);
      g_print ("%s: Got metadata.\n", path);
    }

    tnt_meta_db_set_file_meta (meta_db, file_id, file_meta);
    tnt_file_meta_free (file_meta);
  }

  if (!extract_next)
    g_main_quit (loop);
}

static void
extract_file (const gchar *path)
{
  TntFileID  file_id;
  gchar     *file_name;
  TntStat    tstat;

  if (!tnt_lstat (path, &tstat))
  {
    if (opt_verbose)
      g_print ("%s: Not found.\n", path);
    return;
  }

  if (!tstat.is_file)
  {
    if (opt_verbose)
      g_print ("%s: Not a file.\n", path);
    return;
  }

  if (opt_verbose)
    g_print ("%s: Extracting metadata.\n", path);

  memset (file_id.hash, 0, TNT_FILE_ID_LEN);
  *((gpointer *) file_id.hash) = (gpointer) path;
  file_id.len = tstat.len;

  file_name = g_path_get_basename (path);
  tnt_meta_db_extract_file_meta (meta_db, &file_id, file_name, path, path,
                                 extract_file_done, NULL);
  g_free (file_name);
}

static void
query_done (TntQuery *query, gpointer data)
{
  if (opt_verbose)
    g_print ("Query done: \"%s\".\n", tnt_query_get_string (query));

  queries = g_list_remove (queries, query);
  flux_object_unref (FLUX_OBJECT (query));
}

static void
add_query (const gchar *path)
{
  TntQuery *query;
  gchar    *p;

  p = strrchr (path, '.');
  if (!p)
    return;

  p++;
  if (!*p)
    return;

  query = TNT_QUERY (tnt_local_query_new ());
  tnt_query_set_string (query, p);
  tnt_query_set_match_local (query, TRUE);
  tnt_query_set_match_partial (query, TRUE);

  if (opt_verbose)
    g_print ("Starting query: \"%s\".\n", tnt_query_get_string (query));

  tnt_meta_db_query (meta_db, query, FALSE, query_done, NULL);

  queries = g_list_prepend (queries, query);
}

static void
cancel_query (TntQuery *query)
{
  if (opt_verbose)
    g_print ("Cancelling query: \"%s\".\n", tnt_query_get_string (query));

  queries = g_list_remove (queries, query);
  tnt_meta_db_cancel_query (meta_db, query);
  flux_object_unref (FLUX_OBJECT (query));
}

static void
pause_query (TntQuery *query)
{
  if (opt_verbose)
    g_print ("Pausing query: \"%s\".\n", tnt_query_get_string (query));

  tnt_meta_db_query_set_paused (meta_db, query, TRUE);
}

static void
resume_query (TntQuery *query)
{
  if (opt_verbose)
    g_print ("Resuming query: \"%s\".\n", tnt_query_get_string (query));

  tnt_meta_db_query_set_paused (meta_db, query, FALSE);
}

static void
do_something (void)
{
  gint action;

  action = g_random_int_range (0, 8);

  switch (action)
  {
    case 0:
      /* Add a query */
      if (extract_next)
      {
        gint   skip_back = g_random_int_range (0, QUERY_SKIP_BACK);
        GList *l         = extract_next;
        gint   i;

        for (i = 0; i < skip_back && l->prev; i++)
          l = g_list_previous (l);

        add_query (l->data);
      }
      break;

    case 1:
      /* Cancel a query */
      if (g_list_length (queries))
      {
        gint n = g_random_int_range (0, g_list_length (queries));
        cancel_query (g_list_nth_data (queries, n));
      }
      break;

    case 2:
      /* Pause a query */
      if (g_list_length (queries))
      {
        gint n = g_random_int_range (0, g_list_length (queries));
        pause_query (g_list_nth_data (queries, n));
      }
      break;

    case 3:
      /* Resume a query */
      if (g_list_length (queries))
      {
        gint n = g_random_int_range (0, g_list_length (queries));
        resume_query (g_list_nth_data (queries, n));
      }
      break;

    default:
      /* Extract metadata */
      if (extract_next)
      {
        extract_file (extract_next->data);
        extract_next = g_list_next (extract_next);
      }
      break;
  }
}

static gboolean
fill_queue (gpointer data)
{
  if (tnt_meta_db_get_action_queue_len (meta_db) < QUEUE_LEN_MIN)
  {
    gint i;

    for (i = 0; i < ACTION_BURST_NUM; i++)
      do_something ();
  }

  return TRUE;
}

static void
test_meta_db (void)
{
  TntStorage *storage;

  extract_next = file_list;
  loop = g_main_new (FALSE);

  tnt_dir_delete (SCRATCH_DIR);
  storage = tnt_storage_new (SCRATCH_DIR, TRUE);
  if (!storage)
  {
    g_print ("Could not create scratch directory '%s'.\n", SCRATCH_DIR);
    goto storage_open_failed;
  }

  meta_db = tnt_meta_db_new (storage);
  if (!meta_db)
  {
    g_print ("Could not create meta db in '%s',\n", SCRATCH_DIR);
    goto meta_open_failed;
  }

  g_timeout_add (1000, fill_queue, NULL);
  g_main_run (loop);

  if (opt_verbose)
    g_print ("Terminating...\n");

  flux_object_unref (FLUX_OBJECT (meta_db));
  flux_object_unref (FLUX_OBJECT (storage));
meta_open_failed:
  tnt_dir_delete (SCRATCH_DIR);
storage_open_failed:
  g_main_loop_unref (loop);
}

gint
main (gint argc, gchar *argv [])
{
  gint i;

  tnt_register_classes ();

  if (argc < 2)
  {
    g_print ("Usage: %s [-v] FILE [FILE] ...\n", argv [0]);
    return 1;
  }

  for (i = 1; i < argc; i++)
  {
    if (!strcmp (argv [i], "-v"))
    {
      opt_verbose = TRUE;
    }
    else
    {
      file_list = g_list_prepend (file_list, g_strdup (argv [i]));
    }
  }

  test_meta_db ();

  tnt_list_free_with_data (file_list);
  return 0;
}
